// $Id: Block.java,v 1.2 2004-02-27 17:20:00 msato Exp $
// $OMNI_Release$
// $OMNI_Copyright$
package exc.block;
import exc.object.*;

/**
 * abstract class of a statement block
 */
public class Block extends PropObject {
  /** the nubmer of gnerated block. */
  protected static int BlockCounter = 0;

  int code;		// opcode for this block
  BasicBlock bblock;	// for expressions, statements

  Block prev,next;	// double linked list in BlockList
  BlockList parent;	// block list which this block belongs to.

  LineNo lineno;	// lineno information

  /** contructor for simple block whick contains a basic block */
  public Block(int code,BasicBlock bblock){
    this.code = code;
    this.bblock = bblock;
    if(bblock != null) bblock.parent = this;
    id = BlockCounter++;
  }

  /** constructor to make a copy of block b */
  public Block(Block b){
    this.code = b.code;
    this.lineno = b.lineno;
    if(b.bblock != null){
      this.bblock = b.bblock.copy();
      this.bblock.parent = this;
    }
    id = BlockCounter++;
  }

  /** copy this block */
  public Block copy() { return new Block(this); }

  /** get Opcode for this block */
  public int Opcode() { return code; }
  /** get the BasicBlock */
  public BasicBlock getBasicBlock() { return bblock; }
  /** get the next block of this block */
  public Block getNext() { return next; }
  /** get the previous block */
  public Block getPrev() { return prev; }
  /** get the parent BlockList which this block belongs to. */
  public BlockList getParent() { return parent; }
  /** get the parent Block with which this block belongs to. */
  public Block getParentBlock() { 
    return parent == null? null:parent.getParent(); 
  }

  /** set the line number information */
  public void setLineNo(LineNo ln){ lineno = ln; }
  /** get the line number information */
  public LineNo getLineNo() { return lineno; }

  /** convert to the printable string */
  public String toString(){
    if(code >= Xcode.MAX_CODE) 
      return "Block.<"+code+">#"+id;
    return "Block."+Xcode.getName(code)+"#"+id;
  }

  /** apply BasicBlockVisitor v */
  public void visitBasicBlock(BasicBlockVisitor v){ v.visit(bblock); }
  /** apply BasicBlockVisitor for the body */
  public void visitBody(BasicBlockVisitor v){ v.visit(getBody()); }

  //
  // abstract member function
  //
  /** convert to Xobject */
  public Xobject toXobject() { fatal("toXobject",this); return null; }
  /** get the body */
  public BlockList getBody(){ return null; }
  /** set the body */
  public void setBody(BlockList s) {  fatal("setBody",this); }
  /** get BasicBlock in cond part */
  public BasicBlock getCondBBlock() 
  { fatal("getCondBBlock",this); return null; }
  /** get the BlockList of "then" part */
  public BlockList getThenBody(){ fatal("getThenBody",this); return null; }
  /** get the BlockList of "else" part */
  public BlockList getElseBody(){ fatal("getElseBody",this); return null; }
  /** set the BlockList of "then" part */
  public void setThenBody(BlockList s){ fatal("setThenBody",this); }
  /** set the BlockList of "else" part */
  public void setElseBody(BlockList s){ fatal("setElseBody",this); }

  public BasicBlock getInitBBlock()
  { fatal("getInitBBlock",this); return null;}
  public BasicBlock getIterBBlock()
  { fatal("getIterBBlock",this); return null; }
  public void setInitBBlock(BasicBlock s){ fatal("setInitBBlock",this); }
  public void setIterBBlock(BasicBlock s){ fatal("setIterBBlock",this); }
  public Xobject getLabel() { fatal("getLabel",this); return null; }
  public void setLabel(Xobject x) { fatal("setLabel",this); }

  public Xobject getInfoExpr() { return null; }
  public void setInfoExpr() { fatal("no InfoExpr"); }

  /** add Block s */
  public Block add(Block s) { 
    s.parent = parent;
    s.next = next;
    next = s;
    s.prev = this;
    if(s.next == null){ // tail of list
      parent.tail = s;
    } else {
      s.next.prev = s;
    }
    return s;
  }
  public Block add(BasicBlock bb){ 
    return add(Bcons.BasicBlock(bb));
  }
  public Block add(Xobject s){
    if(code == Xcode.LIST){
      bblock.add(s);
      return this;
    }
    return add(Bcons.Statement(s));
  }

  public Block insert(Block s){
    s.parent = parent;
    s.prev = prev;
    prev = s;
    s.next = this;
    if(s.prev == null){ // head of list
      parent.head = s;
    } else {
      s.prev.next = s;
    }
    return s;
  }
  public Block insert(BasicBlock bb) { 
    return insert(Bcons.BasicBlock(bb));
  }
  public Block insert(Xobject s){
    if(code == Xcode.LIST){
      bblock.insert(s);
      return this;
    }
    return insert(Bcons.Statement(s));
  }

  public void replace(Block b){
    b.parent = parent;
    b.prev = prev;
    b.next = next;
    if(prev != null) prev.next = b;
    else parent.head = b;
    if(next != null) next.prev = b;
    else parent.tail = b;
  }
    
  // remove block from list
  public Block remove(){
    if(prev != null) prev.next = next;
    else  parent.head = next;
    if(next != null) next.prev = prev;
    else parent.tail = prev;
    parent = null;
    next = null;
    prev = null;
    return this;
  }

  // find Id in this context
  public Ident findIdent(String name){
    BlockList b_list;
    Ident id;
    for(b_list = parent; b_list != null; b_list = b_list.getParentList()){
      if((id = b_list.findLocalIdent(name)) != null) return id;
      if(b_list.getParent() instanceof FunctionBlock){
	return ((FunctionBlock)b_list.getParent()).
	  getFile().findIdent(name);
      }
    }
    return null;	// not found
  }

  static void fatal(String msg,Block b){
    fatal(msg+" at "+b);
  }

  static void fatal(String msg){
    System.err.println("Fatal Block:"+msg);
    Thread.dumpStack();
    System.exit(1);
  }

  public static int numberOfBlock(){
    return Block.BlockCounter;
  }

  public static int numberOfBasicBlock(){
    return BasicBlock.BasicBlockCounter;
  }
}


