// $Id: IfBlock.java,v 1.3 2004-07-05 18:14:57 msato Exp $
// $OMNI_Release$
// $OMNI_Copyright$
package exc.block;

import exc.object.*;

/**
 * Block object to express If statement.
 */
public class IfBlock extends Block {
  BlockList then_part;
  BlockList else_part;

  /** constructor */
  public IfBlock(BasicBlock cond,
		 BlockList then_part,BlockList else_part){
    super(Xcode.IF_STATEMENT,cond);
    this.then_part = then_part;
    if(then_part != null) then_part.parent = this;
    this.else_part = else_part;
    if(else_part != null) else_part.parent = this;
  }

  /** constructor to make copy of b */
  public IfBlock(IfBlock b){
    super(b);
    if(b.then_part != null) this.then_part = b.then_part.copy();
    if(b.else_part != null) this.else_part = b.else_part.copy();
  }

  /** return copy of this block */
  public Block copy() { return new IfBlock(this); }

  /** return BasicBlock in condition part */
  public BasicBlock getCondBBlock() { return bblock; }

  /** set BasicBlock in condition part */
  public void setCondBBlock(BasicBlock bb) { 
    bblock = bb; 
    bb.parent = this;
  }

  /** get the BlockList of "then" part */
  public BlockList getThenBody(){ return then_part; }
  /** get the BlockList of "else" part */
  public BlockList getElseBody(){ return else_part; }
  /** set the BlockList of "then" part */
  public void setThenBody(BlockList s){
    then_part = s;
    s.parent = this;
  }
  /** set the BlockList of "else" part */
  public void setElseBody(BlockList s) {
    else_part = s;
    s.parent = this;
  }

  /** apply visitor recursively */
  public void visitBody(BasicBlockVisitor v){ 
    v.visit(then_part); 
    v.visit(else_part); 
  }

  /** convert to Xobject */
  public Xobject toXobject(){
    Xobject then_x = null;
    Xobject else_x = null;
    if(getThenBody() != null) then_x = getThenBody().toXobject();
    if(getElseBody() != null) else_x = getElseBody().toXobject();
    Xobject x = new XobjList(Opcode(),getCondBBlock().toXobject(),
			     then_x,else_x);
    x.setLineNo(getLineNo());
    return x;
  }
}





