import exc.object.*;

/**
 * Object to represnt Instruction, base class
 *
 */
public class Inst extends exc.object.PropObject {

  // kind of Inst
  public static int INFO = 0;		// InstInfo
  public static int PROLOG = 1;		// Inst
  public static int EPILOG = 2;		// Inst
  public static int LABEL = 3;		// Inst
  public static int GOTO = 4;		// Inst

  public static int BINARY = 5;		// BinaryInst
  public static int UNARY = 6;		// UnaryInst
  public static int LOAD = 7;		// MemInst
  public static int STORE = 8;		// MemInst
  public static int SET = 9;		// SetInst
  public static int MOV = 10;		// MovInst

  public static int BINARY_IF = 11;	// BinaryInst
  public static int UNARY_IF = 12;	// UnaryInst
  public static int CALL = 13;		// CallInst
  public static int ARG = 14;		// SetInst

  public static int RETVAL = 15;	// SetInst
  public static int RETURN = 16;	// Inst
  public static int SWITCH = 17;	// SwitchInst

  static String kindName[] = {
    "INFO",
    "PROLOG",
    "EPILOG",
    "LABEL",
    "GOTO",
    "BINARY",
    "UNARY",
    "LOAD",
    "STORE",
    "SET",
    "MOV",
    "BINARY_IF",
    "UNARY_IF",
    "CALL",
    "ARG",
    "RETVAL",
    "RETURN",
    "SWITCH",
    "*"
  };

  static boolean has_dest[] = {
    false, // "INFO",
    false, // "PROLOG",
    false, //"EPILOG",
    false, //"LABEL",
    false, //"GOTO",
    true,  //"BINARY",
    true,  //"UNARY",
    true,  //"LOAD",
    false, //"STORE",
    true, // "SET",
    true, //"MOV",
    false, //"BINARY_IF",
    false, //"UANRY_IF",
    false, //"CALL",
    false, //"ARG",
    false, //"RETVAL",
    false, //"RETURN",
    false, //"SWITCH",
    false //"*"
  };

  int kind;	// kind 
  int val;	// value
  
  /** constructor */
  public Inst(int kind){
    this.kind = kind;
  }

  public Inst(int kind, int val){
    this.kind = kind;
    this.val = val;
  }

  /** return the kind of Inst */
  public int getKind() { return kind; }
  
  /** return the name of kind of Inst */
  public String getKindName() { return kindName[kind]; }

  /** returns true if it has a distination register */
  public boolean hasDest() { return has_dest[kind]; }

  //
  // val field is shared among Label, Register and Value(for misc)
  //
  /** return the label if the kind is GOTO or LABEL */
  public int getLabel() { return val; }
  /** set Label */
  void setLabel(int label) { val = label; }

  /** get Value */
  public int getValue() { return val; }
  /** set Value */
  void setValue(int label) { val = label; }

  /** get register operand */
  public int getRegN() { return val; }
  /** set register operand */
  void setregN(int label) { val = label; }

  /** get destination register */
  int getDest(){ return val; }
  /** get destination register */
  void setDest(int dst) { this.val = dst; }

  /** convert to Inst to String */
  public String toString(){
    return "["+getKindName()+","+val+"]";
  }

  //
  // abstract member function
  //
  /** get opcode */
  public int getOpcode() { fatal("getOpcode"); return 0; }
  /** set opcode */
  public void setOpcode(int op) { fatal("setOpcode"); }

  /** get left operand */
  public Operand getLeft() { fatal("getLeft"); return null; }
  /** set left operand */
  public void setLeft(Operand o) { fatal("setLeft"); }

  /** get right operand */
  public Operand getRight() { fatal("getRight"); return null; }
  /** set right operand */
  public void setRight(Operand o) { fatal("setRight"); }

  /** get operand */
  public Operand getOperand() { fatal("getOperand"); return null; }
  /** set operand */
  public void setOperand(Operand o) { fatal("setOperand"); }

  public void fatal(String msg){
    System.err.println("Fatal Inst:"+msg);
    System.exit(1);
  }

  //
  // static constructor 
  //
  /** make prolog Inst */
  public static Inst Prolog(int nparams){
    return new Inst(PROLOG, nparams);
  }

  /** make epilog Inst */
  public static Inst Epilog(){
    return new Inst(EPILOG);
  }

  /** make Goto Inst */
  public static Inst Goto(int label){
    return new Inst(GOTO,label);
  }

  /** make Label Inst */
  public static Inst Label(int label){
    return new Inst(LABEL,label);
  }

  /** make Binary Inst */
  public static Inst Binary(int op,int dst, Operand left, Operand right){
    return new BinaryInst(BINARY,op,dst,left,right);
  }

  /** make Unary Inst */
  public static Inst Unary(int op,int dst, Operand opd){
    return new UnaryInst(UNARY,op,dst,opd);
  }
  
  /** make Load Inst */
  public static Inst Load(int dst, Operand opd){
    return new MemInst(LOAD,dst,opd);
  }

  /** make Store Inst */
  public static Inst Store(int src, Operand opd){
    return new MemInst(STORE,src,opd);
  }

  /** make Set Inst */
  public static Inst Set(int dst,Operand opd){
    return new SetInst(SET,dst,opd);
  }

  /** make RetVal Inst */
  public static Inst RetVal(Operand opd){
    return new SetInst(RETVAL,0,opd);
  }

  /** make BinaryIf Inst */
  public static Inst BinaryIf(int label, int op, Operand left, Operand right){
    return new BinaryInst(BINARY_IF,op,label,left,right);
  }

  /** make UnaryIf Inst */
  public static Inst UnaryIf(int label, int op, Operand opd){
    return new UnaryInst(UNARY_IF,op,label,opd);
  }

  /** make Swtich Inst */
  //  public static Inst Switch(int dst){
  // return new SwitchInst(dst);
  // }

  public static Inst Info(Xobject x){
    return new InfoInst(INFO,x,0,0);
  }

  public static Inst Info(Xobject x,int x1, int x2){
    return new InfoInst(INFO,x,x1,x2);
  }

}



