import exc.object.*;

/**
 * Object to represent Operand, base class
 */
public class Operand {
  public static int REG = 0;
  public static int LVAR = 1; 	// local variable (offset)
  public static int PARAM = 2;	// parameter variable (offset)
  public static int VAR = 3;	// global variable (symbol)
  public static int LADDR = 4;	// local area (offset)
  public static int ADDR = 5; 	// reg+symbolic name+offset
  public static int ADDR2 = 6; 	// reg+reg
  public static int CONST = 7;  // constant+symbol
  public static int LABEL = 8;	// label

  int kind;	// see above
  int type;	// type, significant for load/store

  /** constructor */
  public Operand(int kind, int type){
    this.kind = kind;
    this.type = type;
  }

  public int getKind(){ return kind; }

  /** get label */
  public int getLabel() { fatal("getLabel"); return 0; }
  /** set value */
  public void setLabel(int offset) { fatal("setLabel"); }
  
  /** get register number */
  public int getRegN() { fatal("getRegN"); return 0; }
  /** set register number */
  public void setRegN(int regn) { fatal("setRegN"); }

  /** get reg 2 number */
  public int getRegN2() { fatal("getRegN2"); return 0; }
  /** set reg 2 number */
  public void setRegN2(int regn) { fatal("setRegN2"); }

  /** get value */
  public int getValue() { fatal("getValue"); return 0; }
  /** set value */
  public void setValue(int offset) { fatal("setValue"); }

  /** get offset */
  public int getOffset() { fatal("getOffset"); return 0;}
  /** set offset */
  public void setOffset(int offset) { fatal("setOffset"); }

  /** get Name */
  public String getName() { fatal("setName"); return null;}
  /** set Name */
  public void setName(String name) { fatal("getName"); }

  void fatal(String msg){
    System.err.println("fatal in Operand: "+msg);
    System.exit(1);
  }

  //
  // static constructor
  // 
  /** make Reg Operand */
  public static Operand Reg(int regn){
    return new RegOperand(regn);
  }

  /** make LocalVar Operand for load/store */
  public static Operand LocalVar(int type, int off){
    return new LocalVarOperand(type,off);
  }

  /** make ParamVar Operand */
  public static Operand Param(int type, int off){
    return new ParamOperand(type,off);
  }

  /** make Variable Operand */
  public static Operand Var(int type, String name){
    return new VarOperand(type,name);
  }

  /** make local memory address Operand */
  public static Operand LocalAddr(int type, int off){
    return new LocalAddrOperand(type, off);
  }

  /** make memory address Operand with (reg+symbol+off) */
  public static Operand Addr(int type, int regn, String name, int off){
    return new AddrOperand(type,regn,name,off);
  }

  /** make memory address operand with (reg1+reg2) */
  public static Operand Addr2(int type,int regn1,int regn2){
    return new Addr2Operand(type,regn1,regn2);
  }

  /** make Constant operand */
  public static Operand Const(int type,int n){
    return new ConstOperand(type,n);
  }

  /** make Constant operand from object */
  public static Operand Const(Xobject e){
    return new ConstOperand(e);
  }

  /** make Label Operand */
  public static Operand Label(int type, int l){
    return new LabelOperand(type,l);
  }

}


