// $Id: BasicType.java,v 1.3 2004-07-05 18:14:58 msato Exp $
// $OMNI_Release$
// $OMNI_Copyright$
package exc.object;

/**
 * Xtype object to present Basic type, such as int, char, ...
 */
public class BasicType extends Xtype {
  /** undefined type */
  public final static int UNDEF = 0;	/* undefined */
  /** void type */
  public final static int VOID = 1;
  /** signed character type (8bit) */
  public final static int CHAR = 2;
  /** unsigned character type (8bit) */
  public final static int UNSIGNED_CHAR = 3;
  /** signed short (16bit) */
  public final static int SHORT = 4;
  /** unsigned short (16bit) */
  public final static int UNSIGNED_SHORT = 5;
  /** signed ??? not used ??? */
  public final static int SIGNED = 6;
  /** signed int */
  public final static int INT = 7;
  /** unsigned int */
  public final static int UNSIGNED_INT = 8;
  /** long int */
  public final static int LONG = 9;
  /** unsigned long int */
  public final static int UNSIGNED_LONG = 10;
  /** longlong int (64bit) */
  public final static int LONGLONG = 11;
  /** unsigned longlong (64bit) */
  public final static int UNSIGNED_LONGLONG = 12;
  /** single precision floating point (32bit) */
  public final static int FLOAT = 13;
  /** double precision floating point (64bit) */
  public final static int DOUBLE = 14;
  /** extended precision floting point (128bit) */
  public final static int LONG_DOUBLE =15;

  /** number of basic types */
  public final static int N_BASIC_TYPE = 16;

  /* extended for Fortran */
  /** single precision floating point complex */
  public final static int FCOMPLEX_FLOAT = 16;
  /** double precision floating point complex */
  public final static int FCOMPLEX_DOUBLE = 17;

  /* for pointer object */
  public final static int ADDR = 20;

  int basic_type;	// member variable

  /** constructor */
  public BasicType(int basic_type,int size,byte align){
    super(Xtype.BASIC,0,size,align,false,false);
    this.basic_type = basic_type;
  }

  /** constructor */
  public BasicType(int basic_type,int size,byte align,boolean ic, boolean iv){
    super(Xtype.BASIC,0,size,align,ic,iv);
    this.basic_type = basic_type;
  }

  /** return basic type */
  public int getBasicType() { return basic_type; }

  public final static String getName(int i){
    return Xtype.basic_type_names[i];
  }
  
  public final static BasicType getType(int i){
    return Xtype.basic_types[i];
  }
 
  public boolean isUnsigned(){
      switch(basic_type){
      case UNSIGNED_CHAR:
      case UNSIGNED_SHORT:
      case UNSIGNED_INT:
      case UNSIGNED_LONG:
      case UNSIGNED_LONGLONG:
	  return true;
      }
      return false;
  }

  public boolean isIntegral(){
      return (basic_type >= CHAR) && (basic_type <= UNSIGNED_LONGLONG);
  }

  public boolean isFloating(){
      return (basic_type >= FLOAT) && (basic_type <= LONG_DOUBLE);
  }

  public static Xtype Conversion(Xtype lt,Xtype rt){
      int b1 = lt.getBasicType();
      int b2 = rt.getBasicType();
      if(b1 > b2) return getType(b1);
      else return getType(b2);
  }

  public Xtype copy(){
    return new BasicType(basic_type,size,align,is_const,is_volatile);
  }

  public String toString(){
    return Xtype.basic_type_names[basic_type];
  }
}



