// $Id: XobjList.java,v 1.2 2004-02-27 17:20:02 msato Exp $
// $OMNI_Release$
// $OMNI_Copyright$
package exc.object;

/**
 * Xobject which contains the list of Xobjects.
 */

public class XobjList extends Xobject {
  XobjArgs args;
  XobjArgs tail;	// for add
  LineNo lineno;

  /** Constructor for the empty XobjList with Xcode.LIST */
  public XobjList(){
    super(Xcode.LIST,null);
  }

  /** Constructor for the empty XobjList with code and type */
  public XobjList(int code,Xtype type){
    super(code,type);
  }

  /** Constructor for the argument list */
  public XobjList(int code,Xtype type,XobjArgs a){
    super(code,type);
    args = a;
  }

  /** Constructor with one argument */
  public XobjList(int code,Xtype type,Xobject a1){
    super(code,type);
    args = new XobjArgs(a1,null);
  }

  /** Constructor with two arguments */
  public XobjList(int code,Xtype type,Xobject a1,Xobject a2){
    super(code,type);
    args = new XobjArgs(a1,new XobjArgs(a2,null));
  }

  /** Constructor with three arguments */
  public XobjList(int code,Xtype type,Xobject a1,Xobject a2,Xobject a3){
    super(code,type);
    args = new XobjArgs(a1, new XobjArgs(a2,new XobjArgs(a3,null)));
  }

  
  /** Constructor with four arguments */
  public XobjList(int code,Xtype type,Xobject a1,Xobject a2,Xobject a3,
		  Xobject a4){
    super(code,type);
    args = new XobjArgs(a1,new XobjArgs(a2, new XobjArgs(a3,new XobjArgs(a4,null))));
  }

  /** Constructor with five arguments */
  public XobjList(int code,Xtype type,Xobject a1,Xobject a2,Xobject a3,
		  Xobject a4,Xobject a5){
    super(code,type);
    args = new XobjArgs(a1,new XobjArgs(a2, new XobjArgs(a3,new XobjArgs(a4, new XobjArgs(a5,null)))));
  }

  /** Constructor with only code */
  public XobjList(int code){
    super(code,null);
  }

  /** Constructor with code and argument */
  public XobjList(int code,XobjArgs a){
    super(code,null);
    args = a;
  }

  /** Constructor with one argument */
  public XobjList(int code,Xobject a1){
    super(code,null);
    args = new XobjArgs(a1,null);
  }
  /** Constructor with two arguments */
  public XobjList(int code,Xobject a1,Xobject a2){
    super(code,null);
    args = new XobjArgs(a1,new XobjArgs(a2,null));
  }

  /** Constructor with three arguments */
  public XobjList(int code,Xobject a1,Xobject a2,Xobject a3){
    super(code,null);
    args = new XobjArgs(a1, new XobjArgs(a2,new XobjArgs(a3,null)));
  }

  /** Constructor with four arguments */
  public XobjList(int code,Xobject a1,Xobject a2,Xobject a3,Xobject a4){
    super(code,null);
    args = new XobjArgs(a1,new XobjArgs(a2, new XobjArgs(a3,new XobjArgs(a4,null))));
  }

  /** Constructor with five arguments */
  public XobjList(int code,Xobject a1,Xobject a2,Xobject a3,
		  Xobject a4,Xobject a5){
    super(code,null);
    args = new XobjArgs(a1,new XobjArgs(a2, new XobjArgs(a3,new XobjArgs(a4, new XobjArgs(a5,null)))));
  }
    
  /** Add object at the end of list */
  public void add(Xobject a){
    if(args == null){
      args = new XobjArgs(a,null);
      return;
    }
    if(tail == null){
      for(tail = args; tail.nextArgs() != null; tail = tail.nextArgs());
    }
    tail.next = new XobjArgs(a,null);
    tail = tail.next;
  }

  /** returns the first argument */
  public Xobject operand(){
    if(args == null) return null;
    return args.getArg();
  }

  /** returns the left argument (the first argument) */
  public Xobject left(){
    if(args == null) return null;
    return args.getArg();
  }

  /** returns the right argment (the second argument) */
  public Xobject right(){
    if(args == null || args.nextArgs() == null) return null;
    return args.nextArgs().getArg();

  }

  /** sets the first argument */
  public void setOperand(Xobject x){
    args.setArg(x);
  }

  /** sets the left (first) argument */
  public void setLeft(Xobject x){
    args.setArg(x);
  }

  /** sets the right (second) argument */
  public void setRight(Xobject x){
    args.nextArgs().setArg(x);
  }

  /** Get the argument list */
  public XobjArgs getArgs() {
    return args;
  }

  /** Sets the argument list */
  public void setArgs(XobjArgs l) {
    args = l; tail = null;
  }

  /** Get the i-th argument */
  public Xobject getArg(int i){
    XobjArgs a = args;
    while(i > 0){
      a = a.nextArgs();
      i--;
    }
    return a.getArg();
  }

  /** Sets the i-th argument */
  public void setArg(int i,Xobject x){
    XobjArgs a = args;
    while(i > 0){
      a = a.nextArgs();
      i--;
    }
    a.setArg(x);
  }

  /** Return the number of arugments. */
  public int Nargs() {
    int i = 0;
    for(XobjArgs a = args; a != null; a = a.nextArgs()) i++;
    return i;
  }

  /** set the line number information. */
  public void setLineNo(LineNo ln){ lineno = ln; }
  /** get the line number information. */
  public LineNo getLineNo() { return lineno; }

  /* returns the copy of this XobjList object. */
  public Xobject copy(){
    XobjList x = new XobjList(code,type);
    for(XobjArgs a = args; a != null; a = a.nextArgs()){
      if(a.getArg() == null) x.add(null);
      else x.add(a.getArg().copy());
    }
    return x;
  }

  /** check the equality of two Xobjects. */
  public boolean equals(Xobject x){
    if(x == null) return false;
    if(!super.equals(x)) return false;
    XobjArgs a1 = args;
    XobjArgs a2 = x.getArgs();
    while(a1 != null){
      if(a2 == null) return false;
      if(a1.getArg() != a2.getArg()){
	if(a1.getArg() == null) return false;
	if(!a1.getArg().equals(a2.getArg())) return false;
      }
      a1 = a1.nextArgs();
      a2 = a2.nextArgs();
    }
    if(a2 != null) return false;
    return true;
  }

  /** convert to the printable string */
  public String toString(){
    String s = "(" + OpcodeName();
    for(XobjArgs a = args; a != null; a = a.nextArgs()){
      if(a.getArg() == null)
	s = s + " ()";
      else 
	s = s + " "+a.getArg();
    }
    return s + ")";
  }
}

