// $Id: Xobject.java,v 1.3 2004-07-05 18:14:58 msato Exp $
// $Release$
// $Copyright$
package exc.object;

/**
 * Abstract class of expression tree objects in Xobject.
 */
public class Xobject extends PropObject {
  int code;
  Xtype type;

  /**
   * Constructs an Xobject with code and type.
   * This constructor is usually called superclass.
   */
  public Xobject(int code,Xtype type){
    this.code = code;
    this.type = type;
  }

  /** Returns the code of this Xobject */
  public int Opcode(){
    return code;
  }
    
  /** Returns the type of this Xobject */
  public Xtype Type(){
    return type;
  }
    
  /** 
   * Returns integer value in XobjInt.
   * Dummy method at base class, it causes fatal. 
   */
  public int getInt(){
    fatal("getInt",this);
    return 0; 
  }

  /**  
   * Returns string in XobjString.
   * Dummy method at base class, it causes fatal. 
   */
  public String getString(){
    fatal("getStr",this);
    return null;
  }

  /**
   * Returns Symbol name in XobjString. it is equal to getString.
   * Dummy method at base class, it causes fatal. 
   */
  public String getSym(){
    fatal("getSym",this);
    return null;
  }

  /**
   * Returns Name of Xobject. 
   * Dummy method at base class.
   */
  public String getName(){
    switch(code){
    case Xcode.FCOMM_VAR:
    case Xcode.FCOMM_VAR_ADDR:
    case Xcode.FCOMM_ARRAY_ADDR:
      return (((XobjList)this).left().getSym());
    }
    fatal("getName",this);
    return null;
  }

  /**  
   * Returns floating value if XobjLong. 
   * Dummy method at base class, it causes fatal. 
   */
  public double getFloat(){
    fatal("getFloat",this);
    return 0.0;
  }

  /**
   * Returns long value (64bits) if XobjLong.
   * Dummy method at base class, it causes fatal. 
   */
  public long getLong(){
    fatal("getLong",this);
    return 0L;
  }

  /**
   * Returns high part of long value  if XobjLong.
   * Dummy method at base class, it causes fatal. 
   */
  public int getIntHIGH(){
    fatal("getIntHIGH",this);
    return 0;
  }

  /**
   * Returns low part of long value if XobjLong.
   * Dummy method at base class, it causes fatal. 
   */
  public int getIntLOW(){
    fatal("getIntLOW",this);
    return 0;
  }
  /** 
   * Return argument list in XobArg.
   * Dummy method at base class, it causes fatal. 
   */
  public Xobject getArg(int i){
    fatal("getArg(int)",this);
    return null;
  }

  /** 
   * Set argument list in XobArg.
   * Dummy method at base class, it causes fatal. 
   */
  public void setArg(int i,Xobject x){
    fatal("setArg(int,Xobject)",this);
  }

  /** 
   * Set number of argument in XobArg.
   * Dummy method at base class, it causes fatal. 
   */
  public int Nargs(){
    fatal("Nargs",this);
    return 0;
  }

  /** 
   * Get an argument list XobArg.
   * Dummy method at base class, it causes fatal. 
   */
  public XobjArgs getArgs(){
    fatal("getArgs",this);
    return null;
  }
  /** 
   * Set an argument list in XobArg.
   * Dummy method at base class, it causes fatal. 
   */
  public void setArgs(XobjArgs l){
    fatal("setArgs",this);
  }
  public Xobject operand(){
    fatal("left",this);
    return null;
  }
  public Xobject left(){
    fatal("left",this);
    return null;
  }
  public Xobject right(){
    fatal("right",this);
    return null;
  }
  public void setOperand(Xobject x){
    fatal("setOperand",this);
  }
  public void setLeft(Xobject x){
    fatal("setLeft",this);
  }
  public void setRight(Xobject x){
    fatal("setRight",this);
  }
  public void add(Xobject a){
    fatal("add",this);
  }
  public Xobject copy(){
    fatal("copy",this);
    return null;
  }
  public void setLineNo(LineNo ln){ fatal("setLineNo",this); }
  public LineNo getLineNo() { return null; }

  /** check the equality of two objects */
  public boolean equals(Object x){
    if(!(x instanceof Xobject)) return false;
    return equals((Xobject)x);
  }	
  public boolean equals(Xobject x){
    if(this == x) return true;
    return (code == x.code && (type == x.type || type.equals(x.type)));
  }

  static void fatal(String msg,Xobject o){
    System.err.println("Fatal Xobj:"+msg+" at "+o);
    Thread.dumpStack();
    System.exit(1);
  }

  public String OpcodeName(){
    String name = (code >= Xcode.MAX_CODE)? 
      ("code<"+Integer.toString(code)+">"): Xcode.getName(code);
    if(type == null) return name;
    else return name+":"+type;
  }

  public String toString(){
    return "<"+OpcodeName()+">";
  }
    
  /** 
   * Return true if the Xobject is a variable.
   */
  public boolean isVariable(){
    switch(code){
    case Xcode.PARAM_VAR:
      if(type.isArray()) return false;	// array parameter
    case Xcode.VAR:
    case Xcode.LVAR:
    case Xcode.FPARAM_VAR:
    case Xcode.FCOMM_VAR:
    case Xcode.REG:
      return true;
    default:
      return false;
    }
  }

  public boolean isLocalVar(){
    switch(code){
    case Xcode.LVAR:
    case Xcode.PARAM_VAR:
    case Xcode.REG:
      return true;
    default:
      return false;
    }
  }

  public boolean isTempVar(){
    return code == Xcode.REG;
  }

  public boolean isVarAddr(){
    switch(code){
    case Xcode.VAR_ADDR:
    case Xcode.LVAR_ADDR:
    case Xcode.PARAM_ADDR:
    case Xcode.FPARAM_VAR_ADDR:
    case Xcode.FCOMM_VAR_ADDR:
      /* case Xcode.REG: */
      return true;
    default:
      return false;
    }
  }

  public boolean isLocalVarAddr(){
    switch(code){
    case Xcode.LVAR_ADDR:
    case Xcode.PARAM_ADDR:
    case Xcode.FPARAM_VAR_ADDR:
      return true;
    default:
      return false;
    }
  }

  public boolean isArrayAddr(){
    switch(code){
    case Xcode.ARRAY_ADDR:
    case Xcode.LARRAY_ADDR:
    case Xcode.PARAM_ADDR:
    case Xcode.FPARAM_ARRAY_ADDR:
    case Xcode.FCOMM_ARRAY_ADDR:
      return true;
    case Xcode.PARAM_VAR:
      if(type.isArray()) return true; // array parameter
    default:
      return false;
    }
  }

  /* for Fortran common */
  public boolean isFCOMM(){
    switch(code){
    case Xcode.FCOMM_VAR_ADDR:
    case Xcode.FCOMM_ARRAY_ADDR:
    case Xcode.FCOMM_VAR:
      return true;
    default:
      return false;
    }
  }

  public Xobject getFCOMM(){  return right().left(); }
  public void setFCOMM(Xobject comm){  right().setLeft(comm);  }

  /** return true if this is an constant object */
  public boolean isConstant(){
    if(!isTerminal()) return false;
    switch(code){
    case Xcode.ARRAY_ADDR:
    case Xcode.LARRAY_ADDR:
    case Xcode.PARAM_ADDR:
    case Xcode.FPARAM_ARRAY_ADDR:
    case Xcode.FCOMM_ARRAY_ADDR:
    case Xcode.FPARAM_VAR_ADDR:
    case Xcode.FCOMM_VAR_ADDR:
    case Xcode.VAR_ADDR:
    case Xcode.LVAR_ADDR:
    case Xcode.INT_CONSTANT:
    case Xcode.LONG_CONSTANT:
    case Xcode.STRING_CONSTANT:
    case Xcode.LONGLONG_CONSTANT:
    case Xcode.FLOAT_CONSTANT:
    case Xcode.MOE_CONSTANT:
    case Xcode.SIZE_OF_EXPR: /* ??? */
    case Xcode.IDENT:  /* ??? */
      return true;
    }
    return false;
  }
  /** return true if this object is zero constant. */
  public boolean isZeroConstant(){
    return code == Xcode.INT_CONSTANT && getInt() == 0;
  }
  /** return true if this object is one constant. */
  public boolean isOneConstant(){
    return code == Xcode.INT_CONSTANT && getInt() == 1;
  }
  /** return true if this object is integer constant */
  public boolean isIntConstant(){
    return code == Xcode.INT_CONSTANT;
  }
  /** return ture if this object is an assignment. */
  public boolean isSet(){
    return code == Xcode.ASSIGN_EXPR;
  }
  /** return true if this object is binary operation. */
  public boolean isBinaryOp(){
    if(code >= Xcode.MAX_CODE) return false;
    return Xcode.kind[code] == 'B';
  }
  /** return true if this object is unary operation. */
  public boolean isUnaryOp(){
    if(code >= Xcode.MAX_CODE) return false;
    return Xcode.kind[code] == 'U';
  }
  /** return ture if this object is an assignment with binary operation */
  public boolean isAsgOp(){
    if(code >= Xcode.MAX_CODE) return false;
    return Xcode.kind[code] == 'A';
  }
  /** return true if this object is a terminal object */
  public boolean isTerminal(){
    if(code >= Xcode.MAX_CODE) return false;
    return Xcode.kind[code] == 'T';
  }

  /** 
   * return true if this object is a logical operation.
   *  EQ, NEQ, GE, GT, LE, LT, AND, OR, NOT 
   */
  public boolean isLogicalOp(){
    switch(code){
    case Xcode.LOG_EQ_EXPR:
    case Xcode.LOG_NEQ_EXPR:
    case Xcode.LOG_GE_EXPR:
    case Xcode.LOG_GT_EXPR:
    case Xcode.LOG_LE_EXPR:
    case Xcode.LOG_LT_EXPR:
    case Xcode.LOG_AND_EXPR:
    case Xcode.LOG_OR_EXPR:
    case Xcode.LOG_NOT_EXPR:
      return true;
    }
    return false;
  }

  /* 
   * allocating new codes 
   */
  static int max_code = Xcode.MAX_CODE;
  public static int newCode() { return max_code++; }

  public boolean isExternalCode() { return code >= Xcode.MAX_CODE; }
}



