// $Id: ompcc.java,v 1.2 2004-02-27 17:20:03 msato Exp $
// $OMNI_Release$
// $OMNI_Copyright$

//
// CC driver
//
package exc.util;
import exc.object.*;
import exc.openmp.*;

public class ompcc extends exc.util.ccDriver {

  protected boolean omp_flag = true; 
  protected boolean doubleAlign = false;

  public ompcc(){ super("ompcc"); }
  public ompcc(String name){ super(name); }

  protected String backend_opts = "";

  public static void main(String args[]){ 
    ompcc driver = new ompcc();
    driver.run(args);
  }

  public boolean parseArg(String arg){ 
    if(arg.equals("-omp")){
      omp_flag = false;
      return true;
    } else if (arg.equals("-doubleAlign") || arg.equals("-malign-double")) {
      doubleAlign = true;
      return true;
    } else {
      backend_opts += " " + arg;
      return false;
    }
  }

  public void printUsage() { 
    System.err.println("options: -omp, disable OpenMP directives");
    System.err.println("options: -doubleAlign or -malign-double, if alignment of double is NOT sizeof double, treat alignment of double is sizeof double");
  }

  public void init(){
    String cflags = getTemplateValue("OMPC_CFLAGS");
    if (cflags != null) {
      cc_opts += " " + cflags;
    }
    if(omp_flag){
      String approvedDate = getTemplateValue("APPROVED_DATE");
      if (approvedDate == null) {
	cpp_opts += " -D_OPENMP ";
      } else {
	cpp_opts += " -D_OPENMP=" + approvedDate + " ";
      }
      ld_opts +=  " "+getTemplateValue("OMPC_MAIN");
      ld_opts +=  " "+getTemplateValue("OMPC_LDFLAGS");
    } else {
      cfront_opts += " -omp ";
    }

    if (doubleAlign) {
      cfront_opts += " -malign-double ";
      cc_opts += " -malign-double ";
    }
    if (backend_opts != "") {
      cc_opts += " " + backend_opts;
    }
  }

  public void run(XobjectFile f){
    if(!omp_flag) return;	// no OpenMP
    if(verbose) System.err.println("OpenMP pass ...");
    OMPtranslate op = new OMPtranslate();
    op.init(f);
    f.iterateDef(op);	// transform OMP pragma 
    if(OMP.errorFlags){
      System.err.println("error in OpenMP. stop.");
      System.exit(1);
    }
  }
}



