// $Id: mustWriteAnalyzer.java,v 1.2 2004-02-27 17:20:04 msato Exp $
// $RWC_Release$
// $RWC_Copyright$

package ompd;

import exc.object.*;
import exc.block.*;
import exc.openmp.*;
import exc.flow.*;
import exc.util.*;

public class mustWriteAnalyzer implements DataFlowSets {
  NumberSet Gen[];
  NumberSet Kill[];
  NumberSet In[];
  NumberSet Out[];

  int max_index;
  BasicBlock head_bb;

  public void run(optBodyEnv env,int num_shared_obj){
    run(env.getCFG(),num_shared_obj);
  }

  public void run(ControlFlowGraph cfg,int num_shared_obj){
    max_index = num_shared_obj;
    
    int nBlock = cfg.numOfBasicBlock();
    Gen = new NumberSet[nBlock];
    Kill = new NumberSet[nBlock];
    In = new NumberSet[nBlock];
    Out = new NumberSet[nBlock];

    head_bb = cfg.getHead();
    setupWriteSet(cfg);

    cfg.solveForwardIntersectFlow(this);
  }

  void setupWriteSet(ControlFlowGraph cfg){
    for(BasicBlock bb = cfg.getHead(); bb != null; bb = bb.topNext()){
      int id = bb.Id();
      if(!bb.isEmpty()){
	NumberSet write_set = NumberSet.empty(max_index);
	for(Statement s = bb.getHead(); s != null; s = s.getNext()){
	  Xobject x = s.getExpr();
	  if(x == null) continue;
	  if(x.Opcode() == RWcheckCode.WCHK){
	    RWcheckCode xx = (RWcheckCode)x;
	    write_set.add(xx.getIndex());
	  }
	}
	Gen[id] = write_set;
      }

      Block b = bb.getParent();
      if((b.Opcode() == Xcode.OMP_FLUSH || b.Opcode() == Xcode.OMP_BARRIER)
	 && ((OMPBlock)b).beginBasicBlock() == bb)
	Kill[id] = NumberSet.full(max_index);	// kill all 
      
      In[id] = NumberSet.full(max_index);
      Out[id] = NumberSet.full(max_index);
    }
  }

  /* interface to DataFlowSets */
  public void initSet(BasicBlock bb){ 
    if(bb == head_bb){
      In[bb.Id()].clear();
      Out[bb.Id()].copy(Gen[bb.Id()]);
    } else {
      Out[bb.Id()].remove(Kill[bb.Id()]);
    }
  }
  public Set emptySet(){ return NumberSet.empty(max_index); }
  public Set GenSet(BasicBlock bb){ return Gen[bb.Id()]; }
  public Set InSet(BasicBlock bb){ return In[bb.Id()]; }
  public Set OutSet(BasicBlock bb){ return Out[bb.Id()]; }
  public Set KillSet(BasicBlock bb){ return Kill[bb.Id()]; }

  /* return result */
  public boolean mustWrite(BasicBlock bb,int index){
    return In[bb.Id()].contain(index);
  }
}

