/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import exc.object.*;
import exc.block.*;
import exc.object.*;

/*
 * Data block for handling begin pragmas.
 *
 * A Translater object contains a linked list of ScmBlock objects
 * and calls process() for translation.
 */
public class ScmBlock implements Names {
    protected ScmBlock next;
    protected Translater translater;

    protected boolean invalid;
    protected String varName;
    protected Xtype varType;
    protected int dimension;
    protected Xobject[] sizeList;
    protected int[] strideList;
    protected Xobject IdList;
    protected XobjList StrideList;

    public ScmBlock(Xobject name, Xobject params, Translater trans,
		    ScmBlock _next)
    {
	next = _next;
	translater = trans;
	invalid = false;
	varName = name.getSym();
	varType = translater.findCommonVariable(varName).Type();
	
	//System.out.println("varType= " + varType);

	if (varType.getKind() != Xtype.ARRAY) {
	    translater.errorMessage(name, "not array");
	    invalid = true;
	    return;
	}

	dimension = varType.getNumDimensions();

	if (params.Nargs() != dimension * 2) {
	    translater.errorMessage(params, "bad parameters");
	    invalid = true;
	    return;
	}

        sizeList = new Xobject[dimension];
	strideList = new int[dimension];

	for (int i = 0; i < dimension; ++i) {
	    sizeList[i] = params.getArg(i);
	    Xobject stride = params.getArg(i + dimension);
	    if (stride instanceof XobjInt)
		strideList[i] = stride.getInt();
	    else {
		translater.errorMessage(params,
					"a stride width must be constant");
		invalid = true;
		return;
	    }
	}

	IdList = genIdList();
	System.out.println(IdList);
	StrideList = genStrideList(params);
    }

    public Xtype getType() { return varType; }

    public int getDimension() { return dimension; }

    public int getStride(int dim) { return strideList[dim]; }

    public ScmBlock lookup(String name) {
	ScmBlock b = this;
	while (b != null)
	    if (b.varName.equals(name))
		return b;
	    else
		b = b.next;

	return null;
    }

    public boolean invalidate(String name, boolean v) {
	ScmBlock b = lookup(name);
	if (b == null)
	    return false;
	else if (b.invalid)
	    return false;
	else {
	    b.invalid = v;
	    return true;
	}
    }

    /* Wraps the given program block with temporary variable declarations.
     *
     * Example: suppose the given block is:
     *
     * for (i = 0; i < N; ++i)
     *     x[i] = y[i + 1];
     *
     * Then, this method returns:
     *
     * { double* x_ptr;
     *   int x_offset_0; int x_size;
     *   x_ptr = scm_alloc(...);
     *   for (i = 0; i < N; ++i) {
     *       ?????????????;		translated by ScmBlock.process().
     *   }
     *   scm_free(x_ptr);
     * }
     *
     */
    public Xobject makeBlock(Xobject block) {
        if (scmf77.isOldFashion){
	StringBuffer code = new StringBuffer();
	code.append(makePointerType()).append(' ')
	    .append(getPointerName()).append(";\n");

	for (int i = 0; i < dimension; ++i)
	    code.append("int ").append(getOffsetName(i))
		.append(" = 0;\nint ").append(getCacheSize(i))
		.append(" = ").append(translater.toString(sizeList[i]))
		.append(";\n");

	extraVarDeclaration(code);
	makeAlloc(code);
	return new XobjList(Xcode.SCM_PRAGMA_CODE,
			    new XobjString(Xcode.IDENT,
					   "\n{\n" + code.toString()),
			    block,
			    new XobjString(Xcode.IDENT,
					   "\n" + makeFree() + "}\n"));
        }

	// IdList = genIdList();
	// System.out.println(IdList);

	Xobject decl = genDecl();
	// System.out.println(decl);

	Xobject v = genAlloc();
	XobjList obj = Xcons.List(v);

	if (block.Opcode() == Xcode.LIST)
	  for (XobjArgs a = block.getArgs(); a != null; a = a.nextArgs())
	    obj.add(a.getArg());
	else
	  obj.add(block);

	obj.add(genFree());

	// System.out.println(obj);

	// BlockList body = Bcons.emptyBody(IdList, decl);
	// Block bb = Bcons.Statement(v);
	// bb.add(block);
	// body.add(bb);
	// System.out.println(body.toXobject());

	return new XobjList(Xcode.COMPOUND_STATEMENT, IdList, decl, obj);
    }

    protected void extraVarDeclaration(StringBuffer code) {
	// do nothing.
    }

    protected void makeAlloc(StringBuffer code) {
	makeAlloc(getPointerName(), code);
    }

    protected void makeAlloc(String ptrName, StringBuffer code) {
	code.append(ptrName).append(" = " + ALLOC + "(sizeof(");
	code.append(makeElementType()).append(") * ");

	/* we don't need the original array size.

	Xtype t = getType();
	while (t.getKind() == Xtype.ARRAY) {
	    code += t.getArrayDim() + ", ";
	    t = t.getRef();
	}
	*/

	int k = 0;
	int dim = dimension;
	while (k < dim) {
	    String size = translater.toString(sizeList[k]);
	    if (getStride(k) > 0)
		code.append("(").append(size).append(")/")
		    .append(getStride(k) + 1);
	    else
		code.append(size);

	    if (++k < dim)
		code.append(" * ");
	}

	code.append(");\n");
    }

    protected String makeFree() {
	return FREE + "(" + getPointerName() + ");\n";
    }

    // The following three methods return the names of temporary variables.

    public String getOffsetName(int dim) {
	return varName + "_offset" + dim;
    }

    public String getCacheSize(int dim) {
	return varName + "_size" + dim;
    }

    /* Returns the name of pointer to an on-chip memory area.
     */
    public String getPointerName() {
        // System.out.println("<DEBUG> getPointerName: " + varName + "_ptr");
	return varName + "_ptr";
    }

    /* Returns the type name of pointer to an array element.
     */
    public String makePointerType() {
	return makeElementType() + "*";
    }

    public String makeElementType() {
	Xtype t = varType;
	while (t.getKind() == Xtype.ARRAY)
	    t = t.getRef();

	String tname = t.toString();
	if (tname.equals("double") || tname.equals("int"))
	    return tname;
	else
	    return "struct _S0";	// the type should be complex.
					// this is very ad hoc!
    }

    /* Gets the C-language name of this array.
     */
    public String getArrayName() {
	Xobject decl = translater.findCommonVariable(varName).getValue();

	if (translater.findCommonVariable(varName).getStorageClass() == 
	    StorageClass.FCOMM){
	  // System.out.println("DEBUG " + decl);

	  Xobject commonName = decl.getArg(1).getArg(0);
	  Xobject memberName = decl.getArg(1).getArg(1);
	  // System.out.println("DEBUG " + commonName + memberName);
	  return commonName.getSym() + "._" + memberName.getInt()
	    + "." + varName;
	}
	else{
	  return varName;
	}
    }

    /*
     * Translates an expression for accessing an array on an on-chip cache.
     *
     * Example: supose that the original expression is:
     *
     *   x(i,j)
     *
     * then, this method translates it into:
     *
     *   x_ptr(i - x_offset0)(j - x_offset1)
     *
     * or
     *
     *   x_ptr((i - x_offset0)/<stride_width0>,
     *         (j - x_offset1)/<stride_width1>)
     *
     * Note: a Fortran array expression x(i,j) is represented by
     * x[j,i] in C.  (i and j are exchanged.)
     */
    public void process(Xobject node, TopdownIterator iterator) {
	Xobject xobj;

	if (invalid || !node.getSym().equals(varName)) {
	    if (next != null)
		next.process(node, iterator);

	    return;
	}

	int dim = dimension;
	boolean first = true;
	Xobject[] index = new Xobject[dim];

	// System.out.println("DEBUG " + dim);

	while (--dim >= 0) {
	    iterator.moveUp();

	    // System.out.println("<DEBUG> ----------------------------- ");
	    if (first) {
		first = false;
		if (node.Opcode() == Xcode.IDENT){ 
		  Xobject p = iterator.getParent();
		  // System.out.println(p);
		  if (p.Opcode() != Xcode.PLUS_EXPR) {
		    processArrayName(iterator);
		    return;
		  }
		} else if (node.Opcode() == Xcode.FPARAM_ARRAY_ADDR){ 
		  Xobject p = iterator.get();
		  if (p.Opcode() != Xcode.PLUS_EXPR) {
		    processArrayName(iterator);
		    return;
		  }
		}
	    }

	    if (node.Opcode() == Xcode.IDENT)
	      iterator.moveUp();

	    xobj = new XobjString(Xcode.VAR, getOffsetName(dim));
	    xobj = new XobjList(Xcode.MINUS_EXPR,
				translater.process(iterator.get().getArg(1)),
				xobj);
	    index[dim] = xobj;
	}

	dim = dimension;
	xobj = index[dim - 1];
	for (int i = dim - 2; i >= 0; --i) {
	    xobj = new XobjList(Xcode.MUL_EXPR, xobj,
			new XobjString(Xcode.VAR, getCacheSize(i)));
	    xobj = new XobjList(Xcode.PLUS_EXPR, xobj, index[i]);

	    // if the stride width is more than zero, then the index is
	    // adjusted.

	    if (strideList[i] > 0)
		xobj = new XobjList(Xcode.DIV_EXPR, xobj,
				    Xcons.IntConstant(strideList[i] + 1));
	}

	xobj = new XobjList(Xcode.PLUS_EXPR,
			    new XobjString(Xcode.VAR, getPointerName()),
			    xobj);
	iterator.substitute(xobj);
	iterator.skipSubtree();
    }

    /*
     * Translates an array name without indexes, which may appear as
     * a subroutine parameter.  It is equivalent to <array name>(1,1,..).
     */
    protected void processArrayName(TopdownIterator iterator) {
	iterator.substitute(new XobjString(Xcode.VAR, getPointerName()));
	iterator.skipSubtree();
    }

//  protected static String getSymbol(Xobject xobj) {
//      return ((XobjString)xobj.getArg(0)).getSym();
//  }

  protected Xobject genIdList() {
    Ident id;

    Xobject id_list = Xcons.emptyIdList();

    Xtype t = Xtype.Pointer(varType.getArrayElementType());
    // Xtype t = new PointerType(100, varType.getRef(),
    //      Machine.pointerSize, Machine.pointerAlign, false, false);
    id = Ident.Local(getPointerName(), t);
    id.Declared();
    id_list.add(id);
    // System.out.println(id);

    for (int i = 0; i < dimension; ++i){
      id = Ident.Local(getOffsetName(i), BasicType.intType);
      id.Declared();
      id_list.add(id);
      // System.out.println(id);
    }

    for (int i = 0; i < dimension; ++i){
      id = Ident.Local(getCacheSize(i), BasicType.intType);
      id.Declared();
      id_list.add(id);
      // System.out.println(id);
    }

    return id_list;
  }


  protected Xobject genDecl() {
    XobjList dec_list = Xcons.List();
    XobjectDef def;
    Xobject val;

    def = XobjectDef.Var(getPointerName(), null);
    // System.out.println(def.getDef());
    dec_list.add(def.getDef());

    for (int i = 0; i < dimension; ++i){
      val = Xcons.IntConstant(0);
      def = XobjectDef.Var(getOffsetName(i), val);
      // System.out.println(def.getDef());
      dec_list.add(def.getDef());
    }

    for (int i = 0; i < dimension; ++i){
      def = XobjectDef.Var(getCacheSize(i), sizeList[i]);
      // System.out.println(def.getDef());
      dec_list.add(def.getDef());
    }

    return dec_list;
  }

  protected Xobject genAlloc() {
    Ident id;
    Xobject obj, tmp, lv, rv;

    id = (Ident) IdList.getArg(0);

    // tmp = Xcons.Symbol(Xcode.LVAR_ADDR, id.Type(), id.getName());
    lv = Xcons.List(Xcode.POINTER_REF, id.Type(), id.getValue());

    tmp = Xcons.binaryOp(Xcode.MUL_EXPR,
	  Xcons.IntConstant(varType.getArrayElementType().getSize()),
	  sizeList[0]);

    // hoge - (be sure to work well in the case of dimension > 1)
    for (int i = 1; i < dimension; ++i){
      tmp = Xcons.binaryOp(Xcode.MUL_EXPR, tmp,sizeList[i]);
    }

    XobjList arg = Xcons.List(tmp);
    tmp = get_scmalloc_func();
    rv = Xcons.Cast(id.Type(), Xcons.functionCall(tmp, arg));

    obj = Xcons.List(Xcode.EXPR_STATEMENT, Xcons.Set(lv, rv));
    // System.out.println(obj);
    
    return obj;
  }


  protected Xobject genFree() {
    Ident id;
    Xtype t;
    Xobject obj, tmp, arg;

    id = (Ident) IdList.getArg(0);
    arg = Xcons.List(Xcons.List(Xcode.POINTER_REF, id.Type(), id.getValue()));

    t = new FunctionType(BasicType.voidType);
    t = Xtype.Pointer(t);

    tmp = Xcons.Symbol(Xcode.FUNC_ADDR, t, FREE);
    tmp = Xcons.functionCall(tmp, arg);

    obj = Xcons.List(Xcode.EXPR_STATEMENT, tmp);
    // System.out.println(obj);
    
    return obj;
  }

  protected Xobject get_scmalloc_func() {
    Xtype t;

    t = Xtype.Pointer(BasicType.voidType);
    t = new FunctionType(t);
    t = Xtype.Pointer(t);

    /* 
    t = new PointerType(100, BasicType.voidType, Machine.pointerSize,
			Machine.pointerAlign, false, false);
    t = new FunctionType(101, t, null, 0, (byte)0, false, false, false);
    t = new PointerType(102, t, Machine.pointerSize,
			Machine.pointerAlign, false, false);
			*/

    return Xcons.Symbol(Xcode.FUNC_ADDR, t, ALLOC);
  }

  public Xobject getIdList(){ return IdList; }

  public XobjList genStrideList(Xobject params) {
    XobjList list = new XobjList();
    for (int i = 0; i < dimension; i++) {
      list.add(params.getArg(i + dimension));
    }
    return list;
  }

  public Xobject getStride2(int dim) { return StrideList.getArg(dim); }
  public Xobject getStrideList() { return StrideList; }

  public Xobject getArrayObj() {
    Xobject decl = translater.findCommonVariable(varName).getValue();
    return decl;
  }


}

