/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import exc.object.*;
import exc.block.*;

public class ScmLoadStore {
  protected boolean is_load;
  protected ScmBlock block;

  protected int dimension;
  protected Xtype eType;

  protected Xobject Xoffset[];
  protected Xobject Xsize[];
  protected Ident Xptr1, Xptr2;
  protected Ident Xitr[];

  protected Xobject OriginList[];
  protected Xobject SizeList[];
  protected Xobject StrideList[];

  public ScmLoadStore(String func, Xobject xobj, ScmBlock block) {
    if (func.equals(Names.LOAD)) is_load = true;
    else is_load = false;

    this.block = block;
    dimension = block.getDimension();
    eType = block.getType().getArrayElementType();

    Xoffset = new Xobject[dimension];
    Xsize   = new Xobject[dimension];

    XobjArgs a = block.getIdList().getArgs().nextArgs();
    for (int i = 0; i < dimension; i++){
      Ident id = (Ident) a.getArg();
      Xoffset[i] = Xcons.List(Xcode.POINTER_REF, id.Type(), id.getValue());
      a = a.nextArgs();
    }

    for (int i = 0; i < dimension; i++){
      Ident id = (Ident) a.getArg();
      Xsize[i] = Xcons.List(Xcode.POINTER_REF, id.Type(), id.getValue());
      a = a.nextArgs();
      //System.out.println("id.getValue=" + id.getValue());
    }

    OriginList = new Xobject[dimension];
    SizeList   = new Xobject[dimension];
    StrideList = new Xobject[dimension];

    if (is_load){
      for (int i = 0; i < dimension; i++){
	OriginList[i] = xobj.getArg(2).getArg(i);
	SizeList[i]   = xobj.getArg(2).getArg(dimension + i);
	StrideList[i] = block.getStride2(i);
      }
    }
    else {
      for (int i = 0; i < dimension; i++){
	StrideList[i] = block.getStride2(i);
      }
    }
  }

  public Xobject genLoadStore() {
    Xobject id_list = genLocalId();
    XobjList body = Xcons.emptyList();
    if (is_load) body.add(genInitObj());

    Ident id = (Ident) block.getIdList().getArg(0);
    Xobject rv = Xcons.List(Xcode.POINTER_REF, id.Type(), id.getValue());
    Xobject p1 = Xcons.List(Xcode.POINTER_REF, Xptr1.Type(), Xptr1.getValue());
    Xobject p2 = Xcons.List(Xcode.POINTER_REF, Xptr2.Type(), Xptr2.getValue());

    body.add(Xcons.List(Xcode.EXPR_STATEMENT, Xcons.Set(p2, rv)));

    rv = block.getArrayObj();

    Xobject tmp;
    for (int i = dimension-1; i > 0; i--){
      tmp = Xcons.List(Xcode.POINTER_REF, Xitr[i].Type(), Xitr[i].getValue());
      tmp = Xcons.binaryOp(Xcode.PLUS_EXPR, Xoffset[i], tmp);
      rv  = Xcons.binaryOp(Xcode.PLUS_EXPR, rv, tmp);
      rv  = Xcons.List(Xcode.POINTER_REF, rv.Type(), rv);
    }
    rv = Xcons.binaryOp(Xcode.PLUS_EXPR, rv, Xoffset[0]);
    rv = Xcons.Set(p1, rv);

    Block b = Bcons.Statement(rv);


  if (is_load) { rv = genFunc(Names.LOAD, p1, p2, eType.getSize(), 
     				Xsize[0], StrideList[0]); }
     else         { rv = genFunc(Names.STORE, p2, p1, eType.getSize(), 
     			Xsize[0], StrideList[0]); }
     b.add(rv);

    tmp = Xcons.binaryOp(Xcode.DIV_EXPR, Xsize[0],
			 Xcons.binaryOp(Xcode.PLUS_EXPR,
					StrideList[0],
					Xcons.IntConstant(1)));
    rv = Xcons.asgOp(Xcode.ASG_PLUS_EXPR, p2, tmp);
    b.add(rv);

    Xobject lv;
    for (int i = 1; i < dimension; i++){
	lv = Xcons.List(Xcode.POINTER_REF, Xitr[i].Type(), Xitr[i].getValue());
	Xobject init = Xcons.Set(lv, Xcons.IntConstant(0));
	Xobject cond = Xcons.binaryOp(Xcode.LOG_LT_EXPR, lv, Xsize[i]);
	Xobject iter = Xcons.asgOp(Xcode.ASG_PLUS_EXPR,
				   lv, Xcons.binaryOp(Xcode.PLUS_EXPR,
						      StrideList[i],
						      Xcons.IntConstant(1)));
	b = Bcons.FOR(init, cond, iter, b);
    }

    /* this conversion seems funny... */
    BlockList bl = new BlockList(b);

    body.add(bl.toXobject());
    // System.out.println("----- <DEBUG> -----\n" + body);

    return new XobjList(Xcode.COMPOUND_STATEMENT, id_list, null, body);
  }

  protected Xobject genLocalId() {
    Xobject id_list = Xcons.emptyIdList();
    Xtype t = Xtype.Pointer(eType);

    Xptr1 = Ident.Local("_ptr", t);
    Xptr2 = Ident.Local("_ptr2", t);
    id_list.add(Xptr1);
    id_list.add(Xptr2);

    Xitr = new Ident[dimension+1];
    Xitr[0] = null;

    for (int i = 1; i < dimension; i++){
      Xitr[i] = Ident.Local(getIteratorName(i), BasicType.intType);
      id_list.add(Xitr[i]);
    }

    return id_list;
  }

  protected XobjList genInitObj() {
    XobjList list = Xcons.List();

    for (int i = 0; i < dimension; i++){
      Xobject lv = Xoffset[i];
      Xobject rv = Xcons.binaryOp(Xcode.MINUS_EXPR, 
				  OriginList[i], Xcons.IntConstant(1));
      list.add(Xcons.List(Xcode.EXPR_STATEMENT, Xcons.Set(lv, rv)));
    }

    for (int i = 0; i < dimension; i++){
      Xobject lv = Xsize[i];
      Xobject rv = SizeList[i];
      list.add(Xcons.List(Xcode.EXPR_STATEMENT, Xcons.Set(lv, rv)));
    }

    return list;
  }

  protected String getIteratorName(int i) { return new String("_i" + i); }

  protected Xobject genFunc(String func, Xobject p1, Xobject p2, int esize, 
			    Xobject size, Xobject stride) {
    XobjList arg = Xcons.List(p1, p2);
    arg.add(Xcons.IntConstant(esize));
    arg.add(size);
    arg.add(stride);

    Xobject obj = Xcons.functionCall(get_FuncObj(func), arg);

    return obj;
  }

  protected Xobject get_FuncObj(String func) {
    Xtype t = new FunctionType(BasicType.voidType);
    return Xcons.Symbol(Xcode.FUNC_ADDR, Xtype.Pointer(t), func);
  }

}
