/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import java.util.*;
import java.io.*;

public class ScmParam{

  public static final String INTERNAL_CLOCK       = "InternalClock";
  public static final String EXTERNAL_CLOCK       = "ExternalClock";
  public static final String MOMERY_SIZE          = "MemorySize";
  public static final String ON_CHIP_MEM_SIZE     = "OnChipMemSize";
  public static final String BUS_BAND_WIDTH       = "BusBandWidth";
  public static final String NUM_INT_REGS         = "NumIntRegs";
  public static final String NUM_FP_REGS          = "NumFpRegs";
  public static final String TLB_SIZE             = "TlbSize";
  public static final String TLB_MISS_PENALTY     = "TlbMissPenalty";
  public static final String OFF_CHIP_MEM_LATENCY = "OffChipMemLatency";
  public static final String STRIDE_LATENCY       = "StrideLatency";
  public static final String ON_CHIP_PAGE_SIZE    = "OnChipPageSize";
  public static final String CACHE_L1D_SIZE       = "CacheL1DSize";
  public static final String CACHE_L1D_LINE       = "CacheL1DLine";
  public static final String CACHE_L1D_ASSOC      = "CacheL1DAssoc";
  public static final String STORE_BUF_SIZE       = "StoreBufSize";
  public static final String STORE_BUF_FLUSH      = "StoreBufFlush";
  public static final String FETCH_SIZE           = "FetchSize";
  public static final String INT_RS_SIZE          = "Int_RSSize";
  public static final String FP_RS_SIZE           = "FP_RSSize";
  public static final String LS_RS_SIZE           = "LS_RSSize";
  public static final String PLS_QUEUE_SIZE       = "PLS_QueueSize";
  public static final String EXE_INT_ASMD         = "ExeInt_ASMD";
  public static final String EXE_FP_AMC           = "ExeFP_AMC";
  public static final String EXE_FP_DS            = "ExeFP_DS";
  public static final String EXE_LS_TOTAL         = "ExeLS_Total";
  
  protected Hashtable m_params = null;
  
  //-------------------------------------------- 
  // Constrator
  public ScmParam(){
  
  }
  public ScmParam(String filename){
    loadParams(filename);
  }
  //--------------------------------------------
  private void initParams(){
    m_params = new Hashtable();
    
    m_params.put(OFF_CHIP_MEM_LATENCY, new Integer(Config.offChipMemLatency));
    m_params.put(BUS_BAND_WIDTH      , new Integer(Config.busBandWidth));
    m_params.put(ON_CHIP_MEM_SIZE    , new Integer(Config.onChipMemSize));
    m_params.put(ON_CHIP_PAGE_SIZE   , new Integer(Config.onChipPageSize));
    m_params.put(STRIDE_LATENCY      , new Integer(Config.strideLatency));
  }

  public int getParam(String key){
    if(m_params == null) initParams();
    Object value = m_params.get(key);
    if(value == null) return -1;
    if(!(value instanceof Integer)) return -1;
    return ((Integer)value).intValue();
  }
  
  public void loadParams(String filename){
    try{
      InputStream is = new FileInputStream(filename);
      loadParams(is);
    }catch(IOException e){
      e.printStackTrace();
      System.exit(1);
    }
  }
  
  public void loadParams(InputStream is){
    initParams();
    
    try{
      BufferedReader reader = new BufferedReader(new InputStreamReader(is));
      
      String line;
      
      while((line = reader.readLine()) != null){
	line = line.trim();
	int si = line.indexOf('#');

	if(si == 0) 
	  continue;
	else if(si >= 0){
	  line = line.substring(0,si);
	}
	
	int space = line.indexOf(' ');
	if(space < 0) continue;
	
	String name = line.substring(0, space);
	String sVal = line.substring(space).trim();
	
	if((name.length() < 1)||(sVal.length() < 1)) continue;
	
	int radix;
	if(sVal.startsWith("0x")){
	  sVal = sVal.substring(2);
	  radix = 16;
	}else if(sVal.startsWith("0")){
	  radix = 8;
	}else{
	  radix = 10;
	}
	
	Integer iVal = null;
	try{
	  iVal = Integer.valueOf(sVal, radix);
	}catch(NumberFormatException e){
	  continue;
	}
	
	m_params.put(name, iVal);
	//	System.out.println(name + " : " + iVal);
      }
    }catch(IOException e){
      e.printStackTrace();
      System.exit(1);
    }
  }
  

  //-------------------------------------------
  // static method

  protected static ScmParam g_instance = null;
  
  public static void loadDefaultParams(String filename){
    g_instance = new ScmParam();
    g_instance.loadParams(filename);
  }
  
  //
  public static int getDefaultParam(String key){
    if(g_instance == null){
      g_instance = new ScmParam();
    }
    return g_instance.getParam(key);
    
  }

  //------------------------------------------
  // main method for debug
  /*
  public static void main(String[] args){
    //ScmParam.loadStaticParams("param_file");
    System.out.println(ScmParam.getStaticParam(ScmParam.OFF_CHIP_MEM_LATENCY));
    System.out.println(ScmParam.getStaticParam(ScmParam.ON_CHIP_MEM_SIZE));
  }
  */
}
