/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import java.util.Stack;
import exc.object.*;

/*
 * Pre-order traversal
 */
public class TopdownIterator extends XobjIterator {
    protected Xobject rootXobject;
    protected XobjArgs currentNode;	// null if at the root
    protected boolean finish;
    protected boolean getIntoSubtree;

    protected Stack nodeStack;

    public TopdownIterator(Xobject x) {
	init(x);
    }

    public void init() {
	init(rootXobject);
    }

    public void init(Xobject x) {
	rootXobject = x;
	nodeStack = new Stack();
	currentNode = null;
	finish = false;
	getIntoSubtree = true;
    }

    private void init2(Xobject x) {
	boolean f = finish;
	init(x);
	finish = f;
    }

    public Xobject getRoot() { return rootXobject; }

    public Xobject getParent() {
	if(nodeStack.empty())
	    if (currentNode == null)
		return null;
	    else
		return rootXobject;
	else
	    return (Xobject)((XobjArgs)nodeStack.peek()).getArg();
    }

    public Xobject get() {
	if (currentNode == null)
	    return rootXobject;
	else
	    return currentNode.getArg();
    }

    public boolean end() { return finish; }

    public void next() {
	if (finish)
	    return;

	Xobject cur = get();
	if (cur != null
	    && cur instanceof XobjList
	    && cur.Opcode() != Xcode.ID_LIST
	    && cur.getArgs() != null
	    && getIntoSubtree)
	{
	    // go down
	    if (currentNode != null)
		nodeStack.push(currentNode);

	    currentNode = cur.getArgs();
	}
	else {
	    // go to the next sibling (or the parent's sibling)
	    if (currentNode == null) {
		finish = true;
		return;
	    }

	    getIntoSubtree = true;
	    currentNode = currentNode.nextArgs();
	    while (currentNode == null)
		if (nodeStack.empty()) {
		    finish = true;
		    return;
		}
		else
		    currentNode = ((XobjArgs)nodeStack.pop()).nextArgs();
	}
    }

    public void skipSubtree() {
	getIntoSubtree = false;
    }

    public void moveUp() {
	if (!nodeStack.empty())
	    currentNode = (XobjArgs)nodeStack.pop();
    }

    public void substitute(Xobject newObj) {
	if(currentNode == null)
	    init2(newObj);
	else
	    currentNode.setArg(newObj);
    }

    public void remove() {
	skipSubtree();
	Xobject p = getParent();
	if (p == null)
	    init2(null);
	else {
	    /* currentNode points to the removed object
	     * even after that object is detached from the tree.
	     */
	    XobjArgs nodes = p.getArgs();
	    XobjArgs cur = currentNode;
	    if (nodes == cur)
		p.setArgs(cur.nextArgs());
	    else {
		while (nodes != null && nodes.nextArgs() != cur)
		    nodes = nodes.nextArgs();

		if (nodes != null)
		    nodes.setNext(cur.nextArgs());
	    }
	}
    }
}
