/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import scmopt.*;
import exc.object.*;
import exc.block.*;
import java.io.StringWriter;

public class Translater implements XobjectDefVisitor, Names {
  private boolean errorFlag = false;
  private XobjectFile env;
  private Xobject fparamIdList = null;
  private Xobject fbodyIdList = null;

  // list of begin and pbegin pragmas
  private ScmBlock blockPragmas = null;

  public void errorMessage(Xobject xobj, String msg) {
    LineNo ln = xobj.getLineNo();
    System.err.println(ln.fileName() + " " + ln.lineNo() + ": " + msg);
    errorFlag = true;
  }

  public boolean error() {
    return errorFlag;
  }

  public XobjectFile getEnv() { return env; }

  public void init(XobjectFile env){
    this.env = env;
  }

  public void doDef(XobjectDef d) {
    if (!d.isFuncDef())
      return;

    Xobject fbody = d.getDef().getArg(3);
    if (fbody.Opcode() != Xcode.COMPOUND_STATEMENT)
      return;

    fparamIdList = d.getDef().getArg(1);	// function parameters
    fbodyIdList = fbody.getArg(0);		// local variables
    Xobject statements = fbody.getArg(2);
    fbody.setArg(2, process(statements));
  }

  protected Xobject process(Xobject statement) {
    TopdownIterator iterator = new TopdownIterator(statement);
    for (iterator.init(); !iterator.end(); iterator.next()){
      process2(iterator);
    }

    return iterator.getRoot();
  }

  protected void process2(TopdownIterator iterator) {
    Xobject s = iterator.get();
    if (s == null)
      return;
    int op = s.Opcode();

    if (op == Xcode.SCM_PRAGMA)
      try {
      String name = getPragmaName(s);
      if (name.equals("uncache_"))
	processUncache(iterator, s);
      else if (name.equals("begin_"))
	processBegin(iterator, s);
      else if (name.equals("escape_"))
	processEscape(iterator, s);
      else if (name.equals("load_"))
	processLoad(iterator, s);
      else if (name.equals("store_"))
	processStore(iterator, s);
      else if (name.equals("pbegin_"))
	processPbegin(iterator, s);
      else if (name.equals("pload_"))
	processPload(iterator, s);
      else if (name.equals("pstore_"))
	processPstore(iterator, s);
      else if (name.equals("swap_"))
	processSwap(iterator, s);
      else if (name.equals("copyto_"))
	processCopyTo(iterator, s);
      else if (name.equals("way_set_"))
	processWaySet(iterator, s);
      else if (name.equals("opt_notreusable_"))
	processOptNotReusable(iterator, s);
      else if (name.equals("opt_reusable_"))
	processOptReusable(iterator, s);
      else if (name.equals("element_"))
	System.out.println("element");
      else
	iterator.remove();
    }
    catch (SyntaxError e) {}
    else if (blockPragmas != null && op == Xcode.IDENT)
      blockPragmas.process(s, iterator);
    else if (blockPragmas != null && op == Xcode.FPARAM_ARRAY_ADDR)
      blockPragmas.process(s, iterator);
  }

  protected void parameterCheck(TopdownIterator iterator,
				Xobject xobj, int num)
       throws SyntaxError
  {
    for (int i = 1; i <= num; ++i)
      if (xobj.getArg(i) == null) {
	System.out.println("syntax error from Translator.java");
	errorMessage(xobj, "syntax error");
	iterator.skipSubtree();
	throw new SyntaxError();
      }
  }

  /*
  protected ScmPBlock toScmPBlock(ScmBlock block, TopdownIterator iterator,
				  Xobject xobj)
       throws SyntaxError
  {
    try {
      return (ScmPBlock)block;
    }
    catch (ClassCastException e) {
      errorMessage(xobj, "pload/pstore/swap not following pbegin");
      iterator.skipSubtree();
      throw new SyntaxError();
    }
  }
  */

  public static String getPragmaName(Xobject xobj) {
    return ((XobjString)xobj.getArg(0)).getSym();
  }

  public void processUncache(TopdownIterator iterator, Xobject xobj)
  {
    String name = xobj.getArg(1).getSym() + "COMMON";
    System.out.println("*** uncache: " + name);
    System.out.println(env.findIdent(name));
    iterator.remove();
  }

  public void processBegin(TopdownIterator iterator, Xobject s)
       throws SyntaxError
  {
    processBegin(false, iterator, s);
  }

  public void processPbegin(TopdownIterator iterator, Xobject s)
       throws SyntaxError
  {
    processBegin(true, iterator, s);
  }

  public void processBegin(boolean isPbegin,
			   TopdownIterator iterator, Xobject s)
       throws SyntaxError
  {
    parameterCheck(iterator, s, 2);
    ScmBlock b;
    if (isPbegin){
      fatal("<pbegin> not implemented yet.");
      b = new ScmPBlock(s.getArg(1), s.getArg(2), this, blockPragmas);
    }
    else
      b = new ScmBlock(s.getArg(1), s.getArg(2), this, blockPragmas);

    ScmBlock old = blockPragmas;
    blockPragmas = b;
    Xobject xobj = process(s.getArg(3));

    iterator.substitute(b.makeBlock(xobj));
    iterator.skipSubtree();
    blockPragmas = old;
  }

  public void processEscape(TopdownIterator iterator, Xobject xobj)
       throws SyntaxError
  {
    parameterCheck(iterator, xobj, 1);
    String name = xobj.getArg(1).getSym();
    Xobject body = xobj.getArg(3);
    if (blockPragmas != null)
      if (blockPragmas.invalidate(name, true)) {
	body = process(body);
	blockPragmas.invalidate(name, false);
      }

    iterator.substitute(body);
    iterator.skipSubtree();
  }

  public void processLoad(TopdownIterator iterator,
			  Xobject xobj) throws SyntaxError
  {
    parameterCheck(iterator, xobj, 2);
    String name = xobj.getArg(1).getSym();
    ScmBlock block = blockPragmas.lookup(name);
    processLoad2(iterator, xobj, block);
  }

  public void processPload(TopdownIterator iterator,
			   Xobject xobj) throws SyntaxError
  {

    // parameterCheck(iterator, xobj, 2);
    // String name = xobj.getArg(1).getSym();
    // ScmBlock block = blockPragmas.lookup(name);
    // ScmPBlock pblock = toScmPBlock(block, iterator, xobj);
    // processLoad2(iterator, xobj, block, new BackgroundArea(pblock));
    fatal("<pload> not implemented yet.");
  }

  public void processLoad2(TopdownIterator iterator, Xobject xobj,
			   ScmBlock block)
  {
    Xobject params = xobj.getArg(2);

    ScmLoadStore ls = new ScmLoadStore(LOAD, xobj, block);
    Xobject st = ls.genLoadStore();
    iterator.substitute(st);
    iterator.skipSubtree();
  }

  public void processStore(TopdownIterator iterator, Xobject xobj)
       throws SyntaxError
  {
    parameterCheck(iterator, xobj, 1);
    String name = xobj.getArg(1).getSym();
    ScmBlock block = blockPragmas.lookup(name);
    processStore2(iterator, xobj, block);
  }

  public void processPstore(TopdownIterator iterator,
			    Xobject xobj) throws SyntaxError
  {
    // parameterCheck(iterator, xobj, 1);
    // String name = xobj.getArg(1).getSym();
    // ScmBlock block = blockPragmas.lookup(name);
    // ScmPBlock pblock = toScmPBlock(block, iterator, xobj);
    // processStore2(iterator, xobj, block, new BackgroundArea(pblock));
    fatal("<pstore> not implemented yet.");
  }

  public void processStore2(TopdownIterator iterator, Xobject xobj,
			    ScmBlock block)
  {
    ScmLoadStore ls = new ScmLoadStore(STORE, xobj, block);
    Xobject st = ls.genLoadStore();
    iterator.substitute(st);
    iterator.skipSubtree();
  }

  public void processWaySet(TopdownIterator iterator, Xobject xobj) throws SyntaxError{
    //parameterCheck(iterator, xobj, 2);
    Xobject XCacheWay = iterator.get().getArg(1).getArg(0);
    Xobject XScmWay   = iterator.get().getArg(1).getArg(1);
    ScmWaySet ws = new ScmWaySet(XCacheWay, XScmWay);
    Xobject st = ws.genWaySet();
    iterator.substitute(st);
    iterator.skipSubtree();
  }

  public void processSwap(TopdownIterator iterator,
			  Xobject xobj) throws SyntaxError
  {
    // parameterCheck(iterator, xobj, 1);
    // String name = xobj.getArg(1).getSym();
    // iterator.substitute(st);
    // iterator.skipSubtree()
    fatal("<swap> not implemented yet.");
    }

  public void processCopyTo(TopdownIterator iterator, Xobject xobj)
       throws SyntaxError
  {
    // parameterCheck(iterator, xobj, 2);
    // iterator.substitute(st);
    // iterator.skipSubtree();
    fatal("<copyto> not implemented yet.");
  }

  public String toString(Xobject xobj) {
    StringWriter sw = new StringWriter();
    new ScmDecompileWriter(sw, getEnv()).print(xobj);
    return sw.toString();
  }

  public Ident findCommonVariable(String name) {
    Ident id = findCommonVariable(fparamIdList, name);
    if (id != null)
      return id;
    else
      return findCommonVariable(fbodyIdList, name);
  }

  private Ident findCommonVariable(Xobject id_list, String name) {
    if (id_list == null || name == null)
      return null;

    for (XobjArgs a = id_list.getArgs(); a != null; a = a.nextArgs()) {
      Ident id = (Ident)a.getArg();
      if (id.getStorageClass() == StorageClass.TAGNAME)
	continue;

      if (name.equals(id.getName()))
	return id;
    }

    return null;
  }

  static void fatal(String msg){
    System.err.println("Fatal <SCM Translator>: " + msg);
    Thread.dumpStack();
    System.exit(1);
  }

  public void processOptNotReusable(TopdownIterator iterator,
				    Xobject xobj) throws SyntaxError
  {
    OptNotReusable opt = new OptNotReusable(this, xobj);
    Xobject st = opt.genOpt();
    st = process(st);
    iterator.substitute(st);
    // iterator.skipSubtree();
  }

  public void processOptReusable(TopdownIterator iterator,
				 Xobject xobj) throws SyntaxError
  {
    OptReusable opt = new OptReusable(this, xobj);
    Xobject st = opt.genOpt2();
    st = process(st);
    iterator.substitute(st);
    // iterator.skipSubtree();
  }

  public void addIdList(Ident id) {
      fbodyIdList.add(id);
  }

  public Xobject getIdList() {
    return fbodyIdList;
  }

}
