/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import exc.object.*;
import scmopt.*;
import java.io.StringWriter;

abstract class OnchipArea {
    abstract String getPointerName();
    abstract String getOffsetName(int i);
    abstract String getCacheSize(int i);
}

class ForegroundArea extends OnchipArea {
    ScmBlockOld block;

    ForegroundArea(ScmBlockOld b) { block = b; }
    String getPointerName() { return block.getPointerName(); }
    String getOffsetName(int i) { return block.getOffsetName(i); }
    String getCacheSize(int i) { return block.getCacheSize(i); }
}

class BackgroundArea extends OnchipArea {
    ScmPBlockOld block;

    BackgroundArea(ScmPBlockOld b) { block = b; }
    String getPointerName() { return block.getBackPointerName(); }
    String getOffsetName(int i) { return block.getBackOffsetName(i); }
    String getCacheSize(int i) { return block.getBackCacheSize(i); }
}

//
public class TranslaterOld extends Translater implements XobjectDefVisitor, Names {
    private boolean errorFlag = false;
    private XobjectFile env;
    private Xobject fparamIdList = null;
    private Xobject fbodyIdList = null;

    // list of begin and pbegin pragmas
    private ScmBlockOld blockPragmas = null;

    public void errorMessage(Xobject xobj, String msg) {
	LineNo ln = xobj.getLineNo();
	System.err.println(ln.fileName() + " " + ln.lineNo() + ": " + msg);
	errorFlag = true;
    }

    public boolean error() {
	return errorFlag;
    }

    public XobjectFile getEnv() { return env; }

    public void init(XobjectFile env){
	this.env = env;
    }

    public void doDef(XobjectDef d) {
	if (!d.isFuncDef())
	    return;

	Xobject fbody = d.getDef().getArg(3);
	if (fbody.Opcode() != Xcode.COMPOUND_STATEMENT)
	    return;

	fparamIdList = d.getDef().getArg(1);	// function parameters
	fbodyIdList = fbody.getArg(0);		// local variables
	Xobject statements = fbody.getArg(2);
	// System.out.println("DEBUG " + statements);
	fbody.setArg(2, process(statements));
    }

    protected Xobject process(Xobject statement) {
	TopdownIterator iterator = new TopdownIterator(statement);
	for (iterator.init(); !iterator.end(); iterator.next()){
	  // System.out.println("DEBUG ----------------------------------");
	    process2(iterator);
	}

	return iterator.getRoot();
    }

    protected void process2(TopdownIterator iterator) {
	Xobject s = iterator.get();
	if (s == null)
	    return;
	int op = s.Opcode();
	// System.out.println("DEBUG " + s.OpcodeName());

	if (op == Xcode.SCM_PRAGMA)
	    try {
		String name = getPragmaName(s);
		if (name.equals("uncache_"))
		    processUncache(iterator, s);
		else if (name.equals("begin_"))
		    processBegin(iterator, s);
		else if (name.equals("escape_"))
		    processEscape(iterator, s);
		else if (name.equals("load_"))
		    processLoad(iterator, s);
		else if (name.equals("store_"))
		    processStore(iterator, s);
		else if (name.equals("pbegin_"))
		    processPbegin(iterator, s);
		else if (name.equals("pload_"))
		    processPload(iterator, s);
		else if (name.equals("pstore_"))
		    processPstore(iterator, s);
		else if (name.equals("swap_"))
		    processSwap(iterator, s);
		else if (name.equals("copyto_"))
		    processCopyTo(iterator, s);
		else if (name.equals("opt_notreusable_"))
		    processOptNotReusable(iterator, s);
		else if (name.equals("opt_reusable_"))
		    processOptReusable(iterator, s);
		else if (name.equals("element_"))
		    System.out.println("element");
		else
		    iterator.remove();
	    }
	    catch (SyntaxError e) {}
	else if (blockPragmas != null && op == Xcode.IDENT)
	    blockPragmas.process(s, iterator);
	else if (blockPragmas != null && op == Xcode.FPARAM_ARRAY_ADDR)
	    blockPragmas.process(s, iterator);
    }

    protected void parameterCheck(TopdownIterator iterator,
				  Xobject xobj, int num)
	throws SyntaxError
    {
	for (int i = 1; i <= num; ++i)
	    if (xobj.getArg(i) == null) {
		errorMessage(xobj, "syntax error");
		iterator.skipSubtree();
		throw new SyntaxError();
	    }
    }

    protected ScmPBlockOld toScmPBlock(ScmBlockOld block, 
				       TopdownIterator iterator, Xobject xobj)
	throws SyntaxError
    {
	try {
	    return (ScmPBlockOld)block;
	}
	catch (ClassCastException e) {
	    errorMessage(xobj, "pload/pstore/swap not following pbegin");
	    iterator.skipSubtree();
	    throw new SyntaxError();
	}
    }

    public static String getPragmaName(Xobject xobj) {
	return ((XobjString)xobj.getArg(0)).getSym();
    }

    public void processUncache(TopdownIterator iterator,
			       Xobject xobj)
    {
	String name = xobj.getArg(1).getSym() + "COMMON";
	System.out.println("*** uncache: " + name);
	System.out.println(env.findIdent(name));
	iterator.remove();
    }

    public void processBegin(TopdownIterator iterator, Xobject s)
	throws SyntaxError
    {
	processBegin(false, iterator, s);
    }

    public void processPbegin(TopdownIterator iterator, Xobject s)
	throws SyntaxError
    {
	processBegin(true, iterator, s);
    }

    public void processBegin(boolean isPbegin,
			     TopdownIterator iterator, Xobject s)
	throws SyntaxError
    {
	parameterCheck(iterator, s, 2);
	ScmBlockOld b;
	if (isPbegin)
	    b = new ScmPBlockOld(s.getArg(1), s.getArg(2),
			     this, blockPragmas);
	else
	    b = new ScmBlockOld(s.getArg(1), s.getArg(2),
			     this, blockPragmas);

	ScmBlockOld old = blockPragmas;
	blockPragmas = b;
	// System.out.println("=======================================");
	Xobject xobj = process(s.getArg(3));
	// System.out.println("---------------------------------------");
	iterator.substitute(b.makeBlock(xobj));
	iterator.skipSubtree();
	blockPragmas = old;
    }

    public void processEscape(TopdownIterator iterator,
			      Xobject xobj)
	throws SyntaxError
    {
	parameterCheck(iterator, xobj, 1);
	String name = xobj.getArg(1).getSym();
	Xobject body = xobj.getArg(3);
	if (blockPragmas != null)
	    if (blockPragmas.invalidate(name, true)) {
		body = process(body);
		blockPragmas.invalidate(name, false);
	    }

	iterator.substitute(body);
	iterator.skipSubtree();
    }

    public void processLoad(TopdownIterator iterator,
			    Xobject xobj) throws SyntaxError
    {
	parameterCheck(iterator, xobj, 2);
	String name = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(name);
	processLoad2(iterator, xobj, block, new ForegroundArea(block));
    }

    public void processPload(TopdownIterator iterator,
			    Xobject xobj) throws SyntaxError
    {
	parameterCheck(iterator, xobj, 2);
	String name = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(name);
	ScmPBlockOld pblock = toScmPBlock(block, iterator, xobj);
	processLoad2(iterator, xobj, block, new BackgroundArea(pblock));
    }

    public void processLoad2(TopdownIterator iterator, Xobject xobj,
			     ScmBlockOld block, OnchipArea area)
    {
	int dim = block.getDimension();
	Xobject params = xobj.getArg(2);
	StringBuffer code = new StringBuffer();
	for (int i = 0; i < dim; ++i)
	    code.append(area.getOffsetName(i)).append(" = -1 + (")
		.append(toString(params.getArg(i))).append(");\n");

	for (int j = 0; j < dim; ++j)
	    code.append(area.getCacheSize(j)).append(" = ")
		.append(toString(params.getArg(j + dim))).append(";\n");

	inlineLoad(code, LOAD, block, dim, area);

	XobjString str = new XobjString(Xcode.IDENT, code.toString());
	Xobject st = new XobjList(Xcode.SCM_PRAGMA_CODE, str, null, null);
	iterator.substitute(st);
	iterator.skipSubtree();
    }

    public void processStore(TopdownIterator iterator,
			     Xobject xobj)
	throws SyntaxError
    {
	parameterCheck(iterator, xobj, 1);
	String name = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(name);
	processStore2(iterator, xobj, block, new ForegroundArea(block));
    }

    public void processPstore(TopdownIterator iterator,
			      Xobject xobj) throws SyntaxError
    {
	parameterCheck(iterator, xobj, 1);
	String name = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(name);
	ScmPBlockOld pblock = toScmPBlock(block, iterator, xobj);
	processStore2(iterator, xobj, block, new BackgroundArea(pblock));
    }

    public void processStore2(TopdownIterator iterator, Xobject xobj,
			      ScmBlockOld block, OnchipArea area)
    {
	int dim = block.getDimension();
	StringBuffer code = new StringBuffer();
	inlineLoad(code, STORE, block, dim, area);

	XobjString str = new XobjString(Xcode.IDENT, code.toString());
	Xobject st = new XobjList(Xcode.SCM_PRAGMA_CODE, str, null, null);
	iterator.substitute(st);
	iterator.skipSubtree();
    }

    public void processSwap(TopdownIterator iterator,
			    Xobject xobj) throws SyntaxError
    {
	parameterCheck(iterator, xobj, 1);
	String name = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(name);
	ScmPBlockOld pblock = toScmPBlock(block, iterator, xobj);
	StringBuffer code = new StringBuffer();

	code.append("{ ");

	code.append(block.makePointerType())
	    .append(" _tmp1;\nint _tmp2;\n");
	code.append("_tmp1 = ").append(pblock.getPointerName())
	    .append(";\n").append(pblock.getPointerName()).append(" = ")
	    .append(pblock.getBackPointerName()).append(";\n")
	    .append(pblock.getBackPointerName()).append(" = _tmp1;\n");

	int dim = block.getDimension();
	for (int i = 0; i < dim; ++i) {
	    code.append("_tmp2 = ").append(pblock.getOffsetName(i))
		.append(";\n");
	    code.append(pblock.getOffsetName(i)).append(" = ")
		.append(pblock.getBackOffsetName(i)).append(";\n");
	    code.append(pblock.getBackOffsetName(i)).append(" = _tmp2;\n");
	    code.append("_tmp2 = ").append(pblock.getCacheSize(i))
		.append(";\n");
	    code.append(pblock.getCacheSize(i)).append(" = ")
		.append(pblock.getBackCacheSize(i)).append(";\n");
	    code.append(pblock.getBackCacheSize(i)).append(" = _tmp2;\n");
	}

	code.append("}\n");

	XobjString str = new XobjString(Xcode.IDENT, code.toString());
	Xobject st = new XobjList(Xcode.SCM_PRAGMA_CODE, str, null, null);
	iterator.substitute(st);
	iterator.skipSubtree();
    }

    protected void inlineLoad(StringBuffer code, String funcName,
			      ScmBlockOld block, int dim, OnchipArea area) {
	int i, k;
	code.append("\n{ ");
	code.append(block.makePointerType()).append(" _ptr;\n");
	code.append(block.makePointerType()).append(" _ptr2 = ")
	    .append(area.getPointerName()).append(";\n");
	for (i = 1; i < dim; ++i)
	    code.append("int _i").append(i).append("; ");

	code.append('\n');
	for (i = dim - 1; i > 0; --i)
	    code.append("for (_i").append(i).append(" = 0; _i").append(i)
	        .append(" < ").append(area.getCacheSize(i))
	        .append("; _i").append(i).append(" += ")
		.append(block.getStride(i) + 1).append(")\n");

	// System.out.println("DEBUG" + code);

	code.append(" { _ptr = ");
	String offset = block.getArrayName();
	// System.out.println("DEBUG " + offset);

	for (i = dim - 1; i > 0; --i)
	    offset = "*(" + offset + " + " + area.getOffsetName(i)
		   + " + _i" + i + ")";

	code.append(offset).append(" + ")
	    .append(area.getOffsetName(i)).append(";\n");
	code.append(' ').append(funcName);
	if (funcName.equals(LOAD))
	    code.append("(_ptr, _ptr2, ");
	else // STORE
	    code.append("(_ptr2, _ptr, ");

	code.append("sizeof(").append(block.makeElementType())
	    .append("), ").append(area.getCacheSize(0));
	if (block.getStride(0) > 0)
	    code.append(" / ").append(block.getStride(0) + 1);

	code.append(", ").append(block.getStride(0)).append(");\n");
	code.append(" _ptr2 += ").append(area.getCacheSize(0));
	if (block.getStride(0) > 0)
	    code.append(" / ").append(block.getStride(0) + 1);

	code.append("; }\n");

	code.append("}\n");
    }

    public void processCopyTo(TopdownIterator iterator, Xobject xobj)
	throws SyntaxError
    {
	parameterCheck(iterator, xobj, 2);

	String srcName = xobj.getArg(1).getSym();
	ScmBlockOld block = blockPragmas.lookup(srcName);
	int dim = block.getDimension();

	Xobject params = xobj.getArg(2);

	if (params.Nargs() != dim + 1) {
	    errorMessage(params, "bad parameters");
	    iterator.skipSubtree();
	    throw new SyntaxError();
	}

	String destName = toString(params.getArg(0));
	String[] offset = new String[dim];
	for (int i = 1; i <= dim; ++i)
	    offset[i - 1] = "-1 + (" + toString(params.getArg(i)) + ")";

	StringBuffer code = new StringBuffer();

	processCopyTo2(code, STORE, block, dim, new ForegroundArea(block),
		       destName, offset);

	XobjString str = new XobjString(Xcode.IDENT, code.toString());
	Xobject st = new XobjList(Xcode.SCM_PRAGMA_CODE, str, null, null);
	iterator.substitute(st);
	iterator.skipSubtree();
    }

    protected void processCopyTo2(StringBuffer code, String funcName,
				  ScmBlockOld block, int dim, OnchipArea area,
				  String destName, String[] offsets) {
	int i, k;
	code.append("\n{ ");
	code.append(block.makePointerType()).append(" _ptr;\n");
	code.append(block.makePointerType()).append(" _ptr2 = ")
	    .append(area.getPointerName()).append(";\n");
	for (i = 1; i < dim; ++i)
	    code.append("int _i").append(i).append("; ");

	code.append('\n');
	for (i = dim - 1; i > 0; --i)
	    code.append("for (_i").append(i).append(" = 0; _i").append(i)
	        .append(" < ").append(area.getCacheSize(i))
	        .append("; _i").append(i).append(" += ")
		.append(block.getStride(i) + 1).append(")\n");

	code.append(" { _ptr = ");
	String offset = destName;
	for (i = dim - 1; i > 0; --i)
	    offset = "*(" + offset + " + " + offsets[i] + " + _i" + i + ")";

	code.append(offset).append(" + ").append(offsets[i]).append(";\n");
	code.append(' ').append(funcName);
	if (funcName.equals(LOAD))
	    code.append("(_ptr, _ptr2, ");
	else // STORE
	    code.append("(_ptr2, _ptr, ");

	code.append("sizeof(").append(block.makeElementType())
	    .append("), ").append(area.getCacheSize(0));
	if (block.getStride(0) > 0)
	    code.append(" / ").append(block.getStride(0) + 1);

	code.append(", ").append(block.getStride(0)).append(");\n");
	code.append(" _ptr2 += ").append(area.getCacheSize(0));
	if (block.getStride(0) > 0)
	    code.append(" / ").append(block.getStride(0) + 1);

	code.append("; }\n");

	code.append("}\n");
    }

    public String toString(Xobject xobj) {
	StringWriter sw = new StringWriter();
	new ScmDecompileWriter(sw, getEnv()).print(xobj);
	return sw.toString();
    }

    public Ident findCommonVariable(String name) {
	Ident id = findCommonVariable(fparamIdList, name);
	if (id != null)
	    return id;
	else
	    return findCommonVariable(fbodyIdList, name);
    }

    private Ident findCommonVariable(Xobject id_list, String name) {
	if (id_list == null || name == null)
	    return null;

	for (XobjArgs a = id_list.getArgs(); a != null; a = a.nextArgs()) {
	    Ident id = (Ident)a.getArg();
	    if (id.getStorageClass() == StorageClass.TAGNAME)
		continue;

	    if (name.equals(id.getName()))
		return id;
	}

	return null;
    }

  public void processOptNotReusable(TopdownIterator iterator,
				    Xobject xobj) throws SyntaxError
  {
    OptNotReusable opt = new OptNotReusable(this, xobj);
    Xobject st = opt.genOpt();
    st = process(st);
    iterator.substitute(st);
    // iterator.skipSubtree();
  }

  public void processOptReusable(TopdownIterator iterator,
				 Xobject xobj) throws SyntaxError
  {
    OptReusable opt = new OptReusable(this, xobj);
    Xobject st = opt.genOpt2();
    st = process(st);
    iterator.substitute(st);
    // iterator.skipSubtree();
  }
    public void addIdList(Ident id) {
      fbodyIdList.add(id);
  }

  public Xobject getIdList() {
    return fbodyIdList;
  }
/*
    public static Xobject getChild(Xobject list, int n) {
	XobjArgs args = ((XobjList)list).getArgs();
	while (n-- > 0)
	    args = args.nextArgs();

	return args.getArg();
    }
*/
}
