/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
package scm;

import java.io.*;
import exc.object.*;
import java.util.Vector;

/**
 * Compiler driver for SCIMA
 *
 * Detailed configurations are specified by scm/Config.java
 */
public class scmf77 extends exc.util.ccDriver {

  public static final String f2cHeader = Config.headerFile;
  public static final boolean isOldFashion = false;
  
  public boolean debugFlag = true;
  
  public String cpp_opts = "";
  public String cc_opts = "";
  public String ffront_opts = "";
  
  public String CPP_command = Config.cpp;
  public String FFRONT_command = Config.parser;

  public Vector compileFiles = new Vector();

  public scmf77() { super("scmf77"); }

  public scmf77(String name){ super(name); }

  public static void main(String args[]){

    scmf77 driver = new scmf77();
    driver.run(args);
  }
  public void run(String[] args) {
    scmf77 driver = new scmf77();
    driver.processArgs(args);
    driver.compileAll();
  }
  public void compileAll(){
    for(int i = 0 ; i < compileFiles.size() ; i++){
      compile((String)compileFiles.get(i));
    }
  }
  public void processArgs(String[] args){
    int i;
    /*
    for(i = 0 ; i < args.length ; i++){
      System.out.println(args[i]);
    }*/

    for(i = 0 ; i < args.length ; ){

      if(args[i].startsWith("-p")){
        ScmParam.loadDefaultParams(args[i+1]);
        i++;
      }else if(args[i].endsWith(".f") || args[i].endsWith(".F")){
        compileFiles.add(args[i]);
      }else if(args[i].endsWith(".x")){
        compileFiles.add(args[i]);
      }
        
      i++;
    }
  }
  
  public void compile(String filename) {
      
    String base = filename.substring(0, filename.lastIndexOf('.'));
    if(filename.endsWith(".F"))
      execCommand(CPP_command + " " + cpp_opts + " " + base
			+ ".F > " + base + ".f");

    if(!filename.endsWith(".x"))
      execCommand(FFRONT_command + " " + ffront_opts + " " + base
		  + ".f " + base + ".x");
    XobjectFile f = new ScmFile();
    f.debugFlag = debugFlag;
    f.addHeaderLine("#include \"" + f2cHeader + "\"");
    f.Input(base + ".x");
    process(f);
	f.Output(base + ".xx");
	f.Decompile(base + ".c");
  }

  public void process(XobjectFile f){
    if(isOldFashion) { 
      LabelRenamer lr = new LabelRenamer();
            f.iterateDef(lr);
            
	    TranslaterOld tr = new TranslaterOld(); 
	    tr.init(f);
	    f.iterateDef(tr);
	    if (tr.error()) {
	      System.err.println("error in SCIMA. stop.");
	      System.exit(1);
	    }
    }
    else {
      Translater tr = new Translater();
	    tr.init(f);
	    f.iterateDef(tr);
	    if (tr.error()) {
	      System.err.println("error in SCIMA. stop.");
	      System.exit(1);
	    }
	}
    }

    public void execCommand(String cmd) {
	int c;
	String  cmdarray[] = { Config.shell, "-c", cmd};
	try {
	    Runtime r = Runtime.getRuntime();
	    Process p = r.exec(cmdarray);
	    InputStream in = p.getErrorStream();
	    while ((c = in.read()) != -1)
		System.err.write(c);

	    System.err.flush();
	    p.waitFor();
	    if(p.exitValue() != 0){
		System.err.println("Error "+p.exitValue());
		System.exit(p.exitValue());
	    }
	}
	catch (IOException e) {
	    System.err.println("IOException: " + e);
	    System.exit(1);
	}
	catch (InterruptedException e) {
	    System.err.println("InterruptedException: " + e);
	    System.exit(1);
	}
    }
}
