package scmopt;

import java.util.Vector;
import scm.*;
import exc.object.*;
import exc.block.*;
import exc.flow.*;


public class XobjModifier{
  protected Optimizer m_optimizer;
  protected Xobject m_root;
  
  public XobjModifier(Xobject root, Optimizer opt){
    m_root = root;
    m_optimizer = opt;
  }
  public Xobject getRoot(){
    return m_root;
  }
  public Xobject loopStripMine(Xobject xFor, Xobject step){

    Block b = Bcons.buildBlock(xFor);
    if(!(b instanceof ForBlock)) return null;
    
    ForBlock forBlock = (ForBlock)b;
    forBlock.Canonicalize();

    loopStripMine lsm = new loopStripMine(m_optimizer, forBlock);
    ForBlock forBlockNew = lsm.getOptLoop(step);

    forBlockNew.Canonicalize();

    Xobject xNew = forBlockNew.toXobject();

    if(!apply(xFor, xNew)) return null;

    return xNew;
  }

  //œK@ꍇ킯
  public Xobject branchOptimize(Xobject xCond, Xobject xBody){
    Xobject xNew = Xcons.List(Xcode.IF_STATEMENT,
                              xBody,
                              xBody.copy());

    if(!apply(xBody, xNew)) return null;
    return xNew;
  }
  
  public Xobject softwarePipeline(Xobject xFor, String[] aNames, 
                                  int[] aDims, boolean[] useDoubleBuf){
    if(xFor.Opcode() != Xcode.FOR_STATEMENT) return null;
    
    Block b = Bcons.buildBlock(xFor);
    if(!(b instanceof ForBlock)) return null;
    
    ForBlock forBlock = (ForBlock)b;
    forBlock.Canonicalize();
    
    Xobject xStep = forBlock.getStep();
    Xobject xLoopVar = forBlock.getLoopVar();
    
    forBlock.setUpperBound(Xcons.binaryOp(Xcode.MINUS_EXPR,
                                          forBlock.getUpperBound(),
                                          xStep));
    

    Xobject xNewFor = forBlock.toXobject();
    Xobject xLoopBody = xNewFor.getArg(3);
    Xobject xEpilogue = xLoopBody.copy();
    Xobject xNew = Xcons.List(xNewFor, xEpilogue);
    
    
    if(!apply(xFor, xNew)) return null;
    
    ForBlock newBlock = (ForBlock)Bcons.buildBlock(xNewFor);
    newBlock.Canonicalize();
    
    int nArrays = aNames.length;
    
    IndexRange range[] = new IndexRange[nArrays];
    for(int i = 0 ; i < nArrays ; i++){
      range[i] = new IndexRange(aNames[i], aDims[i]);
      range[i].process(xLoopBody);
    }
    
    ExprEvaluater eval = new ExprEvaluater();
    eval.buildExprs(xNewFor.getArg(0));
    
    for(int i = 0 ; i < nArrays ; i++){
      Xobject[] xMin = new Xobject[aDims[i]];
      Xobject[] xDif = new Xobject[aDims[i]];
      
      
      for(int j = 0 ; j < aDims[i] ; j++){
        xMin[j] = range[i].getMin(j);
        xDif[j] = XTools.addXobjInt(range[i].getDif(j), Xcons.IntConstant(1));
      }
      
      if(!useDoubleBuf[i]){
        if(range[i].read() || range[i].written()){
          xLoopBody = insertLoad(xLoopBody, aNames[i], xMin, xDif);
          xEpilogue = insertLoad(xEpilogue, aNames[i], xMin, xDif);
        }
        if(range[i].written()){
          xLoopBody = insertStore(xLoopBody, aNames[i]);
          xEpilogue = insertStore(xEpilogue, aNames[i]);
        }
      }else{
        Xobject[] xMin1 = new Xobject[aDims[i]];
        Xobject[] xDif1 = new Xobject[aDims[i]];
        Xobject xNextLoopVar = Xcons.binaryOp(Xcode.PLUS_EXPR, xLoopVar, xStep);
        
        for(int j = 0 ; j < aDims[i] ; j++){
          xMin1[j] = xMin[j].copy();
          xMin1[j] = XTools.substituteXobj(xMin1[j], xLoopVar, forBlock.getLowerBound());
          xDif1[j] = xDif[j].copy();
          xDif1[j] = XTools.substituteXobj(xDif1[j], xLoopVar, forBlock.getLowerBound());
          xMin1[j] = eval.evalExpr(xMin1[j]);
          xDif1[j] = eval.evalExpr(xDif1[j]);
          xMin[j]  = XTools.substituteXobj(xMin[j] , xLoopVar, xNextLoopVar);
          xDif[j]  = XTools.substituteXobj(xDif[j] , xLoopVar, xNextLoopVar);
        }
        
        xNew      = insertLoad  (xNew     , aNames[i], xMin1, xDif1);
        xLoopBody = insertPLoad (xLoopBody, aNames[i], xMin , xDif );
        xLoopBody = insertSwap  (xLoopBody, aNames[i]);
        if(range[i].written()){
          xLoopBody = insertPStore(xLoopBody, aNames[i]);
          xEpilogue = insertStore (xEpilogue, aNames[i]);
        }
      }
      
      
    }
 
    return xNew;
  }
  public Xobject analyzeAndInsertLoadStore(Xobject xBody, String[] aNames, int[] aDims){
    int nArrays = aNames.length;
    
    IndexRange range[] = new IndexRange[nArrays];
    for(int i = 0 ; i < nArrays ; i++){
      range[i] = new IndexRange(aNames[i], aDims[i]);
      range[i].process(xBody);
    }
    
    for(int i = 0 ; i < nArrays ; i++){
      Xobject[] xMin = new Xobject[aDims[i]];
      Xobject[] xDif = new Xobject[aDims[i]];
      
      for(int j = 0 ; j < aDims[i] ; j++){
        xMin[j] = range[i].getMin(j);
        xDif[j] = XTools.addXobjInt(range[i].getDif(j), Xcons.IntConstant(1));
      }
      if(range[i].read() || range[i].written()){
        xBody = insertLoad(xBody, aNames[i], xMin, xDif);
      }
      if(range[i].written()){
        xBody = insertStore(xBody, aNames[i]);
      }
    }
    return xBody;
  }


  ///////////////////////////////////////////////////////////////////////////////////
  // insert pragmas
  
  public Xobject insertPLoad(Xobject xBody, String sName, Xobject[] xIndex, Xobject[] xSize){
    if(xIndex.length != xSize.length) return null;
    
    XobjList blist = new XobjList();
    
    for(int i = 0 ; i < xIndex.length ; i++){
      //FortranłIndexȂ̂łP₳Ȃ΂ȂȂB
      blist.add(XTools.addXobjInt(xIndex[i].copy(), 
                                            Xcons.IntConstant(1)));
    }
    
    for(int i = 0 ; i < xSize.length ; i++){
      blist.add(xSize[i].copy());
    }

    Xobject xLoad = Xcons.List(Xcode.SCM_PRAGMA,
                               Xcons.Symbol(Xcode.IDENT, "pload_"),
                               Xcons.Symbol(Xcode.IDENT, sName),
                               blist);
    Xobject xNew = Xcons.List(xLoad, xBody);
    
    if(!apply(xBody, xNew)) return null;
    return xNew;
  }

  
  public Xobject insertLoad(Xobject xBody, String sName, Xobject[] xIndex, Xobject[] xSize){
    if(xIndex.length != xSize.length) return null;
    
    XobjList blist = new XobjList();
    
    for(int i = 0 ; i < xIndex.length ; i++){
      //FortranłIndexȂ̂łP₳Ȃ΂ȂȂB
      blist.add(XTools.addXobjInt(xIndex[i].copy(), 
                                            Xcons.IntConstant(1)));
    }
    
    for(int i = 0 ; i < xSize.length ; i++){
      blist.add(xSize[i].copy());
    }
    Xobject xLoad = Xcons.List(Xcode.SCM_PRAGMA,
                               Xcons.Symbol(Xcode.IDENT, "load_"),
                               Xcons.Symbol(Xcode.IDENT, sName),
                               blist);
    Xobject xNew = Xcons.List(xLoad, xBody);
    
    if(!apply(xBody, xNew)) return null;
    return xNew;
  }
  
  public Xobject insertSwap(Xobject xBody, String sName){
    Xobject xSwap = Xcons.List(Xcode.SCM_PRAGMA,
                                Xcons.Symbol(Xcode.IDENT, "swap_"),
                                Xcons.Symbol(Xcode.IDENT, sName));
    Xobject xNew = Xcons.List(xBody, xSwap);
    if(!apply(xBody, xNew)) return null;
    return xNew;
  }
  
  public Xobject insertStore(Xobject xBody, String sName){
    Xobject xStore = Xcons.List(Xcode.SCM_PRAGMA,
                                Xcons.Symbol(Xcode.IDENT, "store_"),
                                Xcons.Symbol(Xcode.IDENT, sName));
    Xobject xNew = Xcons.List(xBody, xStore);
    if(!apply(xBody, xNew)) return null;
    return xNew;
  }
  public Xobject insertPStore(Xobject xBody, String sName){
    Xobject xStore = Xcons.List(Xcode.SCM_PRAGMA,
                                Xcons.Symbol(Xcode.IDENT, "pstore_"),
                                Xcons.Symbol(Xcode.IDENT, sName));
    Xobject xNew = Xcons.List(xBody, xStore);
    if(!apply(xBody, xNew)) return null;
    return xNew;

  }
  
  public Xobject insertPbegin(String sName, int[] size, int[] stride){
    return insertPbegin(m_root, sName, size, stride);
  }
  public Xobject insertPbegin(Xobject xBody, String sName, int[] size, int[] stride){
    Xobject[] xSize = new Xobject[size.length];
    for(int i = 0 ; i < size.length ; i++)
      xSize[i] = Xcons.IntConstant(size[i]);

    return insertPbegin(xBody, sName, xSize, stride);
  }
  public Xobject insertPbegin(Xobject xBody, String sName, Xobject[] xSize, int[] stride){
    if(xSize.length != stride.length) return null;
    
    XobjList blist = new XobjList();

    for(int i = 0 ; i < xSize.length ; i++){
      blist.add(xSize[i]);
    }
    for(int i = 0 ; i < stride.length ; i++){
      blist.add(Xcons.IntConstant(stride[i]));
    }
    
    Xobject xNewObj = Xcons.List(Xcode.SCM_PRAGMA,
                                 Xcons.Symbol(Xcode.IDENT, "pbegin_"),
                                 Xcons.Symbol(Xcode.IDENT, sName),
                                 blist,
                                 xBody);
    if(!apply(xBody, xNewObj))return null;
    return xNewObj;

  }
  public Xobject insertBegin(String sName, int[] size, int[] stride){
    return insertBegin(m_root, sName, size, stride);
  }

  public Xobject insertBegin(Xobject xBody, String sName, int[] size, int[] stride){
    if(size.length != stride.length) return null;
    
    XobjList blist = new XobjList();

    for(int i = 0 ; i < size.length ; i++){
      blist.add(Xcons.IntConstant(size[i]));
    }
    for(int i = 0 ; i < stride.length ; i++){
      blist.add(Xcons.IntConstant(stride[i]));
    }
    
    Xobject xNewObj = Xcons.List(Xcode.SCM_PRAGMA,
                                 Xcons.Symbol(Xcode.IDENT, "begin_"),
                                 Xcons.Symbol(Xcode.IDENT, sName),
                                 blist,
                                 xBody);
    if(!apply(xBody, xNewObj))return null;
    return xNewObj;
  }

  private boolean apply(Xobject xOld, Xobject xNew){
    if(xOld == m_root){
      m_root = xNew;
      return true;
    }else{    
      TopdownIterator iter = new TopdownIterator(m_root);
      for(iter.init(); !iter.end(); iter.next()){
        if(iter.get() == xOld){
          iter.substitute(xNew);
          return true;
        }
      }
    }
    return false;
  }
}
