// $Id: shmcc.java,v 1.5 2004-10-15 04:46:04 m-hirano Exp $
// $RWC_Release$
// $RWC_Copyright$

package shm;

import exc.object.*;
import exc.openmp.*;
import exc.tea.*;

public class shmcc extends exc.util.ccDriver {
  protected boolean relocateGlobalFlag = true;

  public shmcc(){ super("shmcc"); }

  public static void main(String args[]){ 
    shmcc driver = new shmcc();
    driver.run(args);
  }

  public boolean parseArg(String arg){
    if(arg.equals("--no-relocate")){
      relocateGlobalFlag = false;
      if(verbose) System.err.println("relocate flag off...");
      return true;
    } else return false;
  }

  public void init(){
    cpp_opts += " -DOMPC";
    if (getTemplateValue("APPROVED_DATE") != null) {
      cpp_opts += " -D_OPENMP=" + getTemplateValue("APPROVED_DATE");
    } else {
      cpp_opts += " -D_OPENMP";
    }
    if(getTemplateValue("CFRONT_FLAGS") != null) 
      cpp_opts += " "+getTemplateValue("CFRONT_FLAGS");
    cpp_opts += " "+getTemplateValue("OMPC_CFLAGS");
    // cc_opts += " "+getTemplateValue("OMPC_CFLAGS");

    ld_opts += " "+getTemplateValue("OMPC_MAIN");
    ld_opts += " "+getTemplateValue("OMPC_LDFLAGS");
  }

  public void run(XobjectFile f){
    if(verbose) System.err.println("transformation pass ... "+
				   f.getSourceFileName());
    f.debugFlag = debugFlag;
    OMP.debugFlag = debugFlag;

    OMP.setMoveAutoFlag(true);
    OMP.leaveThreadPrivate(true);

    TEAanalyzeDecl opTEA = new TEAanalyzeDecl();
    opTEA.init(f);

    OMPtranslateBlock op = new OMPtranslateBlock();
    op.init(f,false);

    f.iterateDef(opTEA);
    f.iterateDef(op);   // transform OMP pragma
    if(OMP.errorFlags){
      System.err.println("error in OpenMP. stop.");
      System.exit(1);
    }

    if(relocateGlobalFlag){
      if(verbose) System.err.println("relocate global data pass ...");
      // relocate global data to shared memory area
      relocateGlobalData opRel = new relocateGlobalData();
      opRel.init(f);
      f.iterateDef(opRel);
      opRel.Finalize();
    }

    op.init(f,true);
    if(verbose) System.err.println("block transformation data ...");
    f.iterateDef(op);   // transform OMP pragma

    String target = getTemplateValue("TARGET");
    if (target.equals("i386-linux")) {
	f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	f.addTailerLine("static void __G_DATA_CTOR() {");
	f.addTailerLine("  asm(\".section	.ctors \");");
	f.addTailerLine("  asm(\".long	 __G_DATA_INIT\");");
	f.addTailerLine("  asm(\".section	.text \");");
	f.addTailerLine("}");
	f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
    } else if (target.equals("x86_64-linux")) {
	f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	f.addTailerLine("static void __G_DATA_CTOR() {");
	f.addTailerLine("  asm(\".section	.ctors \");");
	f.addTailerLine("  asm(\".quad	 __G_DATA_INIT\");");
	f.addTailerLine("  asm(\".section	.text \");");
	f.addTailerLine("}");
	f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
    } else if (target.equals("ia64-linux")) {
	f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	f.addTailerLine("static void __G_DATA_CTOR() {");
	f.addTailerLine("  asm(\".section	.ctors \");");
	f.addTailerLine("  asm(\".align 8\");");
	f.addTailerLine("  asm(\"data8	 @fptr(__G_DATA_INIT)\");");
	f.addTailerLine("  asm(\".section	.text \");");
	f.addTailerLine("}");
	f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
    } else {
	System.err.println("unknown architecture : target = " + target);
	System.exit(1);
    }
  }
}

