// $Id: shmf77.java,v 1.5 2004-10-15 04:46:04 m-hirano Exp $
// $RWC_Release$
// $RWC_Copyright$

package shm;

import exc.object.*;
import exc.openmp.*;
import exc.tea.*;

public class shmf77 extends exc.util.ompf77 {
  protected boolean relocateGlobalFlag = true;

  public shmf77(){ super("shmf77"); }

  public static void main(String args[]){ 
    shmf77 driver = new shmf77();
    driver.run(args);
  }

  public boolean parseArg(String arg){
    if(arg.equals("--no-relocate")){
      relocateGlobalFlag = false;
      if(verbose) System.err.println("relocate flag off...");
      return true;
    } else return super.parseArg(arg);
  }

  public void init() {
    super.init();
    // add shmem option
    ffront_opts += " -saveArray ";
  }

  public void run(XobjectFile f){

    if(!omp_flag) return;	// no OpenMP, nothing to do

    if(verbose) System.err.println("transformation pass ... "+
				   f.getSourceFileName());
    f.debugFlag = debugFlag;
    OMP.debugFlag = debugFlag;

    OMP.setMoveAutoFlag(true);
    OMP.leaveThreadPrivate(true);

    TEAanalyzeDecl opTEA = new TEAanalyzeDecl();
    opTEA.init(f);

    OMPtranslateBlock op = new OMPtranslateBlock();
    op.init(f,false);

    f.iterateDef(opTEA);
    f.iterateDef(op);   // transform OMP pragma
    if(OMP.errorFlags){
      System.err.println("error in OpenMP. stop.");
      System.exit(1);
    }

    if(relocateGlobalFlag){
      if(verbose) System.err.println("relocate global data pass ...");
      // relocate global data to shared memory area
      relocateGlobalData opRel = new relocateGlobalData();
      opRel.init(f);
      f.iterateDef(opRel);
      opRel.Finalize();
    }

    op.init(f,true);
    if(verbose) System.err.println("block transformation data ...");
    f.iterateDef(op);   // transform OMP pragma

    if(relocateGlobalFlag){
      String target = getTemplateValue("TARGET");
      if (target.equals("i386-linux")) {
	  f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	  f.addTailerLine("static void __G_DATA_CTOR() {");
	  f.addTailerLine("  asm(\".section	.ctors \");");
	  f.addTailerLine("  asm(\".long	 __G_DATA_INIT\");");
	  f.addTailerLine("  asm(\".section	.text \");");
	  f.addTailerLine("}");
	  f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
      } else if (target.equals("x86_64-linux")) {
	  f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	  f.addTailerLine("static void __G_DATA_CTOR() {");
	  f.addTailerLine("  asm(\".section	.ctors \");");
	  f.addTailerLine("  asm(\".quad	 __G_DATA_INIT\");");
	  f.addTailerLine("  asm(\".section	.text \");");
	  f.addTailerLine("}");
	  f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
      } else if (target.equals("ia64-linux")) {
	  f.addTailerLine("static char *__g_data_init_f=(char *)__G_DATA_INIT; ");
	  f.addTailerLine("static void __G_DATA_CTOR() {");
	  f.addTailerLine("  asm(\".section	.ctors \");");
	  f.addTailerLine("  asm(\".align 8\");");
	  f.addTailerLine("  asm(\"data8	 @fptr(__G_DATA_INIT)\");");
	  f.addTailerLine("  asm(\".section	.text \");");
	  f.addTailerLine("}");
	  f.addTailerLine("static char *__g_data_ctor_f=(char *)__G_DATA_CTOR; ");
      } else {
	  System.err.println("unknown architecture : target = " + target);
	  System.exit(1);
      }
    }
  }
}


