static char rcsid[] = "$Id";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "exc_platform.h"
#include "grpc.h"
#include "omp.h"
#include "ogr_internal.h"
#include "opengr.h"


typedef struct {
    char *varName;
    int nSessions;
    grpc_sessionid_t *sIdList;
} ogrBarrierVar;


static ogr_lock_t bVarLock;
static ogr_HashTable bVarTable;


int
_ogr_barrierVar_init() {
    ogr_init_lock(&bVarLock);
    (void)ogr_InitHashTable(&bVarTable, OGR_STRING_KEYS);
    return 0;
}


int
_ogr_create_barrier(char *bVar) {
    ogr_HashEntry *ePtr;
    int isNew;
    int ret = -1;

    ogr_set_lock(&bVarLock);

    _ogr_debug(5, "checking barrier variable '%s'.\n",
	       bVar);
    ePtr = ogr_CreateHashEntry(&bVarTable, bVar, &isNew);
    if (isNew == 1) {
	ogrBarrierVar *bVarPtr = (ogrBarrierVar *)
	    malloc(sizeof(ogrBarrierVar));
	if (bVarPtr == NULL) {
	    _ogr_error("Can't create barrier variable '%s'.\n",
		       bVar);
	    goto bailOut;
	}
	(void)memset((void *)bVarPtr, 0, sizeof(ogrBarrierVar));
	bVarPtr->varName = bVar;
	ogr_SetHashValue(ePtr, (ClientData)bVarPtr);
	_ogr_debug(5, "barrier variable '%s' is created.\n",
		   bVar);
	ret = 0;
    }

    bailOut:
    ogr_unset_lock(&bVarLock);

    return ret;
}


int
_ogr_reset_barrier(char *bVar) {
    int ret = -1;
    ogr_HashEntry *ePtr;
    ogr_set_lock(&bVarLock);

    ePtr = ogr_FindHashEntry(&bVarTable, bVar);
    if (ePtr != NULL) {
	ogrBarrierVar *bVarPtr = (ogrBarrierVar *)
	    ogr_GetHashValue(ePtr);
	bVarPtr->nSessions = 0;
	(void)free(bVarPtr->sIdList);
	bVarPtr->sIdList = NULL;
	ret = 0;
    }

    ogr_unset_lock(&bVarLock);

    return ret;
}


int
_ogr_append_session_to_barrier(char *bVar, grpc_sessionid_t sId) {
    int ret = -1;
    ogr_HashEntry *ePtr;
    ogr_set_lock(&bVarLock);

    ePtr = ogr_FindHashEntry(&bVarTable, bVar);
    if (ePtr != NULL) {
	ogrBarrierVar *bVarPtr = (ogrBarrierVar *)
	    ogr_GetHashValue(ePtr);
	if (bVarPtr->sIdList == NULL) {
	    bVarPtr->sIdList = (grpc_sessionid_t *)
		malloc(sizeof(grpc_sessionid_t));
	} else {
	    bVarPtr->sIdList = (grpc_sessionid_t *)
		realloc((void *)bVarPtr->sIdList,
			sizeof(grpc_sessionid_t) * 
			(bVarPtr->nSessions + 1));
	}
	bVarPtr->sIdList[bVarPtr->nSessions] = sId;
	bVarPtr->nSessions++;
	ret = 0;
    }

    ogr_unset_lock(&bVarLock);

    return ret;
}


grpc_sessionid_t *
_ogr_get_session_list_from_barrier(char *bVar, int *nPtr) {
    grpc_sessionid_t *ret = NULL;
    ogr_HashEntry *ePtr;
    ogr_set_lock(&bVarLock);

    ePtr = ogr_FindHashEntry(&bVarTable, bVar);
    if (ePtr != NULL) {
	ogrBarrierVar *bVarPtr = (ogrBarrierVar *)
	    ogr_GetHashValue(ePtr);

	ret = bVarPtr->sIdList;
	if (nPtr != NULL) {
	    *nPtr = bVarPtr->nSessions;
	}
    }

    ogr_unset_lock(&bVarLock);

    return ret;
}
