static char rcsid[] = "$Id: ogr_call.c,v 1.3 2005-02-18 20:59:33 m-hirano Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "exc_platform.h"
#include "grpc.h"
#include "omp.h"
#include "ogr_internal.h"
#include "opengr.h"

/*
 * Implementation note:
 *	Ninf-G/GridRPC API have no va_list parameter API for RPC
 *	invocation. Instead of it, it povides parameter stack
 *	API. But, the parameter stack API is implemented as (void *)
 *	type stack, so all parameter that will be passed must be types
 *	that storage size is <= sizeof(void *). In 32 bit arch
 *	machine, this means WE CAN'T USE double/long long.
 *	Because of this, I modified Ninf-G, implemented:
 *		grpc_call_va(grpc_function_handle_t *, va_list)
 *		grpc_call_async_va(grpc_function_handle_t *, va_list)
 *
 *	m-hirano
 */


void
__rpcSync(char *stubname, char *bVar, ...) {
    va_list args;
    grpc_error_t eCode;
    ogrHostHandle *ohhPtr;

    va_start(args, bVar);

    ohhPtr = _ogr_getHostHandle(stubname);

    if (ohhPtr == NULL) {
	_ogr_error("host handle for '%s' is NULL.\n",
		   stubname);
	return;
    }

    /* !!! host handle lock !!! */
    ogrLockHost(ohhPtr);

    if (ogrHostUsageType(ohhPtr) == ogr_HandleIsUsedAsync ||
	ogrHostUsageType(ohhPtr) == ogr_HandleIsAsyncAcquired) {
	if (ohhPtr->sId != GRPC_SESSIONID_VOID) {
	    (void)grpc_wait(ohhPtr->sId);
	}
    }

    ogrUpdateHostUsage(ohhPtr, ogr_HandleIsWaiting);
    _ogr_debug(1, "sync call start: host '%s' func '%s'.\n",
	       ohhPtr->hostName, ohhPtr->funcName);
    eCode = grpc_call_va(ohhPtr->gfhPtr, args);
    if (eCode != GRPC_NO_ERROR) {
	grpc_perror_np("GridRPC error: "
		       "grpc_call_va()");
	_ogr_error("sync call error: host '%s' func '%s'.\n",
		   ohhPtr->hostName, ohhPtr->funcName);
    }
    va_end(args);

    /*
     * for sync call, always set sid as GRPC_SESSIONID_VOID to avoid
     * illegal grpc_wait() call.
     */
    ohhPtr->sId = GRPC_SESSIONID_VOID;

    ogrUpdateHostUsage(ohhPtr, ogr_HandleIsFree);
    ogrIncrHostUsageNCalls(ohhPtr);

    _ogr_debug(1, "sync call done: host '%s' func '%s'.\n",
	       ohhPtr->hostName, ohhPtr->funcName);

    /* !!! host handle unlock !!! */
    ogrUnlockHost(ohhPtr);

    ogrCondWakeupHHWait(ohhPtr->ofhPtr);

    return;
}


void
__rpcAsync(char *stubname, char *bVar, ...) {
    va_list args;
    grpc_sessionid_t sid;
    grpc_error_t eCode;
    ogrHostHandle *ohhPtr;
    static int firstAsync = 1;
    extern int _ogr_doAsyncWait;

    if (firstAsync == 1) {
	firstAsync = 0;
	_ogr_start_timer("async_call");
    }

    va_start(args, bVar);
    if (_ogr_doAsyncWait != 1) {
	if (bVar != NULL && bVar[0] != '\0') {
	    (void)_ogr_create_barrier(bVar);
	}
    }

    ohhPtr = _ogr_getHostHandle(stubname);
    if (ohhPtr == NULL) {
	_ogr_error("host handle for '%s' is NULL.\n",
		   stubname);
	return;
    }

    /* !!! host handle lock !!! */
    ogrLockHost(ohhPtr);

    if (_ogr_doAsyncWait == 1) {
	if (ogrHostUsageType(ohhPtr) == ogr_HandleIsUsedAsync ||
	    ogrHostUsageType(ohhPtr) == ogr_HandleIsAsyncAcquired) {
	    if (ohhPtr->sId != GRPC_SESSIONID_VOID) {
		(void)grpc_wait(ohhPtr->sId);
	    }
	}
    }

    ogrUpdateHostUsage(ohhPtr, ogr_HandleIsCalling);
    _ogr_debug(1, "async call start: host '%s' func '%s'.\n",
	       ohhPtr->hostName, ohhPtr->funcName);
    eCode = grpc_call_async_va(ohhPtr->gfhPtr, &sid, args);
    if (eCode != GRPC_NO_ERROR) {
	grpc_perror_np("GridRPC error: "
		       "grpc_call_async_va()");
	_ogr_error("async call error: host '%s' func '%s'.\n",
		   ohhPtr->hostName, ohhPtr->funcName);
	ogrHostUsageType(ohhPtr) = ogr_HandleIsFree;
	va_end(args);
	ohhPtr->sId = GRPC_SESSIONID_VOID;
	goto BailOut;
    }
    va_end(args);
    ohhPtr->sId = sid;
    if (_ogr_doAsyncWait != 1) {
	if (bVar != NULL && bVar[0] != '\0') {
	    (void)_ogr_append_session_to_barrier(bVar, sid);
	}
    }
    ogrUpdateHostUsage(ohhPtr, ogr_HandleIsUsedAsync);
    ogrIncrHostUsageNCalls(ohhPtr);

    BailOut:
    /* !!! host handle unlock !!! */
    ogrUnlockHost(ohhPtr);

    _ogr_debug(1, "async call done: host '%s' func '%s' sId:%d.\n",
	       ohhPtr->hostName, ohhPtr->funcName,
	       sid);

    ogrCondWakeupHHWait(ohhPtr->ofhPtr);

    return;
}


void
__rpcWait(int nVars, ...)
{
    va_list args;

    _ogr_debug(1, "async wait all start.\n");

    va_start(args, nVars);

    if (nVars == 0) {
	waitAll:
#if 0
	_ogr_waitAll();
#else
	if (omp_get_thread_num() == 0) {
	    _ogr_start_timer("wait_all");
	    grpc_wait_all();
	    _ogr_stop_timer("wait_all");
	    _ogr_stop_timer("async_call");
	}
#endif
    } else {
	int i;
	grpc_sessionid_t *allSessions = NULL;
	int nSessions = 0;
	char *bVarTmp;
	int nSessionTmp;
	grpc_sessionid_t *sIdListTmp;

	if (nVars == 0) {
	    goto waitAll;
	}

	for (i = 0; i < nVars; i++) {
	    nSessionTmp = 0;
	    bVarTmp = (char *)va_arg(args, char *);
	    sIdListTmp = _ogr_get_session_list_from_barrier(
		bVarTmp,
		&nSessionTmp);
	    if (sIdListTmp == NULL || nSessionTmp == 0) {
		continue;
	    }
	    if (allSessions == NULL) {
		allSessions = (grpc_sessionid_t *)malloc(
		    sizeof(grpc_sessionid_t) * nSessionTmp);
	    } else {
		allSessions = (grpc_sessionid_t *)realloc(
		    (void *)allSessions,
		    sizeof(grpc_sessionid_t) * 
		    (nSessions + nSessionTmp));
	    }
	    (void)memcpy((void *)(&(allSessions[nSessions])),
			 (void *)sIdListTmp,
			 (sizeof(grpc_sessionid_t) * nSessionTmp));
	    nSessions += nSessionTmp;
	    (void)_ogr_reset_barrier(bVarTmp);
	}

	if (nSessions == 0) {
	    goto waitAll;
	}
	(void)grpc_wait_and(allSessions, nSessions);

    }
    _ogr_debug(1, "async wait all done.\n");
}

