static char rcsid[] = "$Id: ogr_init.c,v 1.3 2005-02-18 20:59:34 m-hirano Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
#include "exc_platform.h"
#include "grpc.h"
#include "ogr_hash.h"
#include "omp.h"
#include "ogr_internal.h"


static char *usageTypeStrTab[] = {
    "Free",
    "Acquired",
    "Calling",
    "Async",
    "AsyncAcquired",
    "Waiting",
    "Unusable"
};

#define usageTypeStr(X)	usageTypeStrTab[(int)(X)]


#define FOR_ENTRIES_IN_HASH(h, s, tab) \
	for (h = ogr_FirstHashEntry(tab, s); \
	     h != NULL; \
	     h = ogr_NextHashEntry(s))

ogr_HashTable funcTable;

static char *openGRConfigFile = "./.ogrrc";
static char *ninfGConfigFile = NULL;
static int nRpcHosts = 0;

static ogr_lock_t funcHandleAcqLock;

static ogr_HashTable hostTable;

static int __ogr_Exiting = 0;

static int *nFiPtrList = NULL;
static ogrRpcFuncInfo ***fiPtrList = NULL;

int _ogr_doAsyncWait = 0;

static void
yieldMe(int uSec) {
    struct timeval to;
    if (uSec >= 1000000) {
	to.tv_sec = uSec / 1000000;
	to.tv_usec = uSec - to.tv_sec * 1000000;
    } else {
	to.tv_sec = 0;
	to.tv_usec = uSec;
    }

    (void)select(0, NULL, NULL, NULL, &to);
}


static int
GetToken(char *buf, char **tokens, int max) {
    int n = 0;
    int nonSpace = 0;

    while (*buf != '\0' && n < max) {
	while (isspace((int)*buf) && *buf != '\0') {
	    buf++;
	}
	if (*buf == '\0') {
	    break;
	}
	tokens[n] = buf;

	nonSpace = 0;
	while (!isspace((int)*buf) && *buf != '\0') {
	    nonSpace++;
	    buf++;
	}
	if (*buf == '\0') {
	    if (nonSpace > 0) {
		n++;
	    }
	    break;
	}
	*buf = '\0';
	n++;
	if (*(buf + 1) == '\0') {
	    break;
	} else {
	    buf++;
	}
    }

    return n;
}


#ifndef HAVE_STRDUP
char *
strdup(char *str) {
    int len = strlen(str);
    char *ret = (char *)malloc(len + 1);
    memcpy(ret, str, len);
    ret[len] = '\0';
    return ret;
}
#endif /* HAVE_STRDUP */


static ogrRpcFuncInfo *
allocRpcFuncInfo(char *funcName, char *hostName, int nHandles) {
    ogrRpcFuncInfo *ret = (ogrRpcFuncInfo *)malloc(sizeof(ogrRpcFuncInfo)); 

    if (ret == NULL) {
	return NULL;
    }

    (void)memset((void *)ret, 0, sizeof(ogrRpcFuncInfo));

    ret->funcName = strdup(funcName);
    ret->hostName = strdup(hostName);
    ret->nHandles = nHandles;
    
    ret->gfhPtr = (grpc_function_handle_t *)malloc(
	sizeof(grpc_function_handle_t) * nHandles);

    if (ret->gfhPtr == NULL) {
	(void)free(ret->funcName);
	(void)free(ret->hostName);
	(void)free(ret);
	return NULL;
    }
    (void)memset((void *)(ret->gfhPtr), 0,
		 sizeof(grpc_function_handle_t) * nHandles);

    ogr_init_lock(&(ret->funcInfoLock));
    ret->rpcFuncHandleInited = 0;
    ret->rpcFuncHandleDeleted = 0;
    return ret;
}


static ogrRpcHostInfo *
allocRpcHostInfo(char *hostname) {
    ogrRpcHostInfo *ret = (ogrRpcHostInfo *)malloc(sizeof(ogrRpcHostInfo));

    if (ret == NULL) {
	return NULL;
    }
    ret->hostName = strdup(hostname);
    ret->nHandles = 0;

    (void)ogr_InitHashTable(&(ret->hostFuncTable), OGR_STRING_KEYS);
    ogr_init_lock(&(ret->hostInfoLock));
    return ret;
}


static int
readConfig(void) {
    char *tokens[32];
    int n;
    char buf[1024];
    char *confFile = getenv("OGRRC");
    FILE *fd = NULL;
    int len;
    ogr_HashEntry *ePtr;
    ogr_HashSearch sctx;
    ogrRpcHostInfo *hiPtr;
    int nHosts;

    if (confFile == NULL || confFile[0] == '\0') {
	confFile = openGRConfigFile;
    }

    fd = fopen(confFile, "r");
    if (fd == NULL) {
	_ogr_error("can't open configuration file \"%s\".\n",
		   confFile);
	return 0;
   }

    while (fgets(buf, 1024, fd) != NULL) {
	len = strlen(buf);
	if (len == 1023 && buf[len - 1] != '\n') {
	    _ogr_warning("\"%s\": line too long.\n",
			 confFile);
	}
	n = GetToken(buf, tokens, 32);
	if (n == 0) {
	    continue;
	}
	if (tokens[0][0] == '#') {
	    continue;
	}

	if ((n > 1) &&
	    strcasecmp(tokens[0], "@ninfG-Conf") == 0) {
	    ninfGConfigFile = strdup(tokens[1]);
	    if (access(ninfGConfigFile, R_OK|F_OK) != 0) {
		_ogr_error("can't access \"%s\".\n",
			   ninfGConfigFile);
		(void)fclose(fd);
		ninfGConfigFile = NULL;
		return 0;
	    }
	} else {
	    int adds = 1;
	    int isNew;

	    if (n > 1) {
		int tmpN = atoi(tokens[1]);
		if (tmpN <= 0) {
		    continue;
		} else {
		    adds = tmpN;
		}
	    }

	    isNew = 0;
	    ePtr = ogr_CreateHashEntry(&hostTable, tokens[0], &isNew);
	    if (isNew == 1) {
		hiPtr = allocRpcHostInfo(tokens[0]);
		ogr_SetHashValue(ePtr, (ClientData)hiPtr);
	    }
	    hiPtr->nHandles += adds;
	}
    }
    (void)fclose(fd);

    nHosts = 0;
    FOR_ENTRIES_IN_HASH(ePtr, &sctx, &hostTable) {
	hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);
	_ogr_debug(10, "\thost '%s':\t%d handles\n",
		   hiPtr->hostName,
		   hiPtr->nHandles);
	nHosts += hiPtr->nHandles;
    }
    if (nHosts <= 0) {
	_ogr_error(" no PRC hosts are "
		   "specified in '%s'.\n", confFile);
	return 0;
    }
    nRpcHosts = nHosts;

    if (ninfGConfigFile == NULL) {
	_ogr_error(" no Ninf-G configuration file is "
		   "specified in '%s'.\n", confFile);
	return 0;
    }
    return 1;
}


int
_ogr_init(int argc, char *argv[]) {
    char *envDoOgrAsyncWait = getenv("OGR_ASYNC_WAIT");

    if (envDoOgrAsyncWait != NULL) {
	if (strcasecmp(envDoOgrAsyncWait, "1") == 0 ||
	    strcasecmp(envDoOgrAsyncWait, "Yes") == 0 ||
	    strcasecmp(envDoOgrAsyncWait, "True") == 0) {
	    _ogr_doAsyncWait = 1;
	}
    }

    _ogr_debug_init();
    _ogr_stat_init();
    _ogr_barrierVar_init();
    
    ogr_init_lock(&funcHandleAcqLock);

    (void)ogr_InitHashTable(&hostTable, OGR_STRING_KEYS);
    (void)ogr_InitHashTable(&funcTable, OGR_STRING_KEYS);

    if (readConfig() != 1) {
	return 0;
    }

    if (grpc_initialize(ninfGConfigFile) != GRPC_NO_ERROR) {
	grpc_perror_np("GridRPC error: grpc_initialize()");
	return 0;
    }

    if (grpc_session_info_set_threshold_np(INT_MAX) != GRPC_NO_ERROR) {
	return 1;
    }

    (void)atexit(_ogr_final);

    _ogr_rpc_prologue();

    return 1;
}


static ogrHostHandle *
allocHostHandle(char *funcName, char *hostName) {
    ogrHostHandle *ret = 
	(ogrHostHandle *)malloc(sizeof(ogrHostHandle));

    if (ret == NULL) {
	_ogr_error("can't allocate a host handle for "
		   "'%s' at '%s'.\n", funcName, hostName);
	return NULL;
    }

    (void)memset((void *)ret, 0, sizeof(ogrHostHandle));
    ret->hostName = strdup(hostName);
    ret->funcName = strdup(funcName);
    ogr_init_lock(&(ret->hostLock));
    ret->sId = GRPC_SESSIONID_VOID;
    ret->idx = -1;

    return ret;
}


static ogrFuncHandle *
allocFuncHandle(char *funcName) {
    int i;
    ogrFuncHandle *ret = (ogrFuncHandle *)malloc(sizeof(ogrFuncHandle));

    if (ret == NULL) {
	_ogr_error("can't allocate function handles for '%s'.\n",
		   funcName);
	return NULL;
    }

    (void)memset((void *)ret, 0, sizeof(ogrFuncHandle));
    ret->funcName = strdup(funcName);

    ogr_init_lock(&(ret->funcLock));

    ogr_init_lock(&(ret->fhInitLock));
    ogr_cond_init(&(ret->fhInitCond));

    ogr_init_lock(&(ret->hhWaitLock));
    ogr_cond_init(&(ret->hhWaitCond));

    ret->MRUHandle = -1;
    
    ret->hHandleList = (ogrHostHandle **)
	malloc(sizeof(ogrHostHandle *) * nRpcHosts);
    ret->hUsageList = (ogrHostHandleUsage *)
	malloc(sizeof(ogrHostHandleUsage) * nRpcHosts);

    if (ret->hHandleList == NULL ||
	ret->hUsageList == NULL) {
	_ogr_error("can't allocate function handle array "
		   "for '%s'.\n", funcName);
	return NULL;
    }
    (void)memset((void *)ret->hHandleList, 0,
		 sizeof(ogrHostHandle *) * nRpcHosts);
    (void)memset((void *)ret->hUsageList, 0,
		 sizeof(ogrHostHandleUsage) * nRpcHosts);

    for (i = 0; i < nRpcHosts; i++) {
	ogr_init_lock(&(ogrFunc2HostUsageLock(ret, i)));
    }

    ret->nMaxHandles = nRpcHosts;
    ret->nCurHandles = 0;

    return ret;
}


static ogrRpcFuncInfo *
findRpcFuncInfo(ogrRpcHostInfo *hiPtr, char *funcName) {
    ogrRpcFuncInfo *fiPtr = NULL;
    ogr_HashEntry *ePtr;
    int isNew = 0;

    ogrLockHostInfo(hiPtr);

    ePtr = ogr_CreateHashEntry(&(hiPtr->hostFuncTable), funcName, &isNew);
    if (ePtr == NULL) {
	_ogr_error("can't create RPC function info "
		   "hash entry for %s:%s (n:%d).\n",
		   hiPtr->hostName, funcName,
		   hiPtr->nHandles);
	abort();
    }
    if (isNew == 1) {
	fiPtr = allocRpcFuncInfo(funcName,
				 hiPtr->hostName,
				 hiPtr->nHandles);
	if (fiPtr == NULL) {
	    _ogr_error("can't create RPC function info for "
		       "%s:%s (n:%d).\n",
		       hiPtr->hostName, funcName,
		       hiPtr->nHandles);
	    abort();
	}
	ogr_SetHashValue(ePtr, (ClientData)fiPtr);
    } else {
	fiPtr = (ogrRpcFuncInfo *)ogr_GetHashValue(ePtr);
    }

    ogrUnlockHostInfo(hiPtr);

    return fiPtr;
}


static int
bindGRPCFuncHandle(ogrFuncHandle *fhPtr, ogrRpcFuncInfo *fiPtr) {
    int ret = 0;

    if (fiPtr->rpcFuncHandleInited == 0) {
	/* Initialize GridRPC functon handles */
	ogrHostHandle *newHH;
	grpc_function_handle_t *gfhPtr;
	int addStart;
	int addEnd;
	int i, j;

	if (grpc_function_handle_array_init_np(
		fiPtr->gfhPtr,
		fiPtr->nHandles,
		fiPtr->hostName,
		fiPtr->funcName) != GRPC_NO_ERROR) {
	    grpc_perror_np("GridRPC error: "
			   "grpc_function_handle_array_init_np()");
	    _ogr_error("can't initialize GRPC function handles "
		       "for %s:%s (%d)\n",
		       fiPtr->hostName,
		       fiPtr->funcName,
		       fiPtr->nHandles);
	    abort();
	}

	_ogr_debug(1, "GRPC function handles allocation: "
		   "host '%s': func '%s': handles %d\n",
		   fiPtr->hostName,
		   fiPtr->funcName,
		   fiPtr->nHandles);

	fiPtr->rpcFuncHandleInited = 1;

	/* lock func */
	ogrLockFunc(fhPtr);

	addStart = fhPtr->nCurHandles;
	addEnd = addStart + fiPtr->nHandles;
	gfhPtr = fiPtr->gfhPtr;

	for (j = 0, i = addStart; i < addEnd; i++, j++) {
	    newHH = allocHostHandle(fiPtr->funcName,
				    fiPtr->hostName);
	    if (newHH == NULL) {
		fhPtr->hUsageList[i].type = ogr_HandleIsUnusable;
		fhPtr->hHandleList[i] = NULL;
	    } else {
		fhPtr->hUsageList[i].type = ogr_HandleIsFree;
		newHH->gfhPtr = &(gfhPtr[j]);
		newHH->ofhPtr = fhPtr;
		newHH->idx = i;
		fhPtr->hHandleList[i] = newHH;
	    }
	}

	fhPtr->nCurHandles += fiPtr->nHandles;
	ret = fhPtr->nCurHandles;

	/* unlock func */
	ogrUnlockFunc(fhPtr);
    } else {
	/* lock func */
	ogrLockFunc(fhPtr);

	ret = fhPtr->nCurHandles;

	/* unlock func */
	ogrUnlockFunc(fhPtr);
    }

    return ret;
}


/* 
 * Returns:
 *	n (> 0):	Up to n (index in fhPtr) handles is initialized.
 *	0:		Other threads are initializing rest of all handles.
 */
static int
initHostHandlesOfFuncHandle(ogrFuncHandle *fhPtr, char *funcName) {
    ogr_HashEntry *ePtr = NULL;
    ogr_HashSearch sCtx;
    ogrRpcHostInfo *hiPtr = NULL;
    ogrRpcFuncInfo *fiPtr = NULL;
    int maxThd = omp_get_max_threads();

    int ret = 0;

    FOR_ENTRIES_IN_HASH(ePtr, &sCtx, &hostTable) {
	hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);

	fiPtr = findRpcFuncInfo(hiPtr, funcName);

	/* lock func info */
	if (ogrTryLockFuncInfo(fiPtr) != 0) {
	    /* Check next host info */
	    continue;
	}

	ret = bindGRPCFuncHandle(fhPtr, fiPtr);

	/* unlock func info */
	ogrUnlockFuncInfo(fiPtr);

	if (maxThd > 1) {
	    /*
	     * Maybe other threads are waiting for the handles.
	     * Return and wake them up.
	     */
	    break;
	}
    }

    return ret;
}


/*
 * function handle scheduling work horse
 */
static ogrHostHandle *
getAnyHostHandleFromFuncHandle(ogrFuncHandle *fhPtr) {
    ogrHostHandle *ret = NULL;
    int nAsync;
    int firstAsyncIdx;
    int i;
    int minUsed, minUsedIdx;
    int idx, cIdx;
    int maxN;

    volatile int curNHandles;
    volatile int curMRUHandle;

    ReCheck:

    ret = NULL;
    nAsync = 0;
    firstAsyncIdx = -1;

    /*
     * Checking GRPC function handle initialization
     */

    /* lock func */
    ogrLockFunc(fhPtr);

    curNHandles = fhPtr->nCurHandles;

    /* unlock func */
    ogrUnlockFunc(fhPtr);

    if (curNHandles < fhPtr->nMaxHandles) {
	int tmpN;

	tmpN = initHostHandlesOfFuncHandle(fhPtr, fhPtr->funcName);
	if (tmpN == 0) {
	    /* 
	     * Any other threads did all host handle (including GPRC
	     * func handle) initialization
	     */
	    ogrLockFHInit(fhPtr);
	    _ogr_debug(10, "wait for func(host) handle initialize.\n");
	    ogrCondSleepFHInit(fhPtr);
	    _ogr_debug(10, "func(host) handle available.\n");
	    ogrUnlockFHInit(fhPtr);

	    goto ReCheck;
	}
    }

    /*
     * At here, some host handles needed for this function handle are
     * initialized. Wake all sleepers (if any) up.
     */
    _ogr_debug(10, "wake up sleepers waiting for func(host) handle.\n");
    ogrCondWakeupFHInit(fhPtr);


    /*
     * Start host handle scheduling
     */

    /* lock func */
    ogrLockFunc(fhPtr);

    maxN = fhPtr->nCurHandles;
    curMRUHandle = fhPtr->MRUHandle;

    idx = (curMRUHandle + 1) % maxN;
    minUsedIdx = INT_MAX;
    minUsed = INT_MAX;

    _ogr_debug(5, "searching host handle for '%s' from idx %d/%d.\n",
	       fhPtr->funcName, idx, maxN);

    for (i = 0; i < maxN; i++) {
	cIdx = (idx + i) % maxN;
	if (fhPtr->hHandleList[cIdx] != NULL) {

	    ogrLockHostUsageFromFunc(fhPtr, cIdx);

	    _ogr_debug(15, "check idx %d: '%s' used %4d times.\n",
		       cIdx,
		       usageTypeStr(ogrFunc2HostUsageType(fhPtr, cIdx)),
		       ogrFunc2HostUsageNCalls(fhPtr, cIdx));

	    switch (ogrFunc2HostUsageType(fhPtr, cIdx)) {

		case ogr_HandleIsFree: {
		    /* Use this no argue. */
		    ogrFunc2HostUsageType(fhPtr, cIdx) = 
			ogr_HandleIsAcquired;
		    ret = fhPtr->hHandleList[cIdx];
		    fhPtr->MRUHandle = cIdx;
		    ogrUnlockHostUsageFromFunc(fhPtr, cIdx);
		    goto BailOut;
		}

		case ogr_HandleIsUsedAsync: {
		    if (ogrFunc2HostUsageNCalls(fhPtr, cIdx) < minUsed) {
			minUsed = ogrFunc2HostUsageNCalls(fhPtr, cIdx);
			minUsedIdx = cIdx;
		    }
		    break;
		}

		default: {
		    break;
		}

	    }

	    ogrUnlockHostUsageFromFunc(fhPtr, cIdx);

	}
    }

    if (minUsedIdx != INT_MAX) {
	cIdx = minUsedIdx;
	ogrLockHostUsageFromFunc(fhPtr, cIdx);
	ogrFunc2HostUsageType(fhPtr, cIdx) = ogr_HandleIsAsyncAcquired;
	ogrUnlockHostUsageFromFunc(fhPtr, cIdx);
	ret = fhPtr->hHandleList[cIdx];
	fhPtr->MRUHandle = cIdx;
	goto BailOut;
    }

    /* unlock func */
    ogrUnlockFunc(fhPtr);

    /*
     * No host handle is available. Sleep until someone is done.
     */
    ogrLockHHWait(fhPtr);
    _ogr_debug(10, "wait for free host handle.\n");
    ogrCondSleepHHWait(fhPtr);
    _ogr_debug(10, "host handle available.\n");
    ogrUnlockHHWait(fhPtr);

    goto ReCheck;


    BailOut:

    _ogr_debug(5, "use host handle for '%s' idx %d(%d -> %d, '%s') "
	       "at host '%s'.\n",
	       ret->funcName,
	       cIdx, idx, ret->idx, usageTypeStr(ogrHostUsageType(ret)),
	       ret->hostName);

    /* unlock func */
    ogrUnlockFunc(fhPtr);

    return ret;
}


static ogrFuncHandle *
findFuncHandle(char *funcName) {
    ogr_HashEntry *ePtr;
    ogrFuncHandle *fhPtr;
    int isNew = 0;

    /* !!! Acquire lock !!! */
    ogr_set_lock(&funcHandleAcqLock);

    ePtr = ogr_CreateHashEntry(&funcTable, funcName, &isNew);
    if (ePtr == NULL) {
	_ogr_error("can't create function handle "
		   "hash entry for '%s'.\n", funcName);
	abort();
    }
    if (isNew == 1) {
	fhPtr = allocFuncHandle(funcName);
	if (fhPtr == NULL) {
	    abort();
	}
	ogr_SetHashValue(ePtr, (ClientData)fhPtr);
    } else {
	fhPtr = (ogrFuncHandle *)ogr_GetHashValue(ePtr);
    }

    /* !!! Acquire unlock !!! */
    ogr_unset_lock(&funcHandleAcqLock);

    return fhPtr;
}


ogrHostHandle *
_ogr_getHostHandle(char *funcName) {
    ogrFuncHandle *fhPtr = findFuncHandle(funcName);
    if (fhPtr == NULL) {
	return NULL;
    } else {
	return getAnyHostHandleFromFuncHandle(fhPtr);
    }
}


static void
waitAsyncOnHostHandle(ogrHostHandle *ohhPtr) {
    ogrLockHost(ohhPtr);
    
    if (ohhPtr->sId != GRPC_SESSIONID_VOID) {
	grpc_sessionid_t sId = ohhPtr->sId;
	int dum;
	(void)grpc_wait_or(&sId, 1, &dum);
	_ogr_debug(10, "host '%s' func '%s' idx:%d sid:%d wait done.\n",
		   ohhPtr->hostName, ohhPtr->funcName, 
		   ohhPtr->idx, sId);   
	ohhPtr->sId = GRPC_SESSIONID_VOID;
    }

    ogrUpdateHostUsage(ohhPtr, ogr_HandleIsFree);

    ogrUnlockHost(ohhPtr);
}


static void
waitAsyncOnFuncHandle(ogrFuncHandle *fhPtr) {
    int n;
    int i;
    int nFree;
    ogrHostHandle *ohhPtr;

    ReCheck:
    /* lock func */
    ogrLockFunc(fhPtr);

    ohhPtr = NULL;
    n = fhPtr->nMaxHandles;
    nFree = 0;

    for (i = 0; i < n; i++) {
	ohhPtr = fhPtr->hHandleList[i];
	if (fhPtr->hUsageList[i].type == ogr_HandleIsFree) {
	    nFree++;
	} else if (fhPtr->hUsageList[i].type == ogr_HandleIsUsedAsync) {
	    /*
	     * synchronize async RPC on this handle
	     */
	    /* update usage FIRST */
	    ogrHostUsageType(ohhPtr) = ogr_HandleIsWaiting;
	    /* unlock func */
	    ogrUnlockFunc(fhPtr);
	    /* then wait async */
	    waitAsyncOnHostHandle(ohhPtr);
	    goto ReCheck;
	}
    }

    if (nFree == n) {
	goto BailOut;
    } else {
	ogrUnlockFunc(fhPtr);
	
	goto ReCheck;
    }

    BailOut:
    /* unlock func */
    ogrUnlockFunc(fhPtr);

    return;
}


void
_ogr_waitAll(void) {
    ogr_HashEntry *ePtr;
    ogr_HashSearch sctx;
    ogrFuncHandle *fhPtr;

    FOR_ENTRIES_IN_HASH(ePtr, &sctx, &funcTable) {
	fhPtr = (ogrFuncHandle *)ogr_GetHashValue(ePtr);
	waitAsyncOnFuncHandle(fhPtr);
    }

    return;
}


#ifdef USE_SPROC
typedef void (*cfunc)();
#else
typedef void * (*cfunc)();
#endif /* USE_SPROC */

extern void _ompc_do_parallel(cfunc f, void *args);
static
#ifdef USE_SPROC
void
#else
void *
#endif /* USE_SPROC */
delGFHs(void *dum) {
    int tid = omp_get_thread_num();
    int i;
    int n;
    ogrRpcFuncInfo **fiPtrPtr;
    ogrRpcFuncInfo *fiPtr;

    n = nFiPtrList[tid];
    fiPtrPtr = &(fiPtrList[tid][0]);

    _ogr_debug(5, "enter GRPC function handles destruction.\n");

    for (i = 0; i < n; i++) {
	fiPtr = fiPtrPtr[i];
	if (fiPtr->rpcFuncHandleDeleted == 0) {
	    _ogr_debug(5, "GRPC function handles destruction start: "
		       "host '%s': func '%s': handles %d\n",
		       fiPtr->hostName,
		       fiPtr->funcName,
		       fiPtr->nHandles);
	    (void)grpc_function_handle_array_destruct_np(
		fiPtr->gfhPtr,
		fiPtr->nHandles);
	    _ogr_debug(1, "GRPC function handles destruction done: "
		       "host '%s': func '%s': handles %d\n",
		       fiPtr->hostName,
		       fiPtr->funcName,
		       fiPtr->nHandles);

	    fiPtr->rpcFuncHandleDeleted = 1;
	}
    }
    _ogr_debug(5, "exit GRPC function handles destruction.\n");

#ifndef USE_SPROC
    return NULL;
#endif /* USE_SPROC */
}


void
_ogr_final(void) {
    if (__ogr_Exiting == 0) {
	ogr_HashEntry *ePtr, *fePtr;
	ogr_HashSearch sctx, fsctx;
	ogrRpcHostInfo *hiPtr;
	ogrRpcFuncInfo *fiPtr;
	int i;
	int maxThd = omp_get_max_threads();
	int tIdx;
	int nFiList = 0;

	__ogr_Exiting = 1;

	_ogr_debug(10, "doing finalization.\n");

	grpc_wait_all();

	_ogr_start_timer("final");

	FOR_ENTRIES_IN_HASH(ePtr, &sctx, &hostTable) {
	    hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);
	    FOR_ENTRIES_IN_HASH(fePtr, &fsctx, &(hiPtr->hostFuncTable)) {
		fiPtr = (ogrRpcFuncInfo *)ogr_GetHashValue(fePtr);
		if (fiPtr != NULL) {
		    nFiList++;
		}
	    }
	}

	nFiPtrList = (int *)malloc(sizeof(int) * maxThd);
	memset((void *)nFiPtrList, 0, sizeof(int) * maxThd);

	fiPtrList = (ogrRpcFuncInfo ***)malloc(sizeof(ogrRpcFuncInfo **) *
					       maxThd);
	memset((void *)fiPtrList, 0, sizeof(ogrRpcFuncInfo **) * maxThd);

	for (i = 0; i < maxThd; i++) {
	    fiPtrList[i] = (ogrRpcFuncInfo **)malloc(
		sizeof(ogrRpcFuncInfo *) * nFiList);
	    memset((void *)fiPtrList[i], 0, 
		   sizeof(ogrRpcFuncInfo *) * nFiList);
	}
	
	i = 0;
	FOR_ENTRIES_IN_HASH(ePtr, &sctx, &hostTable) {
	    hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);

	    FOR_ENTRIES_IN_HASH(fePtr, &fsctx, &(hiPtr->hostFuncTable)) {
		fiPtr = (ogrRpcFuncInfo *)ogr_GetHashValue(fePtr);

		if (fiPtr != NULL) {
		    tIdx = i % maxThd;
		    fiPtrList[tIdx][nFiPtrList[tIdx]] = fiPtr;
		    nFiPtrList[tIdx]++;
		    i++;
		}
	    }
	}

	_ompc_do_parallel(delGFHs, NULL);

	(void)grpc_finalize();

	_ogr_stop_timer("final");

	_ogr_stat_final();

	_ogr_debug_final();
    }
}


static char **initFuncs = NULL;
static int nInitFuncs = 0;

static
#ifdef USE_SPROC
void
#else
void *
#endif /* USE_SPROC */
initGFHs(void *dum) {
    int tid = omp_get_thread_num();
    int i;
    int n;
    ogrRpcFuncInfo **fiPtrPtr;
    ogrRpcFuncInfo *fiPtr;
    ogrFuncHandle *fhPtr;

    n = nFiPtrList[tid];
    fiPtrPtr = &(fiPtrList[tid][0]);

    _ogr_debug(5, "enter GRPC function handles pre-creation.\n");

    for (i = 0; i < n; i++) {
	fiPtr = fiPtrPtr[i];
	fhPtr = findFuncHandle(fiPtr->funcName);
	ogrLockFuncInfo(fiPtr);
	(void)bindGRPCFuncHandle(fhPtr, fiPtr);
	ogrUnlockFuncInfo(fiPtr);
    }

    _ogr_debug(5, "exit GRPC function handles pre-creation.\n");

#ifndef USE_SPROC
    return NULL;
#endif /* !USE_SPROC */
}

void
_ogr_rpc_prologue(void) {
    char *envInitFuncs = getenv("OGR_PRE_INIT_FUNCS");

    if (envInitFuncs != NULL && envInitFuncs[0] != '\0') {
	FILE *fd = fopen(envInitFuncs, "r");
	if (fd != NULL) {
	    char buf[1024];
	    int len;
	    char *tokens[10];
	    int n;
	    int i;

	    while (fgets(buf, 1024, fd) != NULL) {
		len = strlen(buf);
		if (len == 1023 && buf[len - 1] != '\n') {
		    _ogr_warning("\"%s\": line too long.\n",
				 envInitFuncs);
		}
		n = GetToken(buf, tokens, 32);
		if (n == 0) {
		    continue;
		}
		if (tokens[0][0] == '#') {
		    continue;
		}

		for (i = 0; i < n; i++) {
		    if (initFuncs == NULL || nInitFuncs == 0) {
			initFuncs = (char **)malloc(sizeof(char *));
		    } else {
			initFuncs = 
			    (char **)realloc((void *)initFuncs,
					     sizeof(char *) * 
					     (nInitFuncs + 1));
		    }
		    initFuncs[nInitFuncs] = strdup(tokens[0]);
		    nInitFuncs++;
		}
	    }
	    fclose(fd);
	}

	if (initFuncs != NULL &&
	    nInitFuncs > 0) {
	    initFuncs[nInitFuncs] = NULL;

	    ogr_HashEntry *ePtr = NULL;
	    ogr_HashSearch sCtx;
	    ogrRpcHostInfo *hiPtr = NULL;
	    ogrRpcFuncInfo *fiPtr = NULL;
	    int nFiList = 0;
	    int maxThd = omp_get_max_threads();
	    int tIdx;
	    int i;
	    int j;

	    for (i = 0; i < nInitFuncs; i++) {
		(void)findFuncHandle(initFuncs[i]);
		FOR_ENTRIES_IN_HASH(ePtr, &sCtx, &hostTable) {
		    hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);
		    fiPtr = findRpcFuncInfo(hiPtr, initFuncs[i]);
		    nFiList++;
		}
	    }

	    nFiPtrList = (int *)malloc(sizeof(int) * maxThd);
	    memset((void *)nFiPtrList, 0, sizeof(int) * maxThd);

	    fiPtrList = (ogrRpcFuncInfo ***)malloc(sizeof(ogrRpcFuncInfo **) *
						   maxThd);
	    memset((void *)fiPtrList, 0, sizeof(ogrRpcFuncInfo **) * maxThd);

	    for (i = 0; i < maxThd; i++) {
		fiPtrList[i] = (ogrRpcFuncInfo **)malloc(
		    sizeof(ogrRpcFuncInfo *) * nFiList);
		memset((void *)fiPtrList[i], 0, 
		       sizeof(ogrRpcFuncInfo *) * nFiList);
	    }

	    j = 0;
	    for (i = 0; i < nInitFuncs; i++) {
		FOR_ENTRIES_IN_HASH(ePtr, &sCtx, &hostTable) {
		    hiPtr = (ogrRpcHostInfo *)ogr_GetHashValue(ePtr);
		    fiPtr = findRpcFuncInfo(hiPtr, initFuncs[i]);
		    tIdx = j % maxThd;
		    fiPtrList[tIdx][nFiPtrList[tIdx]] = fiPtr;
		    nFiPtrList[tIdx]++;
		    j++;
		}
	    }
	    _ogr_start_timer("grpc_init");
	    _ompc_do_parallel(initGFHs, NULL);
	    _ogr_stop_timer("grpc_init");
	}
    }
}
