/* 
 * $Id: ompd.h,v 1.1.1.1 2004-02-06 18:14:56 msato Exp $
 * $RWC_Release$
 * $RWC_Copyright$
 */

/* 
 * Omni OpenMP runtime for distributed memory
 */

#ifndef _OMPD_H
#define _OMPD_H


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#if !(defined(__USE_RHINET__) && defined(__USE_TRUN__))
#include "mpi.h"
#endif
#include "exc_platform.h"
#include "ompc_reduction.h"



#define SUCCESS			0
#define FAILED			1

#define TRUE 			1
#define FALSE 			0

#define MAX(a,b)		(((a)<(b))?(b):(a))
#define MIN(a,b)		(((a)<(b))?(a):(b))


#define MAX_PROC		32
#define PARALLEL_MAX_ARGS  	20


/* barrier operation */
#ifdef __USE_RHINET__
#define OMPD_INITIALIZE(ac,av)			_ompd_rhinet_init(ac,av)
#ifdef __USE_TRUN__
#define OMPD_FINALIZE()				_ompd_trun_barrier()
#define OMPD_TCP_BARRIER()			_ompd_trun_barrier()
#else /* __USE_TRUN__ */
#define OMPD_FINALIZE()				MPI_Finalize()
#define OMPD_TCP_BARRIER()			MPI_Barrier(_ompd_comm)
#endif /* __USE_TRUN__ */
#define OMPD_SERVER_INIT()			
#define OMPD_BARRIER()				_ompd_rhinet_barrier()
#define OMPD_WAIT_ALL()				_ompd_rhinet_barrier()
#define OMPD_FLUSH_ALL()			_ompd_flush_any()
#define OMPD_FLUSH_SYNC()			_rmem_sync()
#define OMPD_ENTER_CRITICAL(lp)			_ompd_rhinet_critical_lock((lp))
#define OMPD_EXIT_CRITICAL(lp)			_ompd_rhinet_critical_unlock((lp))
#define OMPD_DYNAMIC_SCHED_INIT(l,u,s,c)	_ompd_rhinet_dsched_init((l),(u),(s),(c))
#define OMPD_DYNAMIC_SCHED_NEXT(g,b,e)		_ompd_rhinet_dsched_next((g),(b),(e))
#define OMPD_ORDERED_INIT(l,s)			_ompd_rhinet_ordrd_init((l),(s))
#define OMPD_ORDERED_BEGIN(i)			_ompd_rhinet_ordrd_begin((i))
#define OMPD_ORDERED_END()			_ompd_rhinet_ordrd_end()
#define OMPD_REDUCTION(d,s,t,o)			_ompd_rhinet_reduction((d),(s),(t),(o))
#define OMPD_ATOMIC_LOCK()			_ompd_rhinet_atomic_lock()
#define OMPD_ATOMIC_UPDATE(a,s)			_ompd_rhinet_atomic_update((a),(s))
#define OMPD_BCAST(p,nb)			_ompd_rhinet_broadcast0((p),(nb))
#define OMPD_BCAST_WORD(p,nw)			_ompd_rhinet_broadcast0((char *)(p),(nw)*sizeof(int))

/* lock */
typedef int	_ompc_lock_t;
typedef int	_ompc_nest_lock_t;

#define	OMPSM_INIT_SLOCK(l)			_ompd_rhinet_lock_alloc ((l))
#define	OMPSM_DESTROY_SLOCK(l)			_ompd_rhinet_lock_free ((l))
#define	OMPSM_LOCK_SLOCK(l)			_ompd_rhinet_lock (*(l))
#define	OMPSM_UNLOCK_SLOCK(l)			_ompd_rhinet_unlock (*(l))
#define	OMPSM_TEST_SLOCK(l)			_ompd_rhinet_trylock (*(l))
#define	OMPSM_INIT_NLOCK(l)			_ompd_rhinet_lock_alloc ((l))
#define	OMPSM_DESTROY_NLOCK(l)			_ompd_rhinet_lock_free ((l))
#define	OMPSM_LOCK_NLOCK(l)			_ompd_rhinet_nlock (*(l))
#define	OMPSM_UNLOCK_NLOCK(l)			_ompd_rhinet_nunlock (*(l))
#define	OMPSM_TEST_NLOCK(l)			_ompd_rhinet_ntrylock (*(l))

#else /* __USE_RHINET__ */
#define OMPD_INITIALIZE(ac,av)			_ompd_mpi_init(ac,av)
#define OMPD_FINALIZE()				MPI_Finalize()
#define OMPD_SERVER_INIT()			_ompd_init_server ();
#define OMPD_BARRIER()				MPI_Barrier(_ompd_comm)
#define OMPD_WAIT_ALL()				MPI_Barrier(_ompd_comm)
#define OMPD_FLUSH_ALL()			_ompd_flush_any()
#define OMPD_FLUSH_SYNC()			
#define OMPD_ENTER_CRITICAL(lp)			_ompd_critical_lock((lp))
#define OMPD_EXIT_CRITICAL(lp)			_ompd_critical_unlock((lp))
#define OMPD_DYNAMIC_SCHED_INIT(l,u,s,c)	_ompd_loop_init_index((l))
#define OMPD_DYNAMIC_SCHED_NEXT(g,b,e)		_ompd_loop_next((g),(b),(e))
#define OMPD_ORDERED_INIT(l,s)			_ompd_init_ordered((l),(s))
#define OMPD_ORDERED_BEGIN(i)			_ompd_ordered_begin((i))
#define OMPD_ORDERED_END()			_ompd_ordered_end()
#define OMPD_REDUCTION(d,s,t,o)			_ompd_reduction((d),(s),(t),(o))
#define OMPD_ATOMIC_LOCK()			printf("not implement\n")
#define OMPD_ATOMIC_UPDATE(a,s)			printf("not implement\n")
#define OMPD_BCAST(p,nbytes)			MPI_Bcast(p,nbytes,MPI_BYTE,0,_ompd_comm)
#define OMPD_BCAST_WORD(p,nwords)		MPI_Bcast(p,nwords,MPI_INT,0,_ompd_comm)

/* lock */
typedef void	*_ompc_lock_t;
#if 0
typedef int	_ompc_nest_lock_t;
#endif
#define	OMPSM_INIT_SLOCK(l)			printf("not implement\n")
#define	OMPSM_DESTROY_SLOCK(l)			printf("not implement\n")
#define	OMPSM_LOCK_SLOCK(l)			printf("not implement\n")
#define	OMPSM_UNLOCK_SLOCK(l)			printf("not implement\n")
#define	OMPSM_TEST_SLOCK(l)			printf("not implement\n")
#define	OMPSM_INIT_NLOCK(l)			printf("not implement\n")
#define	OMPSM_DESTROY_NLOCK(l)			printf("not implement\n")
#define	OMPSM_LOCK_NLOCK(l)			printf("not implement\n")
#define	OMPSM_UNLOCK_NLOCK(l)			printf("not implement\n")
#define	OMPSM_TEST_NLOCK(l)			printf("not implement\n")

#endif /* __USE_RHINET__ */


enum sched_kind {
    SCHED_NONE = 0, 
    SCHED_STATIC, 
    SCHED_DYNAMIC, 
    SCHED_GUIDED 
};

/* shared data */
struct ompd_globals {
    int			debug_flag;
    int			ompd_mode;
    enum sched_kind	runtime_sched_kind;
    int 		runtime_chunk_size;
};

extern struct ompd_globals _ompd_globals;


/* thread invocation message */
typedef void* (*cfunc)();

extern struct ompd_thd_info {
    cfunc	func;
    int		narg;
    void	*_ompd_thd_args[PARALLEL_MAX_ARGS];
} _ompd_thd_info;


/* runtime library communicator */
#if !(defined(__USE_RHINET__) && defined(__USE_TRUN__))
extern MPI_Comm	_ompd_comm;
#endif
#ifndef __USE_RHINET__
extern char	_ompd_node_processor_name[MPI_MAX_PROCESSOR_NAME];
#endif /* __USE_RHINET__ */


/*
 * non-nested thread management
 */
extern int _ompc_n_node;
extern int _ompc_node_id;
#define NODE_IS_MASTER		(_ompc_node_id == 0)

extern int _ompc_debug_flag;
extern int _ompc_in_parallel; /* parallel nest level */
#define IN_PARALLEL		(_ompc_in_parallel != 0)
#define NOT_IN_PARALLEL		(_ompc_in_parallel == 0)
#define IN_NESTED_PARALLEL	(1<_ompc_in_parallel)

extern volatile int _ompc_nested;	/* nested enable/disable */
extern volatile int _ompc_dynamic;	/* dynamic enable/disable */
extern volatile int _ompc_max_threads;	/* max number of thread */
extern volatile int _ompc_num_threads;
extern int _ompc_n_proc;

extern enum sched_kind _ompc_runtime_sched_kind;
extern int _ompc_runtime_chunk_size; /* default */

#define SCHED_NONE 	0
#define SCHED_STATIC 	1
#define SCHED_DYNAMIC 	2
#define SCHED_GUIDED 	3
#define SCHED_RUNTIME 	4

/* ompd runtime mode */
#define OMPD_MODE_NONE 0
#define OMPD_NO_CACHE  1
#define OMPD_TDL_MODE  2

extern int _ompd_mode;

#ifdef __USE_RHINET__
enum locktype {
  LOCK_SYS = 1,
  LOCK_CRITICAL,
  LOCK_ORDERED,
  LOCK_ATOMIC,
  LOCK_TAIL
};
#endif


void		_ompc_fatal (char * msg);
void		_ompc_terminate (int);


/* remote memory access interface */
void		_rmem_init ();
int		_rmem_read_int (int *);
void		_rmem_write_int (int *,int);
void		_rmem_sync ();
void		_rmem_read (char *, int);
void		_rmem_write (char *, int);
void		_rmem_write_block (char *, int, int, int);
void		_rmem_read_block (char *, int, int, int);

void		_ompd_read_any (void *, int);
void		_ompd_write_any (void *, int);
void		_ompd_flush_any ();

void		_ompd_mpi_init (int *, char ***);

/* other interface */
#ifdef __USE_RHINET__
void		_ompd_rhinet_init (int *, char ***);

void		_ompd_rhinet_barrier ();

void		_ompd_rhinet_critical_lock (_ompc_lock_t *);
void		_ompd_rhinet_critical_unlock (_ompc_lock_t *);

void		_ompd_rhinet_dsched_init (int, int, int, int);
void		_ompd_rhinet_dsched_next (int, int *, int *);

void		_ompd_rhinet_ordrd_init (int, int);
void		_ompd_rhinet_ordrd_begin (int);
void		_ompd_rhinet_ordrd_end ();

void		_ompd_rhinet_atomic_lock ();
void		_ompd_rhinet_atomic_update (char *, int);

void		_ompd_rhinet_reduction (void *, void *, int, int);

void		_ompd_rhinet_broadcast0 (char *, int);

void		_ompd_rhinet_lock_alloc (int *);
void		_ompd_rhinet_lock_free (int *);
void		_ompd_rhinet_lock_init (int);
void		_ompd_rhinet_lock_destroy (int);
void		_ompd_rhinet_lock (int);
void		_ompd_rhinet_unlock (int);
int		_ompd_rhinet_trylock (int);
void		_ompd_rhinet_nlock_init (int);
void		_ompd_rhinet_nlock_destroy (int);
void		_ompd_rhinet_nlock (int);
void		_ompd_rhinet_nunlock (int);
int		_ompd_rhinet_ntrylock (int);

#else /* __USE_RHINET__ */
/* ompd server functions */
void		_ompd_init_server ();

void		_ompd_critical_lock (_ompc_lock_t *);
void		_ompd_critical_unlock (_ompc_lock_t *);

void		_ompd_loop_init_index (int);
void 		_ompd_loop_next (int, int *, int *);
void		_ompd_init_ordered (int, int);
void		_ompd_ordered_begin (int);
void		_ompd_ordered_end ();

void		_ompd_atomic_lock ();
void		_ompd_atomic_update (char *, int);

void		_ompd_reduction(void *, void *, int, int);

_ompc_lock_t	_ompd_lock_alloc (void);
void		_ompd_lock_free (_ompc_lock_t);

void		_ompd_lock ();
void		_ompd_unlock();

#ifdef __USE_TRUN__
int		_ompd_trun_barrier ();
#endif /* __USE_TRUN__ */

#endif /* __USE_RHINET__ */

#endif /* OMPD_H */
