static char rcsid[] = "$Id: ompd_RWcode.c,v 1.1.1.1 2004-02-06 18:14:57 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "ompd.h"
#include "ompd_server.h"
#include "ompd_object.h"

/* 
 * RW check codes, which are generated by the compiler.
 * - allocate descriptor:
 *    for variable,
 *      _ompd_D_v(desc_addr,var_addr,size)
 *    for array,
 *      _ompd_D_a(desc_addr,array_addr,size,elem_size,ndim,d1,d2,d3,d4)
 * - read/write check codes:
 *    access by pointer
 *      _ompd_R/W_p(context,address,size)
 *    access the entire data (variable/array)
 *      _ompd_R/W_v(context,desc)
 *    access the region of array
 *      _ompd_R/W_a(context,desc)
 * - setting access region:
 *    setting index
 *      _ompd_acc_index(dim,index)
 *    setting "for" range
 *      _ompd_acc_for(dim,lb,ub,step,dir)
 *    setting "forall" range
 *      _ompd_acc_forall(dim,lb,ub,step,dir,sched,chunk)
 *    set expression
 *      _ompd_add_expr(a,b,base,indx,offset)
 *    set for exprssion
 *      _ompd_add_for(lb_a,lb_b,lb_base,lb_indx,lb_offset,
 *                    ub_a,ub_b,ub_base,ub_indx,ub_offset,step,dir)
 *    end of setting region
 *      _ompd_acc_region(ndim)
 * - flush:
 *    flush for consistency
 *      _ompd_flush(desc_addr)
 *    flush for random access
 *      _ompd_flushR(desc_addr)
 */
#define ACC_RGN_WORK_SIZE  200
#define N_ACC_RGN_WORK    50
#define N_ACC_EXPR_WORK    50

/* tempoary to store region arguments */
static ompd_access_range _ompd_acc_range_work[ACC_RGN_WORK_SIZE];
static ompd_access_range *_ompd_acc_range_work_p = _ompd_acc_range_work;
static ompd_access_range *current_rp;

static int _ompd_acc_n_expr = 0;
static ompd_access_expr _ompd_acc_exprs[N_ACC_EXPR_WORK];

/* work space */
static int _ompd_acc_n_rgn = 0;
static ompd_access_range *_ompd_acc_rgns[N_ACC_RGN_WORK];

char *_ompd_malloc(int size)
{
    char *p;
    p = malloc(size);
    if(p == NULL) _ompc_fatal("ompd_malloc failed\n");
    bzero(p,size);
    return p;
}

/* 
 * allocate descriptor:
 */
void _ompd_D_v(ompd_obj **dpp, char *addr, int size)
{
    ompd_obj *dp;
    dp = (ompd_obj *)_ompd_malloc(sizeof(*dp));
    dp->ndim = 0; 	/* variable */
    dp->addr = addr;
    dp->size = size;
    dp->cached = dp->modified = FALSE;
    *dpp = dp;
}

void _ompd_D_a(ompd_obj **dpp, char *addr,int size,int elem_size,
	       int ndim,int d1,int d2,int d3,int d4)
{
    ompd_array *dp;
    int i,s;
	
    dp = (ompd_array *)_ompd_malloc(sizeof(*dp));
    dp->ndim = ndim;
    dp->size = size;
    dp->addr = addr;
    dp->modified = dp->cached = FALSE;

    dp->elem_size = elem_size;

    switch(ndim){
    default: 
	_ompc_fatal("ompd_array_alloc: ndim >= 4");
	break;
    case 4:
	dp->dim_size[3] = d4;
    case 3:
	dp->dim_size[2] = d3;
    case 2:
	dp->dim_size[1] = d2;
    case 1:
	dp->dim_size[0] = d1;
	break;
    }
    s = elem_size;
    for(i = 0; i < ndim; i++){
	dp->dim_esize[i] = s;
	s *= dp->dim_size[i];
    }

    if(_ompd_mode == OMPD_TDL_MODE){
	dp->type = OMPD_TDL_ARRAY;
	_ompd_init_TDL_array(dp);
#ifdef not
    } else if(_ompd_mode == OMPD_DIST_MODE){
	dp->type = OMPD_DIST_ARRAY;
	_ompd_init_darray_default(dp);
#endif
    } else {
	dp->type = OMPD_RMEM_ARRAY;
    }
    *dpp = (ompd_obj *)dp;
}

/* 
 * setting access region:
 */
void _ompd_acc_index(int dim,int index)
{
    ompd_access_range *rp = &(_ompd_acc_range_work_p[dim]);
    rp->type = ACCESS_INDEX;
    rp->lb = index;
    rp->expr = NULL;
}

void _ompd_acc_for(int dim,int lb,int ub,int step,int dir)
{
    ompd_access_range *rp = &(_ompd_acc_range_work_p[dim]);
    current_rp = rp;
    rp->type = ACCESS_FOR;
    rp->lb = lb;
    rp->ub = ub;
    rp->step = step;
    rp->dir = dir;
    rp->scale = 1;
    rp->offset = 0;
    rp->expr = NULL;
}

void _ompd_acc_forall(int dim,int lb,int ub,int step,int dir,
		      int sched,int chunk)
{
    ompd_access_range *rp = &(_ompd_acc_range_work_p[dim]);
    extern int _ompc_last_loop_b,_ompc_last_loop_e;

    current_rp = rp;
    rp->scale = 1;
    rp->offset = 0;
    rp->expr = NULL;

    switch(sched){
    case SCHED_STATIC:
	if(chunk > 0){	/* cyclic */
	    rp->type = ACCESS_FOR_CYCLIC;
	    rp->lb = lb;
	    rp->ub = ub;
	    rp->step = step;
	    rp->dir = dir;
	    rp->chunk = chunk;
	}
	/* fall through */
    case SCHED_NONE:	/* default schedule, block */
	rp->type = ACCESS_FOR_BLOCK;
	rp->lb = lb;
	rp->ub = ub;
	rp->step = step;
	rp->dir = dir;
	break;

    case SCHED_DYNAMIC:
    case SCHED_GUIDED:
    case SCHED_RUNTIME:
	/* use the last schedule */
	rp->type = ACCESS_FOR;
	rp->lb = _ompc_last_loop_b;
	rp->ub = _ompc_last_loop_e;
	rp->step = step;
	rp->dir = dir;
	break;
    }
}

void _ompd_add_expr(int a, int b, int *addr ,ompd_array *dp, int off)
{
    ompd_access_expr *ep,*eq;

    if(current_rp->expr == NULL && addr == NULL){
	current_rp->scale = a;
	current_rp->offset = b;
    } else {
	ep = &_ompd_acc_exprs[_ompd_acc_n_expr++];
	ep->is_for = FALSE;
	ep->a = a;
	ep->b = b;
	ep->addr = addr;
	ep->desc = dp;
	ep->off = off;
	ep->expr = NULL;
	if(current_rp->expr == NULL) current_rp->expr = ep;
	else {
	    for(eq = current_rp->expr; eq != NULL; eq = eq->expr){
		if(eq->expr == NULL){
		    eq->expr = ep;
		    break;
		}
	    }
	}
    }
}

void _ompd_add_for(int a, int b, int *addr,ompd_array *dp, int off,
		   int a1, int b1, int *addr1, ompd_array *dp1, int off1,
		   int step,int dir)
{
    ompd_access_expr *ep,*eq;
    ep = &_ompd_acc_exprs[_ompd_acc_n_expr++];

    ep->is_for = TRUE;

    ep->a = a;
    ep->b = b;
    ep->addr = addr;
    ep->desc = dp;
    ep->off = off;

    ep->a1 = a1;
    ep->b1 = b1;
    ep->addr1 = addr1;
    ep->desc1 = dp1;
    ep->off1 = off1;

    ep->step = step;
    ep->dir = dir;
    ep->expr = NULL;

    if(current_rp->expr == NULL) current_rp->expr = ep;
    else {
	for(eq = current_rp->expr; eq != NULL; eq = eq->expr){
	    if(eq->expr == NULL){
		eq->expr = ep;
		break;
	    }
	}
    }
}

void _ompd_acc_range(int ndim)
{
    /* save the current region */
    _ompd_acc_rgns[_ompd_acc_n_rgn++] = _ompd_acc_range_work_p;
    _ompd_acc_range_work_p += ndim;
}

/* 
 * read/write 
 */
/* for variable */
void _ompd_R_v(int context,ompd_obj *dp)
{
    enum ompd_array_type atype;
    
    if(_ompd_mode == OMPD_NO_CACHE){
	if(context == CTX_IN_MASTER) return;
	_rmem_read(dp->addr,dp->size);
	return;
    }

    /* variable is not distributed */
    if(dp->ndim == 0) atype = OMPD_RMEM_ARRAY;
    else atype = ((ompd_array *)dp)->type;

    switch(atype){
    case OMPD_RMEM_ARRAY:
	if(context == CTX_IN_MASTER) break;
	if(dp->size == 0) _ompc_fatal("_ompd_R_v, unknown size");
	if(dp->cached) return;
	dp->cached = TRUE;
	/* fetch remote memory */
	_rmem_read(dp->addr,dp->size);
	break;
#ifdef not /* not yet */
    case OMPD_DIST_ARRAY:
	switch(context){
	case CTX_IN_PARALLEL:
	    break;
	case CTX_IN_MASTER:
	     break; /* nothing */
	case CTX_IN_FORALL:
	default:
	    _ompc_fatal("_ompd_R_v: variable, bad context");
	}
	break;
#endif
    default:
	_ompc_fatal("_ompd_R_v: unknown type array");
    }
}

void _ompd_W_v(int context, ompd_obj *dp)
{
    if(_ompd_mode == OMPD_NO_CACHE){
	if(context == CTX_IN_MASTER) return;
	_rmem_write(dp->addr,dp->size);
	return;
    }

    if(dp->ndim == 0){
	dp->modified = TRUE;
	dp->cached = TRUE;
    } else
	_ompc_fatal("_ompd_W_v: ndim != 0 not implemented yet");
}

/* for array. */
void _ompd_R_a(int context, ompd_array *dp)
{
    int i, n_rgn;
    ompd_range rgn_args[MAX_DIM];	/* work space */
    enum ompd_array_type atype;

    /* reset working space */
    n_rgn = _ompd_acc_n_rgn;
    _ompd_acc_n_rgn = 0;
    _ompd_acc_n_expr = 0;
    _ompd_acc_range_work_p = _ompd_acc_range_work;

    if(_ompd_mode == OMPD_NO_CACHE){
	if(context == CTX_IN_MASTER) return;
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,_ompd_acc_rgns[i],ACCESS_RMEM_READ);
	return;
    }

    if(dp->is_random_acc_mode) atype == OMPD_RMEM_ARRAY;
    else atype = dp->type;

    switch(atype){
    case OMPD_RMEM_ARRAY:
	if(context == CTX_IN_MASTER) break;
	for(i = 0; i < n_rgn; i++){
	    if(_ompd_setup_rgn(dp->ndim,_ompd_acc_rgns[i],rgn_args))
		_ompd_read_rmem_array(dp,rgn_args);
	    else 
		_ompd_access_array(dp,_ompd_acc_rgns[i],FALSE);
	}
	break;

    case OMPD_TDL_ARRAY:
	switch(context){
	case CTX_IN_PARALLEL:
	    _ompd_read_TDL_array(dp,n_rgn,_ompd_acc_rgns);
	    break;
	case CTX_IN_MASTER:
	    /* store back all array into master */
	    _ompd_master_TDL_array(dp);
	     break; 
	case CTX_IN_FORALL:
	default:
	    _ompc_fatal("_ompd_R_a: bad context");
	}
	break;
#ifdef not /* not yet */
    case OMPD_DIST_ARRAY:
	switch(context){
	case CTX_IN_PARALLEL:
	    _ompd_read_dist_array(dp,n_rgn,_ompd_acc_rgns);
	    break;
	case CTX_IN_MASTER:
	    _ompd_master_dist_array(dp,n_rgn,_ompd_acc_rgns);
	     break; 
	case CTX_IN_FORALL:
	default:
	    _ompc_fatal("_ompd_R_a: bad context");
	}
#endif
    default:
	_ompc_fatal("_ompd_R_a: unknown type array");
    }
}

void _ompd_W_a(int context, ompd_array *dp)
{
    int i, n_rgn;
    ompd_range rgn_args[MAX_DIM];	/* work space */
    enum ompd_array_type atype;

    /* reset working space */
    n_rgn = _ompd_acc_n_rgn;
    _ompd_acc_n_rgn = 0;
    _ompd_acc_n_expr = 0;
    _ompd_acc_range_work_p = _ompd_acc_range_work;

    if(_ompd_mode == OMPD_NO_CACHE){
	if(context == CTX_IN_MASTER) return;
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,_ompd_acc_rgns[i],ACCESS_RMEM_WRITE);
	return;
    }

    if(dp->is_random_acc_mode) atype == OMPD_RMEM_ARRAY;
    else atype = dp->type;

    switch(atype){
    case OMPD_RMEM_ARRAY:
	if(context == CTX_IN_MASTER) break;
	for(i = 0; i < n_rgn; i++){
	    if(_ompd_setup_rgn(dp->ndim,_ompd_acc_rgns[i],rgn_args))
		_ompd_write_rmem_array(dp,rgn_args);
	    else
		_ompd_access_array(dp,_ompd_acc_rgns[i],ACCESS_RMEM_WRITE);
	}
	break;
    case OMPD_TDL_ARRAY:
	switch(context){
	case CTX_IN_PARALLEL:
	    _ompd_write_TDL_array(dp,n_rgn,_ompd_acc_rgns);
	    break;
	case CTX_IN_MASTER:
	    /* store back all array into master */
	    _ompd_master_TDL_array(dp);
	     break; 
	case CTX_IN_FORALL:
	default:
	    _ompc_fatal("_ompd_R_a: bad context");
	}
	break;
   default:
	_ompc_fatal("_ompd_W_a: unknown type array");
    }
}

/*
 * flush object
 */
void _ompd_flush(ompd_obj *dp)
{
    if(_ompd_mode == OMPD_NO_CACHE) return;	/* nothing */

    if(dp->ndim == 0){	/* variable */
	if(dp->modified){
	    _rmem_write(dp->addr,dp->size);
	    dp->modified = FALSE;
	}
	dp->cached = FALSE;
    } else {
	ompd_array *dq = (ompd_array *)dp;
	if(dq->is_random_acc_mode || dq->type == OMPD_RMEM_ARRAY)
	    _ompd_flush_rgn_array(dq,TRUE);
    }
    _rmem_sync();
}

void _ompd_flushR(ompd_obj *dp)
{
    if(_ompd_mode == OMPD_NO_CACHE) return;	/* nothing */

    _ompc_fatal("_ompd_flushR: not implemented yet");
}

/*
 * schedule for ompd
 */
void _ompd_block_sched(int *b_p,int *e_p,int s)
{
    int b,e,blk_s;
    int id,n_thd;

    id = _ompc_node_id;
    n_thd = _ompc_n_node;
    b = *b_p;
    e = *e_p;
    if(s > 0){
	blk_s = (e-b+n_thd-1)/n_thd;
	blk_s = ((blk_s+s-1)/s)*s;
	b += blk_s*id;
	e = b + blk_s;
	if(e >= *e_p) e = *e_p;
    } else if(s < 0){
	blk_s = (e-b-n_thd+1)/n_thd;
	blk_s = ((blk_s+s+1)/s)*s;
	b += blk_s*id;
	e = b + blk_s;
	if(e <= *e_p) e = *e_p;
    } else return;	/* s == 0 ??? */
    *b_p = b;
    *e_p = e;
}

void _ompd_cyclic_sched(int *b_p,int *chunk_p,int *s_p)
{
    int b,c,s;
    int id,n_thd;

    id = _ompc_node_id;
    n_thd = _ompc_n_node;
    b = *b_p;
    s = *s_p;
    c = *chunk_p*s;	/* scale chunk with step */
    *b_p = b + id*c;
    *s_p = s*n_thd;	/* return stride */
    *chunk_p = c;
}

/*
 * For setting region
 */
int _ompd_setup_rgn(int ndim,ompd_access_range *ap, ompd_range *rp)
{
    int i,b,e,s,offset,scale,stride,chunk;
    int id,n_thd;

    for(i = 0; i < ndim; i++, ap++, rp++){
	if(ap->expr != NULL) return FALSE; /* complex region */
	switch(ap->type){
	case ACCESS_INDEX:
	    rp->lb = ap->lb;
	    rp->step = 0; /* if step == 0, point */
	    break;

	case ACCESS_FOR:
	    scale = ap->scale;
	    offset = ap->offset;
	    b = rp->lb = scale*ap->lb+offset;
	    e = rp->ub = scale*ap->ub+offset;
	    s = ap->step*scale;
	    if(s != 1){
		rp->stride = s;
		rp->blk_n = 1;
	    } else {
		rp->stride = 0;		/* single block */
		rp->blk_n = e-b;
	    }
	    rp->step = 1;		/* no scale */
	    break;

	case ACCESS_FOR_BLOCK:
	    b = ap->lb;
	    e = ap->ub;
	    s = ap->step;
	    scale = ap->scale;
	    offset = ap->offset;
	    if(s == 0) return;
	    _ompd_block_sched(&b,&e,s);
	    rp->lb = b*scale+offset;
	    rp->ub = e*scale+offset;
	    s = scale*s;
	    if(s != 1){
		rp->stride = s;
		rp->blk_n = 1;
	    } else {
		rp->stride = 0;	/* signle block */
		rp->blk_n = rp->ub-rp->lb;
	    }
	    rp->step = 1;
	    break;

	case ACCESS_FOR_CYCLIC:
	    b = ap->lb;
	    e = ap->ub;
	    s = ap->step;
	    scale = ap->scale;
	    offset = ap->offset;
	    stride = s;
	    chunk = ap->chunk;
	    _ompd_cyclic_sched(&b,&chunk,&stride);
	    /* scaling */
	    rp->lb = scale*b+offset;
	    rp->ub = scale*e+offset;
	    rp->stride = scale*stride;
	    s = s*scale;
	    if(s == 1){
		rp->step = 1;	/* no scale */
		rp->blk_n = chunk;
	    } else {
		rp->step = s;
		rp->blk_n = chunk;
	    }
	    break;
	}
    }
    return TRUE;
}

/*
 * for debug 
 */
static ompd_range rgn_work[MAX_DIM];

void _ompd_set_range(int dim_i,int lb,int ub,int stride, int blk_n)
{
    ompd_range *rp = &rgn_work[dim_i];

    rp->lb = lb;
    rp->ub = ub;
    if(stride == blk_n){
	rp->stride = 0;
	rp->blk_n = rp->ub-rp->lb;
    } else {
	rp->stride = stride;
	rp->blk_n = blk_n;
    }
    rp->step = 1;
}

void _ompd_read_array_no_cache(ompd_obj *dp)
{
    _ompd_array_rmem_block((ompd_array *)dp,rgn_work,TRUE);
}

void _ompd_write_array_no_cache(ompd_obj *dp)
{
    _ompd_array_rmem_block((ompd_array *)dp,rgn_work,FALSE);
}





