static char rcsid[] = "$Id: ompd_array.c,v 1.1.1.1 2004-02-06 18:14:57 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "ompd.h"
#include "ompd_object.h"

/* prototype */

static enum rgn_overlap _ompd_rgn_overlap(int ndim,
					  ompd_range *rp,ompd_range *rq);

/* region overlap status */
enum rgn_overlap { 
    RGN_P,	/* p include q, or equal */
    RGN_Q, 	/* q include p */
    RGN_OVERLAP, /* regions may overlap */
    RGN_EMPTY 	/* empty, regions may not overlap */
};

void _ompd_read_rgn_array(ompd_array *dp,ompd_range *rgn_arg)
{
    int i;
    enum rgn_overlap r;

    if(dp->modified){
	r = _ompd_rgn_overlap(dp->ndim,dp->rgn_modified,rgn_arg);
	if(r == RGN_P) return;
	if(r != RGN_EMPTY) _ompd_flush_rgn_array(dp,FALSE);
    }
    if(dp->cached){
	r = _ompd_rgn_overlap(dp->ndim,dp->rgn_cached,rgn_arg);
	if(r == RGN_P) return;
    }
    
    _ompd_array_rmem_block(dp,rgn_arg,TRUE);

    for(i = 0; i < dp->ndim; i++) dp->rgn_cached[i] = rgn_arg[i];
}

void _ompd_write_rgn_array(ompd_array *dp,ompd_range *rgn_arg)
{
    int i;
    enum rgn_overlap r;

    if(dp->modified){
	r = _ompd_rgn_overlap(dp->ndim,dp->rgn_modified,rgn_arg);
	if(r == RGN_P) return; 
	if(r != RGN_Q) _ompd_flush_rgn_array(dp,FALSE);
    }
    dp->modified = TRUE;
    for(i = 0; i < dp->ndim; i++) dp->rgn_modified[i] = rgn_arg[i];
}

/* flush modified part to remote memory */
void _ompd_flush_rgn_array(ompd_array *dp,int flush_all)
{
    int i;

    if(dp->modified){
	_ompd_array_rmem_block(dp,dp->rgn_modified,FALSE);
	if(!flush_all && dp->cached &&
	   _ompd_rgn_overlap(dp->ndim,dp->rgn_modified,dp->rgn_cached)
	   == RGN_P){
	    for(i = 0; i < dp->ndim; i++) 
		dp->rgn_cached[i] = dp->rgn_modified[i];
	}
	dp->modified = FALSE;
    }
    if(flush_all) dp->cached = FALSE;
}

static struct {
    ompd_array *desc;
    ompd_range *rgn;
    int rw;
    int blocked_dim;
    int base,blk_size,stride,count;
    int last_offset,last_size;	/* for remainder block */
} _rmem_block_cmd;

static void _ompd_array_rmem_block_cmd(int dim,char *addr);

/* 
 * schedule remote memory access communcation
 * using stride memory access 
 */
void _ompd_array_rmem_block(ompd_array *dp,ompd_range *rgn_arg, int rw)
{
    int i,k;
    ompd_range *rp;
    int base,count,blk_size,stride,esize;
    int last_size,last_offset;

    /* 
     * schedule 
     */
    _rmem_block_cmd.desc = dp;
    _rmem_block_cmd.rgn = rgn_arg;
    _rmem_block_cmd.rw = rw;

    /* init, sub-region is element. */
    base = 0;
    blk_size = dp->elem_size;
    stride = 0;
    count = 1;
    last_size = 0;
    last_offset = 0;
    
#ifdef not
    { 
	int i;
	printf("_ompd_array_rmem_block: ndim=%d\n",dp->ndim);
	for(i = 0; i < dp->ndim; i++)
	    printf("\tdim=%d,lb=%d,ub=%d,stride=%d,blk_n=%d,step=%d\n",
		   i,rgn_arg[i].lb,rgn_arg[i].ub,
		   rgn_arg[i].stride,rgn_arg[i].blk_n,rgn_arg[i].step);
    }
#endif

    for(i = 0; i < dp->ndim; i++){
	rp = &rgn_arg[i];
	esize = dp->dim_esize[i];

	if(rp->step == 0){ 	/* point, just change base */
	    base += rp->lb*esize;
	    continue;
	}
	if(count != 1) break;	

	/*
	 * sub-region must be single-block 
	 */

	if(rp->step != 1){
	    /* scaling in the block, give up!!! */
	    break;
	}

	if(base == 0 && blk_size == esize){
	    /* sub-region is just the same as element */
	    base = rp->lb*esize;
	    blk_size = esize*rp->blk_n;
	    if(rp->stride != 0){ /* not single block */
		count = (rp->ub-rp->lb)/rp->stride;
		k = rp->ub - (count*rp->stride+rp->lb);
		if(k > 0){
		    if(k >= rp->blk_n) count++;
		    else {
			last_size = k*esize;
			last_offset = count*stride;
		    }
		} 
		stride = rp->stride*esize;
	    }
	    /* printf("marge1: base=%d, blk_size=%d, stride=%d,count=%d\n",
		   base,blk_size,stride,count); */
	} else if(rp->stride == 0){	/* single block */
	    base += rp->lb*esize;
	    count = rp->blk_n;
	    stride = esize;
	    /* printf("marge2: base=%d, blk_size=%d, stride=%d,count=%d\n",
		   base,blk_size,stride,count); */
	} else break;
    }

    _rmem_block_cmd.blocked_dim = i-1;
    _rmem_block_cmd.base = base;
    _rmem_block_cmd.count = count;
    _rmem_block_cmd.blk_size = blk_size;
    _rmem_block_cmd.stride = stride;
    _rmem_block_cmd.last_offset = last_offset;
    _rmem_block_cmd.last_size = last_size;

#ifdef not
    printf("block_cmd[%d]:base=%d,count=%d,blk_n=%d,stride=%d,last=(%d,%d)\n",
	   i-1,base,count,blk_size,stride,last_offset,last_size);
#endif
    /* execute command! */
    _ompd_array_rmem_block_cmd(dp->ndim-1,dp->addr);
}

static void _ompd_array_rmem_block_cmd(int dim,char *addr)
{
    int i,j,k,esize;
    ompd_range *rp;

    if(dim < 0){
	esize = _rmem_block_cmd.desc->elem_size;
	if(_rmem_block_cmd.rw) _rmem_read(addr,esize);
	else _rmem_write(addr,esize);
    } else if(dim == _rmem_block_cmd.blocked_dim){
	addr +=_rmem_block_cmd.base;
	if(_rmem_block_cmd.rw){
	    _rmem_read_block(addr,_rmem_block_cmd.count,
			     _rmem_block_cmd.blk_size,
			     _rmem_block_cmd.stride);
	    if(_rmem_block_cmd.last_size > 0)
		_rmem_read_block(addr+_rmem_block_cmd.last_offset,
				 1,_rmem_block_cmd.last_size,0);
	} else {
	    _rmem_write_block(addr,_rmem_block_cmd.count,
			      _rmem_block_cmd.blk_size,
			      _rmem_block_cmd.stride);
	    if(_rmem_block_cmd.last_size > 0)
		_rmem_write_block(addr+_rmem_block_cmd.last_offset,
				  1,_rmem_block_cmd.last_size,0);
	}
    } else {
	rp = &(_rmem_block_cmd.rgn[dim]);
	esize = _rmem_block_cmd.desc->dim_esize[dim];
	if(rp->step == 0) /* point */
	    _ompd_array_rmem_block_cmd(dim-1,addr+esize*rp->lb);
	else if(rp->stride == 0){	/* single block */
	    addr += rp->lb*esize;
	    for(i = 0; i < rp->blk_n; i++)
		_ompd_array_rmem_block_cmd(dim-1,addr+i*rp->step*esize);
	} else {
	    for(j = rp->lb; j < rp->ub; j += rp->stride){
		for(i = 0; i < rp->blk_n; i++){
		    k = j+i*rp->step;
		    if(k >= rp->ub) break;
		    _ompd_array_rmem_block_cmd(dim-1,addr+k*esize);
		}
	    }
	}
    }
}

/* compute region overlap : rp-rq */
static enum rgn_overlap _ompd_rgn_overlap(int ndim,
					  ompd_range *rp,ompd_range *rq)
{
    int i;
    int include_p,include_q;
    ompd_range *p,*q;

    include_p = TRUE;
    include_q = TRUE;
    for(i = 0; i < ndim; i++){
	p = &rp[i];
	q = &rq[i];
	if(p->step == 0){
	    if(q->step == 0){
		if(p->lb == q->lb) continue;
		else return RGN_EMPTY;
	    } else {  /* p is signle-element, q is range */
		if(q->stride == 0 && q->step == 1){
		    if(p->lb >= q->lb && p->lb < q->ub){
			/* q include p */
			include_p = FALSE;
			continue;
		    } /* else */
		    return RGN_EMPTY;
		} /* else */
		return RGN_OVERLAP;
	    }
	} else if(q->step == 0){
	    /* q is single-element, p is range */
	    if(p->stride == 0 && p->step == 1){
		    if(q->lb >= p->lb && q->lb < p->ub){
			/* p include q */
			include_q = FALSE;
			continue;
		    } /* else */
		    return RGN_EMPTY;
	    } /* else */
	    return RGN_OVERLAP;
	}

	if(p->stride != 0 || p->step != 1 ||
	   q->stride != 0 || q->step != 1) 
	    return RGN_OVERLAP;  /* may be overlap */

	if(p->ub <= q->lb || p->lb >= q->ub) 
	    return RGN_EMPTY;
	if(p->lb <= q->lb && q->ub <= p->ub){ /* p include q */
	    include_q = FALSE;
	} else if(q->lb <= p->lb && p->ub <= q->ub){ /* q include p */
	    include_p = FALSE;
	} else return RGN_OVERLAP;
    }
    if(include_p) return RGN_P;
    else if(include_q) return RGN_Q;
    return RGN_OVERLAP;
}


/*
 * read/write array using rmem
 */
static void _ompd_array_rmem_rec(int dim,char *addr,
				 ompd_array *dp,ompd_range *rgn_arg,int rw);

void _ompd_read_rmem_array(ompd_array *dp,ompd_range *rgn_arg)
{
#ifdef not /* debug */
    { 
	int i;
	printf("_ompd_read_rmem_array: ndim=%d\n",dp->ndim);
	for(i = 0; i < dp->ndim; i++)
	    printf("\tdim=%d,lb=%d,ub=%d,stride=%d,blk_n=%d,step=%d\n",
		   i,rgn_arg[i].lb,rgn_arg[i].ub,
		   rgn_arg[i].stride,rgn_arg[i].blk_n,rgn_arg[i].step);
    }
#endif
    _ompd_array_rmem_rec(dp->ndim-1,dp->addr,dp,rgn_arg,TRUE);
}

void _ompd_write_rmem_array(ompd_array *dp,ompd_range *rgn_arg)
{
    _ompd_array_rmem_rec(dp->ndim-1,dp->addr,dp,rgn_arg,FALSE);
}

static void _ompd_array_rmem_rec(int dim,char *addr,
				 ompd_array *dp,ompd_range *rgn_arg,int rw)
{
    int i,j,k,esize;
    ompd_range *rp;

    if(dim < 0){
	if(rw) _rmem_read(addr,dp->elem_size);
	else _rmem_write(addr,dp->elem_size);
    } else {
	rp = &rgn_arg[dim];
	esize = dp->dim_esize[dim];
	if(rp->step == 0) /* point */
	    _ompd_array_rmem_rec(dim-1,addr+esize*rp->lb,dp,rgn_arg,rw);
	else if(rp->stride == 0){	/* single block */
	    addr += rp->lb*esize;
	    for(i = 0; i < rp->blk_n; i++)
		_ompd_array_rmem_rec(dim-1,addr+i*rp->step*esize,
				     dp,rgn_arg,rw);
	} else {
	    for(j = rp->lb; j < rp->ub; j += rp->stride){
		for(i = 0; i < rp->blk_n; i++){
		    k = j+i*rp->step;
		    if(k >= rp->ub) break;
		    _ompd_array_rmem_rec(dim-1,addr+k*esize,dp,rgn_arg,rw);
		}
	    }
	}
    }
}



