static char rcsid[] = "$Id: ompd_array_access.c,v 1.1.1.1 2004-02-06 18:14:57 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "ompd.h"
#include "ompd_object.h"

static void _ompd_access_array_dim(ompd_array *dp,
				   int dim,char *base,
				   ompd_access_range *rp);
static void _ompd_access_array_expr(ompd_access_expr *ep,int indx,
				    ompd_array *dp,
				    int dim,char *base,
				    ompd_access_range *rp);


static int access_mode;

void _ompd_access_array(ompd_array *dp,
			ompd_access_range *rp,
			int mode)
{
    access_mode = mode;

    /*
    printf("%d: access_mode=%d, base=0x%x\n",_ompc_node_id,mode,dp->addr);
    print_access_range(stdout,dp->ndim, rp);
    fflush(stdout);
    */

    _ompd_access_array_dim(dp,dp->ndim,dp->addr,rp);
}

static void _ompd_access_array_dim(ompd_array *dp,
				   int dim,char *base,
				   ompd_access_range *rpp)
{
    int indx,b,e,s,chunk,stride,ee,bb,k;
    ompd_access_range *rp;

    if(dim == 0){	/* access element */
	/* printf("%d: access addr=0x%x mode=%d\n",
	   _ompc_node_id,base,access_mode);*/
	switch(access_mode){
	case ACCESS_RMEM_READ:
	    _rmem_read(base,dp->elem_size);
	    return;
	case ACCESS_RMEM_WRITE:
	    _rmem_write(base,dp->elem_size);
	    return;
	case ACCESS_INSPECT_HOME:
	    _ompd_TDL_inspect_home(base,dp->elem_size);
	    return;
	case ACCESS_INSPECT_SHARING:
	    _ompd_TDL_inspect_sharing(base,dp->elem_size);
	    return;
	case ACCESS_INSPECT_REF:
	    _ompd_TDL_inspect_ref(base,dp->elem_size);
	    return;
	case ACCESS_INSPECT_MODIFY:
	    _ompd_TDL_inspect_modify(base,dp->elem_size);
	    return;
	default:
	    _ompc_fatal("_ompd_access_array: bad mode");
	}
    }
    dim--;
    rp = rpp+dim;
    /* printf("%d: dim access type=%d, base=0x%x, esize=%d\n",
	   _ompc_node_id,rp->type,base,dp->dim_esize[dim]); */
    switch(rp->type){
    case ACCESS_INDEX:
	indx = rp->lb;
	_ompd_access_array_dim(dp,dim,
			       base+dp->dim_esize[dim]*indx,rpp);
	break;
	
    case ACCESS_FOR:
	b = rp->lb;
	e = rp->ub;
	s = rp->step;
	goto do_loop;

    case ACCESS_FOR_BLOCK:
	b = rp->lb;
	e = rp->ub;
	s = rp->step;
	_ompd_block_sched(&b,&e,s);
	if(rp->expr == NULL){
	    b = rp->scale*b + rp->offset;
	    e = rp->scale*e + rp->offset;
	    s = rp->scale*s;
	}
    do_loop:
	if(rp->dir){
	    for(indx = b; indx < e; indx += s)
		_ompd_access_array_expr(rp->expr,indx,dp,dim,base,rpp);
	} else {
	    for(indx = b; indx > e; indx += s)
		_ompd_access_array_expr(rp->expr,indx,dp,dim,base,rpp);
	}
	break;

    case ACCESS_FOR_CYCLIC:
	b = rp->lb;
	e = rp->ub;
	s = rp->step;
	chunk = rp->chunk;
	stride = s;
	_ompd_cyclic_sched(&b,&chunk,&stride);
	/* scale ??? */
	k = b;
	while(1){
	    bb = k;
	    ee = k+chunk;
	    if(rp->dir){
		if(ee > e) ee = e;
		for(indx = bb; indx < ee; indx += s)
		    _ompd_access_array_expr(rp->expr,indx,dp,dim,base,rpp);
	    } else {
		if(ee < e) ee = e;
		for(indx = bb; indx > ee; indx += s)
		    _ompd_access_array_expr(rp->expr,indx,dp,dim,base,rpp);
	    }
	    if(ee == e) return;
	}
	break;
    }
}

static void _ompd_access_array_expr(ompd_access_expr *ep,int indx,
				    ompd_array *dp,
				    int dim,char *base,
				    ompd_access_range *rpp)
{
    int lb,ub,i;

    /* printf("%d: expr access indx=%d\n",_ompc_node_id,indx); */
    if(ep != NULL){
	if(ep->is_for){
	    lb = ep->a*indx+ep->b;
	    if(ep->addr != NULL) lb = ep->addr[lb]+ep->off;
	    ub = ep->a1*indx+ep->b1;
	    if(ep->addr1 != NULL) ub = ep->addr1[ub]+ep->off1;
	    if(ep->dir){
		for(i = lb; i < ub; i += ep->step)
		    _ompd_access_array_expr(ep->expr,i,dp,dim,base,rpp);
	    } else {
		for(i = lb; i > ub; i += ep->step)
		    _ompd_access_array_expr(ep->expr,i,dp,dim,base,rpp);
	    }
	    return;
	}
	indx = ep->a*indx+ep->b;
	indx = ep->addr[indx]+ep->off;
    }
    _ompd_access_array_dim(dp,dim,
			   base+dp->dim_esize[dim]*indx,rpp);
}

/* 
 * copy and compare 
 */
static ompd_access_expr *ompd_copy_access_expr(ompd_access_expr *ep);

ompd_access_range **ompd_copy_access_range(int n_rgn,ompd_access_range *rp[])
{
    int i;
    ompd_access_range **rq;

    rq = (ompd_access_range **)_ompd_malloc(n_rgn*sizeof(*rp));
    bcopy(rp,rq,n_rgn*sizeof(*rq));
    for (i = 0; i < n_rgn; ++i) {
	rq[i] = (ompd_access_range *)_ompd_malloc(sizeof(**rp));
	bcopy(rp[i],rq[i],sizeof(**rq));
	if(rp[i]->expr != NULL)
	    rq[i]->expr = ompd_copy_access_expr(rp[i]->expr);
    }
    return rq;
}

static ompd_access_expr *ompd_copy_access_expr(ompd_access_expr *ep)
{
    ompd_access_expr *eq;
    eq = (ompd_access_expr *)_ompd_malloc(sizeof(*ep));
    bcopy(ep,eq,sizeof(*ep));
    if(eq->expr != NULL) eq->expr = ompd_copy_access_expr(ep->expr);
    return ep;
}

static int
ompd_compare_access_expr(ompd_access_expr *ep,ompd_access_expr *eq);

int ompd_compare_access_range(ompd_access_range *rp,ompd_access_range *rq)
{
    if(rp->type == rq->type &&
       rp->lb == rq->lb &&  rp->ub == rq->ub && rp->step == rq->step &&
       rp->dir == rq->dir && rp->chunk == rp->chunk){
	if(rp->expr == rq->expr){
	    if(rp->expr == NULL)
		return (rp->scale == rq->scale && rp->offset == rq->offset);
	    else return FALSE;
	} else
	    return ompd_compare_access_expr(rp->expr,rq->expr);
    } else return FALSE;
}

static int
ompd_compare_access_expr(ompd_access_expr *ep,ompd_access_expr *eq)
{
    if(ep == NULL || eq == NULL){
	return (ep == eq);
    } else {
	if(ep->is_for != eq->is_for ||
	   ep->a != eq->a || ep->b != eq->b ||
	   ep->desc != eq->desc) return FALSE;
	if(ep->desc != NULL &&  ep->off != eq->off) return FALSE;
	if(!ep->is_for) return TRUE;
	if(ep->a1 != eq->a1 || ep->b1 != eq->b1 ||
	   ep->desc1 != eq->desc1) return FALSE;
	if(ep->desc1 != NULL &&  ep->off1 != eq->off1) return FALSE;
	return ompd_compare_access_expr(ep->expr,eq->expr);
    }
}

static void
ompd_free_access_expr(ompd_access_expr *ep);

void
ompd_free_access_range(int n_rgn, ompd_access_range *rp[])
{
    int i;
    for (i = 0; i < n_rgn; ++i) {
	if(rp[i]->expr != NULL)
	    ompd_free_access_expr(rp[i]->expr);
	free(rp[i]);
    }
    free(rp);
}

static void
ompd_free_access_expr(ompd_access_expr *ep)
{
    if(ep->expr != NULL)
	ompd_free_access_expr(ep->expr);
    free(ep);
}

void print_access_range(FILE *fp,int n_dim, ompd_access_range *rp)
{
    int i;

    fprintf(fp,"{n_dim=%d: ",n_dim);
    for(i = 0; i < n_dim; i++,rp++){
	fprintf(stdout,"\n");
	switch(rp->type){
	case ACCESS_INDEX:  
	    fprintf(fp,"\t[INDEX: %d]",rp->lb);
	    break;
	case ACCESS_FOR:
	    fprintf(fp,"\t[FOR (%d:%d:%d,%d)",rp->lb,rp->ub,
		    rp->step,rp->dir);
	    print_access_expr(fp,rp->expr);
	    fprintf(fp,"]");
	    break;
	case ACCESS_FOR_BLOCK:
	    fprintf(fp,"\t[FORALL_BLOCK (%d:%d:%d,%d)",rp->lb,rp->ub,
		    rp->step,rp->dir);
	    print_access_expr(fp,rp->expr);
	    fprintf(fp,"]");
	    break;
	case ACCESS_FOR_CYCLIC:
	    fprintf(fp,"\t[FORALL_CYCLIC (%d:%d:%d,%d,%d)",rp->lb,rp->ub,
		    rp->step,rp->dir,rp->chunk);
	    print_access_expr(fp,rp->expr);
	    fprintf(fp,"]");
	    break;
	}
    }
    fprintf(fp,"}\n");
}

void print_access_expr(FILE *fp,ompd_access_expr *ep)
{
    if(ep == NULL){
	fprintf(fp,"<null>");
	return;
    }
    if(ep->is_for){
	fprintf(fp,"\n\t\t<for (%d,%d,0x%x+%d):(%d,%d,0x%x+%d):%d,%d ",
		ep->a,ep->b,ep->addr,ep->off,
		ep->a1,ep->b1,ep->addr1,ep->off1,
		ep->step,ep->dir);
	print_access_expr(fp,ep->expr);
	fprintf(fp,">");
    } else {
	fprintf(fp,"\n\t\t<(%d,%d,0x%x+%d)>",ep->a,ep->b,ep->addr,ep->off);
    }
}

