/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
#include <stdlib.h>
#include "ompd.h"
#include "omp.h"

/*
 * OMP standard library function
 */

int 
omp_get_thread_num ()
{
  if ((NOT_IN_PARALLEL) || (IN_NESTED_PARALLEL)) {
    return 0;
  } else {
    return _ompc_node_id;
  }
}


int 
omp_get_thread_num_ ()
{
  return omp_get_thread_num (); 
}


int
omp_get_num_threads ()
{
  if ((NOT_IN_PARALLEL) || (IN_NESTED_PARALLEL)) {
    return 1;
  } else {
    return _ompc_n_node;
  }
}


int	
omp_get_num_threads_ ()
{
  return omp_get_num_threads (); 
}


void
omp_set_num_threads (int num)
{
  if(num <= 0){
    fprintf (stderr, "omp_set_num_threads: argument must be positive.\n");
    _ompc_fatal ("ompc_set_num_threads");
  }
  if ( _ompc_max_threads < num ) {
    fprintf (stderr, "set fail: %d exceeds max parallel number %d\n",
	     num, _ompc_max_threads);
    _ompc_fatal ("ompc_set_num_threads");
  }
  fprintf (stderr, "cannot change the number of thread in this mode\n");
}


void 
omp_set_num_threads_ (int *num)
{
  omp_set_num_threads (*num); 
}


int
omp_get_max_threads ()
{
  return _ompc_num_threads;
}


int
omp_get_max_threads_ ()
{
  return omp_get_max_threads (); 
}


int
omp_get_num_procs ()
{
    return _ompc_n_proc;
}


int
omp_get_num_procs_ ()
{
  return omp_get_num_procs (); 
}


int
omp_in_parallel ()
{
  return (IN_PARALLEL);
}


int
omp_in_parallel_ ()
{	
  return omp_in_parallel ();
}


void
omp_set_dynamic (int dynamic_thds)
{
  _ompc_dynamic = dynamic_thds;
  if (dynamic_thds) {
    fprintf (stderr,"dynamic allocation is not supported\n");
  }
}


void
omp_set_dynamic_ (int *dynamic_thds)
{
  omp_set_dynamic (*dynamic_thds); 
}


int
omp_get_dynamic ()
{
  return 0 /*_ompc_dynamic*/;
}


int 
omp_get_dynamic_() 
{
  return omp_get_dynamic (); 
}


void
omp_set_nested (int n_nested)
{
  _ompc_nested = n_nested;
  if (n_nested) {
    fprintf (stderr,"nested parallelism is not supported\n");
  }
}


void 
omp_set_nested_ (int *n_nested)
{
  omp_set_nested (*n_nested); 
}


int
omp_get_nested ()
{
    return 0 /*_ompc_nested */; 
}


int omp_get_nested_ () 
{
  return omp_get_nested (); 
}


/*
 * Lock Functions
 */
void
omp_init_lock (omp_lock_t *lock)
{
  OMPSM_INIT_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_init_lock_ (_omf77Lock_t *lock)
{
  OMPSM_INIT_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_destroy_lock (omp_lock_t *lock)
{
  OMPSM_DESTROY_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_destroy_lock_ (_omf77Lock_t *lock)
{
  OMPSM_DESTROY_SLOCK ((_ompc_lock_t *)lock);
}


void 
omp_set_lock (omp_lock_t *lock)
{
  OMPSM_LOCK_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_set_lock_ (_omf77Lock_t *lock)
{
  OMPSM_LOCK_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_unset_lock (omp_lock_t *lock)
{
  OMPSM_UNLOCK_SLOCK ((_ompc_lock_t *)lock);
}


void
omp_unset_lock_ (_omf77Lock_t *lock)
{
  OMPSM_UNLOCK_SLOCK ((_ompc_lock_t *)lock);
}


int
omp_test_lock (omp_lock_t *lock)
{
  return OMPSM_TEST_SLOCK ((_ompc_lock_t *)lock);
}


int
omp_test_lock_ (_omf77Lock_t *lock)
{
  return OMPSM_TEST_SLOCK ((_ompc_lock_t *)lock);
}


void 
omp_init_nest_lock (omp_nest_lock_t *lock)
{
  OMPSM_INIT_NLOCK ((_ompc_nest_lock_t *)lock);
}

void 
omp_init_nest_lock_ (_omf77Lock_t *lock)
{
  OMPSM_INIT_NLOCK ((_ompc_nest_lock_t *)lock);
}


void 
omp_destroy_nest_lock (omp_nest_lock_t *lock)
{
  OMPSM_DESTROY_NLOCK ((_ompc_nest_lock_t *)lock);
}


void 
omp_destroy_nest_lock_ (_omf77Lock_t *lock) 
{
  OMPSM_DESTROY_NLOCK ((_ompc_nest_lock_t *)lock);
}


void 
omp_set_nest_lock(omp_nest_lock_t *lock)
{
  OMPSM_LOCK_NLOCK((_ompc_nest_lock_t *)lock);
}


void 
omp_set_nest_lock_(_omf77Lock_t *lock)
{
  OMPSM_LOCK_NLOCK((_ompc_nest_lock_t *)lock);
}


void
omp_unset_nest_lock (omp_nest_lock_t *lock)
{
  OMPSM_UNLOCK_NLOCK ((_ompc_nest_lock_t *)lock);
}


void 
omp_unset_nest_lock_ (_omf77Lock_t *lock)
{
  OMPSM_UNLOCK_NLOCK ((_ompc_nest_lock_t *)lock);
}


int
omp_test_nest_lock (omp_nest_lock_t *lock)
{
  return OMPSM_TEST_NLOCK ((_ompc_nest_lock_t *)lock);
}


int
omp_test_nest_lock_ (_omf77Lock_t *lock)
{
  return OMPSM_TEST_NLOCK ((_ompc_nest_lock_t *)lock);
}


#if 0
/*
 * global memory allocater
 */
void *
ompsm_galloc (int sz, int mode, int arg)
{
  return (void *) OMPSM_GALLOC(sz, mode, arg);
}
#endif
