/* 
 * $Id: ompd_object.h,v 1.1.1.1 2004-02-06 18:14:57 msato Exp $
 * $RWC_Release$
 * $RWC_Copyright$
 */
#ifndef _OMPD_OBJECT_H
#define _OMPD_OBJECT_H

#define CTX_IN_PARALLEL 0
#define CTX_IN_FORALL 1
#define CTX_IN_MASTER 2

#define MAX_DIM 4

typedef struct _ompd_obj_descriptor {
    char modified;	/* modified or not */
    char cached;
    short ndim;		/* if ndim == 0, variable */
    char *addr;		/* base address of this object */
    int size;		/* variable size in byte */
    struct _ompd_obj_descriptor *link;	/* free link */
} ompd_obj;

typedef struct _ompd_range {
    int lb,ub; 		/* upper and lower bound, lb <= i < ub */
    int blk_n;		/* block length */
    int stride;		/* stride. if stride == 0, single block. */
    int step;		/* if step != 1, for scaled access, 
			 *  block length = blk_n*step.
			 *  if step == 0, single point */
} ompd_range;

/* array type */
enum ompd_array_type {
    OMPD_NONE = 0,
    OMPD_RMEM_ARRAY = 1,	/* use remote memory and cache */
    OMPD_DIST_ARRAY = 2,	/* distributed arrry */
    OMPD_TDL_ARRAY = 3		/* TDL array */
};

typedef struct _ompd_array_descriptor {
    /* header */
    char modified;		/* modified or not */
    char cached;
    short ndim;			/* if ndim == 0, variable */
    char *addr;			/* base address of this object */
    int size;			/* total size */

    struct _ompd_array_descriptor *link;	/* free link */
    char is_random_acc_mode;	/* random access, force rmem mode */
    char type;			/* type of array */
    short elem_size;		/* element size in byte */
    int dim_size[MAX_DIM];
    int dim_esize[MAX_DIM];	/* total sub-element size */

    /* for remote memory mode */
    ompd_range rgn_cached[MAX_DIM],rgn_modified[MAX_DIM];

    /* for distributed array */
    int dist_dim; 	/* distributed dimension */
    int dist_offset;	/* offset, scaling not supported */
    int dist_blk_size;	/* block size */

    /* for TDL array */
    void *TDL_info;
} ompd_array;

enum ompd_access_type {
    ACCESS_INDEX,
    ACCESS_FOR,
    ACCESS_FOR_BLOCK,
    ACCESS_FOR_CYCLIC
};

typedef struct _ompd_access_range {
    enum ompd_access_type type;
    int lb,ub;
    int step;
    int dir;	/* direction */
    int chunk;
    int scale,offset;
    struct _ompd_access_expr *expr;
} ompd_access_range;

/* a*i+b or addr[a*i+b]+off */
typedef struct _ompd_access_expr {
    int is_for;
    int a,b,off;
    int *addr;
    ompd_array *desc;
    int a1,b1,off1;
    int *addr1;
    ompd_array *desc1;
    int step,dir;
    struct _ompd_access_expr *expr;
} ompd_access_expr;

enum ompd_access_mode {
    ACCESS_RMEM_READ = 0,
    ACCESS_RMEM_WRITE,
    ACCESS_INSPECT_HOME,
    ACCESS_INSPECT_SHARING,
    ACCESS_INSPECT_REF,
    ACCESS_INSPECT_MODIFY
};

/* prototype */
/* interface for memory runtime */
void _ompd_D_v(ompd_obj **dpp,char *addr,int size);
void _ompd_D_a(ompd_obj **dpp,char *addr,int size,int elem_size,int ndim,
	       int d1,int d2,int d3,int d4);
void _ompd_acc_index(int dim_i,int i);
void _ompd_acc_for(int dim_i, int lb, int ub, int step,int dir);
void _ompd_acc_forall(int dim_i, int lb, int ub, int step, int dir,
		      int schd, int chunk);
void _ompd_acc_range(int ndim);
void _ompd_R_v(int context, ompd_obj *dp);
void _ompd_W_v(int context, ompd_obj *dp);
void _ompd_R_a(int context, ompd_array *dp);
void _ompd_W_a(int context, ompd_array *dp);
void _ompd_flush(ompd_obj *dp);
void _ompd_flushR(ompd_obj *dp);

int _ompd_setup_rgn(int ndim,ompd_access_range *ap, ompd_range *rp);
void _ompd_block_sched(int *b_p,int *e_p,int s);
void _ompd_cyclic_sched(int *b_p,int *chunk_p,int *s_p);

char *_ompd_malloc(int size);

/* remote memory array */
void _ompd_read_rgn_array(ompd_array *dp,ompd_range *rgn_arg);
void _ompd_write_rgn_array(ompd_array *dp,ompd_range *rgn_arg);
void _ompd_flush_rgn_array(ompd_array *dp,int flush_all);

void _ompd_array_rmem_block(ompd_array *dp,ompd_range *ranges, int rw);
void _ompd_read_rmem_array(ompd_array *dp,ompd_range *rgn_arg);
void _ompd_write_rmem_array(ompd_array *dp,ompd_range *rgn_arg);

/* remote memory function */
/* raw mode intrface for debug */
int _rmem_read_int(int *addr);
void _rmem_write_int(int *addr,int w);

/* simple read with remote memory interface */
void _rmem_sync(void);
void _rmem_read(char *addr,int size);
void _rmem_write(char *addr,int size);
void _rmem_write_block(char *addr,int count,int size,int stride);
void _rmem_read_block(char *addr,int count,int size,int stride);

/* coherent memory interface */
void _ompd_read_any(void *addr, int size);
void _ompd_write_any(void *addr, int size);
void _ompd_flush_any(void);

void _ompd_access_array(ompd_array *dp,	ompd_access_range *rp,int mode);
void print_access_range(FILE *fp,int n_dim, ompd_access_range *rp);
void print_access_expr(FILE *fp,ompd_access_expr *ep);

void _ompd_init_TDL();
void _ompd_init_TDL_array(ompd_array *dp);
void _ompd_TDL_inspect_home(void *addr,int size);
void _ompd_TDL_inspect_sharing(void *addr,int size);
void _ompd_TDL_inspect_ref(void *addr,int size);
void _ompd_TDL_inspect_modify(void *addr,int size);
void _ompd_read_TDL_array(ompd_array *dp,int n_rgn,ompd_access_range **rp);
void _ompd_write_TDL_array(ompd_array *dp,int n_rgn,ompd_access_range **rp);
void _ompd_master_TDL_array(ompd_array *dp);


#endif /* _OMPD_OBJECT_H */

