/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
#ifdef __USE_RHINET__

#include <stdio.h>
#include <stdlib.h>
#include <api.h>
#include "ompd.h"


#define DEBUG
#define DEBUGMSG(X)		;
#define BUG_4BYTE_ACCESS
#define BUG_RN_PULL

#define WORD_SIZE		4
#define WWORD_SIZE		(WORD_SIZE * 2)

#define	PAGESIZE		4096
#define PAGEALIGN(X)		((int)(X) & ~(PAGESIZE-1))
#define	PAGEOFFSET(X)		((int)(X) & (4096-1))

#define MSGSIZE			128
#define MSGBUFFSIZE		10

#define MAX_MSG_READ_SIZE	(MSGSIZE)
#define MAX_MSG_WRITE_SIZE	(MSGSIZE - sizeof(int) * 2)
#define MAX_MSG_RPCN_SIZE	(MSGSIZE - sizeof(int) * 3)

extern void _start;
extern void _etext;
extern void __data_start;
extern void _edata;
extern void __bss_start;
extern void _end;


extern int _ompc_debug_flag;
extern int _ompc_n_node;
extern int _ompc_node_id;


static int buf[PAGESIZE/sizeof(int)];

enum msgstatus {
  MSG_DONE = 0,
  MSG_SEND,
  MSG_REPLY
};

enum msgtype {
  MSG_FUNC_NONE = 0,
  MSG_FUNC_BARRIER,
  MSG_FUNC_READ,
  MSG_FUNC_WRITE,
  MSG_FUNC_READ_DMA,
  MSG_FUNC_RPC,
  MSG_FUNC_RPCN
};

struct t_msg {
  /* don't send */
  int				node;
  struct t_msg			*dist_addr;

  /* send */
  int				data[MSGSIZE / sizeof(int)];
  enum msgtype			type;
  int				repl_addr;
  volatile enum msgstatus	status;
#ifdef BUG_4BYTE_ACCESS
  int				dmy;
#endif  
};

struct t_msgbuf {
  int	cur;			      /* current message position */
  struct t_msg msg[MSGBUFFSIZE];      /* receive buffer */
};


static struct {
  int			gid;
  int			gid_addr;
  int			pdarea_top;
  int			pdarea_tail;
  STATUS		status;
#ifdef BUG_4BYTE_ACCESS
  int	dmy;			      /* may be not need this padding */
#endif  

  int			msgbuf_gid;
  int			msgbuf_top;
  int			msgbuf_tail;
  int			*msgbuf_dist_pos;
  struct t_msgbuf	*msgbuf;
  
} rhinet;


static struct t_barrier {
  int		cur;
  volatile int	cnt[2];
} barr = {0, {0, 0}};


int	*rmacnt;


static int		_ompd_trun_init ();
static int		_ompd_trun_getline ();

static void		_rhinet_read (int, int *, int *, int);
static void		_rhinet_write (int, int *, int *, int);
static void		_rhinet_rma_sync ();
static void		_rhinet_rma_clear ();
static int		_rhinet_is_pindown(char *, int size);

static struct t_msg *	_ompd_rhinet_msg_getbuff (int);
static void		_ompd_rhinet_msg_push (int, void *, void *, int);
static void		_ompd_rhinet_msg_send (struct t_msg *, int);
static void		_ompd_rhinet_msg_reply (struct t_msg *, int);
static void		_ompd_rhinet_msg_free (struct t_msg *);
static void		_ompd_rhinet_msg_wait (struct t_msg *);
static void		_ompd_rhinet_msg_rmem_read (int, char *, char *, int);
static void		_ompd_rhinet_msg_handler_rmem_read (struct t_msg *);
static void		_ompd_rhinet_msg_rmem_write (int, char *, char *, int);
static void		_ompd_rhinet_msg_handler_rmem_write (struct t_msg *);
static void		_ompd_rhinet_msg_rmem_dmaread (int, int *, int *, int);
static void		_ompd_rhinet_msg_handler_rmem_dmaread (struct t_msg *);
static void		_ompd_rhinet_msg_barrier ();
static void		_ompd_rhinet_msg_handler_barrier (struct t_msg *);
static int		_ompd_rhinet_msg_rpc_s (int, int (*)(int, void *), void *);
static void		_ompd_rhinet_msg_rpc_a (int, void (*)(int, void *), void *);
static void		_ompd_rhinet_msg_handler_rpc (struct t_msg *);
static int		_ompd_rhinet_msg_rpcn_s (int, int (*)(int, void *, int), void *, int);
static void		_ompd_rhinet_msg_rpcn_a (int, void (*)(int, void *, int), void *, int);
static void		_ompd_rhinet_msg_handler_rpcn (struct t_msg *);
static void		_ompd_rhinet_msg_handler (struct t_msg *);
static void		_ompd_rhinet_poll ();
static void		_ompd_rhinet_msg_sync ();

static void		_rhinet_lock_initialize();
static int 		_rhinet_lock_get_locknum ();
static void 		_rhinet_lock_free_locknum (int);
static void		_rhinet_lock_init (int, void *);
static void		_rhinet_lock_destroy (int, void *);
static void		_rhinet_lock (int, void *);
static void		_rhinet_unlock (int, void *);
static int		_rhinet_trylock (int, void *);
static void		_rhinet_nlock (int, void *);
static void		_rhinet_nunlock (int, void *);
static int		_rhinet_ntrylock (int, void *);

static void		_ompd_rhinet_dsched (int, void *);

static any_type		_ompd_to_anytype (void *, int);
static void		_ompd_from_anytype (any_type, void *, int);
static void		_ompd_rhinet_rdctn (int, void *, int);

static void		_ompd_rhinet_read (int, char *, int);
static void		_ompd_rhinet_write (int, char *, int);


void
_ompd_rhinet_init (int *argc, char ***argv)
{
  int 		rv, ptr, sz, node, msg;
  struct t_msg	*mp;


#ifdef __USE_TRUN__
  rv = _ompd_trun_init ();
  if (rv != SUCCESS) {
    _ompc_fatal ("_ompd_rhinet_init : _ompd_trun_init failed");
  }
  if (_ompc_debug_flag) {
    char namebuf[1024];
    rv = gethostname (namebuf, 1024);
    if (rv < 0) {
      _ompc_fatal ("can not get hostname");
    } else {
      fprintf (stderr,"Process %d on %s\n", _ompc_node_id, namebuf);
    }
  }
#else /* __USE_TRUN__ */
  _ompd_mpi_init (argc, argv);
#endif /* __USE_TRUN__ */

  DEBUGMSG(fprintf (stderr, "<%d> _ompd_rhinet_init\n", _ompc_node_id));
  DEBUGMSG(fprintf (stderr, "<%d> text segment : 0x%08x - 0x%08x\n", _ompc_node_id, &_start, &_etext));
  DEBUGMSG(fprintf (stderr, "<%d> data segment : 0x%08x - 0x%08x\n", _ompc_node_id, &__data_start, &_edata));
  DEBUGMSG(fprintf (stderr, "<%d> bss  segment : 0x%08x - 0x%08x\n", _ompc_node_id, &__bss_start, &_end));

  /* initialize rhinet */
  rv = rn_initialize (NULL, 0, _ompc_node_id);
  if (rv < 0) {
    _ompc_fatal ("rn_initialize failed");
  }

  /* pindown all data and bss area */
  ptr = PAGEALIGN(&rhinet);
  sz  = PAGEALIGN(((int)&rhinet + sizeof(rhinet) + PAGESIZE - 1)) - ptr;
  DEBUGMSG(fprintf (stderr,"<%d> pindown 0x%08x - 0x%08x\n", _ompc_node_id, ptr, ptr+sz));
  rn_pindown ((void *)ptr, sz);

  rhinet.pdarea_top  = (int)&__data_start;
  rhinet.pdarea_tail = (int)&_end;
  rhinet.gid         = 1;
  rhinet.gid_addr    = PAGEALIGN(rhinet.pdarea_top);

  sz  = PAGEALIGN(((int)&_end + PAGESIZE - 1)) - rhinet.gid_addr;

  DEBUGMSG(fprintf (stderr, "<%d> set_gid %d = 0x%08x - 0x%08x\n", _ompc_node_id, rhinet.gid, rhinet.gid_addr, rhinet.gid_addr+sz));
  rhinet.status = 0;
  rn_set_gid (rhinet.gid, (void *)rhinet.gid_addr, &rhinet.status);
  while (rhinet.status == 0) {
  }
  if (rhinet.status != PrimDone) {
    _ompc_fatal ("rn_set_gid failed");
  }
  DEBUGMSG(fprintf (stderr, "<%d> pindown 0x%08x - 0x%08x\n", _ompc_node_id, rhinet.gid_addr, rhinet.gid_addr+sz));
  rn_pindown ((void *)rhinet.gid_addr, sz);


  /* pindown message buffer */
  sz = sizeof(struct t_msgbuf) * _ompc_n_node;
  rhinet.msgbuf = (struct t_msgbuf *) malloc (sz + PAGESIZE);
  rhinet.msgbuf_dist_pos = (int *) malloc (sizeof(int) * _ompc_n_node);
  if (rhinet.msgbuf == NULL  ||  rhinet.msgbuf_dist_pos == NULL) {
    _ompc_fatal ("_ompd_rhinet_init : can not allocate message buffer");
  }
  rhinet.msgbuf      = (struct t_msgbuf *) PAGEALIGN(((int)rhinet.msgbuf +
						      PAGESIZE - 1));
  rhinet.msgbuf_gid  = 2;
  rhinet.msgbuf_top  = (int)rhinet.msgbuf;
  rhinet.msgbuf_tail = rhinet.msgbuf_top + sz;

  sz = PAGEALIGN((rhinet.msgbuf_tail + PAGESIZE - 1)) - rhinet.msgbuf_top;

  DEBUGMSG(fprintf (stderr, "<%d> set_gid %d = 0x%08x - 0x%08x\n", _ompc_node_id, rhinet.msgbuf_gid, rhinet.msgbuf_top, rhinet.msgbuf_top+sz));
  rhinet.status      = 0;
  rn_set_gid (rhinet.msgbuf_gid, (void *)rhinet.msgbuf_top, &rhinet.status);
  while (rhinet.status == 0) {
  }
  if (rhinet.status != PrimDone) {
    _ompc_fatal ("rn_set_gid failed");
  }
  DEBUGMSG(fprintf (stderr, "<%d> pindown 0x%08x - 0x%08x\n", _ompc_node_id, rhinet.msgbuf_top, rhinet.msgbuf_top+sz));
  rn_pindown ((void *)rhinet.msgbuf_top, sz);


  /* setup remote memory access counter */
  rmacnt = (int *) malloc (sizeof(int) * _ompc_n_node);
  if (rmacnt == NULL) {
    _ompc_fatal ("_ompd_rhinet_init : can not allocate RMA check buffer");
  }
  for (node=0; node<_ompc_n_node; node++) {
    rmacnt[node] = 0;
  }

  /* setup message buffer */
  for (node=0; node<_ompc_n_node; node++) {
    rhinet.msgbuf[node].cur = 0;
    for (msg=0; msg<MSGBUFFSIZE; msg++) {
      mp = &rhinet.msgbuf[node].msg[msg];
      mp->status = MSG_DONE;
      mp->node = node;
    }
    rhinet.msgbuf_dist_pos[node] = 0;
  }

  /* initialize lock variable */
  OMPD_TCP_BARRIER ();
  _rhinet_lock_initialize ();
  if (_ompc_node_id == 0) {
    _ompd_rhinet_lock_init (LOCK_SYS);
    _ompd_rhinet_lock_init (LOCK_CRITICAL);
    _ompd_rhinet_lock_init (LOCK_ORDERED);
    _ompd_rhinet_lock_init (LOCK_ATOMIC);
  }

  /* initialize global variable */
  _ompc_n_proc      = _ompc_n_node;
  _ompc_max_threads = _ompc_n_node;
  _ompc_num_threads = _ompc_n_node;

  _ompd_rhinet_barrier ();
  DEBUGMSG(fprintf (stderr, "<%d> _ompd_rhinet_init done\n", _ompc_node_id));
}


static void
_rhinet_read (int node, int *dst, int *src, int size)
{
#ifdef BUG_4BYTE_ACCESS
  int sz_align = WWORD_SIZE;
#else
  int sz_align = WORD_SIZE;
#endif

  if ((size % sz_align) != 0) {
    printf ("<%d> : _rhinet_read : can not read %d byte data\n",
	    _ompc_node_id, size);
    for (;;) {
      _ompd_rhinet_poll ();
    }
  }

  /* copy from own node */
  if (node == _ompc_node_id) {
    if (dst != src) {
      memcpy (dst, src, size);
    }

  /* copy from other node */
  } else {
#ifdef BUG_RN_PULL
    _ompd_rhinet_msg_rmem_dmaread (node, dst, src, size);
#else
    if ((int)dst < rhinet.pdarea_top ||	rhinet.pdarea_tail < (int)dst + size ||
	(int)src < rhinet.pdarea_top ||	rhinet.pdarea_tail < (int)src + size) {
      _ompc_fatal ("_rhinet_read : out of pin down area");
    }
    rhinet.status = 0;
    DEBUGMSG(fprintf (stderr, "<%d> rn_pull (0x%x, %d, %d, 0x%x(0x%x-0x%x), 0, 0x%x, %d\n", _ompc_node_id, dst, node, rhinet.gid, (int)src - rhinet.gid_addr, src, rhinet.gid_addr, &rhinet.status, size));
    rn_pull (dst, node, rhinet.gid, (int)src - rhinet.gid_addr, 0,
	     &rhinet.status, size);
    while (rhinet.status == 0) {
    }
    if (rhinet.status != PrimDone) {
      printf ("_rhinet_read : rn_pull failed(error code = %d)\n", 
	      rhinet.status);
    }
#endif
  }
}


static void
_rhinet_write (int node, int *dst, int *src, int size)
{
#ifdef BUG_4BYTE_ACCESS
  int sz_align = WWORD_SIZE;
#else
  int sz_align = WORD_SIZE;
#endif

  if ((size % sz_align) != 0) {
    printf ("<%d> : _rhinet_write : can not write %d byte data\n",
	    _ompc_node_id, size);
    for (;;) {
      _ompd_rhinet_poll ();
    }
  }

  /* copy from own node */
  if (node == _ompc_node_id) {
    if (dst != src) {
      memcpy (dst, src, size);
    }

  /* copy from other node */
  } else {
    if ((int)dst < rhinet.pdarea_top || rhinet.pdarea_tail < (int)dst+size ||
	(int)src < rhinet.pdarea_top ||	rhinet.pdarea_tail < (int)src + size) {
      _ompc_fatal ("_rhinet_write : out of pin down area");
    }
    rhinet.status = 0;
    DEBUGMSG(fprintf (stderr, "<%d> rn_push (0x%x, %d, %d, 0x%x(0x%x-0x%x), 0, 0x%x, %d\n", _ompc_node_id, src, node, rhinet.gid, (int)dst - rhinet.gid_addr, dst, rhinet.gid_addr, &rhinet.status, size));
    rmacnt[node] = 1;
    rn_push (src, node, rhinet.gid, (int)dst - rhinet.gid_addr, 0,
	     &rhinet.status, size);
    while (rhinet.status == 0) {
    }
    if (rhinet.status != PrimDone) {
      printf ("_rhinet_write : rn_push failed(error code = %d)\n", 
	      rhinet.status);
    }
  }
}



static void
_rhinet_rma_sync ()
{
  int	node;


  for (node=0; node<_ompc_n_node; node++) {
    if (rmacnt[node] != 0) {
#ifdef BUG_4BYTE_ACCESS
      _rhinet_read(node, buf, buf, WWORD_SIZE);
#else
      _rhinet_read(node, buf, buf, WORD_SIZE);
#endif
    }
  }
  _rhinet_rma_clear ();
}



static void
_rhinet_rma_clear ()
{
  int	node;


  for (node=0; node<_ompc_n_node; node++) {
    rmacnt[node] = 0;
  }
}



static int
_rhinet_is_pindown (char *addr, int size)
{
  if ((int)addr < rhinet.pdarea_top ||
      rhinet.pdarea_tail < (int)addr + size) {
    return FALSE;
  } else {
    return TRUE;
  }
}



/* ---------------------- message passing routine ------------------------- */
static struct t_msg *
_ompd_rhinet_msg_getbuff (int node)
{
  struct t_msgbuf	*msgb;
  struct t_msg		*msg;


  msgb = &rhinet.msgbuf[_ompc_node_id];
  for (;;) {
    msg  = &msgb->msg[msgb->cur];
    if (msg->status == MSG_DONE) {
      msg->node      = node;
      msg->dist_addr = &msgb->msg[rhinet.msgbuf_dist_pos[node]];
      msg->repl_addr = (int)msg - rhinet.msgbuf_top;

      msgb->cur = (msgb->cur + 1) % MSGBUFFSIZE;
      rhinet.msgbuf_dist_pos[node] = ((rhinet.msgbuf_dist_pos[node] + 1) %
				      MSGBUFFSIZE);
      break;
    }
    _ompd_rhinet_poll ();
  }

  return msg;
}


static void
_ompd_rhinet_msg_push (int node, void *dst, void *src, int size)
{
#ifdef BUG_4BYTE_ACCESS
  if ((size % WWORD_SIZE) != 0) {
    size = size - (size % WWORD_SIZE);
  }
#endif  
#ifdef DEBUG
  if ((int)src < rhinet.msgbuf_top || rhinet.msgbuf_tail < (int)src+size ||
      (int)dst < rhinet.msgbuf_top || rhinet.msgbuf_tail < (int)dst+size) {
    _ompc_fatal ("_ompd_rhinet_msg_push : out of pin down area");
  }
#endif

  rhinet.status = 0;
  DEBUGMSG(fprintf (stderr, "<%d> rn_push (0x%x, %d, %d, 0x%x(0x%x-0x%x), 0, 0x%x, %d)\n", _ompc_node_id, src, node, rhinet.msgbuf_gid, dst - rhinet.msgbuf_top, dst, rhinet.msgbuf_top, &rhinet.status, size));
  rn_push (src, node, rhinet.msgbuf_gid, (int)dst - rhinet.msgbuf_top,
	   0, &rhinet.status, size);
  while (rhinet.status == 0) {
  }
  DEBUGMSG(fprintf (stderr, "<%d> rn_push done\n", _ompc_node_id))
  if (rhinet.status != PrimDone) {
    _ompc_fatal ("_ompd_rhinet_msg_push :  rn_push failed");
  }
}


static void
_ompd_rhinet_msg_send (struct t_msg *msg, int size)
{
  int	src, dst, ofs;


  _ompd_rhinet_poll ();

  if (size % WORD_SIZE) {
    size = size + WORD_SIZE - size % WORD_SIZE;
  }
  ofs = sizeof(struct t_msg) - size;
  src = (int)msg + ofs;
  dst = (int)msg->dist_addr + ofs;

  msg->status = MSG_SEND;
  _ompd_rhinet_msg_push (msg->node, (void *)dst, (void *)src, size);
}


static void
_ompd_rhinet_msg_reply (struct t_msg *msg, int size)
{
  int	src, dst, ofs;


  if (size % WORD_SIZE) {
    size = size + WORD_SIZE - size % WORD_SIZE;
  }
  ofs = sizeof(struct t_msg) - size;
  src = (int)msg + ofs;
  dst = rhinet.msgbuf_top + (int)msg->repl_addr + ofs;

  msg->status = MSG_REPLY;
  _ompd_rhinet_msg_push (msg->node, (void *)dst, (void *)src, size);
}


static void
_ompd_rhinet_msg_free (struct t_msg *msg)
{
  int	src, dst, ofs, size;


  if((int)(&(rhinet.msgbuf[_ompc_node_id])) <= (int)msg  &&
     (int)msg < (int)(&(rhinet.msgbuf[_ompc_node_id+1]))) {
    msg->status = MSG_DONE;

  } else {
    size = (int)msg + sizeof(struct t_msg) - (int)&msg->status;
    if (size % WORD_SIZE) {
      size = size + WORD_SIZE - size % WORD_SIZE;
    }
    ofs = sizeof(struct t_msg) - size;
    src = (int)msg + ofs;
    dst = rhinet.msgbuf_top + (int)msg->repl_addr + ofs;

    msg->status = MSG_DONE;
    _ompd_rhinet_msg_push (msg->node, (void *)dst, (void *)src, size);
  }
}


static void
_ompd_rhinet_msg_wait (struct t_msg *msg)
{
  while (msg->status == MSG_SEND) {
    _ompd_rhinet_poll ();
  }
}


static void
_ompd_rhinet_msg_rmem_read (int node, char *dst, char *src, int size)
{
  struct t_msg	*msg;
  int		sz;

  int pos  = (MSGSIZE / sizeof(int)) - 2;
  int tail = MSGSIZE / sizeof(int);


  if (MAX_MSG_READ_SIZE < size) {
    _ompc_fatal ("_ompd_rhinet_msg_rmem_read : "
		 "size of remote memory access is too large.");
  }

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_READ;
  msg->data[pos]   = (int)src;
  msg->data[pos+1] = (int)size;
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[pos];

  _ompd_rhinet_msg_send (msg, sz);
  _ompd_rhinet_msg_wait(msg);

  memcpy (dst, (char *)&msg->data[tail] - size, size);
  _ompd_rhinet_msg_free (msg);

  return;
}


static void
_ompd_rhinet_msg_handler_rmem_read (struct t_msg *msg)
{
  char	*addr;
  int	size, sz;

  int pos  = (MSGSIZE / sizeof(int)) - 2;
  int tail = MSGSIZE / sizeof(int);


  addr = (char *)msg->data[pos];
  size = msg->data[pos+1];
  memcpy ((char *)&msg->data[tail] - size, addr, size);

  sz = (int)msg + sizeof(struct t_msg) - ((int)&msg->data[tail] - size);

  _ompd_rhinet_msg_reply (msg, sz);
}


static void
_ompd_rhinet_msg_rmem_write (int node, char *dst, char *src, int size)
{
  struct t_msg	*msg;
  int		sz;

  int pos = (MSGSIZE / sizeof(int)) - 2;


  if (MAX_MSG_WRITE_SIZE < size) {
    _ompc_fatal ("_ompd_rhinet_msg_rmem_write : "
		 "size of remote memory access is too large.");
  }

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_WRITE;
  msg->data[pos]   = (int)dst;
  msg->data[pos+1] = size;
  memcpy ((char *)&msg->data[pos] - size, src, size);
  sz = (int)msg + sizeof(struct t_msg) - ((int)&msg->data[pos] - size);

  _ompd_rhinet_msg_send (msg, sz);
}


static void
_ompd_rhinet_msg_handler_rmem_write (struct t_msg *msg)
{
  char	*dst, *src;
  int	sz;

  int pos = (MSGSIZE / sizeof(int)) - 2;


  sz  = msg->data[pos+1];
  dst = (char *)msg->data[pos];
  src = (char *)&msg->data[pos] - sz;
  if (dst == 0) {
    _ompc_fatal ("distination is null");
  }
  memcpy (dst, src, sz);

  _ompd_rhinet_msg_free (msg);
}


static void
_ompd_rhinet_msg_rmem_dmaread (int node, int *dst, int *src, int size)
{
  struct t_msg	*msg;
  int		sz;

  int pos  = (MSGSIZE / sizeof(int)) - 3;


  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_READ_DMA;
  msg->data[pos]   = (int)dst;
  msg->data[pos+1] = (int)src;
  msg->data[pos+2] = (int)size;
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[pos];

  _ompd_rhinet_msg_send (msg, sz);
  _ompd_rhinet_msg_wait(msg);
}


static void
_ompd_rhinet_msg_handler_rmem_dmaread (struct t_msg *msg)
{
  int sz;

  int pos  = (MSGSIZE / sizeof(int)) - 3;


  _rhinet_write (msg->node, (int *)msg->data[pos],
		 (int *)msg->data[pos+1], msg->data[pos+2]);

  _ompd_rhinet_msg_free (msg);
}


static void
_ompd_rhinet_msg_barrier ()
{
  int		node, n, sz;
  struct t_msg	*msg;

  int pos = (MSGSIZE / sizeof(int)) - 1;

  for (n=1; n<_ompc_n_node; n++) {
    node = (n + _ompc_node_id) % _ompc_n_node;
    msg  = _ompd_rhinet_msg_getbuff (node);

    msg->type      = MSG_FUNC_BARRIER;
    msg->data[pos] = barr.cur;
    sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[pos];
    _ompd_rhinet_msg_send (msg, sz);
  }
  while (barr.cnt[barr.cur] != _ompc_n_node - 1) {
    _ompd_rhinet_poll ();
  }
  barr.cnt[barr.cur] = 0;
  barr.cur = (barr.cur + 1) % 2;
}


static void
_ompd_rhinet_msg_handler_barrier (struct t_msg *msg)
{
  int sz;

  int pos = (MSGSIZE / sizeof(int)) - 1;


  barr.cnt[msg->data[pos]] ++;

  _ompd_rhinet_msg_free (msg);
}


static int
_ompd_rhinet_msg_rpc_s (int node, int (*func)(int, void *), void *arg)
{
  struct t_msg	*msg;
  int		sz, ret;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  if (node == _ompc_node_id) {
    _ompd_rhinet_poll ();
    return func (_ompc_node_id, arg);
  }

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_RPC;
  msg->data[pos]   = (int)func;
  msg->data[pos+1] = (int)arg;
  msg->data[pos+2] = 1;
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[pos];

  _ompd_rhinet_msg_send (msg, sz);
  _ompd_rhinet_msg_wait(msg);
  ret = msg->data[tail];
  _ompd_rhinet_msg_free (msg);

  return ret;
}


static void
_ompd_rhinet_msg_rpc_a (int node, void (*func)(int, void *), void *arg)
{
  struct t_msg	*msg;
  int		sz;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  if (node == _ompc_node_id) {
    _ompd_rhinet_poll ();
    func (_ompc_node_id, arg);
    return;
  }

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_RPC;
  msg->data[pos]   = (int)func;
  msg->data[pos+1] = (int)arg;
  msg->data[pos+2] = 0;
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[pos];

  _ompd_rhinet_msg_send (msg, sz);
}


static void
_ompd_rhinet_msg_handler_rpc (struct t_msg *msg)
{
  int	(*func)(int, void *);
  void	*arg;
  int	sz, rv, sync;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  func = (int (*)())msg->data[pos];
  arg  = (void *)msg->data[pos+1];
  sync = msg->data[pos+2];

  if (sync) {
    msg->data[tail] = func(msg->node, arg);
    sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[tail];
    _ompd_rhinet_msg_reply (msg, sz);

  } else {
    func(msg->node, arg);
    _ompd_rhinet_msg_free (msg);
  }
}


static int
_ompd_rhinet_msg_rpcn_s (int node,
			 int (*func)(int, void *, int),
			 void *arg, int size)
{
  struct t_msg	*msg;
  int		sz, dpos, ret;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  if (size < 0 || MAX_MSG_RPCN_SIZE < size) {
    _ompc_fatal ("_ompd_rhinet_msg_rpcn_s : rpc data size is too large.");
  }

  if (node == _ompc_node_id) {
    _ompd_rhinet_poll ();
    return func (_ompc_node_id, arg, size);
  }

  dpos = pos - (size + sizeof(int) - 1) / sizeof(int);

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_RPCN;
  msg->data[pos]   = (int)func;
  msg->data[pos+1] = size;
  msg->data[pos+2] = 1;
  memcpy (&msg->data[dpos], arg, size);
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[dpos];

  _ompd_rhinet_msg_send (msg, sz);
  _ompd_rhinet_msg_wait(msg);
  ret = msg->data[tail];
  _ompd_rhinet_msg_free (msg);

  return ret;
}


static void
_ompd_rhinet_msg_rpcn_a (int node,
			 void (*func)(int, void *, int),
			 void *arg, int size)
{
  struct t_msg	*msg;
  int		sz, dpos;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  if (size < 0 || MAX_MSG_RPCN_SIZE < size) {
    _ompc_fatal ("_ompd_rhinet_msg_rpcn_s : rpc data size is too large.");
  }

  if (node == _ompc_node_id) {
    _ompd_rhinet_poll ();
    func (_ompc_node_id, arg, size);
    return;
  }

  dpos = pos - (size + sizeof(int) - 1) / sizeof(int);

  msg = _ompd_rhinet_msg_getbuff (node);

  msg->type        = MSG_FUNC_RPCN;
  msg->data[pos]   = (int)func;
  msg->data[pos+1] = size;
  msg->data[pos+2] = 0;
  memcpy (&msg->data[dpos], arg, size);
  sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[dpos];

  _ompd_rhinet_msg_send (msg, sz);
}


static void
_ompd_rhinet_msg_handler_rpcn (struct t_msg *msg)
{
  int	(*func)(int, void *, int);
  void	*arg;
  int	sz, rv, sync, size, dpos;

  int pos  = (MSGSIZE / sizeof(int)) - 3;
  int tail = (MSGSIZE / sizeof(int)) - 1;


  func = (int (*)())msg->data[pos];
  size = msg->data[pos+1];
  sync = msg->data[pos+2];

  dpos = pos - (size + sizeof(int) - 1) / sizeof(int);
  arg  = (void *)&msg->data[dpos];

  if (sync) {
    msg->data[tail] = func(msg->node, arg, size);
    sz = (int)msg + sizeof(struct t_msg) - (int)&msg->data[tail];
    _ompd_rhinet_msg_reply (msg, sz);

  } else {
    func(msg->node, arg, size);
    _ompd_rhinet_msg_free (msg);
  }
}


static void
_ompd_rhinet_msg_handler (struct t_msg *msg)
{
  switch (msg->type) {	
  case MSG_FUNC_BARRIER:
    _ompd_rhinet_msg_handler_barrier (msg);
    break;
  case MSG_FUNC_READ:
    _ompd_rhinet_msg_handler_rmem_read (msg);
    break;
  case MSG_FUNC_WRITE:
    _ompd_rhinet_msg_handler_rmem_write (msg);
    break;
  case MSG_FUNC_READ_DMA:
    _ompd_rhinet_msg_handler_rmem_dmaread (msg);
    break;
  case MSG_FUNC_RPC:
    _ompd_rhinet_msg_handler_rpc (msg);
    break;
  case MSG_FUNC_RPCN:
    _ompd_rhinet_msg_handler_rpcn (msg);
    break;
    
  default:
    _ompc_fatal ("_ompd_rhinet_msg_handler : receive unknown message.");
    break;
  }
}


static void
_ompd_rhinet_poll ()
{
  struct t_msgbuf	*msgb;
  struct t_msg		*msg;
  int			rpl_cnt, n, node;


  do {
    rpl_cnt = 0;

    for (n=1; n<_ompc_n_node; n++) {
      node = (n + _ompc_node_id) % _ompc_n_node;
      msgb = &rhinet.msgbuf[node];
      msg  = &msgb->msg[msgb->cur];
      if (msg->status == MSG_SEND) {
	_ompd_rhinet_msg_handler (msg);
	msgb->cur = (msgb->cur + 1) % MSGBUFFSIZE;
	rpl_cnt ++;
      }
    }
  } while (rpl_cnt);
}


static void
_ompd_rhinet_msg_sync ()
{
  struct t_msgbuf	*msgb;
  struct t_msg		*msg;
  int			i;


  msgb = &rhinet.msgbuf[_ompc_node_id];
  for (i=0; i<MSGBUFFSIZE; i++) {
    msg  = &msgb->msg[i];
    while (msg->status != MSG_DONE) {
      _ompd_rhinet_poll ();
    }
  }
}


/* --------------------------- lock function  ----------------------------- */
struct _rhinet_lock {
  int			locknum;
  int			locknode;
  int			nestcount;
  struct _rhinet_lock	*next;

  int			*waitlist;
  int			waitcnt;
  int			waittop;
};

static struct _rhinet_lock *lockp = NULL;


struct _lock_free {
  int			num;
  struct _lock_free	*np;
};


static struct _lock_pool {
  int			next;
  struct _lock_free	*lp;
} lock_pool;


#ifdef BUG_4BYTE_ACCESS
static long long		result_true = TRUE;
static volatile long long	lock_done;
#else
static long long		result_true = TRUE;
static volatile int		lock_done;
#endif  


static void
_rhinet_lock_initialize ()
{
  int tmp;


  lock_pool.lp = NULL;
  tmp = LOCK_TAIL + _ompc_n_node - 1;
  lock_pool.next = tmp - tmp % _ompc_n_node + _ompc_node_id;
}


static int 
_rhinet_lock_get_locknum ()
{
  struct _lock_free	*lp;
  int 			ret;


  if (lock_pool.lp) {
    lp           = lock_pool.lp;
    ret          = lp->num;
    lock_pool.lp = lock_pool.lp->np;
    free (lp);

  } else {
    ret = lock_pool.next;
    lock_pool.next = lock_pool.next + _ompc_n_node;
  }

  return ret;
}


static void 
_rhinet_lock_free_locknum (int n)
{
  struct _lock_free	*lp;


  if (lock_pool.next = n + _ompc_n_node) {
    lock_pool.next = lock_pool.next - _ompc_n_node;
  } else {
    lp = (struct _lock_free *) malloc (sizeof(struct _lock_free *));
    if (lp == NULL) {
      _ompc_fatal ("_rhinet_lock_free_locknum : can not allocate memory");
    }
    lp->num      = n;
    lp->np       = lock_pool.lp;
    lock_pool.lp = lp;
  }
}


static void
_rhinet_lock_init (int node, void *arg)
{
  struct _rhinet_lock	*l;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      _ompc_fatal ("_rhinet_lock_init : lock variable is already initialize.");
    }
  }

  l = (struct _rhinet_lock *) malloc (sizeof (struct _rhinet_lock));
  if (l == NULL) {
    _ompc_fatal ("_rhinet_lock_init : can not allocate memory.");
  }
  l->waitlist = (int *) malloc (sizeof(int) * _ompc_n_node);
  if (l->waitlist == NULL) {
    _ompc_fatal ("_rhinet_lock_init : can not allocate memory.");
  }

  l->locknum   = n;
  l->locknode  = -1;
  l->nestcount = 0;
  l->waittop   = 0;
  l->waitcnt   = 0;
  l->next      = lockp;
  lockp        = l;
}


static void
_rhinet_lock_destroy (int node, void *arg)
{
  struct _rhinet_lock	*l, *p;

  int n = (int)arg;


  p = NULL;
  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (p) {
	p->next = l->next;
      } else {
	lockp = l->next;
      }
      free (l->waitlist);
      free (l);
      return;
    }
    p = l;
  }

  _ompc_fatal ("_rhinet_lock_destroy : not initialized lock variable.");
}


static void
_rhinet_lock (int node, void *arg)
{
  struct _rhinet_lock	*l;
  int			cur;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode == -1) {
	l->locknode = node;
	_rhinet_write (node, (int *)&lock_done, (int *)&result_true,
		       sizeof(result_true));
      } else {
	cur = (l->waittop + l->waitcnt) % _ompc_n_node;
	l->waitlist[cur] = node;
	l->waitcnt += 1;
      }
      return;
    }
  }

  _ompc_fatal ("_rhinet_unlock : not initialized lock variable.");
}


static void
_rhinet_unlock (int node, void *arg)
{
  struct _rhinet_lock	*l;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode != node) {
	_ompc_fatal ("_rhinet_unlock : recive unlock request,"
		     " but lock variable is locked another node.");
      }
      if (0 < l->waitcnt) {
	l->locknode = l->waitlist[l->waittop];
	l->waittop  = (l->waittop + 1) % _ompc_n_node;
	l->waitcnt -= 1;
	_rhinet_write (l->locknode, (int *)&lock_done, (int *)&result_true,
		       sizeof(result_true));
      } else {
	l->locknode = -1;
      }
      return;
    }
  }

  _ompc_fatal ("_rhinet_unlock : not initialized lock variable.");
}


static int
_rhinet_trylock (int node, void *arg)
{
  struct _rhinet_lock	*l;
  int			ret;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode == -1) {
	l->locknode = node;
	return TRUE;		      /* return Non-Zero value */
      } else {
	return FALSE;
      }
    }
  }

  _ompc_fatal ("_rhinet_trylock : not initialized lock variable.");
}


static void
_rhinet_nlock (int node, void *arg)
{
  struct _rhinet_lock	*l;
  int			cur;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode == node) {
	l->nestcount += 1;
	_rhinet_write (node, (int *)&lock_done, (int *)&result_true,
		       sizeof(result_true));
      } else if (l->locknode == -1) {
	l->locknode  = node;
	l->nestcount = 1;
	_rhinet_write (node, (int *)&lock_done, (int *)&result_true,
		       sizeof(result_true));
      } else {
	cur = (l->waittop + l->waitcnt) % _ompc_n_node;
	l->waitlist[cur] = node;
	l->waitcnt += 1;
      }
      return;
    }
  }

  _ompc_fatal ("_rhinet_nunlock : not initialized lock variable.");
}


static void
_rhinet_nunlock (int node, void *arg)
{
  struct _rhinet_lock	*l;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode != node) {
	_ompc_fatal ("_rhinet_nunlock : receive unlock request,"
		     " but lock variable is locked another node.");
      }
      if (1 < l->nestcount) {
	l->nestcount -= 1;
      } else if (l->nestcount == 1) {
	if (0 < l->waitcnt) {
	  l->locknode  = l->waitlist[l->waittop];
	  l->nestcount = 1;
	  l->waittop   = (l->waittop + 1) % _ompc_n_node;
	  l->waitcnt  -= 1;
	  _rhinet_write (l->locknode, (int *)&lock_done, (int *)&result_true,
			 sizeof(result_true));
	} else {
	  l->locknode  = -1;
	  l->nestcount = 0;
	}
      } else {
        _ompc_fatal ("_rhinet_nunlock : illegal nestcount.");
      }
      return;
    }
  }

  _ompc_fatal ("_rhinet_nunlock : not initialized lock variable.");
}


static int
_rhinet_ntrylock (int node, void *arg)
{
  struct _rhinet_lock	*l;
  int			ret;

  int n = (int)arg;


  for (l=lockp; l != NULL; l=l->next) {
    if (l->locknum == n) {
      if (l->locknode == node) {
	l->nestcount += 1;
	return l->nestcount;
      } else if (l->locknode == -1) {
	l->locknode   = node;
	l->nestcount = 1;
	return 1;
      } else {
	return 0;
      }
    }
  }

  _ompc_fatal ("_rhinet_ntrylock : not initialized lock variable.");
}


void
_ompd_rhinet_lock_alloc (int *locknum)
{
  *locknum = _rhinet_lock_get_locknum ();
  _ompd_rhinet_lock_init (*locknum);
}


void
_ompd_rhinet_lock_free (int *locknum)
{
  _ompd_rhinet_lock_destroy (*locknum);
  _rhinet_lock_free_locknum (*locknum);
  *locknum = 0;
}


void
_ompd_rhinet_lock_init (int locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_lock_init, (void *)locknum);
}


void
_ompd_rhinet_lock_destroy (int locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_lock_destroy, (void *)locknum);
}


void
_ompd_rhinet_lock (locknum)
{
  lock_done = FALSE;
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_lock, (void *)locknum);
  while (lock_done == FALSE) {	
    _ompd_rhinet_poll ();
  }
}


void
_ompd_rhinet_unlock (locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_unlock, (void *)locknum);
}


int
_ompd_rhinet_trylock (locknum)
{
  return _ompd_rhinet_msg_rpc_s (locknum % _ompc_n_node,
				 _rhinet_trylock, (void *)locknum);
}


void
_ompd_rhinet_nlock_init (int locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_lock_init, (void *)locknum);
}


void
_ompd_rhinet_nlock_destroy (int locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_lock_destroy, (void *)locknum);
}


void
_ompd_rhinet_nlock (locknum)
{
  lock_done = FALSE;
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_nlock, (void *)locknum);
  while (lock_done == FALSE) {	
    _ompd_rhinet_poll ();
  }
}


void
_ompd_rhinet_nunlock (locknum)
{
  _ompd_rhinet_msg_rpc_a (locknum % _ompc_n_node,
			  _rhinet_nunlock, (void *)locknum);
}


int
_ompd_rhinet_ntrylock (locknum)
{
  return _ompd_rhinet_msg_rpc_s (locknum % _ompc_n_node,
				 _rhinet_ntrylock, (void *)locknum);
}


/* -------------------------- barrier function  --------------------------- */
void
_ompd_rhinet_barrier ()
{
#if 0 /* not need this implement */
  _ompd_rhinet_msg_sync ();
#endif
  _ompd_rhinet_msg_barrier ();
  _rhinet_rma_clear ();
}


/* ------------------------------ critical -------------------------------- */
void
_ompd_rhinet_critical_lock (_ompc_lock_t *lp)
{
  int	i, node;

  int	*p = (int *)lp;


  if (*p == 0) {
    _ompd_rhinet_lock (LOCK_CRITICAL);
    if (*p == 0) {
      *p = _rhinet_lock_get_locknum ();
      _ompd_rhinet_lock_init (*p);

#ifdef BUG_4BYTE_ACCESS
      for (i=1; i<_ompc_n_node; i++) {
	node = (i + _ompc_node_id) % _ompc_n_node;
	_ompd_rhinet_msg_rmem_write (node, (char *)p, (char *)p, sizeof(int));
      }
      _ompd_rhinet_msg_sync ();
#else
      for (i=1; i<_ompc_n_node; i++) {
	node = (i + _ompc_node_id) % _ompc_n_node;
	_rhinet_write (node, (char *)p, (char *)p, sizeof(int));
      }
      _rhinet_rma_sync ();
#endif
    }
    _ompd_rhinet_unlock (LOCK_CRITICAL);
  }
  _ompd_rhinet_lock (*p);
}


void
_ompd_rhinet_critical_unlock (_ompc_lock_t *lp)
{
  int	*p = (int *)lp;


  if (*p == 0) {
    _ompc_fatal ("_ompd_critical_unlock : lock pointer is not initialized.");
  }
  _ompd_rhinet_unlock (*p);
}


/* -------------------------- dynamic scheduler  -------------------------- */
static int	loop_index;
static int	loop_end;
static int	loop_step;
static int	loop_chunksz;

struct t_sched {
  int	lb;
  int	ub;
  int	flag;
#ifdef BUG_4BYTE_ACCESS
  int	dmy;
#endif
};

static volatile struct t_sched	sched_buff, sched_dst;


static void
_ompd_rhinet_dsched (int node, void *arg)
{
  int	l, s, e;
  int	guided = (int)arg;


  if (guided) {
    l = ((loop_end - loop_index) / loop_step) / _ompc_n_node;
    l = ((l + loop_chunksz - 1) / loop_chunksz) * loop_chunksz;
    if (0 < loop_chunksz) {
      if (l < loop_chunksz) {
	l = loop_chunksz;
      }
    } else {
      if(loop_chunksz < l) {
	l = loop_chunksz;
      }
    }
    s = loop_index;
    e = loop_index + l * loop_step;

  } else {
    s = loop_index;
    e = loop_index + loop_chunksz * loop_step;
  }

  sched_buff.lb   = s;
  sched_buff.ub   = e;
  sched_buff.flag = TRUE;
  loop_index = e;

  _rhinet_write (node, (int *)&sched_dst, (int *)&sched_buff,
		 sizeof(struct t_sched));
}


void
_ompd_rhinet_dsched_init (int lb, int ub, int step, int chunksz)
{
  if (_ompc_node_id == 0) {
    loop_index   = lb;
    loop_end     = ub;
    loop_step    = step;
    loop_chunksz = chunksz;
  }
}


void
_ompd_rhinet_dsched_next (int guide, int *lb, int *ub)
{
  sched_dst.flag = FALSE;
  _ompd_rhinet_msg_rpc_a (0, _ompd_rhinet_dsched, (void *)guide);
  while (sched_dst.flag == FALSE) {
    _ompd_rhinet_poll ();
  }
  *lb = sched_dst.lb;
  *ub = sched_dst.ub;
}


/* -------------------------- ordered section ----------------------------- */
static int	ordrd_id;
static int	ordrd_step;

struct t_ordrd {
  int	current_id;
#ifdef BUG_4BYTE_ACCESS
  int	dmy;
#endif
};

static volatile struct t_ordrd  ordrd_buff;


void
_ompd_rhinet_ordrd_init (int lb, int step)
{
  ordrd_buff.current_id = lb;
  ordrd_step = step;
}


void
_ompd_rhinet_ordrd_begin (int id)
{
  int	i, n;


  while (ordrd_buff.current_id != id) {
    _ompd_rhinet_poll ();
  }
  _ompd_rhinet_lock (LOCK_ORDERED);
  ordrd_buff.current_id = id + ordrd_step;
  for (i=1; i<_ompc_n_node; i++) {
    n = (i + _ompc_node_id) % _ompc_n_node;
    _rhinet_write (n, (int *)&ordrd_buff, (int *)&ordrd_buff,
		   sizeof(struct t_ordrd));
  }
}


void
_ompd_rhinet_ordrd_end ()
{
  _rhinet_rma_sync ();
  _ompd_rhinet_unlock (LOCK_ORDERED);
}


/* ----------------------------- reduction -------------------------------- */
static int	rdctn_cur;
static int	rdctn_type[2];
static int	rdctn_op[2];
static int 	rdctn_cnt[2] = { 0, 0 };
static any_type	rdctn_outd[2];
static any_type	*rdctn_buff[2] = { NULL, NULL };

struct t_rdctn {
  any_type	data;
  int		flag;
#ifdef BUG_4BYTE_ACCESS
  int	dmy;
#endif
};

static volatile struct t_rdctn	rdctn_dst;


#define DO_REDUCTION_INTEGRAL(out,array,op,type_t,t) {	\
  any_type v;						\
  int i;						\
  v.t = out.t;						\
  switch (op) { 					\
  case OMPC_REDUCTION_PLUS: 				\
  case OMPC_REDUCTION_MINUS: 				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t += array[i].t;				\
    }							\
    break; 						\
  case OMPC_REDUCTION_MUL:				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t *= array[i].t;				\
    } 							\
    break; 						\
  case OMPC_REDUCTION_BITAND: 				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t &= array[i].t;				\
    } 							\
    break; 						\
  case OMPC_REDUCTION_BITOR: 				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t |= array[i].t;				\
    } 							\
    break; 						\
  case OMPC_REDUCTION_BITXOR: 				\
    for (i=0; i<_ompc_n_node; i++) {			\
      v.t ^= array[i].t;				\
    } 							\
    break; 						\
  case OMPC_REDUCTION_LOGAND: 				\
    if (!v.t) {						\
      break; 						\
    }							\
    for (i=0; i<_ompc_n_node; i++) { 			\
      if (!array[i].t) {				\
        v.t = 0;					\
        break;						\
      }							\
    } 							\
    break;						\
  case OMPC_REDUCTION_LOGOR: 				\
    if (v.t) {						\
      break; 						\
    }							\
    for (i=0; i<_ompc_n_node; i++) { 			\
      if (array[i].t) {					\
        v.t = 1;					\
        break;						\
      }							\
    } 							\
    break;						\
  case OMPC_REDUCTION_MIN: 				\
    for (i=0; i<_ompc_n_node; i++) {			\
      if (array[i].t < v.t) {				\
        v.t = array[i].t;				\
      }							\
    }							\
    break;						\
  case OMPC_REDUCTION_MAX:				\
    for (i=0; i<_ompc_n_node; i++) {			\
      if (v.t < array[i].t) {				\
        v.t = array[i].t;				\
      }							\
    }							\
    break;						\
  default:						\
    _ompc_fatal ("_ompd_reduction: bad op\n"); 		\
  } 							\
  out.t = v.t;		 				\
}


#define DO_REDUCTION_FLOAT(out,array,op,type_t,t) {	\
  any_type v;						\
  int i;						\
  v.t = out.t;						\
  switch (op) { 					\
  case OMPC_REDUCTION_PLUS: 				\
  case OMPC_REDUCTION_MINUS: 				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t += array[i].t;				\
    }							\
    break; 						\
  case OMPC_REDUCTION_MUL:				\
    for (i=0; i<_ompc_n_node; i++) { 			\
      v.t *= array[i].t;				\
    } 							\
    break; 						\
  case OMPC_REDUCTION_LOGAND: 				\
    if (!v.t) {						\
      break; 						\
    }							\
    for (i=0; i<_ompc_n_node; i++) { 			\
      if (!array[i].t) {				\
        v.t = 0;					\
        break;						\
      }							\
    } 							\
    break;						\
  case OMPC_REDUCTION_LOGOR: 				\
    if (v.t) {						\
      break; 						\
    }							\
    for (i=0; i<_ompc_n_node; i++) { 			\
      if (array[i].t) {					\
        v.t = 1;					\
        break;						\
      }							\
    } 							\
    break;						\
  case OMPC_REDUCTION_MIN: 				\
    for (i=0; i<_ompc_n_node; i++) {			\
      if (array[i].t < v.t) {				\
        v.t = array[i].t;				\
      }							\
    }							\
    break;						\
  case OMPC_REDUCTION_MAX:				\
    for (i=0; i<_ompc_n_node; i++) {			\
      if (v.t < array[i].t) {				\
        v.t = array[i].t;				\
      }							\
    }							\
    break;						\
  default:						\
    _ompc_fatal ("_ompd_reduction: bad op\n"); 		\
  } 							\
  out.t = v.t;		 				\
}


#define DO_REDUCTION_COMPLEX(out,array,op,type_t,t) {	\
  any_type v, vv;					\
  int i;						\
  v.t = out.t;						\
  switch (op) {						\
  case OMPC_REDUCTION_PLUS:				\
  case OMPC_REDUCTION_MINUS:				\
    for (i=0; i<_ompc_n_node; i++) {			\
      v.t.re += array[i].t.re;				\
      v.t.im += array[i].t.im;				\
    }							\
    break; 						\
  case OMPC_REDUCTION_MUL:				\
    for (i=0; i<_ompc_n_node; i++) {			\
      vv.t.re = v.t.re*array[i].t.re-v.t.im*array[i].t.im;\
      vv.t.im = v.t.re*array[i].t.im+v.t.im*array[i].t.re;\
      v.t.re = vv.t.re;					\
      v.t.im = vv.t.im;					\
    }							\
    break;						\
  default:						\
    _ompc_fatal("_ompc_reduction: bad op\n");		\
  }							\
  out.t = v.t;						\
}


static void
_ompd_rhinet_rdctn (int node, void *arg, int size)
{
  int	cur, i, n, sz;

  struct t_rdctn *r = (struct t_rdctn *)arg;


  cur = r->flag;
  if (rdctn_buff[cur] == NULL) {
    rdctn_buff[cur] = (any_type *) malloc (sizeof(any_type) * _ompc_n_node);
    if (rdctn_buff[cur] == NULL) {
      _ompc_fatal ("_ompd_rhinet_init : can not allocate reduction buffer");
    }
  }
  rdctn_buff[cur][node] = r->data;
  rdctn_cnt[cur] += 1;

  if (rdctn_cnt[cur] == _ompc_n_node) {
    switch (rdctn_type[cur]) {
    case OMPC_REDUCTION_CHAR:
      DO_REDUCTION_INTEGRAL (rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], char,c);
      break;
    case OMPC_REDUCTION_UNSIGNED_CHAR:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], unsigned char,uc);
      break;
    case OMPC_REDUCTION_SHORT:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], short,s);
      break;
    case OMPC_REDUCTION_UNSIGNED_SHORT:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], unsigned short,us);
      break;
    case OMPC_REDUCTION_SIGNED:
    case OMPC_REDUCTION_INT:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], int, i);
      break;
    case OMPC_REDUCTION_UNSIGNED_INT:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], unsigned int, ui);
      break;
    case OMPC_REDUCTION_LONG:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], long,l);
      break;
    case OMPC_REDUCTION_UNSIGNED_LONG:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], unsigned long,ul);
      break;
    case OMPC_REDUCTION_LONGLONG:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], long long, ll);
      break;
    case OMPC_REDUCTION_UNSIGNED_LONGLONG:
      DO_REDUCTION_INTEGRAL(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], unsigned long long, ull);
      break;
    case OMPC_REDUCTION_FLOAT:
      DO_REDUCTION_FLOAT(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], float,f);
      break;
    case OMPC_REDUCTION_DOUBLE:
      DO_REDUCTION_FLOAT(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], double,d);
      break;
#ifdef HAS_LONGDOUBLE
    case OMPC_REDUCTION_LONG_DOUBLE:
      DO_REDUCTION_FLOAT(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], long double,ld);
      break;
#endif /* HAS_LONGDOUBLE */
    case OMPC_REDUCTION_COMPLEX_FLOAT:
      DO_REDUCTION_COMPLEX(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], ompc_complex_float,cf);
      break;
    case OMPC_REDUCTION_COMPLEX_DOUBLE:
      DO_REDUCTION_COMPLEX(rdctn_outd[cur], rdctn_buff[cur], rdctn_op[cur], ompc_complex_double,cd);
      break;
    default:
      _ompc_fatal("_ompc_reduction: bad type");
    }

    rdctn_dst.data = rdctn_outd[cur];
    rdctn_dst.flag = TRUE;
    sz = sizeof(rdctn_dst) - sizeof(rdctn_dst) % WWORD_SIZE;
    for (i=0; i<_ompc_n_node; i++) {
      n = (i + _ompc_node_id) % _ompc_n_node;
#ifdef BUG_4BYTE_ACCESS
      sz = sizeof(rdctn_dst);
      _rhinet_write (n, (int *)&rdctn_dst, (int *)&rdctn_dst, sz);
#else
      _rhinet_write (n, (int *)&rdctn_dst, (int *)&rdctn_dst,
		     size(rdctn_dst));
#endif
    }

    rdctn_cnt[cur] = 0;
  }
}


static any_type
_ompd_to_anytype (void *p, int type)
{
  any_type	ret;


  switch (type) {
  case OMPC_REDUCTION_CHAR:
  case OMPC_REDUCTION_UNSIGNED_CHAR:
    ret.uc = *((unsigned char *)p);
    break;
  case OMPC_REDUCTION_SHORT:
  case OMPC_REDUCTION_UNSIGNED_SHORT:
    ret.us = *((unsigned short *)p);
    break;
  case OMPC_REDUCTION_SIGNED:
  case OMPC_REDUCTION_INT:
  case OMPC_REDUCTION_LONG:
  case OMPC_REDUCTION_UNSIGNED_INT:
  case OMPC_REDUCTION_UNSIGNED_LONG:
    ret.ul = *((unsigned long *)p);
    break;
  case OMPC_REDUCTION_LONGLONG:
  case OMPC_REDUCTION_UNSIGNED_LONGLONG:
    ret.ull = *((unsigned long long *)p);
    break;
  case OMPC_REDUCTION_FLOAT:
    ret.f = *((float *)p);
    break;
  case OMPC_REDUCTION_DOUBLE:
    ret.d = *((double *)p);
    break;
#ifdef HAS_LONGDOUBLE
  case OMPC_REDUCTION_LONG_DOUBLE:
    ret.ld = *((long double *)p);
    break;
#endif /* HAS_LONGDOUBLE */
  case OMPC_REDUCTION_COMPLEX_FLOAT:
    ret.cf = *((ompc_complex_float *)p);
    break;
  case OMPC_REDUCTION_COMPLEX_DOUBLE:
    ret.cd = *((ompc_complex_double *)p);
    break;
  default:
    _ompc_fatal("_ompc_reduction: bad type");
  }

  return ret;
}


static void
_ompd_from_anytype (any_type a, void *p, int type)
{
  switch (type) {
  case OMPC_REDUCTION_CHAR:
  case OMPC_REDUCTION_UNSIGNED_CHAR:
    *((unsigned char *)p) = a.uc;
    break;
  case OMPC_REDUCTION_SHORT:
  case OMPC_REDUCTION_UNSIGNED_SHORT:
    *((unsigned short *)p) = a.us;
    break;
  case OMPC_REDUCTION_SIGNED:
  case OMPC_REDUCTION_INT:
  case OMPC_REDUCTION_LONG:
  case OMPC_REDUCTION_UNSIGNED_INT:
  case OMPC_REDUCTION_UNSIGNED_LONG:
    *((unsigned long *)p) = a.ul;
    break;
  case OMPC_REDUCTION_LONGLONG:
  case OMPC_REDUCTION_UNSIGNED_LONGLONG:
    *((unsigned long long *)p) = a.ull;
    break;
  case OMPC_REDUCTION_FLOAT:
    *((float *)p) = a.f;
    break;
  case OMPC_REDUCTION_DOUBLE:
    *((double *)p) = a.d;
    break;
#ifdef HAS_LONGDOUBLE
  case OMPC_REDUCTION_LONG_DOUBLE:
    *((long double *)p) = a.ld;
    break;
#endif /* HAS_LONGDOUBLE */
  case OMPC_REDUCTION_COMPLEX_FLOAT:
    *((ompc_complex_float *)p) = a.cf; 
    break;
  case OMPC_REDUCTION_COMPLEX_DOUBLE:
    *((ompc_complex_double *)p) = a.cd;
    break;
  default:
    _ompc_fatal("_ompc_reduction: bad type");
  }
}


void
_ompd_rhinet_reduction (void *in_p, void *out_p, int type, int op)
{
  struct t_rdctn	src;


  if (_ompc_node_id == 0) {
    rdctn_outd[rdctn_cur] = _ompd_to_anytype (out_p, type);
    rdctn_type[rdctn_cur] = type;
    rdctn_op[rdctn_cur]   = op;
  }

  src.data = _ompd_to_anytype (in_p, type);
  src.flag = rdctn_cur;
  rdctn_dst.flag = FALSE;

  _ompd_rhinet_msg_rpcn_a (0, _ompd_rhinet_rdctn, &src, sizeof(src));

  while (rdctn_dst.flag == FALSE) {
    _ompd_rhinet_poll ();
  }
  _ompd_from_anytype (rdctn_dst.data, out_p, type);
  rdctn_cur = (rdctn_cur + 1) % 2;
}




/* ----------------------------- broadcast -------------------------------- */
void
_ompd_rhinet_broadcast0 (char *p, int size)
{
  int i, n;


  if (_ompc_node_id == 0) {
    for (i=1; i<_ompc_n_node; i++) {
      n = (i + _ompc_node_id) % _ompc_n_node;
      _ompd_rhinet_write (n, p, size);
    }
#if 0 /* not needed, sync at _ompd_rhinet_barrier */
    _ompd_rhinet_msg_sync ();
    _rhinet_rma_sync ();
#endif
  }

  _ompd_rhinet_barrier ();
}


/* --------------------------- atomic update ------------------------------ */
void
_ompd_rhinet_atomic_lock ()
{
  _ompd_rhinet_lock (LOCK_ATOMIC);
}


void
_ompd_rhinet_atomic_update (char *addr, int size)
{
  if (_ompc_node_id != 0) {
    _ompd_rhinet_write (0, addr, size);
    _ompd_rhinet_msg_sync ();
    _rhinet_rma_sync ();
  }
  _ompd_rhinet_unlock (LOCK_ATOMIC);
}


/* --------------- remote memory access runtime library ------------------- */
static void
_ompd_rhinet_read (int node, char *addr, int size)
{
  int p, sz, sz_align, rv;


  if (_ompc_node_id == node) {
    _ompd_rhinet_poll ();
    return;
  }

  rv = _rhinet_is_pindown (addr, size);
  if (rv == TRUE) {		      /* addr is pindown */
    p = (int)addr;
#ifdef BUG_4BYTE_ACCESS
    if (p % WWORD_SIZE) {
      sz = MAX_MSG_READ_SIZE - (p % WWORD_SIZE);
      sz = MIN(size, sz);
      _ompd_rhinet_msg_rmem_read (node, (char *)p, (char *)p, sz);
      p    = p + sz;
      size = size - sz;
      if (size == 0) {
	return;
      }
#else /* BUG_4BYTE_ACCESS */
    if (p % WORD_SIZE) {
      sz = MAX_MSG_READ_SIZE - (p % WORD_SIZE);
      sz = MIN(size, sz);
      _ompd_rhinet_msg_rmem_read (node, (char *)p, (char *)p, sz);
      p    = p + sz;
      size = size - sz;
      if (size == 0) {
	return;
      }
#endif /* BUG_4BYTE_ACCESS */
    }

#ifdef BUG_4BYTE_ACCESS
    sz_align = WWORD_SIZE;
#else
    sz_align = WORD_SIZE;
#endif
    if (size % sz_align) {
      if (size < MAX_MSG_READ_SIZE) {
	_ompd_rhinet_msg_rmem_read (node, (char *)p, (char *)p, size);
      } else {
	sz   = size % sz_align;
	size = size - sz;
	_rhinet_read (node, (int *)p, (int *)p, size);

	p = p + size;
	_ompd_rhinet_msg_rmem_read (node, (char *)p, (char *)p, sz);
      }

    } else {
      _rhinet_read (node, (int *)p, (int *)p, size);
    }

  } else {			      /* addr is not pindown */
    while (0 < size) {
      sz = MIN(size, MAX_MSG_READ_SIZE);
      _ompd_rhinet_msg_rmem_read (node, addr, addr, sz);
      addr += sz;
      size -= sz;
    }
  }
}


static void
_ompd_rhinet_write (int node, char *addr, int size)
{
  int p, sz, sz_align, rv;


  if (_ompc_node_id == node) {
    _ompd_rhinet_poll ();
    return;
  }

  rv = _rhinet_is_pindown (addr, size);
  if (rv == TRUE) {		      /* addr is pindown */
    p = (int)addr;
#ifdef BUG_4BYTE_ACCESS
    if (p % WWORD_SIZE) {
      sz = MAX_MSG_WRITE_SIZE - (p % WWORD_SIZE);
      sz = MIN(size, sz);
      _ompd_rhinet_msg_rmem_write (node, (char *)p, (char *)p, sz);
      p    = p + sz;
      size = size - sz;
      if (size == 0) {
	return;
      }
#else /* BUG_4BYTE_ACCESS */
    if (p % WORD_SIZE) {
      sz = MAX_MSG_WRITE_SIZE - (p % WORD_SIZE);
      sz = MIN(size, sz);
      _ompd_rhinet_msg_rmem_write (node, (char *)p, (char *)p, sz);
      p    = p + sz;
      size = size - sz;
      if (size == 0) {
	return;
      }
#endif /* BUG_4BYTE_ACCESS */
    }

#ifdef BUG_4BYTE_ACCESS
    sz_align = WWORD_SIZE;
#else
    sz_align = WORD_SIZE;
#endif
    if (size % sz_align) {
      if (size < MAX_MSG_WRITE_SIZE) {
	_ompd_rhinet_msg_rmem_write (node, (char *)p, (char *)p, size);
      } else {
	sz   = size % sz_align;
	size = size - sz;
	_rhinet_write (node, (int *)p, (int *)p, size);

	p = p + size;
	_ompd_rhinet_msg_rmem_write (node, (char *)p, (char *)p, sz);
      }

    } else {
      _rhinet_write (node, (int *)p, (int *)p, size);
    }

  } else {			      /* addr is not pindown */
    while (0 < size) {
      sz = MIN(size, MAX_MSG_READ_SIZE);
      _ompd_rhinet_msg_rmem_write (node, addr, addr, sz);
      addr += sz;
      size -= sz;
    }
  }
}


/* remote memory access interface */
int
_rmem_read_int (int *addr)
{
  int	rv, tmp;


  if (_ompc_node_id == 0) {
    _ompd_rhinet_poll ();
    return *addr;
  }

#if defined(BUG_RN_PULL) || defined(BUG_4BYTE_ACCESS)
  _ompd_rhinet_msg_rmem_read (0, (char *)&tmp, (char *)addr, sizeof(int));
  return tmp;
#else
  rv = _rhinet_is_pindown (addr, sizeof(int));
  if (rv == TRUE) {		      /* addr is pindown */
    _rhinet_read (0, buf, addr, sizeof(int));
    return *((int *)buf);
  } else {			      /* addr is not pindown */
    _ompd_rhinet_msg_rmem_read (0, (char *)&tmp, (char *)addr, sizeof(int));
    return tmp;
  }
#endif
}


void
_rmem_write_int (int *addr, int w)
{
  int *p, rv;


  if (_ompc_node_id == 0) {
    _ompd_rhinet_poll ();
    *addr = w;
    return;
  }

#ifdef BUG_4BYTE_ACCESS
  _ompd_rhinet_msg_rmem_write (0, (char *)addr, (char *)&w, sizeof(int));
#else
  rv = _rhinet_is_pindown (addr, sizeof(int));
  if (rv == TRUE) {		      /* addr is pindown */
    p  = (int *)buf;
    *p = w;
    _rhinet_write (0, addr, p, sizeof(int));
  } else {
    _ompd_rhinet_msg_rmem_write (0, (char *)addr, (char *)&w, sizeof(int));
  }
#endif
}


void
_rmem_read (char *addr, int size)
{
  _ompd_rhinet_read (0, addr, size);
}


void
_rmem_write (char *addr, int size)
{
  _ompd_rhinet_write (0, addr, size);
}


void
_rmem_sync ()
{
  _rhinet_rma_sync ();
  _ompd_rhinet_msg_sync ();
  _ompd_rhinet_poll ();
}


void
_rmem_read_block (char *addr, int count, int size, int stride)
{
  int	iter;


  for (iter=0; iter<count; iter++) {
    _ompd_rhinet_read (0, addr, size);
    addr += stride;
  }
}


void
_rmem_write_block (char *addr, int count, int size, int stride)
{
  int	iter;


  for (iter=0; iter<count; iter++) {
    _ompd_rhinet_write (0, addr, size);
    addr += stride;
  }
}


/* ------------------------------ for trun -------------------------------- */
#ifdef __USE_TRUN__
static char *_ompd_barrier_host;
static int  _ompd_barrier_port;


static int
_ompd_trun_init ()
{
  char *p, *str;
  int i;


  if (_ompd_barrier_host) {
    free (_ompd_barrier_host);
    _ompd_barrier_host = NULL;
  }

  p = getenv ("NODENO");
  if (p == NULL) {
    return FAILED;
  }
  _ompc_node_id = atoi (p);
  _ompc_n_node  = atoi (rindex(p, '/') + 1);
  if (_ompc_n_node < 0  ||
      _ompc_node_id < 0  ||  _ompc_n_node <= _ompc_node_id) {
    return FAILED;
  }
  
  p = getenv ("BARRIER_PORT");
  if (p == NULL) {
    return FAILED;
  }
  _ompd_barrier_host = strdup (p);
  p = strchr (_ompd_barrier_host, ':');
  if (p == NULL) {
    return FAILED;
  }
  *p = (int)NULL;
  _ompd_barrier_port = atoi(p+1);

  return SUCCESS;
}


int
_ompd_trun_barrier ()
{
  static char 		barmsg[] = "barrier";
  static char		rplmsg[] = "done";

  struct hostent	*h_ep;
  struct sockaddr_in	cli_addr;
  char			*addr;

  int			sock, rv, len;
  char			buf[1024];


  if (_ompd_barrier_host == NULL) {
    _ompc_fatal ("_ompd_trun_barrier : unknown hostname");
    return FAILED;
  }
  if (_ompd_barrier_port == 0) {
    _ompc_fatal ("_ompd_trun_barrier : illegal port number");
    return FAILED;
  }

  /* create socket */
  for (;;) {
    sock = socket (AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
      _ompc_fatal ("_ompd_trun_barrier : can not open socket");
      return FAILED;
    }

    /* get host address */
    h_ep = gethostbyname (_ompd_barrier_host);
    if (h_ep == NULL) {
      _ompc_fatal ("_ompd_trun_barrier : can not translate ip-address from hostname");
      close (sock);
      return FAILED;
    }
    addr = inet_ntoa (*((struct in_addr *)*h_ep->h_addr_list));

    /* connect */
    memset ((char *)&cli_addr, 0, sizeof (struct sockaddr_in));
    cli_addr.sin_family		= AF_INET;
    cli_addr.sin_addr.s_addr	= inet_addr (addr);
    cli_addr.sin_port		= htons (_ompd_barrier_port);

    rv = connect (sock, (struct sockaddr *)&cli_addr, sizeof (cli_addr));
    if (rv == 0) {
      break;
    } else { 
      close (sock);		      /* close socket, and retry */
    }
  } 

  /* barrier */
  sprintf (buf, "%s\n", barmsg);
  len = strlen (buf);
  rv = send (sock, buf, len, 0);
  if (rv != len) {
    _ompc_fatal ("_ompd_trun_barrier : can not send message to socket");
    close (sock);
    return FAILED;
  }

  rv = _ompd_trun_getline (sock, buf, sizeof(buf));
  if (strcmp (buf, rplmsg) != 0) {
    _ompc_fatal ("_ompd_trun_barrier : can not receive message from socket");
    close (sock);
    return FAILED;
  }

  /* close */
  close (sock);


  return SUCCESS;
}


static int 
_ompd_trun_getline (int fd, char *buf, int sz)
{
  int	cnt, rv;


  cnt = 0;
  do {
    rv = recv (fd, buf, 1, 0);
    if (rv < 0) {
      *buf = (int)NULL;		      /* read error */
      return 0;
    } else if (rv == 0) {
      *buf = (int)NULL;
      return cnt;
    } else if (*buf == '\n') {
      *buf = (int)NULL;
      return cnt;
    } else {
      buf ++;
      cnt ++;
      sz --;
    }
  } while (sz);

  return cnt;
}
#endif /* __USE_TRUN__ */
#endif /* __USE_RHINET__ */
