static char rcsid[] = "$Id: ompd_rmem.c,v 1.1.1.1 2004-02-06 18:14:58 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#include "ompd.h"
#include "ompd_server.h"
#include "ompd_object.h"

#define RMEM_N_ENTRY 1024
#define RMEM_HASH_SIZE 256	/* must be power of 2 */
#define RMEM_HASH_MASK (RMEM_HASH_SIZE-1)
#define RMEM_HASH_CODE(addr) ((((unsigned int)addr)>>4)&RMEM_HASH_MASK)

char _rmem_buffer[RMEM_N_ENTRY*sizeof(ompd_packet)];

/* hash table */
typedef struct _rmem_entry {
    struct _rmem_entry *next,*prev; /* hash backet link */
    struct _rmem_entry *dirty_link;
    void *addr;
    char size;
} rmem_entry;

static rmem_entry *_rmem_hash_table[RMEM_HASH_SIZE];
static rmem_entry _rmem_entries[RMEM_N_ENTRY];

static rmem_entry *_rmem_dirty_link;
static int _rmem_entry_lastp;		/* LRU last pointer */
static int _rmem_entry_freep;		/* free pointer */

static rmem_entry *_rmem_free_entry();
static void _rmem_cache_flush(int cleanup);


void _rmem_init()
{
    /* initialize rmem entry */
    _rmem_dirty_link = NULL;
    _rmem_cache_flush(TRUE);  /* cleanup */
}


#ifndef __USE_RHINET__
/* 
 * handlers
 */
void _rmem_READ_handler(int fd,ompd_packet *pktp)
{
    switch(pktp->size){
    case 0:	/* for sync */
	break;
    case 1:
	pktp->data.c = *((byte1 *)pktp->addr);
	break;
    case 2:
	pktp->data.s = *((byte2 *)pktp->addr);
	break;
    case 4:
	pktp->data.w = *((word4 *)pktp->addr);
	break;
    case 8:
	pktp->data.d = *((word8 *)pktp->addr);
	break;
    default:
	fprintf(stderr,"rmem_process_cmd, bad size=%d",pktp->size);
	exit(1);
    }
    _write_n(fd,(char *)pktp,sizeof(*pktp));
}

void _rmem_WRITE_handler(int fd,ompd_packet *pktp)
{
    switch(pktp->size){
    case 1:
	*((byte1 *)pktp->addr) = pktp->data.c;
	break;
    case 2:
	*((byte2 *)pktp->addr) = pktp->data.s;
	break;
    case 4:
	*((word4 *)pktp->addr) = pktp->data.w;
	break;
    case 8:
	*((word8 *)pktp->addr) = pktp->data.d;
	break;
    default:
	fprintf(stderr,"rmem_before_read, bad size=%d",pktp->size);
	exit(1);
    }
}

void _rmem_WRITE_FLUSH_handler(int fd,ompd_packet *pktp)
{
    int s,r,i;
    ompd_packet *pktq;

    s = sizeof(*pktp)*pktp->data.size;
    _read_n(fd,_rmem_buffer,s);

    pktq = (ompd_packet *)_rmem_buffer;
    for(i = 0; i < pktp->data.size; i++){
	switch(pktq->size){
	case 1:
	    *((byte1 *)pktq->addr) = pktq->data.c;
	    break;
	case 2:
	    *((byte2 *)pktq->addr) = pktq->data.s;
	    break;
	case 4:
	    *((word4 *)pktq->addr) = pktq->data.w;
	    break;
	case 8:
	    *((word8 *)pktq->addr) = pktq->data.d;
	    break;
	default:
	    fprintf(stderr,"flush read, bad size=%d",pktq->size);
	    exit(1);
	}
	pktq++;
    }
}


/* 
 * raw mode interface. (used for debug)
 */
int _rmem_read_int(int *addr)
{
    ompd_packet pkt;

    if(NODE_IS_MASTER)	return *addr;

    pkt.cmd = RMEM_READ;
    pkt.id = _ompc_node_id;
    pkt.addr = (char *)addr;
    pkt.size = 4;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    return pkt.data.w;
}

void _rmem_write_int(int *addr,int w)
{
    ompd_packet pkt;

    if(NODE_IS_MASTER){
	*addr = w;
	return;
    }
    pkt.cmd = RMEM_WRITE;
    pkt.id = _ompc_node_id;
    pkt.size = 4;
    pkt.addr = (void *)addr;
    pkt.data.w = w;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
}

/*
 * simple read for memery server
 */
void _rmem_sync()
{
    ompd_packet pkt;
    if(NODE_IS_MASTER) return;
    pkt.cmd = RMEM_READ;
    pkt.id = _ompc_node_id;
    pkt.addr = 0;	/* sync */
    pkt.size = 0;	/* sync */
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
}

void _rmem_read(char *addr,int size)
{
    ompd_packet pkt;

    /* printf("%d _rmem_read: addr=0x%x, size=%d\n",_ompc_node_id,addr,size);*/

    if(NODE_IS_MASTER) return;

    pkt.cmd = RMEM_READ;
    pkt.id = _ompc_node_id;
    pkt.addr = addr;
    pkt.size = size;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));

    switch(size){
    case 1:
	*((byte1 *)addr) = pkt.data.c;
	break;
    case 2:
	*((byte2 *)addr) = pkt.data.s;
	break;
    case 4:
	*((word4 *)addr) = pkt.data.w;
	break;
    case 8:
	*((word8 *)addr) = pkt.data.d;
	break;
    default:
	fprintf(stderr,"rmem_read, bad size=%d",size);
	exit(1);
    }
}

void _rmem_write(char *addr,int size)
{
    ompd_packet pkt;

    if(NODE_IS_MASTER) return;

    pkt.cmd = RMEM_WRITE;
    pkt.id = _ompc_node_id;
    pkt.size = size;
    pkt.addr = addr;
    
    switch(size){
    case 1:
	pkt.data.c = *((byte1 *)addr);
	break;
    case 2:
	pkt.data.s = *((byte2 *)addr);
	break;
    case 4:
	pkt.data.w = *((word4 *)addr);
	break;
    case 8:
	pkt.data.d = *((word8 *)addr);
	break;
    default:
	fprintf(stderr,"_rmem_write, bad size=%d",size);
	exit(1);
    }

    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
}

/*
 * block mode interface.
 */
void _rmem_write_block(char *addr,int count,int size,int stride)
{
    ompd_packet pkt;
    int i;

    if(NODE_IS_MASTER) return;
    pkt.cmd = RMEM_WRITE_BLOCK;
    pkt.id = _ompc_node_id;
    pkt.size = size;
    pkt.addr = (void *)addr;
    pkt.data.args[0] = count;
    pkt.data.args[1] = stride;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    for(i = 0; i < count; i++){
	_write_n(_ompd_sock_fd,addr,size);
	addr += stride;
    }
}


void _rmem_WRITE_BLOCK_handler(int fd,ompd_packet *pktp)
{
    char *addr;
    int size,stride,count,i;

    addr = (char *)pktp->addr;
    size = pktp->size;
    count = pktp->data.args[0];
    stride = pktp->data.args[1];

    /* printf("WRITE_BLOCK_h: size=%d,count=%d,stride=%d,addr=0x%x\n",
	   size,count,stride,addr); */

    for(i = 0; i < count; i++){
	_read_n(fd,addr,size);
	addr += stride;
    }
}


void _rmem_read_block(char *addr,int count,int size,int stride)
{
    ompd_packet pkt;
    int i;

    if(NODE_IS_MASTER) return;
    pkt.cmd = RMEM_READ_BLOCK;
    pkt.id = _ompc_node_id;
    pkt.size = size;
    pkt.addr = (void *)addr;
    pkt.data.args[0] = count;
    pkt.data.args[1] = stride;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    for(i = 0; i < count; i++){
	_read_n(_ompd_sock_fd,addr,size);
	addr += stride;
    }
}


void _rmem_READ_BLOCK_handler(int fd,ompd_packet *pktp)
{
    char *addr;
    int size,stride,count,i;

    addr = (char *)pktp->addr;
    size = pktp->size;
    count = pktp->data.args[0];
    stride = pktp->data.args[1];

    /* printf("READ_BLOCK_h: size=%d,count=%d,stride=%d,addr=0x%x\n",
	   size,count,stride,addr); */

    for(i = 0; i < count; i++){
	_write_n(fd,addr,size);
	addr += stride;
    }
}
#endif

/*
 * coherent memory interface.
 */
void _ompd_read_any(void *addr, int size)
{
    ompd_packet pkt;
    rmem_entry *rp,**rpp;

    if(NODE_IS_MASTER) return; /* on master, do nothing */

    /* printf("before_read: addr=%x size=%d\n",addr,size); */

    rpp = &_rmem_hash_table[RMEM_HASH_CODE(addr)];
    for(rp = *rpp; rp != NULL; rp = rp->next){
	if(rp->addr == addr) return; /* check size ? */
    }

#ifndef __USE_RHINET__
    /* not found, remote read */
    pkt.cmd = RMEM_READ;
    pkt.id = _ompc_node_id;
    pkt.size = size;
    pkt.addr = addr;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
#else
    _rmem_read (addr, size);
#endif

    /* find free entry */
    rp = _rmem_free_entry();

    rp->addr = addr;
    rp->size = size;
    rp->dirty_link = NULL;

#ifndef __USE_RHINET__
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));

    switch(size){
    case 1:
	*((byte1 *)addr) = pkt.data.c;
	break;
    case 2:
	*((byte2 *)addr) = pkt.data.s;
	break;
    case 4:
	*((word4 *)addr) = pkt.data.w;
	break;
    case 8:
	*((word8 *)addr) = pkt.data.d;
	break;
    default:
	fprintf(stderr,"rmem_read_any, bad size=%d",size);
	exit(1);
    }
#endif

    /* link to backet */
    rp->next = *rpp;
    rp->prev = NULL;
    if(*rpp != NULL) (*rpp)->prev = rp;
    *rpp = rp;
}

void _ompd_write_any(void *addr, int size)
{
    ompd_packet pkt;
    rmem_entry *rp,**rpp;

    if(NODE_IS_MASTER) return; /* on master, do nothing */

    /* printf("after_write: addr=%x size=%d\n",addr,size); */
    
    rpp = &_rmem_hash_table[RMEM_HASH_CODE(addr)];
    for(rp = *rpp; rp != NULL; rp = rp->next){
	if(rp->addr == addr){
	    /* check size ? */
	    if(rp->dirty_link == NULL){
		rp->dirty_link = _rmem_dirty_link;
		_rmem_dirty_link = rp;
	    }
	    return;
	}
    }
    
    /* find free entry */
    rp = _rmem_free_entry();

    rp->addr = addr;
    rp->size = size;
    rp->dirty_link = _rmem_dirty_link;
    _rmem_dirty_link = rp;

    /* link backet */
    rp->next = *rpp;
    rp->prev = NULL;
    if(*rpp != NULL) (*rpp)->prev = rp;
    *rpp = rp;
}

void _ompd_flush_any()
{
    _rmem_cache_flush(TRUE);
}

/* 
 *  find free entry:
 *  if free entry is available, then use it.
 *  otherwise, search entry to be freed in LRU order.
 *  A dirty entry cannot be freed without flush.
 *  If the number of dirty entry reached to RMEM_MAX_COUNT,
 *  do flush!!
 */
#define RMEM_MAX_COUNT 10 

static rmem_entry *_rmem_free_entry()
{
    int count;
    rmem_entry *rp;

    if(_rmem_entry_freep < RMEM_N_ENTRY){
	return &_rmem_entries[_rmem_entry_freep++];
    }
    count = 0;
    while(1){
	if(count++ > RMEM_MAX_COUNT) _rmem_cache_flush(FALSE);
	if(_rmem_entry_lastp >= RMEM_N_ENTRY) 
	    _rmem_entry_lastp = 0; /* wrap round */
	rp = &_rmem_entries[_rmem_entry_lastp++];
	if(rp->dirty_link == NULL) break;
    }
    /* remove rp */
    if(rp->next != NULL) 
	rp->next->prev = rp->prev;
    if(rp->prev != NULL) rp->prev->next = rp->next;
    else _rmem_hash_table[RMEM_HASH_CODE(rp->addr)] = rp->next;
    return rp;
}

static void _rmem_cache_flush(int cleanup)
{
    int i,s;
    rmem_entry *rp,*rq;
    ompd_packet *pktp,pkt;
    void *addr;

    if(_rmem_dirty_link != NULL){
#ifdef __USE_RHINET__
	for(rp = _rmem_dirty_link; rp != NULL; rp = rq) {
	  _rmem_write (rp->addr, rp->size);
	  rq = rp->dirty_link;
	  rp->dirty_link = NULL;  /* clear link */
	}
#else
	pktp = (ompd_packet *)_rmem_buffer;
	i = 0;
	for(rp = _rmem_dirty_link; rp != NULL; rp = rq){
	    addr = pktp->addr = rp->addr;
	    switch(pktp->size = rp->size){
	    case 1:
		pktp->data.c = *((byte1 *)addr);
		break;
	    case 2:
		pktp->data.s = *((byte2 *)addr);
		break;
	    case 4:
		pktp->data.w = *((word4 *)addr);
		break;
	    case 8:
		pktp->data.d = *((word8 *)addr);
		break;
	    default:
		fprintf(stderr,"flush pack, bad size=%d",rp->size);
		exit(1);
	    }
	    i++;
	    pktp++;
	    rq = rp->dirty_link;
	    rp->dirty_link = NULL;  /* clear link */
	}
	pkt.cmd = RMEM_WRITE_FLUSH;
	pkt.id = _ompc_node_id;
	pkt.data.size = i;

	_write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
	s = sizeof(pkt)*i;
	_write_n(_ompd_sock_fd,_rmem_buffer,s);
#endif
	_rmem_dirty_link = NULL;
    }

    if(cleanup){
	_rmem_entry_lastp = 0;
	_rmem_entry_freep = 0;

	/* clean up hash table */
	for(i = 0; i < RMEM_HASH_SIZE; i++)
	    _rmem_hash_table[i] = NULL;
    }
}
