static char rcsid[] = "$Id: ompd_runtime.c,v 1.1.1.1 2004-02-06 18:14:58 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
#include <stdlib.h>
#include "ompd.h"

int _ompc_sched;
int _ompc_last_loop_b;
int _ompc_last_loop_e;
int _ompc_loop_step;

/* private variable */
int _ompc_last_flag;
int _ompc_loop_sched_index;
int _ompc_loop_end;
int _ompc_loop_chunk_size;
int _ompc_loop_stride;
int _ompc_loop_id;
int _ompc_section_indx;
int _ompc_last_section_indx;

/* private copy */
enum sched_kind _ompc_runtime_sched_kind;
int _ompc_runtime_chunk_size; /* default */


/*
 * compiler runtime
 */
void _ompc_static_bsched(int *lb, int *ub, int *step);


void
_ompc_barrier ()
{
  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return;
  }

  OMPD_FLUSH_ALL ();
  OMPD_BARRIER ();
}


void
_ompc_default_sched (int *lb, int *ub, int *step)
{
  _ompc_static_bsched (lb, ub, step);
}


/* static scheduling: cyclic */
void
_ompc_static_csched (int *lb, int *up, int *step)
{
  int n,s;
  int n_thd,id;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return;
  }

  id = _ompc_node_id;
  n_thd = _ompc_n_node;
  _ompc_last_flag = 0;
  s = *step;

  /* how many iteration */
  if (s > 0) {
    n = (*up-*lb+s-1)/s;
  } else {
    n = (*up-*lb+s+1)/s;
  }

  *lb += id*s;   /* adjust low bound */
  *step = s*n_thd;
  if (n > 0 && ((n-1)%n_thd) == id) {
    _ompc_last_flag = 1;
  }
}


/* static scheduling: block */
void 
_ompc_static_bsched (int *lb, int *ub, int *step)
{
  int n_thd, id;
  int b,e,ee,s,blk_s;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return;
  }

  s = *step;
  b = *lb;
  ee = e = *ub;
  id = _ompc_node_id;
  n_thd = _ompc_n_node;
  _ompc_last_flag = 0;

  if (s > 0) {
    blk_s = (e-b+n_thd-1)/n_thd;
    blk_s = ((blk_s+s-1)/s)*s;
    b += blk_s*id;
    e = b + blk_s;
    if (e >= ee) {
      e = ee;
      if (ee > b) {
	_ompc_last_flag = 1;
      }
    }

  } else if (s < 0) {
    blk_s = (e-b-n_thd+1)/n_thd;
    blk_s = ((blk_s+s+1)/s)*s;
    b += blk_s*id;
    e = b + blk_s;
    if (e <= ee) {
      e = ee;
      if (ee < b) {
	_ompc_last_flag = 1;
      }
    }

  } else {
    return;
  }

  *lb = b;
  *ub = e;
}


/*
 * static schedule
 */
void 
_ompc_static_sched_init (int lb, int up, int step, int chunk_size)
{
  struct ompc_thread *tpp;
  int	n;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    _ompc_loop_sched_index = lb;
    _ompc_loop_end = up;
    return;
  }

  if (chunk_size < 0) {
    printf("check size is non-positive\n");
    _ompc_fatal("_ompc_static_sched_init");
  } else if (chunk_size == 0) {
    if (step > 0) {
      n = (up - lb + step - 1) / step;
    } else {
      n = (lb - up - step - 1) / -step;
    }
    chunk_size = (n + _ompc_n_node - 1) / _ompc_n_node;
  }

  chunk_size *= step;
  _ompc_loop_sched_index = lb+chunk_size*_ompc_node_id;
  _ompc_loop_chunk_size = chunk_size;
  _ompc_loop_stride = chunk_size*_ompc_n_node;
  _ompc_loop_end = up;
  _ompc_last_flag = 0;
}


int
_ompc_static_sched_next (int *lb, int *ub)
{
  int b,e;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    b = _ompc_loop_sched_index;
    e = _ompc_loop_end;
    if (b == e) {
      return FALSE;
    } else {
      *lb = b;
      *ub = e;
      _ompc_loop_sched_index = e;
      return TRUE;
    }
  }

  b = _ompc_loop_sched_index;
  _ompc_loop_sched_index += _ompc_loop_stride;
  e = b+_ompc_loop_chunk_size;

  if (_ompc_loop_chunk_size > 0) {
    if (b >= _ompc_loop_end) {
      return FALSE; 
    }
    if (e >= _ompc_loop_end) {
      e = _ompc_loop_end;
      _ompc_last_flag = 1;
    }

  } else {
    if (b <= _ompc_loop_end) {
      return FALSE;
    }
    if (e <= _ompc_loop_end) {
      e = _ompc_loop_end;
      _ompc_last_flag = 1;
    }
  }
  *lb = b;
  *ub = e;

  return TRUE;
}


/* 
 * dynamic schedule
 */
void
_ompc_dynamic_sched_init (int lb, int up, int step, int chunk_size)
{
  int id, n;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) { /* not parallel */
    _ompc_loop_sched_index = lb;
    _ompc_loop_end = up;
    return;	/* stride is not used */
  }

  if (chunk_size < 0) {
    printf ("check size is non-positive\n");
    _ompc_fatal ("_ompc_dynamic_sched_init");
  }
  if (chunk_size == 0) {
    chunk_size = 1;
  }
  _ompc_loop_chunk_size = chunk_size*step;
  _ompc_loop_end = up;
  _ompc_loop_sched_index = lb;
  _ompc_last_flag = 0;

  OMPD_BARRIER ();		/* wait for the previous "nowait" loop */
  OMPD_DYNAMIC_SCHED_INIT (lb,up,step,chunk_size);
  OMPD_BARRIER ();
}


int
_ompc_dynamic_sched_next (int *lb, int *ub)
{
  return _ompc_dynamic_sched_next_chunk (lb, ub, FALSE);
}


int 
_ompc_dynamic_sched_next_chunk (int *lb,int *ub,int guided)
{
  int b,e,l,c;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {  /* not in parallel */
    b = _ompc_loop_sched_index;
    e = _ompc_loop_end;
    if (b == e) {
      return FALSE;
    } else {
      *lb = b;
      *ub = e;
      _ompc_loop_sched_index = e;
      return TRUE;
    }
  } 
  c = _ompc_loop_chunk_size;

  /* get my chunk, set b and e */
  OMPD_DYNAMIC_SCHED_NEXT (guided, &b, &e);
    
  if (c > 0) {
    if (_ompc_loop_sched_index >= _ompc_loop_end) {
      return FALSE;
    }
  } else {
    if (_ompc_loop_sched_index <= _ompc_loop_end) {
      return FALSE;
    }
  }

  _ompc_loop_sched_index = e;

  /* adjust the last iteration */
  if (c > 0) {
    if (b >= _ompc_loop_end) {
      return FALSE; 
    }
    if (e >= _ompc_loop_end) {
      e = _ompc_loop_end;
      _ompc_last_flag = 1;
    }
  } else {
    if (b <= _ompc_loop_end) {
      return FALSE;
    }
    if (e <= _ompc_loop_end){
      e = _ompc_loop_end;
      _ompc_last_flag = 1;
    }
  }
  *lb = b;
  *ub = e;

  return TRUE;
}


/* 
 * guided schedule
 */
void
_ompc_guided_sched_init (int lb, int up, int step, int chunk_size)
{
  /* same as _ompc_dynamic_sched_init */
  _ompc_dynamic_sched_init (lb, up, step, chunk_size);
}


int
_ompc_guided_sched_next (int *lb, int *ub)
{
  return _ompc_dynamic_sched_next_chunk (lb, ub, TRUE);
}


/* 
 * runtime schedule 
 */
void 
_ompc_set_runtime_schedule (char *s)
{
  char *cp;
  cp = s;


  while (isspace((int)*cp)) {
    cp++;
  }
  if(*cp == 0) {
    return;
  }

  if (strncmp(cp,"static",6) == 0) {
    cp += 6;
    _ompc_runtime_sched_kind = SCHED_STATIC;
  } else if (strncmp(cp,"dynamic",7) == 0) {
    cp += 7;
    _ompc_runtime_sched_kind = SCHED_DYNAMIC;
  } else if (strncmp(cp,"guided",6) == 0){
    cp += 6;
    _ompc_runtime_sched_kind = SCHED_GUIDED;
  }

  while (isspace((int)*cp)) {
    cp++;
  }
  if (*cp == 0) {
    return;
  }
  if (*cp != ',') {
    goto err;
  }
  cp++;
  while (isspace((int)*cp)) {
    cp++;
  }
  if (!isdigit((int)*cp)) {
    goto err;
  }
  sscanf (cp,"%d",&_ompc_runtime_chunk_size);
  if (_ompc_runtime_chunk_size < 0) {
    _ompc_runtime_sched_kind = SCHED_NONE;
    goto err;
  }
  return;

err:
  fprintf (stderr,"OMP_SCHEDULE ='%s'",s);
  _ompc_fatal ("bad OMP_SCHEDULE");
}


void
_ompc_runtime_sched_init (int lb, int up, int step)
{
  int chunk_size,n_thd;


  chunk_size = _ompc_runtime_chunk_size;
  switch (_ompc_runtime_sched_kind) {
  case SCHED_DYNAMIC:
  case SCHED_GUIDED:
    if (chunk_size <= 0) {
      chunk_size = 1;
    }
    _ompc_dynamic_sched_init(lb, up, step, chunk_size);
    break;
  case SCHED_STATIC:
  case SCHED_NONE:
  default:
    if (NOT_IN_PARALLEL) {
      n_thd = 1;
    } else {
      n_thd = _ompc_n_node;
    }
    if (chunk_size <= 0) {
      chunk_size = (up - lb)/(step*n_thd) + (((up - lb)%(step*n_thd))?(1):(0));
      if (chunk_size <= 0) {
	chunk_size = 1;
      }
    }
    _ompc_static_sched_init(lb, up, step, chunk_size);
    break;
  }
}


int 
_ompc_runtime_sched_next (int *lb, int *ub)
{
    switch(_ompc_runtime_sched_kind){
    case SCHED_DYNAMIC:
	return _ompc_dynamic_sched_next_chunk(lb, ub,FALSE);
    case SCHED_GUIDED:
	return _ompc_dynamic_sched_next_chunk(lb, ub, TRUE);
    case SCHED_STATIC:
    case SCHED_NONE:
    default:
	return _ompc_static_sched_next(lb, ub);
    }
}


/* 
 * ordered
 */
void
_ompc_set_loop_id (int i)
{
  _ompc_loop_id = i;
}


void
_ompc_init_ordered (int lb,int step)
{
  int n;


  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return; /* not in parallel */
  }
    
  OMPD_BARRIER ();
  OMPD_ORDERED_INIT (lb,step);
  OMPD_BARRIER ();
}


void 
_ompc_ordered_begin ()
{
  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return;	/* sequential */
  }
  OMPD_ORDERED_BEGIN (_ompc_loop_id);
}


void 
_ompc_ordered_end ()
{
  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return;	/* sequential */
  }
  OMPD_ORDERED_END ();
}


/*
 * sections directives. section_id is allocated in round-robin manner.
 */
void
_ompc_section_init (int n_sections)
{
  _ompc_section_indx = _ompc_node_id;
  _ompc_last_section_indx = n_sections - 1;
  _ompc_last_flag = 0;
}


int
_ompc_section_id ()
{
  int id;
  id = _ompc_section_indx;

  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    _ompc_section_indx += 1;
  } else {
    _ompc_section_indx += _ompc_n_node;
  }
  if (id == _ompc_last_section_indx) {
    _ompc_last_flag = 1;
  }
  return id;
}


int 
_ompc_is_last ()
{
  return (_ompc_last_flag || NOT_IN_PARALLEL || IN_NESTED_PARALLEL);
}


/*
 * single construct
 */
int
_ompc_do_single ()
{
  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return TRUE; /* not in parallel */
  } else {
    /* single is the same as "master" */
    return NODE_IS_MASTER;
  }
}


int 
_ompc_is_master ()
{
  if (NOT_IN_PARALLEL || IN_NESTED_PARALLEL) {
    return TRUE;
  } else {
    return NODE_IS_MASTER;
  }
}


void
_ompc_enter_critical (_ompc_lock_t *p)
{
  OMPD_ENTER_CRITICAL(p);
}


void
_ompc_exit_critical (_ompc_lock_t *p)
{
  OMPD_EXIT_CRITICAL (p);
}


void
_ompc_flush (char *dst,int nbyte)
{
  OMPD_FLUSH_ALL ();
  OMPD_FLUSH_SYNC ();
}


void 
_ompc_reduction (void *in_p,void *out_p,int type, int op)
{
  int n_node;


  if (NOT_IN_PARALLEL) {
    switch (type) {
    case OMPC_REDUCTION_CHAR:
      *((char *)out_p) = *((char *)in_p);
      break;
    case OMPC_REDUCTION_UNSIGNED_CHAR:
      *((unsigned char *)out_p) = *((unsigned char *)in_p);
      break;
    case OMPC_REDUCTION_SHORT:
      *((short *)out_p) = *((short *)in_p);
      break;
    case OMPC_REDUCTION_UNSIGNED_SHORT:
      *((unsigned short *)out_p) = *((unsigned short *)in_p);
      break;
    case OMPC_REDUCTION_SIGNED:
    case OMPC_REDUCTION_INT:
    case OMPC_REDUCTION_LONG:
      *((long *)out_p) = *((long *)in_p);
      break;

    case OMPC_REDUCTION_UNSIGNED_INT:
    case OMPC_REDUCTION_UNSIGNED_LONG:
      *((unsigned long *)out_p) = *((unsigned long *)in_p);
      break;

    case OMPC_REDUCTION_FLOAT:
      *((float *)out_p) = *((float *)in_p);
      break;
    case OMPC_REDUCTION_DOUBLE:
      *((double *)out_p) = *((double *)in_p);
      break;
    case OMPC_REDUCTION_LONG_DOUBLE:
    case OMPC_REDUCTION_LONGLONG:
    case OMPC_REDUCTION_UNSIGNED_LONGLONG:
      _ompc_fatal("_ompc_reduction: bad type");
    }
  } else {
    OMPD_REDUCTION (in_p, out_p, type, op);
  }
}


void
_ompc_atomic_lock ()
{
  OMPD_ATOMIC_LOCK ();
}


void
_ompc_atomic_update (char *addr, int size)
{
  OMPD_ATOMIC_UPDATE (addr, size);
}
