static char rcsid[] = "$Id: ompd_server.c,v 1.1.1.1 2004-02-06 18:14:58 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */

#ifndef __USE_RHINET__
#include "ompd.h"
#include "ompd_server.h"
#include <pthread.h>
#include <netdb.h>

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

/* 
 * one-sided remote memory/control module on TCP/IP
 */

struct sockaddr_in _ompd_server_sin;
int _ompd_sock_fd;	/* descriptor */

struct ompd_client {
    int fd;
    struct sockaddr_in sin;
    struct ompd_client *next;	/* queue for suspened */
};

struct ompd_client _ompd_clients[MAX_PROC];

cfunc _ompd_msg_handlers[OMPD_MAX_CMD];

typedef struct ompd_lock {
    pthread_mutex_t mutex;
    int flag;			/* lock flag */
    struct ompd_client *head;
    struct ompd_client *tail;
} _ompd_lock_t;

/* conditional var for master */
static pthread_cond_t _ompd_master_cond_var;

/* for dynamic loop scheduling */
pthread_mutex_t _ompd_loop_mutex;
int _ompd_dynamic_index;
int _ompd_ordered_index;
int _ompd_ordered_step;

/* for atomic update */
pthread_mutex_t _ompd_atomic_mutex;

static void *_ompd_server_main(void *);
static void _setup_ompd_server_socket(int backlog);
static int _ompd_process_command(int fd);
static void _ompd_init_msg_handler(void);

static void _ompd_suspend(_ompd_lock_t *lp,int id);
static void _ompd_resume(_ompd_lock_t *lp);
static int _ompd_server_lock(int id,_ompd_lock_t *lp);
static void _ompd_server_unlock(_ompd_lock_t *lp);
static int _ompd_server_critical_lock(int id,_ompc_lock_t *p);
static void _ompd_server_loop_next(int guided,int *begin,int *end);


void 
_read_n (int fd, char *p, int size)
{
  int r;
  while(size > 0) {
    r = read(fd,p,size);
    if(r < 0) {
      fprintf(stderr,"readn: read failed\n");
      exit(1);
    }
    size -= r;
    p += r;
  }
}


void
_write_n (int fd, char *p, int size)
{
  int r;
  while(size > 0) {
    r = write (fd,p,size);
    if(r < 0){
      fprintf(stderr,"writen: write failed\n");
      exit(1);
    }
    size -= r;
    p += r;
  }
}


void _ompd_init_server()
{
    int i,s,r;
    static pthread_attr_t attr;
    static pthread_t tid;
    int sinlen;

    if(NODE_IS_MASTER) _setup_ompd_server_socket(_ompc_n_node);

    MPI_Bcast(&_ompd_server_sin,sizeof(struct sockaddr_in),
	      MPI_BYTE,0,_ompd_comm);

    sinlen = sizeof(struct sockaddr_in);
    if(NODE_IS_MASTER){

	_ompd_init_msg_handler();

	for(i = 1; i < _ompc_n_node; i++){
	    s = accept(_ompd_sock_fd,&_ompd_clients[i].sin,&sinlen);
	    if(s < 0){
		perror("accept failed");
		exit(1);
	    }
	    _ompd_clients[i].fd = s;
	}

	/* start rmem server with other thread */
	pthread_attr_init(&attr);
	r = pthread_create(&tid,&attr,_ompd_server_main,NULL);
	if(r != 0){
	    fprintf(stderr,"cannot create rmem_server thread");
	    exit(1);
	}

	pthread_cond_init(&_ompd_master_cond_var,NULL);
	pthread_mutex_init(&_ompd_loop_mutex,NULL);
	pthread_mutex_init(&_ompd_atomic_mutex,NULL);
    } else {
	_ompd_sock_fd = socket(AF_INET, SOCK_STREAM, 0);
	if(_ompd_sock_fd < 0){
	    perror("socket failed");
	    exit(1);
	}
	r = connect(_ompd_sock_fd,&_ompd_server_sin,sinlen);
	if(r < 0){
	    perror("connect failed");
	    exit(1);
	}
    }
}

static void _setup_ompd_server_socket(int backlog)
{
    int sinlen,r;
    struct sockaddr_in sin;
    char hostname[MAXHOSTNAMELEN];
    struct hostent *hp;

    _ompd_sock_fd = socket(AF_INET, SOCK_STREAM, 0);
    if(_ompd_sock_fd < 0){
	perror("socket failed");
	exit(1);
    }

    r = gethostname(hostname,MAXHOSTNAMELEN);
    if(r < 0){
	perror("hostname");
	exit(1);
    }

    hp = gethostbyname(hostname);
    if(hp == NULL){
	perror("gethostbyname");
	exit(1);
    }

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_port = htons(0);
    /* sin.sin_addr.s_addr = INADDR_ANY;*/
    bcopy(hp->h_addr,&sin.sin_addr.s_addr,hp->h_length);

    sinlen = sizeof(sin);

    r = bind(_ompd_sock_fd, (struct sockaddr *) & sin, sizeof(sin));
    if (r < 0){
	perror("bind");
	exit(1);
    }

    r = listen(_ompd_sock_fd,backlog);  /* set back log */
    if (r < 0){
	perror("listen");
	exit(1);
    }

    r = getsockname(_ompd_sock_fd, 
		    (struct sockaddr *)&_ompd_server_sin, &sinlen);
    if(r < 0){
	perror("getsockname");
	exit(1);
    }
}

#define DEF_HANDLER(cmd,handler) _ompd_msg_handlers[cmd] = (cfunc)handler

static void _ompd_init_msg_handler()
{
    DEF_HANDLER(MSG_NONE,_ompd_MSG_NONE_handler);
    DEF_HANDLER(MSG_LOOP_NEXT,_ompd_MSG_LOOP_NEXT_handler);
    DEF_HANDLER(MSG_LOCK,_ompd_MSG_LOCK_handler);
    DEF_HANDLER(MSG_UNLOCK,_ompd_MSG_UNLOCK_handler);
    DEF_HANDLER(MSG_CRITICAL_LOCK,_ompd_MSG_CRITICAL_LOCK_handler);
    DEF_HANDLER(MSG_CRITICAL_UNLOCK,_ompd_MSG_CRITICAL_UNLOCK_handler);
    DEF_HANDLER(MSG_ORDERED_BEGIN,_ompd_MSG_ORDERED_BEGIN_handler);
    DEF_HANDLER(MSG_ORDERED_END,_ompd_MSG_ORDERED_END_handler);
    DEF_HANDLER(MSG_ATOMIC_UPDATE,_ompd_MSG_ATOMIC_UPDATE_handler);

    DEF_HANDLER(RMEM_READ,_rmem_READ_handler);
    DEF_HANDLER(RMEM_WRITE,_rmem_WRITE_handler);
    DEF_HANDLER(RMEM_WRITE_FLUSH,_rmem_WRITE_FLUSH_handler);

    DEF_HANDLER(RMEM_READ_BLOCK,_rmem_READ_BLOCK_handler);
    DEF_HANDLER(RMEM_WRITE_BLOCK,_rmem_WRITE_BLOCK_handler);
}

static void *_ompd_server_main(void *arg)
{
    fd_set client_fds,rfds;
    int i,r,fd;
    int max_fd;
    ompd_packet pkt;
    cfunc handler;

    /* printf("ompd_server_main start ...\n"); */

    FD_ZERO(&client_fds);
    max_fd = 0;
    for(i = 1; i < _ompc_n_node; i++){
	fd = _ompd_clients[i].fd;
	FD_SET(fd,&client_fds);
	if(fd > max_fd) max_fd = fd;
    }

    while(1){ /* loop forever */
	rfds = client_fds;
	r = select(max_fd+1,&rfds,NULL,NULL,NULL);
	if(r <= 0){
	    perror("select");
	    exit(1);
	}

	for(i = 1; i < _ompc_n_node; i++){
	    fd = _ompd_clients[i].fd;
	    if(FD_ISSET(fd,&rfds)){
		_read_n(fd,(char *)&pkt,sizeof(pkt));
#ifdef not
		r = read(fd,&pkt,sizeof(pkt));
		if(r == 0){  /* EOF */
		    close(fd);
		    FD_CLR(fd,&client_fds);  /* remove */
		    continue;
		} else if(r != sizeof(pkt)){
		    fprintf(stderr,"ompd packet read failed\n");
		    exit(1);
		}
#endif
		if((handler = _ompd_msg_handlers[pkt.cmd]) == NULL){
		    fprintf(stderr,"ompd packet unkown cmd=%d\n",pkt.cmd);
		    exit(1);
		}
		/* printf("req (fd=%d, cmd=%d, id=%d)\n",fd,pkt.cmd,pkt.id); */
		/* call handler */
		(*handler)(fd,&pkt);
	    }
	}
    }
    /* NOT REACHED HERE */
}

void _ompd_MSG_NONE_handler(int fd,ompd_packet *pktp)
{
    /* do nothing */
}

void _ompd_loop_init_index(int start_index)
{
    _ompd_dynamic_index = start_index;
}

void _ompd_loop_next(int guided,int *begin,int *end)
{
    ompd_packet pkt;
    if(NODE_IS_MASTER){
	_ompd_server_loop_next(guided,begin,end);
    } else {
	pkt.cmd = MSG_LOOP_NEXT;
	pkt.id = _ompc_node_id;
	pkt.data.w = guided;
	_write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
	_read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    }
}

void _ompd_MSG_LOOP_NEXT_handler(int fd,ompd_packet *pktp)
{
    _ompd_server_loop_next(pktp->data.w,
			   &(pktp->data.args[0]),&(pktp->data.args[1]));
    _write_n(fd,(char *)pktp,sizeof(*pktp));
}

static void _ompd_server_loop_next(int guided,int *begin,int *end)
{
    int b,e,l,c;
    extern int _ompc_loop_chunk_size,_ompc_loop_end;

    c = _ompc_loop_chunk_size;
    pthread_mutex_lock(&_ompd_loop_mutex);
    if(guided){
	b = _ompd_dynamic_index;
	l = (_ompc_loop_end - b)/_ompc_n_node;
	l = ((l+c)/c)*c;
	if(c > 0){
	    if(c > l) l = c;
	} else {
	    if(c < l) l = c;
	}
	e = b + l;
	_ompd_dynamic_index = e;
    } else {
	b = _ompd_dynamic_index;
	e = b + c;
	_ompd_dynamic_index = e;
    }
    pthread_mutex_unlock(&_ompd_loop_mutex);
    *begin = b;
    *end = e;
}

void _ompd_init_ordered(int start,int step)
{
    _ompd_ordered_index = start;
    _ompd_ordered_step = step;
}

void _ompd_ordered_begin(int loop_id)
{
    _ompc_fatal("_ompd_ordered_begin not implmented yet");
#ifdef not
    OMPSM_WAIT((volatile int)_ompc_loop_id != (volatile int)__G__->ordered_id);
#endif
}

void _ompd_ordered_end()
{
    _ompc_fatal("_ompd_ordered_begin not implmented yet");
#ifdef not
    __G__->ordered_id += __G__->ordered_step;
#endif
}

void _ompd_MSG_ORDERED_BEGIN_handler(int fd,ompd_packet *pktp)
{
}

void _ompd_MSG_ORDERED_END_handler(int fd,ompd_packet *pktp)
{
}


/*
 * LOCK/UNLOCK
 */

_ompc_lock_t _ompd_lock_alloc()
{
    _ompd_lock_t *lp;
    lp = (_ompd_lock_t *)malloc(sizeof(_ompd_lock_t));
    if(lp == NULL) _ompc_fatal("_ompd_lock_alloc: cannot malloc");
    pthread_mutex_init(&lp->mutex,NULL);
    lp->flag = FALSE;
    lp->head = lp->tail = NULL;
    return (_ompc_lock_t)lp;
}

void _ompd_lock_free(_ompc_lock_t l)
{
    _ompd_lock_t *lp = (_ompd_lock_t *)l;
    pthread_mutex_destroy(&lp->mutex);
}

static void _ompd_suspend(_ompd_lock_t *lp,int id)
{
    struct ompd_client *q;

    q = &_ompd_clients[id];
    q->next = NULL;
    if(lp->head == NULL){
	lp->head = lp->tail = q;
    } else {
	lp->tail->next = q;
	lp->tail = q;
    }
}

static void _ompd_resume(_ompd_lock_t *lp)
{
    ompd_packet pkt;
    struct ompd_client *q;
    int id;

    if((q = lp->head) == NULL) return;
    lp->head = q->next;
    id = q - &_ompd_clients[0];
    if(id == 0){ 	/* master */
	pthread_cond_signal(&_ompd_master_cond_var);
    } else {
	pkt.cmd = MSG_NONE;
	_write_n(q->fd,(char *)&pkt,sizeof(pkt));
    }
}

static int _ompd_server_lock(int id,_ompd_lock_t *lp)
{
    int r;
    pthread_mutex_lock(&lp->mutex);
    if(lp->flag){  /* already locked */
	if(id == 0) pthread_cond_wait(&_ompd_master_cond_var,&lp->mutex);
	else _ompd_suspend(lp,id);
	r = FALSE;
    } else {
	lp->flag = TRUE; /* lock it! */
	r = TRUE;
    }
    pthread_mutex_unlock(&lp->mutex);
    return r;
}

static void _ompd_server_unlock(_ompd_lock_t *lp)
{
    pthread_mutex_lock(&lp->mutex);
    if(lp->head != NULL) _ompd_resume(lp);
    else lp->flag = FALSE; 	/* unlock */
    pthread_mutex_unlock(&lp->mutex);
}

void _ompd_lock(_ompd_lock_t *lp)
{
    ompd_packet pkt;

    if(NODE_IS_MASTER){
	_ompd_server_lock(0,lp);
	return;
    }
    pkt.cmd = MSG_LOCK;
    pkt.id = _ompc_node_id;
    pkt.addr = (char *)lp;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
}

void _ompd_MSG_LOCK_handler(int fd,ompd_packet *pktp)
{
    if(_ompd_server_lock(pktp->id,(_ompd_lock_t *)(pktp->addr)))
	_write_n(fd,(char *)pktp,sizeof(*pktp));
}

void _ompd_unlock(_ompd_lock_t *lp)
{
    ompd_packet pkt;
    if(NODE_IS_MASTER){
	_ompd_server_unlock(lp);
    } else {
	pkt.cmd = MSG_UNLOCK;
	pkt.id = _ompc_node_id;
	pkt.addr = (char *)lp;
	_write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    }
}

void _ompd_MSG_UNLOCK_handler(int fd,ompd_packet *pktp)
{
    _ompd_server_unlock(pktp->addr);
}

static int _ompd_server_critical_lock(int id,_ompc_lock_t *p)
{
    _ompd_lock_t *lp;
    if(*p == NULL){
	/* use atomic_mutex */
	pthread_mutex_lock(&_ompd_atomic_mutex);
	if(*p == NULL){
	    *p = (_ompc_lock_t)_ompd_lock_alloc();
	}
	pthread_mutex_unlock(&_ompd_atomic_mutex);
    }
    return _ompd_server_lock(id,(_ompd_lock_t *)*p);
}


#ifndef __USE_RHINET__
void _ompd_critical_lock(_ompc_lock_t *p)
{
    ompd_packet pkt;

    if(NODE_IS_MASTER){
	_ompd_server_critical_lock(0,(_ompc_lock_t *)p);
	return;
    } 
    pkt.cmd = MSG_CRITICAL_LOCK;
    pkt.id = _ompc_node_id;
    pkt.addr = (char *)p;
    _write_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
    _read_n(_ompd_sock_fd,(char *)&pkt,sizeof(pkt));
}
#endif

void _ompd_MSG_CRITICAL_LOCK_handler(int fd,ompd_packet *pktp)
{
    if(_ompd_server_critical_lock(pktp->id,(_ompc_lock_t *)(pktp->addr)))
	_write_n(fd,(char *)pktp,sizeof(*pktp));
}

#ifndef __USE_RHINET__
void _ompd_critical_unlock(_ompc_lock_t *p)
{
    ompd_packet pkt;
    if(NODE_IS_MASTER){
	_ompd_server_unlock(*((_ompd_lock_t **)p));
    } else {
	pkt.cmd = MSG_CRITICAL_UNLOCK;
	pkt.id = _ompc_node_id;
	pkt.addr = (char *)p;
	write(_ompd_sock_fd,&pkt,sizeof(pkt));
    }
}
#endif

void _ompd_MSG_CRITICAL_UNLOCK_handler(int fd,ompd_packet *pktp)
{
    _ompd_server_unlock(*((_ompd_lock_t **)(pktp->addr)));
}

#ifdef not
void _ompd_atomic_update(char *addr,int size)
{
    if(NODE_IS_MASTER){
	pthread_mutex_unlock(&_ompd_atomic_mutex);
    } else {
	pkt.cmd = MSG_ATOMIC_UPDATE;
	pkt.id = _ompc_node_id;
	pkt.addr = (char *)lp;
	_write_n(_ompd_sock_fd,&pkt,sizeof(*pktp));
	_read_n(_ompd_sock_fd,&pkt,sizeof(*pktp));
    }
}
#endif

void _ompd_MSG_ATOMIC_UPDATE_handler(int fd,ompd_packet *pktp)
{
}

/*
 *  reduction
 */
static int _ompd_MPI_initialized = FALSE;
static int _ompd_MPI_data_types[OMPC_REDUCTION_MAX_TYPES];
static int _ompd_MPI_reduction_ops[OMPC_REDUCTION_MAX_OPS];

static void _ompd_init_reduction_table()
{
    _ompd_MPI_data_types[OMPC_REDUCTION_CHAR] = MPI_CHAR;
    _ompd_MPI_data_types[OMPC_REDUCTION_UNSIGNED_CHAR] = MPI_UNSIGNED_CHAR;
    _ompd_MPI_data_types[OMPC_REDUCTION_SHORT] = MPI_SHORT;
    _ompd_MPI_data_types[OMPC_REDUCTION_UNSIGNED_SHORT] = MPI_UNSIGNED_SHORT;
    _ompd_MPI_data_types[OMPC_REDUCTION_SIGNED] = MPI_INT;
    _ompd_MPI_data_types[OMPC_REDUCTION_INT] = MPI_INT;
    _ompd_MPI_data_types[OMPC_REDUCTION_UNSIGNED_INT] = MPI_UNSIGNED;
    _ompd_MPI_data_types[OMPC_REDUCTION_LONG] = MPI_LONG;
    _ompd_MPI_data_types[OMPC_REDUCTION_UNSIGNED_LONG] = MPI_UNSIGNED_LONG;
    _ompd_MPI_data_types[OMPC_REDUCTION_LONGLONG] = MPI_LONG_LONG_INT;
    /* _ompd_MPI_data_types[OMPC_REDUCTION_UNSIGNED_LONGLONG] */
    _ompd_MPI_data_types[OMPC_REDUCTION_FLOAT] = MPI_FLOAT;
    _ompd_MPI_data_types[OMPC_REDUCTION_DOUBLE] = MPI_DOUBLE;
    _ompd_MPI_data_types[OMPC_REDUCTION_LONG_DOUBLE] = MPI_LONG_DOUBLE;

    _ompd_MPI_reduction_ops[OMPC_REDUCTION_PLUS] = MPI_SUM;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_MINUS] = MPI_SUM;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_MUL] = MPI_PROD;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_BITAND] = MPI_BAND;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_BITOR] = MPI_BOR;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_BITXOR] = MPI_BXOR;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_LOGAND] = MPI_LAND;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_LOGOR] = MPI_LOR;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_MIN] = MPI_MIN;
    _ompd_MPI_reduction_ops[OMPC_REDUCTION_MAX] = MPI_MAX;
}

void _ompd_reduction(void *in_p,void *out_p,int type, int op)
{
    if(!_ompd_MPI_initialized) _ompd_init_reduction_table();
    MPI_Allreduce(in_p,out_p,1,
		  _ompd_MPI_data_types[type],
		  _ompd_MPI_reduction_ops[op],
		  _ompd_comm);
}
#endif /* __USE_RHINET__ */
