
#include "ompd.h"
#include "ompd_server.h"
#include "ompd_object.h"
#include "ompd_tdl.h"

/* 
 * ompd TDL interface 
 */
void _ompd_init_TDL()
{
    _TDL_init("ompd_tdl");
}

static void _ompd_block_TDL_array(ompd_array *dp);

void _ompd_init_TDL_array(ompd_array *dp)
{
    ompd_TDL_info *info;
    info = (ompd_TDL_info *)_ompd_malloc(sizeof(*info));
    /* initially, all data is located in HOME */
    info->current_desc = TDL_desc_create(dp->addr,dp->size,TDL_MASTER_ALL);
    info->num_sched = 0;
    {
	int i;
	for (i = 0; i < MAX_TDL_SCHED; ++i)
	    info->sched[i].used = FALSE;
    }
    dp->TDL_info = (void *)info;
    /* remap to block distribution */
    _ompd_block_TDL_array(dp);
}

static TDL_desc access_desc;

void _ompd_TDL_inspect_home(void *addr,int size)
{
    TDL_inspect_home(addr,size,access_desc);
}

void _ompd_TDL_inspect_sharing(void *addr,int size)
{
    TDL_inspect_sharing(addr,size,access_desc);
}

void _ompd_TDL_inspect_ref(void *addr,int size)
{
    TDL_inspect_ref(addr,size,access_desc);
}

void _ompd_TDL_inspect_modify(void *addr,int size)
{
    TDL_inspect_modify(addr,size,access_desc);
}

extern ompd_access_range **
ompd_copy_access_range(int n_rgn,ompd_access_range *rp[]);

extern int
ompd_compare_access_range(ompd_access_range *rp,
			  ompd_access_range *rq);

extern void
ompd_free_access_range(int n_rgn, ompd_access_range *rp[]);

void
_ompd_TDL_insert_sched(ompd_TDL_info *info, int rw_flag, int n_rgn,
		       ompd_access_range **rp, TDL_desc desc)
{
    ompd_TDL_comm_sched *s = info->sched;
    int num_sched = info->num_sched;
    if (s[num_sched].used) {
	TDL_desc_free(s[num_sched].desc);
	ompd_free_access_range(s[num_sched].n_rgn, s[num_sched].rp);
    }
    s[num_sched].used = TRUE;
    s[num_sched].rw_flag = rw_flag;
    s[num_sched].n_rgn = n_rgn;
    s[num_sched].rp = ompd_copy_access_range(n_rgn, rp);
    s[num_sched].desc = desc;

    debug(printf("_ompd_TDL_insert_sched: [%d] rw %d rgn %d %p desc %p\n",
		 num_sched, rw_flag, n_rgn, rp, desc));
    debug(fflush(stdout));

    if (++info->num_sched >= MAX_TDL_SCHED)
	info->num_sched = 0;
}

TDL_desc
_ompd_TDL_check_sched(ompd_TDL_info *info, int rw_flag, int n_rgn,
		      ompd_access_range **rp)
{
    ompd_TDL_comm_sched *s = info->sched;
    int i = 0;
    while (s[i].used && i < MAX_TDL_SCHED)
    {
	int j;
	if (n_rgn == s[i].n_rgn && s[i].rw_flag == rw_flag) {
	    for (j = 0; j < n_rgn; ++j) {
		assert(s[i].rp[j]);
		assert(rp[j]);
		if (!ompd_compare_access_range(s[i].rp[j], rp[j]))
		    break;
	    }
	    if (j == n_rgn)
		return s[i].desc;
	}
	++i;
    }
    return NULL;
}

void _ompd_read_TDL_array(ompd_array *dp,int n_rgn,
			  ompd_access_range *rp[])
{
    ompd_TDL_info *info = (ompd_TDL_info *)dp->TDL_info;
    /*
     * for the same access pattern, re-use the same access_desc
     */
    if ((access_desc = _ompd_TDL_check_sched(info, OMPD_TDL_READ, n_rgn, rp)) == NULL)
    {
	int i;

	debug(puts("_ompd_read_TDL_array: inspecter"));
	debug(fflush(stdout));

	/*  Create a new TDL descriptor. */
	access_desc = TDL_desc_create_desc(info->current_desc);
	/* inspect home */
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,rp[i],ACCESS_INSPECT_HOME);
	TDL_desc_fix_home(access_desc);
	/* inspect ref */
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,rp[i],ACCESS_INSPECT_REF);
	TDL_desc_fix_ref(access_desc);

	/* insert this communication schedule. */

	_ompd_TDL_insert_sched(info, OMPD_TDL_READ, n_rgn, rp, access_desc);
    }
    /* finally, remap_ref */
    TDL_remap_ref(info->current_desc,access_desc);

    /*  It is not necessary to free the descriptor, because it is now
	cached. */
}

void _ompd_write_TDL_array(ompd_array *dp,int n_rgn,
			   ompd_access_range **rp)
{
    ompd_TDL_info *info = (ompd_TDL_info *)dp->TDL_info;
    /*
     * for the same access pattern and the same mapping,
     *     re-use the same access_desc 
     */
    if ((access_desc = _ompd_TDL_check_sched(info, OMPD_TDL_WRITE,
					     n_rgn, rp)) == NULL)
    {
	int i;

	debug(puts("_ompd_write_TDL_array: inspecter"));
	fflush(stdout);

	access_desc = TDL_desc_create_desc(info->current_desc);
	/* use the same mapping */
	TDL_desc_copy_home(info->current_desc, access_desc);
    
	/* inspect sharing */
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,rp[i],ACCESS_INSPECT_SHARING);
	TDL_desc_fix_sharing(access_desc);

	/* inspect modify */
	for(i = 0; i < n_rgn; i++)
	    _ompd_access_array(dp,rp[i],ACCESS_INSPECT_MODIFY);
	TDL_desc_fix_modify(access_desc);

	/* insert this communication schedule. */

	_ompd_TDL_insert_sched(info, OMPD_TDL_WRITE, n_rgn, rp, access_desc);
    }
    /*  If distribution is changed, it is necessary to re-fix
	descriptor for modification. */
    if (!TDL_desc_compare_home(info->current_desc, access_desc)) {

	debug(puts("_ompd_write_TDL_array: distribution changed"));
	debug(fflush(stdout));

	TDL_desc_copy_home(info->current_desc, access_desc);
	TDL_desc_fix_modify(access_desc);
    }
    TDL_update_modify(access_desc);

    /*  It is not necessary to free the descriptor, because it is now
	cached. */
}

void _ompd_flush_TDL_array(ompd_array *dp)
{
    /* nothing ??? */
}

/* move all data to master */
void _ompd_master_TDL_array(ompd_array *dp)
{
    TDL_desc home_desc;
    ompd_TDL_info *info = (ompd_TDL_info *)dp->TDL_info;

    home_desc = TDL_desc_create_desc(info->current_desc);
    TDL_desc_dist(home_desc,TDL_MASTER);
    TDL_remap_home_force(info->current_desc,home_desc);
    TDL_desc_free(home_desc);
}

/* change distribution to block */
static void _ompd_block_TDL_array(ompd_array *dp)
{
    TDL_desc home_desc;
    ompd_TDL_info *info = (ompd_TDL_info *)dp->TDL_info;

    home_desc = TDL_desc_create_desc(info->current_desc);
    TDL_desc_dist(home_desc,TDL_BLOCK_ALL);
    TDL_remap_home_force(info->current_desc,home_desc);
    TDL_desc_free(home_desc);
}
