static char rcsid[] = "$Id: ompd_thread.c,v 1.1.1.1 2004-02-06 18:14:59 msato Exp $";
/* 
 * $RWC_Release$
 * $RWC_Copyright$
 */
/* 
 * RWC Omni OpenMP compiler Run-time libraries 
 * for distributed memory 
 */

#include <unistd.h>
#include <stdlib.h>
#include "ompd.h"

int _ompc_debug_flag = FALSE;
int _ompc_n_proc;

volatile int _ompc_nested;	/* nested enable/disable */
volatile int _ompc_dynamic;	/* dynamic enable/disable */
volatile int _ompc_max_threads;	/* max number of thread */
volatile int _ompc_num_threads;	/* number of team member? */

/* private node data */
int _ompc_n_node = -1;
int _ompc_node_id = -1;

/* parallel region nest level */
int _ompc_in_parallel;

/* thread invocation message */
struct ompd_thd_info _ompd_thd_info;

int  _ompd_mode = OMPD_MODE_NONE;

/* shared data */
struct ompd_globals _ompd_globals;

static void _ompd_slave_main(void);
static void _ompd_init_env(void);
static void _ompd_BCAST_globals();

/* 
 * initialize library
 */
void
_ompd_init (int *argc_p,char ***argv_p)
{
  char	*cp;
  int	t, r, val;


  OMPD_INITIALIZE (argc_p, argv_p);

  /* initialize rmem entry */
  _rmem_init();

  if (!NODE_IS_MASTER) {
    _ompd_slave_main ();
    OMPD_FINALIZE ();
    if(_ompc_debug_flag) {
      printf("<%d> slave thread terminate.\n", _ompc_node_id);
    }
    exit(0);
    /* NO REACHED HERE */
  }

  /* 
   * Master only run through down.
   */
  _ompd_init_env ();

  /* broadcast global data */
  _ompd_BCAST_globals ();

  OMPD_SERVER_INIT ();

  if(_ompd_mode == OMPD_TDL_MODE) {
    _ompd_init_TDL ();
  }

  if(_ompc_debug_flag) {
    printf("init end(Master)\n");
  }
}


static void
_ompd_init_env ()
{
  char	*cp;
  int	t, r, val;
  long	lnp;


  cp = getenv ("OMPC_DEBUG");
  if (cp != NULL) {
    _ompd_globals.debug_flag = TRUE;
    fprintf (stderr,"debug flag on ...\n");
  }
    
  cp = getenv("OMPC_NUM_PROCS");   /* processor */
  if ( cp != NULL ) {
    sscanf(cp, "%d", &val);
    if(val <= 0) {
      _ompc_fatal("bad OMPC_NUM_PROCS(<= 0)");
    }
    _ompc_n_proc = val;
  }

  _ompc_nested = 0;		      /* nested enable/disable */
  _ompc_dynamic = 0;		      /* dynamic enable/disable */
  _ompc_max_threads = _ompc_n_proc;   /* max number of thread, default */

  cp = getenv ("OMP_SCHEDULE");
  if (cp != NULL) {
    _ompc_set_runtime_schedule(cp);
    _ompd_globals.runtime_chunk_size = _ompc_runtime_chunk_size;
    _ompd_globals.runtime_sched_kind = _ompc_runtime_sched_kind;
  }

  cp = getenv ("OMP_DYNAMIC");
  if (cp != NULL && 
      (strcmp(cp,"TRUE") == 0 || strcmp(cp,"true") == 0)) {
    _ompc_dynamic = 1;
  }

  cp = getenv ("OMP_NESTED");
  if (cp != NULL && 
      (strcmp(cp,"TRUE") == 0 || strcmp(cp,"true") == 0)) {
    _ompc_nested = 1;
  }

  cp = getenv ("OMP_NUM_THREADS");    /* a number of team member */
  if ( cp == NULL ) {
    _ompc_num_threads = _ompc_n_proc;
  } else {
    sscanf (cp, "%d", &val);
    if (val <= 0) {
      _ompc_fatal("bad OMP_NUM_THREADS(<= 0)");
    }
    _ompc_num_threads = val;
    if (_ompc_num_threads > _ompc_max_threads) {
      _ompc_max_threads = _ompc_num_threads;
    }
  }

  if ((cp = getenv("OMPD_MODE")) != NULL) {
    if (strcmp(cp,"nocache") == 0) {
      printf("nocache mode ...\n");
      fflush(stdout);
      _ompd_globals.ompd_mode = OMPD_NO_CACHE;
    } else if (strcmp(cp,"tdl") == 0) {
      printf("TDL mode ...\n");
      _ompd_globals.ompd_mode = OMPD_TDL_MODE;
    } else {
      fprintf (stderr,"Unknown OMPD_MODE: '%s'\n",cp);
    }
  } 
}

/* finalize */
void 
_ompd_finalize ()
{
  struct ompd_thd_info *tp = &_ompd_thd_info;

  tp->func = NULL;
  OMPD_BCAST ((char *)tp, sizeof(_ompd_thd_info));
  OMPD_FINALIZE ();
}


void
_ompc_fatal (char * msg)
{
    fprintf(stderr, "<%d> OMPD FATAL: %s\n", _ompc_node_id, msg);
#ifdef __USE_RHINET__
    for (;;) ;
#endif
    exit(-1);
}


static void 
_ompd_BCAST_globals ()
{
  OMPD_BCAST ((char *)&_ompd_globals, sizeof(_ompd_globals));
  _ompc_debug_flag = _ompd_globals.debug_flag;
  _ompd_mode = _ompd_globals.ompd_mode;
  _ompc_runtime_chunk_size = _ompd_globals.runtime_chunk_size;
  _ompc_runtime_sched_kind =_ompd_globals.runtime_sched_kind;
}


/*
 * slave main
 */
static void 
_ompd_slave_main ()
{
  struct ompd_thd_info *tp = &_ompd_thd_info;

  /* first receive global data from host */
  _ompd_BCAST_globals ();

  OMPD_SERVER_INIT ();

  if(_ompd_mode == OMPD_TDL_MODE) {
    _ompd_init_TDL();
  }

  _ompc_in_parallel = 1;  /* always in parallel on slave */

  for(;;) {
    /* wait task invocation */
    OMPD_BCAST ((char *)tp, sizeof(_ompd_thd_info));
    if(tp->func == (cfunc)NULL) {
      break; /* exit */
    }

    (*tp->func)(tp->_ompd_thd_args);

    OMPD_FLUSH_ALL ();
    OMPD_WAIT_ALL ();
  }
}


/* called from compiled code. */
void 
_ompc_do_parallel (cfunc f, void **args, int narg)
{
  struct ompd_thd_info *tp = &_ompd_thd_info;
  int i,r,rr;


  _ompc_in_parallel ++;
  if(_ompc_in_parallel > 1) {
    (*f)(args);
  } else {
    tp->func = f;
    if (args != NULL) {
      tp->narg = narg;
    } else {
      tp->narg = 0;
    }
    for (i=0; i<tp->narg; i++) {	
      if (PARALLEL_MAX_ARGS < i) {
	_ompc_fatal ("_ompc_do_parallel : argument buffer overflow");
      }
      tp->_ompd_thd_args[i] = args[i];
    }
    OMPD_BCAST ((char *)tp, sizeof(_ompd_thd_info));

    (*f)(args);

    OMPD_WAIT_ALL ();
  }
  _ompc_in_parallel --;
}

#ifdef not_yet
void
_ompc_do_parallel_if(int cond, cfunc f, void *args)
{
  struct ompc_thread *tp;
  if(cond) {
    _ompc_do_parallel(f,args);
  } else {
    tp = _ompc_current_thread();
    if(args == NULL) {
      (*f)(tp);
    } else {
      (*f)(args, tp);
    }
  }
}

#endif


void
_ompc_terminate (int exitcode)
{
  _ompd_finalize();
  exit (exitcode);
}
