#include <stdlib.h>
#include <stdio.h>

/*
 *  Convert a in COO format to b in CSR format.
 */

void
COO_to_CSR(n, N, ai, aj, a, idx, ib, ie, b)
    int n, N;
    int *ai, *aj;
    int *idx, *ib, *ie;
    double *a, *b;
{
    int *freq, *od;
    int i;

    freq = (int*)calloc(n, sizeof(int));
    od   = (int*)calloc(n, sizeof(int));

    for (i = 0; i < N; ++i)
	++freq[ai[i]];

    ib[0] = 0;
    for (i = 1; i < n; ++i)
	ib[i] = ie[i - 1] = ib[i - 1] + freq[i - 1];
    ie[n - 1] = ib[n - 1] + freq[n - 1];

    for (i = 0; i < N; ++i) {
	int aii = ai[i];
	int j = ib[aii] + od[aii];

	idx[j] = aj[i];
	b[j] = a[i];
	++od[aii];
    }

    free(freq);
    free(od);
}

/*
 *  Convert a in COO format to b in CSC format.
 */

void
COO_to_CSC(n, N, ai, aj, a, idx, ib, ie, b)
    int n, N;
    int *ai, *aj;
    int *idx, *ib, *ie;
    double *a, *b;
{
    int *freq, *od;
    int i;

    freq = (int*)calloc(n, sizeof(int));
    od   = (int*)calloc(n, sizeof(int));

    for (i = 0; i < N; ++i)
	++freq[aj[i]];

    ib[0] = 0;
    for (i = 1; i < n; ++i)
	ib[i] = ie[i - 1] = ib[i - 1] + freq[i - 1];
    ie[n - 1] = ib[n - 1] + freq[n - 1];

    for (i = 0; i < N; ++i) {
	int aji = aj[i];
	int j = ib[aji] + od[aji];

	idx[j] = ai[i];
	b[j] = a[i];
	++od[aji];
    }

    free(freq);
    free(od);
}

/*
 *
 */

void
print_vec(n, a)
    int n;
    double *a;
{
    int i, j;

    puts("VEC:");
/*
    for (i = 0; i < n; ++i)
	printf("%g ", a[i]);
	*/
    for (i = 0; i < n; ++i) {
	for (j = 0; j < 31; ++j)
	    printf("%g\n", a[i + j]);
	puts("");
	i += j;
    }
    puts("");
}

void
print_COO(N, ai, aj, a)
    int N;
    int *ai, *aj;
    double *a;
{
    int i;

    puts("VAL:");
    for (i = 0; i < N; ++i) {
	printf("%g ", a[i]);
    }
    puts("");
    puts("IDX:");
    for (i = 0; i < N; ++i) {
	printf("%d ", ai[i]);
    }
    puts("");
    puts("IDY:");
    for (i = 0; i < N; ++i) {
	printf("%d ", aj[i]);
    }
    puts("");
}

void
print_CSR(n, idx, ib, ie, a)
    int n;
    int *idx, *ib, *ie;
    double* a;
{
    int i;
    int N = ie[n - 1];

    puts("VAL:");
    for (i = 0; i < N; ++i) {
	printf("%g ", a[i]);
    }
    puts("");
    puts("IDX:");
    for (i = 0; i < N; ++i) {
	printf("%d ", idx[i]);
    }
    puts("");
    puts("IB:");
    for (i = 0; i < n; ++i) {
	printf("%d ", ib[i]);
    }
    puts("");
    puts("IE:");
    for (i = 0; i < n; ++i) {
	printf("%d ", ie[i]);
    }
    puts("");
}

#if 0
#include "psc98_mat.h"

print_PSC98(n, idx, a)
    int n;
    psc98_idx idx;
    psc98_mat a;
{
    int i, nz = 0;
    for (i = 0; i < n; ++i) {
	int j;
	printf("%d:", i);
	for (j = 0; idx[i][j] != -1 && j < NZMAX; ++j) {
	    printf(" (%d; %g)", idx[i][j], a[i][j]);
/*	    printf("%d %d\n", i, idx[i][j]); */
	    ++nz;
	}
	puts("");
    }
    printf("num nzero = %d\n", nz);
}
#endif

#if 0

#define n	5
#define N	14

double val[N] = {
    11, 51, 31, 32, 34, 52, 13, 23, 33, 14, 24, 42, 55, 44
};
int indx[N] = {
     0,  4,  2,  2,  2,  4,  0,  1,  2,  0,  1,  3,  4,  3
};
int indy[N] = {
     0,  0,  0,  1,  3,  1,  2,  2,  2,  3,  3,  1,  4,  3
};

main()
{
    int idx[N], ib[n], ie[n];
    double b[N];

    print_COO(N, indx, indy, val);

    COO_to_CSR(n, N, indx, indy, val, idx, ib, ie, b);

    print_CSR(n, idx, ib, ie, b);

    COO_to_CSC(n, N, indx, indy, val, idx, ib, ie, b);

    print_CSR(n, idx, ib, ie, b);
}

#endif
