/*
 *  $Id: mult-test.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include <assert.h>
#include <stdlib.h>
#include <tdl.h>

extern void
prob1(int n, int m, int *ai, int *aj, double *a,
      int *size, int *nzero);

extern void
COO_to_CSR(int n, int N, int *ai, int *aj, double *a,
	   int *idx, int *ib, int *ie, double *b);

extern void
mult(int size, int *idx, int *ib, double *a, double *b, double *c,
     TDL_desc idx_desc, TDL_desc ib_desc,
     TDL_desc a_desc, TDL_desc b_desc, TDL_desc c_desc);

#define N1     5
#define N2     5
#define N     (N1 * N2)
#define MAX_NZERO  (5 * N)

/*
 *  Distributed (global) arrays
 */

int *ib, *idx;
TDL_desc ib_desc, idx_desc;
double *a, *b, *c;
TDL_desc a_desc, b_desc, c_desc;

int *ai, *aj;
double *el;

int
main(argc, argv)
    int argc;
    char* argv[];
{
    int n1 = N1, n2 = N2;
    int mat_size, nzero;
    int i;

    TDL_init(&argc, &argv);

    /**  Initialization  **/

    ai = (int*)malloc(MAX_NZERO * sizeof(int));
    aj = (int*)malloc(MAX_NZERO * sizeof(int));
    el = (double*)malloc(MAX_NZERO * sizeof(double));

    /**  Set up  **/

    prob1(n1, n2, ai, aj, el, &mat_size, &nzero);
    printf("mat_size = %d, nzero = %d\n", mat_size, nzero);
/* print_COO(nzero, ai, aj, el); */

    assert(nzero <= MAX_NZERO);
    assert(mat_size <= N);

    /**  Initialization  **/

    ib = (int*)malloc(sizeof(int) * (mat_size + 1));
    ib_desc = TDL_desc_create(ib, sizeof(int) * (mat_size + 1), TDL_BLOCK);
    idx = (int*)malloc(sizeof(int) * nzero);
    idx_desc = TDL_desc_create(idx, sizeof(int) * nzero, TDL_BLOCK);
    a = (double*)malloc(sizeof(double) * nzero);
    a_desc = TDL_desc_create(a, sizeof(double) * nzero, TDL_BLOCK);
    b = (double*)malloc(sizeof(double) * mat_size);
    b_desc = TDL_desc_create(b, sizeof(double) * mat_size, TDL_BLOCK);
    c = (double*)malloc(sizeof(double) * mat_size);
    c_desc = TDL_desc_create(c, sizeof(double) * mat_size, TDL_BLOCK);

    /**  Convert to CSR format  **/

    COO_to_CSR(mat_size, nzero, ai, aj, el, idx, ib, &ib[1], a);
/* print_CSR(mat_size, idx, ib, &ib[1], a); */

    /**  Setup vector. **/

    for (i = 0; i < mat_size; ++i)
	if (TDL_IS_VALID(b_desc, &b[i]))
	    b[i] = 1;

    for (i = 0; i < mat_size; ++i)
	if (TDL_IS_VALID(c_desc, &c[i]))
	    c[i] = 0;

    /**  MAIN  **/

    mult(mat_size, idx, ib, a, b, c,
	 idx_desc, ib_desc, a_desc, b_desc, c_desc);

    /**  Finalization  **/

    free(ib);
    TDL_desc_free(ib_desc);
    free(idx);
    TDL_desc_free(idx_desc);
    free(a);
    TDL_desc_free(a_desc);
    free(b);
    TDL_desc_free(b_desc);
    free(c);
    TDL_desc_free(c_desc);

    free(ai);
    free(aj);
    free(el);

    TDL_finalize();

    exit(0);
}
