/*
 *  $Id: mult.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include <stdio.h>
#include <tdl.h>

/*
 *  If TIME_ONLY_COMPUTATION is defined, only computation is timed.
 */

#ifndef TIME_ONLY_COMPUTATION
#include "ttime.h"
#define timing(x) x
#else
#define timing(x)
#endif

/*
 *  MAIN
 */

/* #define TEST */

#define NUM_ITER	3

static double t1[NUM_ITER];
static double t2[NUM_ITER];

extern void
vec_mult_CSR_IE(int n, int *idx, int *ib,
		double *a, double *b, double *c,
#ifdef TIME_ONLY_COMPUTATION
		double *time,
#endif
		TDL_desc idx_desc, TDL_desc ib_desc,
		TDL_desc a_desc, TDL_desc b_desc, TDL_desc c_desc);

void
mult(int size, int *idx, int *ib, double *a, double *b, double *c,
     TDL_desc idx_desc, TDL_desc ib_desc,
     TDL_desc a_desc, TDL_desc b_desc, TDL_desc c_desc)
{
    int i;
    timing(ttime_t tt1);
    timing(ttime_t tt2);

    /**  MAIN  **/

    TDL_barrier();

    timing(tgettime(tt1));
    vec_mult_CSR_IE(size, idx, ib, a, b, c,
#ifdef TIME_ONLY_COMPUTATION
		    &t1[0],
#endif
		    idx_desc, ib_desc, a_desc, b_desc, c_desc);
    timing(tgettime(tt2));
    timing(t1[0] = ttime_diff_sec(tt1, tt2));
    for (i = 1; i < NUM_ITER - 1; i += 2) {
	timing(tgettime(tt1));
	vec_mult_CSR_IE(size, idx, ib, a, c, b,
#ifdef TIME_ONLY_COMPUTATION
			&t1[i],
#endif
			idx_desc, ib_desc, a_desc, c_desc, b_desc);
	timing(tgettime(tt2));
	timing(t1[i] = ttime_diff_sec(tt1, tt2));
	timing(tgettime(tt1));
	vec_mult_CSR_IE(size, idx, ib, a, b, c,
#ifdef TIME_ONLY_COMPUTATION
			&t1[i + 1],
#endif
			idx_desc, ib_desc, a_desc, b_desc, c_desc);
	timing(tgettime(tt2));
	timing(t1[i + 1] = ttime_diff_sec(tt1, tt2));
    }
    if (i < NUM_ITER) {
	timing(tgettime(tt1));
	vec_mult_CSR_IE(size, idx, ib, a, c, b,
#ifdef TIME_ONLY_COMPUTATION
			&t1[i],
#endif
			idx_desc, ib_desc, a_desc, c_desc, b_desc);
	timing(tgettime(tt2));
	timing(t1[i] = ttime_diff_sec(tt1, tt2));
    }

#ifdef TEST
    {
	double *r;
	TDL_desc r_desc;

	if (NUM_ITER & 1) {
	    r = c;
	    r_desc = TDL_desc_create_desc(c_desc);
	    TDL_desc_dist(r_desc, TDL_MASTER);
	    TDL_remap_home(c_desc, r_desc);
	}
	else {
	    r = b;
	    r_desc = TDL_desc_create_desc(b_desc);
	    TDL_desc_dist(r_desc, TDL_MASTER);
	    TDL_remap_home(b_desc, r_desc);
	}
	TDL_fence();

	if (TDL_gThreadNum == 0)
	    for (i = 0; i < size; i += 1)
		printf("[0] r[%d] = %g\n", i, r[i]);

	TDL_desc_free(r_desc);
    }
#endif

    /**  Compute each interval time and gather.  **/

#ifdef MPI
    MPI_Reduce(t1, t2, NUM_ITER, MPI_DOUBLE, MPI_MAX, 0, TDL_gComm);
    if (TDL_gThreadNum == 0) {
	double min = t2[0];
	for (i = 1; i < NUM_ITER; ++i) {
	    if (min > t2[i])
		min = t2[i];
	}
	printf("%g\n", t2[0]);
	printf("%g\n", min);
    }
#elif defined (RHiNET)
    for (i = 0; i < NUM_ITER; ++i) {
	printf("[%d] %g\n", TDL_gThreadNum, t1[i]);
    }
#endif
}
