/*
 *  TIMER routine
 *
 *  $Id: ttime.h,v 1.1.1.1 2004-02-06 18:15:01 msato Exp $
 */

#ifndef __TTIME_H__
#define __TTIME_H__

/*
 *   #include <stdio.h>
 *   #include <ttime.h>
 *
 *   ttime_t t1, t2;
 *
 *   tgettime(t1);
 *   ....
 *   tgettime(t2);
 *
 *   printf("%g msec\n", ttime_diff_msec(t1, t2));
 */

#if !defined(__TTIME_GETTIMEOFDAY__) && !defined(__TTIME_MPI_WTIME__)

/***  SUN SOLARIS  ***/
#if defined(sun) && defined(__SVR4)
#include <sys/time.h>

typedef hrtime_t ttime_t;
#define tgettime(t1)		{ (t1) = gethrtime(); }
#define ttime_diff_msec(t1, t2)	(((double)(t2) - (double)(t1)) * .000001)

/***  SGI IRIX  ***/
#elif defined(sgi)
#include <time.h>

typedef struct timespec ttime_t;
#define tgettime(t1)	clock_gettime(CLOCK_SGI_CYCLE, &(t1))
#define ttime_diff_msec(t1, t2)	\
	(((double)(t2.tv_sec) - (double)(t1.tv_sec)) * 1000.0	\
	+ ((double)(t2.tv_nsec) - (double)(t1.tv_nsec)) * .000001)

/***  EMX  ***/
#elif defined(EMX)

typedef double ttime_t;

#define tgettime(t1)	(double)em_utime()
#define ttime_diff_msec(t1, t2)	(((t2) - (t1)) * .0000012)

/***  gettimeofday(2)  ***/
#else
#include <sys/time.h>

typedef struct timeval ttime_t;
#define tgettime(t1)	gettimeofday(&(t1), NULL)
#define ttime_diff_msec(t1, t2)	\
	(((double)(t2.tv_sec) - (double)(t1.tv_sec)) * 1000.0	\
	+ ((double)(t2.tv_usec) - (double)(t1.tv_usec)) * .001)

#endif  /**  END OF MACHINE DEPENDENT PART  **/

#elif defined(__TTIME_GETTIMEOFDAY__)

#include <sys/time.h>

typedef struct timeval ttime_t;
#define tgettime(t1)	gettimeofday(&(t1), NULL)
#define ttime_diff_msec(t1, t2)	\
	(((double)(t2.tv_sec) - (double)(t1.tv_sec)) * 1000.0	\
	+ ((double)(t2.tv_usec) - (double)(t1.tv_usec)) * .001)

#else /** defined(__TTIME_MPI_WTIME__) **/

/*
 *  TIMER routine for MPI
 */

typedef double ttime_t;
#define tgettime(t1)		{ (t1) = MPI_Wtime(); }
#define ttime_diff_msec(t1, t2)	(((double)(t2) - (double)(t1)) * .001)
#define ttime_diff_sec(t1, t2)	(((double)(t2) - (double)(t1)))

#endif /** !defined(__TTIME_GETTIMEOFDAY__) || !defined(__TTIME_MPI_WTIME__) **/

#ifndef ttime_diff_sec
#define ttime_diff_sec(t1, t2)	(ttime_diff_msec(t1, t2) * .001)
#endif  /**  ttime_diff_sec  **/

#endif  /** __TTIME_H__ **/
