/*
 *  $Id: tdl.h,v 1.1.1.1 2004-02-06 18:14:59 msato Exp $
 */

#ifndef __tdl__
#define __tdl__

#if !defined(RHiNET) && !defined(MPI)
#define MPI
#endif

#if defined(MPI) && !defined(MPI2)
#define MPI1
#endif

#if !defined(MPI) && defined(MPI2)
#define MPI
#endif

#if defined(MPI) && defined(RHiNET)
#error Error: both MPI and RHiNET are specified!
#endif

/**  Include a header file for a messaging library  **/

#if defined(MPI)
#include <mpi.h>
#elif defined(RHiNET)
#include "rhinet.h"
#else
#error tdl.h: Not support the message library
#endif

/* */

#include "tdl_list.h"

/*
 *  Constants
 */

/*  Page size shoule be the power of two. */

#define TDL_PAGE_SIZE		(4 * 1024)
#define TDL_PAGE_UNDEF		(-1)
#define TDL_SHARE_UNDEF		(-1)

/*
 *  Access count and sharing processes.
 */

struct _TDL_access_count {
    int ref;
    unsigned long *ac;
};

typedef struct _TDL_access_count *TDL_Access_count;

typedef unsigned long TDL_Share_type;

struct _TDL_share {
    int ref;
    TDL_Share_type *sh;
};

typedef struct _TDL_share *TDL_Share;

/*
 *  List of messages
 */

struct _TDL_messages {
    int *num;		   /**  Num. of messages       **/
    unsigned long *offset; /**  Array of message addr. **/
    unsigned long *size;   /**  Array of message size  **/
};

typedef struct _TDL_messages *TDL_Messages;

struct _TDL_msg_list {
    int ref;
    Send_el_list *sl;
#ifdef MPI1
    TDL_Messages msgs;	/*  Message list for all processors  */
#endif
};

typedef struct _TDL_msg_list *TDL_Msg_list;

/*
 *  Descriptor for distributed array
 */

struct _TDL_desc {
    int size;		/**  size (byte)    **/
    int psize;		/**  num. of pages  **/
    void* base;		/**  base address   **/
    TDL_Access_count tac;	/**  access counter & home  **/
    TDL_Access_count tprevhome;	/**  previous home  **/
    TDL_Share tsh;	/**  sharing information    **/

    /**  For MPI and MPI-2  **/

    TDL_Msg_list send_list;	/**  send list		    **/
    TDL_Msg_list recv_list;	/**  receive list	    **/

#ifdef RHiNET
    /**  For RHiNET  **/
    GID	area;
#endif
};

typedef struct _TDL_desc *TDL_desc;

/*
 *  Distribution
 */

enum dist {
    TDL_MASTER,
    TDL_MASTER_ALL,
    TDL_BLOCK,
    TDL_BLOCK_ALL,
    TDL_CYCLIC
};

typedef enum dist TDL_dist;

/*
 *  Function declarations
 */

/*
 *  Initialization/Finalization/Barrier
 */

void
TDL_init(int *argc, char **argv[]
#ifdef RHiNET
	 , int num_threads, int thread_num
	 , char *daemonhost, int daemonport
#endif
    );

void
_TDL_init(char *progname);

void
TDL_finalize();

int
TDL_barrier();

int
TDL_fence();

/*
 *  TDL_malloc
 */

void*
TDL_malloc(int);

/*
 *  Descriptor for distributed array
 */

TDL_desc
TDL_desc_create(void *base, int byte_size, TDL_dist dist
#ifdef RHiNET
		, GID gid
#endif
    );

TDL_desc
TDL_desc_create_desc(TDL_desc);

void
TDL_desc_free(TDL_desc);

void
TDL_desc_print(TDL_desc);

void
TDL_desc_print_abs(TDL_desc);

void
TDL_desc_dist(TDL_desc desc, TDL_dist dist);

void
TDL_desc_copy_dist(TDL_desc, TDL_desc);

void
TDL_desc_copy_home(TDL_desc, TDL_desc);

int
TDL_desc_compare_dist(TDL_desc, TDL_desc);

int
TDL_desc_compare_home(TDL_desc, TDL_desc);

/*
 *  Inspect
 */

void
TDL_inspect_home(void *addr, int size, TDL_desc desc);

void
TDL_inspect_sharing(void *addr, int size, TDL_desc desc);

void
TDL_inspect_ref(void *addr, int size, TDL_desc desc);

void
TDL_inspect_modify(void *addr, int size, TDL_desc desc);

/*
 *  Fix descriptor
 */

void
TDL_desc_fix_home(TDL_desc desc);

void
TDL_desc_fix_sharing(TDL_desc desc);

void
TDL_desc_fix_ref(TDL_desc desc);

void
TDL_desc_fix_modify(TDL_desc desc);

void
TDL_desc_merge(TDL_desc, TDL_desc);

/*
 *  Remap
 */

void
TDL_remap(TDL_desc old, TDL_desc new);

int
TDL_remap_home(TDL_desc old, TDL_desc new);

int
TDL_remap_home_force(TDL_desc old, TDL_desc new);

void
TDL_remap_ref(TDL_desc old, TDL_desc new);

void
TDL_remap_modify(TDL_desc old, TDL_desc new);

void
TDL_remap_ref_modify(TDL_desc old, TDL_desc new);

/*
 *  Update
 */

void
TDL_update(TDL_desc desc);

void
TDL_update_modify(TDL_desc desc);

/*
 *  Loop schedule
 */

void
TDL_loop_schedule(TDL_dist dist,
		  int start, int end, int step,
		  int *l_start, int *l_end, int *l_step);

/*
 *  I/O
 */

#include <stdio.h>

void
TDL_desc_read(FILE *fd, TDL_desc desc);

void
TDL_desc_write(FILE *fd, TDL_desc desc);

/*
 *  External variables
 */

#ifdef MPI
extern MPI_Comm TDL_gComm;
#endif

extern int TDL_gNumThreads;
extern int TDL_gThreadNum;
    
/*
 *  Macros
 */

#define TDL_NUM_PAGES_FROM_SIZE(N) (((N) + TDL_PAGE_SIZE - 1) / TDL_PAGE_SIZE)
#define TDL_PAGE(addr, desc) \
    (/*(unsigned long)*/(((char*)(addr) - (char*)TDL_BASE(desc)) / TDL_PAGE_SIZE))
#define TDL_PAGE_MASK(addr)	(void*)((unsigned long)(addr) & ~(TDL_PAGE_SIZE - 1))
#define TDL_PAGE_ALIGN(addr)	(TDL_PAGE_MASK((char*)(addr) + TDL_PAGE_SIZE - 1))
#define TDL_PAGE_OFFSET(page)	((page) * TDL_PAGE_SIZE)
#define TDL_DISP(addr, desc) \
    (/*(unsigned long)*/((char*)(addr) - (char*)TDL_BASE(desc)))
#define TDL_ADDR(offset, desc) \
    ((char*)TDL_BASE(desc) + /*(unsigned long)*/(offset))

#define TDL_SIZE(desc)		((desc)->size)
#define TDL_NUM_PAGES(desc)	((desc)->psize)
#define TDL_BASE(desc)		((desc)->base)
#define TDL_HOME(desc)		(TDL_ACCESS_COUNT(desc))
#define TDL_PREV_BASE(desc)	((desc)->tprevhome)
#define TDL_PREV_HOME(desc)	(TDL_PREV_BASE(desc)->ac)
#define TDL_ACCESS_BASE(desc)	((desc)->tac)
#define TDL_ACCESS_COUNT(desc)	(TDL_ACCESS_BASE(desc)->ac)
#define TDL_SHARE_BASE(desc)	((desc)->tsh)
#define TDL_SHARING(desc)	(TDL_SHARE_BASE(desc)->sh)
#define TDL_SEND_BASE(desc)	((desc)->send_list)
#define TDL_SEND_LIST(desc)	(TDL_SEND_BASE(desc)->sl)
#define TDL_RECV_BASE(desc)	((desc)->recv_list)
#define TDL_RECV_LIST(desc)	(TDL_RECV_BASE(desc)->sl)
#ifdef MPI1
#define TDL_SEND_MLIST(desc)	(TDL_SEND_BASE(desc)->msgs)
#define TDL_RECV_MLIST(desc)	(TDL_RECV_BASE(desc)->msgs)
#endif /*  MPI1  */
#ifdef RHiNET
#define TDL_AREA(desc)		((desc)->area)
#endif /*  RHiNET  */
#define TDL_REF(x)		((x)->ref)

#define TDL_IS_HOME_PAGE(desc, i) \
    (TDL_HOME(desc)[i] == TDL_gThreadNum)
#define TDL_IS_SHARED_WITH(desc, i, rank) \
    (TDL_SHARING(desc)[i] & (1 << (rank)))
#define TDL_IS_PRIVATE(desc, i)	\
    (TDL_SHARING(desc)[i] == 1 << TDL_gThreadNum)
#define TDL_IS_SHARED(desc, i) \
    (!TDL_IS_PRIVATE(desc, i) && TDL_IS_SHARED_WITH(desc, i, TDL_gThreadNum))
#define TDL_IS_HOME(desc, addr) \
    (TDL_HOME(desc)[TDL_PAGE(addr, desc)] == TDL_gThreadNum)
#define TDL_IS_VALID(desc, addr) \
    (TDL_IS_SHARED_WITH(desc, TDL_PAGE(addr, desc), TDL_gThreadNum))

#ifdef MPI1
/*
 *  Message Type
 */

#define TDL_UPDATE		1000
#define TDL_UPDATE_REF		1001
#define TDL_UPDATE_MODIFY	1002
#define TDL_REMAP		1003
#define TDL_REMAP_HOME		1004
#define TDL_CREATE_RECV1	1005
#define TDL_CREATE_RECV2	1006
#define TDL_CREATE_RECV3	1007
#define TDL_CREATE_SKIP		3
#endif

#endif  /** __tdl__  **/
