/*
 *  $Id: tdl_io.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include "tdl.h"
#include "tdl_debug.h"

/*
 *  Read/write a TDL descriptor
 *
 *  ac, sh
 *  num_send_list
 *  send_list
 *  num_recv_list
 *  recv_list
 *  ...
 */

/*
 *  READ
 */

void
TDL_desc_read(fd, desc)
    FILE *fd;
    TDL_desc desc;
{
    int i;
    int num_send_list, num_recv_list;

    debug(printf("[%d] NUM_PAGES: %d\n", TDL_gThreadNum, TDL_NUM_PAGES(desc)));

    for (i = 0; i < TDL_NUM_PAGES(desc); ++i) {
	int j;
	if (fscanf(fd, "%lu %lu", &TDL_HOME(desc)[i],
		   &TDL_SHARING(desc)[i]) == EOF)
	    fprintf(stderr, "TDL_desc_read: Error (1)\n"), exit(1);

	debug(printf("[%d] HOME: %ld, DIR %lx\n", TDL_gThreadNum,
		     TDL_HOME(desc)[i], TDL_SHARING(desc)[i]));

	if (fscanf(fd, "%d", &num_send_list) == EOF)
	    fprintf(stderr, "TDL_desc_read: Error (2)\n"), exit(1);

	debug(printf("[%d] NUM_SL: %d\n", TDL_gThreadNum, num_send_list));

	for (j = 0; j < num_send_list; ++j) {
	    struct _send_el p;
	    if (fscanf(fd, "%lx %lu", &p.addr, &p.size) == EOF)
		fprintf(stderr, "TDL_desc_read: Error (3)"), exit(1);
	    TDL_SEND_LIST(desc)[i]
		= cons_send_el(&p, TDL_SEND_LIST(desc)[i]);
	}

	debug(print_send_el(TDL_SEND_LIST(desc)[i]));

	if (fscanf(fd, "%d", &num_recv_list) == EOF)
	    fprintf(stderr, "TDL_desc_read: Error (4)\n"), exit(1);

	debug(printf("[%d] NUM_RL: %d\n", TDL_gThreadNum, num_recv_list));

	for (j = 0; j < num_recv_list; ++j) {
	    struct _send_el p;
	    if (fscanf(fd, "%lx %lu", &p.addr, &p.size) == EOF)
		fprintf(stderr, "TDL_desc_read: Error (5)"), exit(1);
	    TDL_RECV_LIST(desc)[i]
		= cons_send_el(&p, TDL_RECV_LIST(desc)[i]);
	}

	debug(print_send_el(TDL_RECV_LIST(desc)[i]));
    }
}

/*
 *  WRITE
 */

void
TDL_desc_write(fd, desc)
    FILE *fd;
    TDL_desc desc;
{
    int i;

    for (i = 0; i < TDL_NUM_PAGES(desc); ++i) {
	int num_send_list, num_recv_list;
	Send_el_list ls = TDL_SEND_LIST(desc)[i];

	fprintf(fd, "%lu %lu\n", TDL_HOME(desc)[i], TDL_SHARING(desc)[i]);

	num_send_list = length_send_el(ls);
	fprintf(fd, "%d\n", num_send_list);

	while (ls) {
	    Send_el p;
	    p = car_send_el(ls);
	    fprintf(fd, "%lx %lu\n", p->addr, p->size);
	    ls = cdr_send_el(ls);
	}

	ls = TDL_RECV_LIST(desc)[i];
	num_recv_list = length_send_el(ls);
	fprintf(fd, "%d\n", num_recv_list);

	while (ls) {
	    Send_el p;
	    p = car_send_el(ls);
	    fprintf(fd, "%lx %lu\n", p->addr, p->size);
	    ls = cdr_send_el(ls);
	}
    }
}

