/*
 *  $Id: tdl_list.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include "tdl_list.h"

/*
 *  Structure-dependent functions
 */

int
eq_send_el(el1, el2)
    Send_el el1, el2;
{
    return (el1->addr == el2->addr
	    && el1->size == el2->size);
}

Send_el_list
cons_send_el(el, ls)
    Send_el el;
    Send_el_list ls;
{
    Send_el_list t = (Send_el_list)malloc(sizeof(struct _send_el_list));
    if (t == NULL) {
	fprintf(stderr, "cons_send_el(): Can't allocate memory\n");
	exit(1);
    }

    t->next = ls;
    if ((t->el = (Send_el)malloc(sizeof(struct _send_el))) == NULL) {
	fprintf(stderr, "cons_send_el(): Can't allocate memory\n");
	exit(1);
    }
    *t->el = *el;
    return t;
}

void
delete_send_el(ls)
    Send_el_list ls;
{
    while (ls) {
	Send_el_list next_ls;

	next_ls = cdr_send_el(ls);
	free((void*)ls->el);
	free((void*)ls);
	ls = next_ls;
    }
}

void
print_send_el(ls)
    Send_el_list ls;
{
    int i = 0;

    while (ls) {
	Send_el p;
	p = car_send_el(ls);
	printf("send_el_list[%d] = (%lx, %lu)\n",
	       i, p->addr, p->size);
	i++;
	ls = cdr_send_el(ls);
    }
}

Send_el_list
insert_sorted_send_el(Send_el el, Send_el_list ls)
{
    Send_el_list prev_ls = (Send_el_list)NULL;
    Send_el_list saved_ls = ls;
    Send_el p;

    while (ls) {
	p = car_send_el(ls);
	if (el->addr <= p->addr + p->size)
	    break;
	prev_ls = ls;
	ls = cdr_send_el(ls);
    }
    if (ls && el->addr + el->size >= p->addr) {

	/*  Merge an element. */

	unsigned long new_addr, new_size;
	unsigned long t1, t2;

	new_addr = (el->addr < p->addr) ? el->addr : p->addr;

	t1 = el->addr + el->size;
	t2 = p->addr + p->size;
	new_size = (t1 > t2) ? (t1 - new_addr) : (t2 - new_addr);

	p->addr = new_addr;
	p->size = new_size;

	ls = saved_ls;
    }
    else {

	/*  Insert an element. */

	ls = cons_send_el(el, ls);

	if (prev_ls) {
	    prev_ls->next = ls;
	    ls = saved_ls;
	}
    }    
    return ls;
}

/*
 *  Structure-independent functions
 */

Send_el
car_send_el(ls)
    Send_el_list ls;
{
    return ls->el;
}

Send_el_list
cdr_send_el(ls)
    Send_el_list ls;
{
    return ls->next;
}

Send_el_list
append_send_el(ls1, ls2)
    Send_el_list ls1, ls2;
{
    if (ls1)
	return cons_send_el(car_send_el(ls1),
			    append_send_el(cdr_send_el(ls1), ls2));
    else
	return ls2;
}

Send_el
nth_send_el(n, ls)
    int n;
    Send_el_list ls;
{
    int i;
  
    for (i = 0; i < n; i++)
	ls = cdr_send_el(ls);
    return car_send_el(ls);
}

int
length_send_el(ls)
    Send_el_list ls;
{
    int i = 0;

    while (ls) {
	i++;
	ls = cdr_send_el(ls);
    }
    return i;
}

Send_el_list
remove_send_el(el, ls)
    Send_el el;
    Send_el_list ls;
{
    if (ls) {
	if (eq_send_el(el, car_send_el(ls)))
	    return remove_send_el(el, cdr_send_el(ls));
	else
	    return cons_send_el(car_send_el(ls),
				remove_send_el(el, cdr_send_el(ls)));
    }
    else
	return ls;
}

Send_el_list
uniq_send_el(ls)
    Send_el_list ls;
{
    if (ls)
	return cons_send_el(car_send_el(ls),
			    uniq_send_el(remove_send_el(car_send_el(ls),
							cdr_send_el(ls))));
    else
	return ls;
}

Send_el_list
reverse_send_el(ls)
    Send_el_list ls;
{
    Send_el_list rev_ls = (Send_el_list)NULL;
    while(ls){
	rev_ls = cons_send_el(car_send_el(ls), rev_ls);
	ls = cdr_send_el(ls);
    }
    return rev_ls;
}

Send_el_list
copy_send_el(ls)
    Send_el_list ls;
{
    if (ls)
	return cons_send_el(car_send_el(ls), copy_send_el(cdr_send_el(ls)));
    else
	return ls;
}

Send_el_list
member_send_el(e, ls)
    Send_el e;
    Send_el_list ls;
{
    while(ls) {
	if (eq_send_el(e, car_send_el(ls)))
	    break;
	ls = cdr_send_el(ls);
    }
    return ls;
}


#if 0
int
main(argc, argv)
    int argc;
    char* argv[];
{
    struct _send_el a1, a2;
    Send_el_list li = NULL, li2 = NULL;

    a1.addr = (unsigned long)&a1;
    a1.size = 4;
    a2.addr = (unsigned long)&a2;
    a2.size = 8;

    li = cons_send_el(&a1, li);
    li = cons_send_el(&a2, li);

    print_send_el(li);
    li2 = reverse_send_el(li);
    print_send_el(li2);
    delete_send_el(li2);
    li2 = copy_send_el(li);
    print_send_el(li2);
    print_send_el(li);

    exit(0);
}
#endif

