/*
 *  $Id: tdl_sort.c,v 1.1.1.1 2004-02-06 18:15:00 msato Exp $
 */

#include "tdl_list.h"

#define GET_ADDR(a)	((unsigned long)((a).addr))

#define STACK_SIZE	40

struct {
    int left, right;
} stack[STACK_SIZE];

/*
 *  Descriptions:
 *
 *  isort() sorts the array 'a' in ascending order by quick sort or
 *  insertion sort.
 */

void
isort(n, a)
    int n;
    Send_el a;
{
    int limit = 10;
    int sp = 0;

    /**  Push the initial range to the stack. **/

    stack[sp].left = 0;
    stack[sp].right = n - 1;

    do {
	int left, right;
	int i, j;

	/**  Pop from stack. **/

	left = stack[sp].left;
	right = stack[sp].right;
	--sp;

	while (right - left >= limit) {
	    struct _send_el t;

	    /**  Quick sort  **/

	    {
		int center;
		struct _send_el w1, w2, w3;
		unsigned long w1_addr, w2_addr, w3_addr;

		/*
		 *  Choose samples from left-most, right-most and
		 *  the center.
		 */ 

		w1 = a[left];
		w2 = a[right];
		center = (left + right) / 2;
		w3 = a[center];
		a[center] = a[left + 1];

		/*  Sort 'w1', 'w2' and 'w3'. */

		w1_addr = GET_ADDR(w1);
		w2_addr = GET_ADDR(w2);
		w3_addr = GET_ADDR(w3);

		if (w1_addr < w2_addr) {
		    if (w2_addr < w3_addr) {
			a[left] = w1; a[left + 1] = w2; a[right] = w3;
		    }
		    else if (w1_addr < w3_addr) {
			a[left] = w1; a[left + 1] = w3; a[right] = w2;
		    }
		    else {
			a[left] = w3; a[left + 1] = w1; a[right] = w2;
		    }
		}
		else {
		    if (w1_addr < w3_addr) {
			a[left] = w2; a[left + 1] = w1; a[right] = w3;
		    }
		    else if (w2_addr < w3_addr) {
			a[left] = w2; a[left + 1] = w3; a[right] = w1;
		    }
		    else {
			a[left] = w3; a[left + 1] = w2; a[right] = w1;
		    }
		}
	    }

	    {
		unsigned long pivot;
		
		/*
		 *  Pivot is the middle value among the three
		 *  samples.
		 */

		pivot = GET_ADDR(a[left + 1]);

		/**  Quick sort  **/

		i = left + 1;
		j = right;
		do {
		    while (GET_ADDR(a[++i]) < pivot) ;
		    while (GET_ADDR(a[--j]) > pivot) ;
		    if (i < j) {
			t = a[i];
			a[i] = a[j];
			a[j] = t;
		    }
		} while (i < j);

		t = a[left + 1];
		a[left + 1] = a[j];
		a[j] = t;
	    }

	    /*
	     *  Push the wider range to the stack.  Narrow region is
	     *  sorted in advance. 
	     */

	    if (sp >= STACK_SIZE - 1)
		break;

	    ++sp;
	    if (j - left <= right - j) {
		stack[sp].left = j + 1;
		stack[sp].right = right;
		right = j - 1;
	    }
	    else {
		stack[sp].left = left;
		stack[sp].right = j - 1;
		left = j + 1;
	    }
	}

	/**  Insertion sort  **/

	for (i = left + 1; i <= right; ++i) {
	    struct _send_el w;
	    unsigned long w_addr;

	    w = a[i];
	    w_addr = GET_ADDR(w);

	    j = i - 1;
	    while ((w_addr < GET_ADDR(a[j])) && (j >= left)) {
		a[j + 1] = a[j];
		--j;
	    }
	    a[j + 1] = w;
	}

    } while (sp >= 0);
}
