#include  <stdio.h>
#include  <sys/types.h>
#include  <sys/time.h>
#include  <string.h>
#include  <math.h>

#include  "omsc.h"

#define  XSIZE   10
#define  YSIZE   1022
#define  XS (XSIZE+2)
#define  YS (YSIZE+2)
#define  LAP_ALLOC_SIZE  (XS * YS * sizeof(double) * 2 + SCASH_PAGE_SIZE)

#define  XMASK  ((1<<9) - 1)
#define  XRUPS  ((XS + XMASK) & (~XMASK))

#define  PI   3.1415927
#define  NITER 10

double   time1, time2, time3;
double   second();
double  *u;
double  *uu;
double  *sum;

int   pe;  /* My PE Number */
int   ps;  /* Partition Size = Number of PE */ 

inline double   
second()
{
    double  ret;

    struct  timeval  tv;
    gettimeofday(&tv, NULL);
    
    ret = tv.tv_usec;
    ret = ret / 1000000.0;
    ret = ret + tv.tv_sec;
    return   ret;
}

void 
laplace_alloc()
{
    int     alloc_size = sizeof(double) * XS * YS;
    caddr_t   up;
    caddr_t   uup;
    int       sp;

    ps = scash_ps();

    sp = alloc_size / ps;

    u = (double *)scash_alloc(alloc_size);
    if (u == NULL) 
	scash_exit(1);  
    scash_distribute_caddr((char**)&u);
    uu = (double *)scash_alloc(alloc_size);
    if (uu == NULL) 
	scash_exit(1); 
    scash_distribute_caddr((char**)&uu);
    
    up = (caddr_t)&u[0];
    uup = (caddr_t)&uu[0];
/*
    {
        int  i;
        for (i = 0 ; i < ps ; i++) {
            scash_home(i, up , sp / SCASH_PAGE_SIZE);
            scash_home(i, uup, sp / SCASH_PAGE_SIZE);
	    up += sp;
	    uup += sp;
        }
    }
*/
#if 1
    scash_home_first(up, alloc_size/SCASH_PAGE_SIZE);
    scash_home_first(uup, alloc_size/SCASH_PAGE_SIZE);
#endif

    sum = (double *)scash_alloc( sizeof(double) * ps);
    scash_distribute_double((char**)&sum);
}

void
laplace(int iteration)
{
    int   x, y;
    int   y_to, y_from;
    double *tmpu;

    y_from = (YS / ps) * pe;
    y_to = y_from + (YS / ps);
    if (y_from == 0)  y_from++;
    if (y_to == YS)   y_to--;

    scash_barrier(iteration);

    time1 = second();
    do {
	if (pe==2) scash_prf("laplace iter=%d", iteration);

	for (y = y_from ; y < y_to ; y++) {
	    scash_poll();
	    for (x = 1 ; x <= XSIZE ; x++){
		uu[y * XS + x] = (u[ (y-1) * XS + x] + u[(y+1) * XS + x] + u[y * XS + x-1] + u[y * XS + x+1])/4.0;
	    }
	}
	tmpu = uu;
	uu = u;
	u = tmpu;

	scash_barrier(iteration + 2);
    } while (--iteration) ;
    time2 = second();
}

void 
laplace_init_x(int  y)
{
    int  x;

    scash_poll();
    u[y * XS ] = uu[y * XS ] = 0.0;
    u[y * XS + XSIZE + 1] = uu[y * XS + XSIZE + 1] = 0.0;
     
    if ( y == 0 || y == XSIZE + 1) {
	for (x = 1 ; x <= XSIZE ; x++)
	    u[y * XS + x] = uu[y * XS + x] = 0.0;
    } else {
	for (x = 1 ; x <= XSIZE ; x++)
	    u[y * XS + x] = sin((double)(y-1)/YSIZE*PI) + cos((double)(x-1)/XSIZE*PI);
	    uu[y * XS + x] = 0.0;
    }
    scash_poll();
}

void
laplace_init()
{
    int  y;

    ps = scash_ps();
    pe = scash_pe();
    
    for (y = (YS / ps) * pe; y < (YS / ps) * (pe+1) ; y++)
        laplace_init_x(y);
    scash_barrier(1);
}

void
laplace_sum()
{
    int   y_from, y_to;
    int   x, y;
    double *local_sum ;
    int     node;

    local_sum = sum + pe;
    *local_sum = 0;

    y_from = YS / ps * pe;
    y_to = y_from + (YS / ps);
    if (y_from == 0)  y_from++;
    if (y_to == YS)   y_to--;

    for (y = y_from ; y < y_to ; y++)  
	for (x = 1 ; x <= XSIZE ; x++)  
	    *local_sum += (uu[y * XS + x] - u[y * XS + x]);
    scash_prf("local_sum = %f", *local_sum);
    scash_barrier(1);

    if (scash_pe() == 0) {
	for (node = 1 ; node < ps ; node++)
	    local_sum[0] += local_sum[node];
	printf("Checksum = %f\n", local_sum[0]);
    }
}

int
main(int  argc,
     char** argv)
{
    int x,y;

    scash_initialize(argc, argv, LAP_ALLOC_SIZE);

    if (0 == scash_pe())  laplace_alloc();
    
    scash_barrier(1);
    laplace_init();
    scash_barrier(1);

    if (0 == scash_pe())
        printf("laplace:init -- done\n");

    // scash_profile_reset();
    scash_tlog_start();
    laplace(NITER);
    scash_tlog_end();
    // scash_profile_print();

    laplace_sum();
    scash_barrier(1);
    
    if (0 == scash_pe()) {
        time3 = second();
        printf("Total time = %g\nCalc time = %g\nChecksum time= %g\n",
	       time3 - time1, time2 - time1, time3 - time2);
    }
    scash_exit(0);
    exit(0);
}










