/* 
 * $Id: omsc.h,v 1.7 2006-10-24 13:40:57 ojima Exp $
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
#ifndef __scash_h
#define __scash_h

#include <sys/types.h>
#include <signal.h>

#define TRUE 1
#define FALSE 0

#if defined(__alpha__) || defined(__ia64__) || defined(__x86_64__)
typedef  unsigned long u_int64;
typedef  long int64;
#else
typedef  unsigned long long u_int64;
typedef  long long int64;
#endif

#ifdef USE_PM
typedef long long addrHandle;
#endif

#ifdef USE_MPI
typedef caddr_t addrHandle;
extern int omsc_num_proc;
#endif

/*
 *  Page Size depends the CPU's MMU. 
 */
#ifdef i386
#define  SCASH_PAGE_SIZE  (4096)
#endif
#ifdef __alpha__
#define  SCASH_PAGE_SIZE  (8192)
#endif
#ifdef __ia64__
#define  SCASH_PAGE_SIZE  (16384)
#endif
#ifdef __x86_64__
#define  SCASH_PAGE_SIZE  (4096)
#endif

#define  SCASH_WORD_SIZE  (4)
#define  SCASH_SUCCESS  0
#define  SCASH_MAX_PE   (64)
#ifdef __ia64__
#define  SCASH_MAX_PAGE (160 * 1024)
#else
#define  SCASH_MAX_PAGE (128 * 1024)
#endif /* __ia64__ */
#define  SCASH_MAX_LOCK (256)

#define  SCASH_LOCK_BUSY (0x00)
#define  SCASH_LOCK_OK   (0x01)
#define  SCASH_LOCK_OFF  (0xFFFF)

/* lock structure */
typedef  struct  scash_lock_s {
    int    id;
    u_int  head;
    u_int  q_cnt;
    u_char *q;
} scash_lock_t;

/* profile */
typedef  struct  scash_profile_s {
    int      barrier_cnt;
    u_int64  barrier_time;
    u_int64  barrier_1st_sync;
    u_int64  barrier_flush;
    u_int64  barrier_inval;
    int      inval_msg;
    u_int64  barrier_1st_size;
    u_int64  barrier_flush_size;
    u_int64  barrier_inval_size;
    int      first_flag;
    int      flush_flag;
    int      inval_flag;
    int      barrier_inval_cnt;
    int      barrier_flush_cnt;

    int      rma_cnt;
    u_int64  rma_time;
    char     barrier_flag;
    int      barrier_msg_cnt;
    u_int64      barrier_msg_size;
    u_int64  lock_get_time;
    int      lock_get_cnt;
    u_int64  handler_time;
    int      handler_cnt;
    u_int64  prefetch_wait;
    u_int64  rma_wait;
    u_int64      total_msg_size;
    char     get_msg_size;

    double   handler_time_d;

    u_int64  loop_cnt;

    int      resend_cnt;
    int      resend_diff_size;

    int zsolve;
    double rma_start, rma_end;
    int msg_cnt;
    double msg_send_time;
    int diff_send_cnt;
    double diff_send_time;
} scash_profile_t;

/* for prefetching */
#define START -1
#define END -2
#define BARRIER -3

typedef  struct  scash_prefetch_s {
    int    get_history;   
    int    prefetch_on;
    int    *access_history;
    int    access_index;    
} scash_prefetch_t;

typedef struct scash_s {
    int    pid;   /* Process ID */
    int    ps;    /* Partition Size */
    int    pe;    /* PE Number */

    int    proc_id;  /* processor id in SMP box */
    int    node;  /* node id for SMP */
    int    map_fd;	/* mapfile file descriptor */
    int    mtu;   /* MTU of SCASH Network */

    size_t  dsm_size;	/* dsm size */
    int     dsm_pages;	/* number of pages */
    caddr_t dsm_addr;	/* dsm area size */
    caddr_t shadow_addr;	/* shadow area addr */

    addrHandle shadow_addr_handle[SCASH_MAX_PE];
    int pe_node[SCASH_MAX_PE];

    volatile int     home_ok;

    volatile int     lock_stat;
    int              lock_cnt;
    int              lock_hint;
    scash_lock_t     lock[SCASH_MAX_LOCK];

    struct   sigaltstack sig_stack;
    struct   sigaction   sig_act;
    sigset_t sig_mask;
    sigset_t sig_blockio;

    scash_profile_t  profile; /* for profiling */
    scash_prefetch_t prefetch; /* for prefetching */
} scash_t;

extern  scash_t  scash;	/* global structure */

extern int scash_debug; /* debug flag */
extern int scash_tlog_flag; /* for loging */

/*
 * Public Interfaces
 */
void scash_initialize(int, char**, int);
void scash_barrier_init(void);
void scash_lock_init(void);
void scash_sig_handler_set(void);
void scash_poll(void);

caddr_t scash_alloc(size_t);
caddr_t scash_alloc_page_align(size_t);
char*   scash_alloc_omni_shmem(size_t);
void scash_page_align(void);

int scash_pe(void);
int scash_ps(void);
int scash_proc(void);
int scash_smp(void);

void scash_distribute_caddr(caddr_t*);
void scash_distribute_char(char*);
void scash_distribute_int(int*);
void scash_distribute_uint(unsigned*);

void scash_home(int, caddr_t, int);
void scash_home_fo(int, caddr_t, int);
void scash_home_first(caddr_t, int);
int scash_refresh(caddr_t, int);
void scash_omp_flush(caddr_t, int);

void scash_lock_init(void);
void scash_recv_lock(int  id, int  request_node);
void scash_recv_try_lock(int  id, int  request_node);
void scash_recv_unlock(int  id, int  request_node);
int scash_try_lock(int  id);
int scash_lock(int  id);
int scash_unlock(int  id);

void scash_barrier(int  id);
void scash_barrier_sync(int  id);

void scash_check_page(u_int);

void scash_prefetch(void);
int omsc_prefetch_page(int, addrHandle, addrHandle, int);

void scash_fatal(char *msg,...);
void scash_prf(char *msg,...);

int scash_exit(int status);

unsigned long long scash_rdtsc(void);
double scash_dtime();

/***
   +----------+   0x00000000
   |          |
   | Text     |
   |          |
   +----------+
   |          |
   | Data     |
   |          |
   +----------+
   |          |
   |          |
   +----------+
   |          |
   | Shared   |
   | Lib      |
   |          |
   +----------+  SCASH_DEFAULT_ADDR 
   | header   |
   | OmniShmem| 
   | pagetbl  |
   +----------+
   |          |  Protect NONE(1 Page)
   +----------+
   |          |
   |  Shadow  |
   |          |
   +----------+
   |          |  Protect NONE(1 Page)
   +----------+
   |          |
   |  DSM     |
   |          |
   +----------+
   |          |
   +----------+
   |          |
   |  Stack   |
   |          |
   +----------+
***/

#ifdef USE_PM
#define OMSC_INITIALIZE(argc,argv)	scash_score_initialize(argc,argv)
#define OMSC_INIT()		scash_score_init()
#define OMSC_TERMINATE(status) 	sc_terminate(status)
#define OMSC_NUM_PROC()		scash_score_num_proc()
#define OMSC_TMP_FILE(x1,x2,x3) sc_create_temporary_file2(x1,x2,x3)

#define OMSC_BARRIER() 	   	sc_barrier()
#define OMSC_RELEASE_BUF()  	scash_pm_release_buffer()

#define OMSC_WRITE(dst_node,dst_addr,src_addr,size) \
                                scash_pm_write(dst_node,dst_addr,src_addr,size)
#define OMSC_IS_WRITE_DONE()	scash_pm_is_write_done()

#define OMSC_READ(src_node,src_addr,dst_addr,size) \
                                scash_pm_read(src_node,src_addr,dst_addr,size)
#define OMSC_IS_READ_DONE()	scash_pm_is_read_done()

#define OMSC_GET_SEND_BUF(dst,bufpp,size) scash_pm_get_send_buf(dst,bufpp,size)
#define OMSC_TRUNC_BUF(size)	scash_pm_trunc_buf(size)
#define OMSC_SEND()		scash_pm_send()
#define OMSC_IS_SEND_DONE()	scash_pm_is_send_done()
#define OMSC_RECV(addr,size)	scash_pm_recv(addr,size)
#endif

#ifdef USE_MPI
void omsc_initialize(int argc, char **argv);
void omsc_init(void);
void omsc_terminate(int status);
void omsc_barrier(void);
void omsc_start_mpi_handler(void);
void omsc_write(int dst_node, addrHandle dst_addr,
		addrHandle src_addr,int size);
int omsc_is_write_done();
int omsc_read(int src_node, addrHandle src_addr, 
	      addrHandle dst_addr, int size);
int omsc_is_read_done();
int omsc_get_send_buf(int dst, caddr_t *bufpp, int size);
int omsc_trunc_buf(int size);
int omsc_send(void);
int omsc_is_send_done(void);
int omsc_recv(caddr_t *msg, u_int *size);
void omsc_release_buffer(void);
void scash_idle(void);
void scash_busy(void);
void scash_memory_lock_init(void);
void omsc_init_mapfile(size_t map_file_size);
void omsc_send_raw_data(int dst, char *p,int size);
void omsc_recv_raw_data(int src, char *p,int size);

#define OMSC_INITIALIZE(argc,argv)	omsc_initialize(argc,argv)
#define OMSC_INIT()		omsc_init()
#define OMSC_INIT_MAPFILE(size) omsc_init_mapfile(size)
#define OMSC_TERMINATE(status) 	omsc_terminate(status)
#define OMSC_NUM_PROC()		omsc_num_proc

#define OMSC_BARRIER() 	   	omsc_barrier()
#define OMSC_SEND_RAW_DATA(dst,p,size)  omsc_send_raw_data(dst,p,size)
#define OMSC_RECV_RAW_DATA(src,p,size)  omsc_recv_raw_data(src,p,size)

#define OMSC_RELEASE_BUF()  	omsc_release_buffer()/* */

#define OMSC_WRITE(dst_node,dst_addr,src_addr,size) \
                                omsc_write(dst_node,dst_addr,src_addr,size)
#define OMSC_IS_WRITE_DONE()	omsc_is_write_done()

#define OMSC_READ(src_node,src_addr,dst_addr,size) \
                                omsc_read(src_node,src_addr,dst_addr,size)
#define OMSC_IS_READ_DONE()	omsc_is_read_done()

#define OMSC_GET_SEND_BUF(dst,bufpp,size) omsc_get_send_buf(dst,bufpp,size)
#define OMSC_TRUNC_BUF(size)	omsc_trunc_buf(size)
#define OMSC_SEND()		omsc_send()
#define OMSC_IS_SEND_DONE()	omsc_is_send_done()
#define OMSC_RECV(addr,size)	omsc_recv(addr,size)
#endif

#endif  /* __scash_h*/
