/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_barrier.c,v 1.7 2006-10-24 13:40:57 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"
#include  "tlog-omsc.h"

static unsigned  scash_barrier_a2a_cnt = 0;
int  scash_barrier_a2a_recv[4][SCASH_MAX_PE];
unsigned  short  scash_barrier_cnt;

void scash_barrier_sync(int barrier_id);

void  
scash_barrier_init()
{
    int  i, j;

    for (i = 0 ; i < 4 ; i++) {
	for (j = 0 ; j < SCASH_MAX_PE ; j++) {
	    scash_barrier_a2a_recv[i][j] = -1;
	}
    }
}

/*
 * barrier on invalidate protocol
 */
void
scash_barrier(int  barrier_id)
{
#ifdef PROFILE
    u_int64 start, end;
    u_int64 sync_s, sync_e;
    u_int64 flush_s, flush_e;
    u_int64 inval_s, inval_e;
    int diff;

    start = scash_rdtsc();
#endif

    if (scash.ps==1) return;	/* single processor case */

    //if(scash_tlog_flag) tlog_BARRIER_IN();

    //scash_prf("scash_barrier : %d", barrier_id);

#ifdef PROFILE
    sync_s = scash_rdtsc();
    scash.profile.first_flag = TRUE;
#endif
    scash.profile.barrier_flag = TRUE;
    scash_barrier_sync(barrier_id);
#ifdef PROFILE
    scash.profile.first_flag = FALSE;
    sync_e = scash_rdtsc();
    diff = sync_e - sync_s;
    scash.profile.barrier_1st_sync += diff;

    flush_s = scash_rdtsc();
#endif
    if(scash_tlog_flag) tlog_FLUSH_ALL_IN();
    scash.profile.flush_flag = TRUE;
    scash_flush_all();
    scash_barrier_sync(barrier_id);
    scash.profile.flush_flag = FALSE;
#ifdef PROFILE
    if(scash_tlog_flag) tlog_FLUSH_ALL_OUT();
    flush_e = scash_rdtsc();
    diff = flush_e - flush_s;
    scash.profile.barrier_flush += diff;

    inval_s = scash_rdtsc();
#endif
    if(scash_tlog_flag) tlog_PAGE_INV_IN();
    scash.profile.inval_flag = TRUE;
    scash_flush_page_invalidate();
    scash_barrier_sync(barrier_id);
    scash.profile.inval_flag = FALSE;
    scash.profile.barrier_flag = 0;
    if(scash_tlog_flag) tlog_PAGE_INV_OUT();
#ifdef PROFILE
    inval_e = scash_rdtsc();
    diff = inval_e - inval_s;
    scash.profile.barrier_inval += diff;
#endif

    if(scash_tlog_flag) tlog_BARRIER_OUT();

    //scash_prf("scash_barrier end");

#ifdef PROFILE
    end = scash_rdtsc();
    diff = end-start;

    scash.profile.barrier_time += diff;
    scash.profile.barrier_cnt++;
#endif

    /* Prefetch */
    if(scash.prefetch.get_history){
	scash.prefetch.access_history[++scash.prefetch.access_index] = BARRIER;
    }
    if(scash.prefetch.prefetch_on){
	scash_prefetch();
    }
}

void
scash_message_broadcast(int    msg_id,
			int    body_size,
			void*  body)
{
    int i, dest;

    if(scash_debug) 
	scash_prf("scash_message_broadcast id=%s, size=%d ...",
		  scash_msg_name(msg_id),body_size);

    dest = scash.pe+1;
    for(i = 0; i < (scash.ps-1); i++){
	if(dest >= scash.ps) dest = 0;
	scash_message_send(msg_id, dest, body_size, body);
	dest++;
    }
}

int bsync_id = 1;
/*
 *  All to All barrier 
 */

void
scash_barrier_sync(int  barrier_id)
{
    int i,check_cnt,id;
    scash_msg_barrier_a2a_t  barrier_msg;

    if (barrier_id < 0) {
	scash_fatal("scash_barrier:illegal barrier_id");
    }

    if(scash_debug) scash_prf("scash_barrier_a2a id=%d ...",barrier_id);

    /* check send done */
    while (!OMSC_IS_SEND_DONE()){
	scash_poll();
    }

    scash_barrier_a2a_cnt++;
    scash_barrier_a2a_recv[scash_barrier_a2a_cnt % 4][scash.pe] = barrier_id;

    barrier_msg.barrier_cnt = scash_barrier_a2a_cnt;
    barrier_msg.barrier_id  = barrier_id;

    scash_message_broadcast(SCASH_MSG_BARRIER_OK, sizeof(barrier_msg), 
			    &barrier_msg);

    do {
	check_cnt = 0;
	for (i=0; i < scash.ps ; i++) {
	    id = scash_barrier_a2a_recv[scash_barrier_a2a_cnt%4][i];
	    if (id == barrier_id) check_cnt++;
	    else if(id != -1)
		scash_fatal("scash invalid barrier id %d(%d)!!", 
			    id, barrier_id); 
	    scash_poll();
	}
    } while(check_cnt != scash.ps);
    
    /* reset this barrier */
    for (i=0; i < scash.ps ; i++) {
        scash_barrier_a2a_recv[scash_barrier_a2a_cnt % 4][i] = -1;
    }

    scash_poll();
    if(scash_debug) scash_prf("scash_barrier_a2a ... end");
}

/* EOF */
