/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_init.c,v 1.4 2006-10-24 13:40:57 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"

scash_t   scash;	 /* global SCASH structure */

/* local prototype */
static void scash_init(int  shmem_size);
static void scash_shmem_init(caddr_t  map_addr, int  size);

/*
 * initialize omni-scash system
 */
void 
scash_initialize(int argc, char **argv, int shmem_size) 
{
    if(scash_debug) scash_prf("scash_initialize ...");
    OMSC_INITIALIZE(argc, argv);

    scash_init(shmem_size);

    if(scash_debug) scash_prf("scash_initialize ... end");

#ifdef USE_MPI
    omsc_start_mpi_handler();
    /* From here, MPI cannot be called directly. */
#endif    
}

/* 
 * internal main initialize routine 
 */
static void
scash_init(int  shmem_size)
{
    if(scash_debug) scash_prf("scash_init shmem_size=0x%x ...",shmem_size);
    
    if (OMSC_NUM_PROC() == 1) { /* single processor case */
        scash.ps = 1;
        scash.pe = 0;
        scash.proc_id = 0;
        scash.node = 0;
	return;
    }

    OMSC_INIT();

#ifdef USE_PM
    scash_score_init();
    if(scash_tlog_flag) scash_tlog_initialize();
#endif

    scash_barrier_init();

    /* initialize memory */
    scash_shmem_init((caddr_t)SCASH_DEFAULT_ADDR, shmem_size);

    scash_lock_init();
    scash_sig_handler_set();

    sigemptyset(&scash.sig_blockio);
    sigaddset(&scash.sig_blockio, SIGIO);

    /* In case of PM:
     * We should flush kernel buffer caches if we pindown
     * large physical memory.
     */
    if(scash_debug) scash_prf("scash_init ... sync");
    sync();

    OMSC_BARRIER();
    
    if(scash_debug) scash_prf("scash_init ... end");
}

static void
scash_shmem_init(caddr_t  map_addr, int  map_size)
{
    size_t file_map_size;
    caddr_t  mapp;
    unsigned int pagetbl_size;

    if(scash_debug) scash_prf("scash_shmem_alloc : addr = %p, size = %d", map_addr, map_size);

    scash.dsm_size = SCASH_ROUNDUP(map_size, SCASH_PAGE_SIZE);
    scash.dsm_pages = scash.dsm_size/(SCASH_PAGE_SIZE);
    pagetbl_size = sizeof(page_table_t) + sizeof(page_t)*(scash.dsm_pages-1);
    pagetbl_size = SCASH_ROUNDUP(pagetbl_size,SCASH_PAGE_SIZE);
    file_map_size = scash.dsm_size + pagetbl_size;

    if ( (scash.dsm_size/SCASH_PAGE_SIZE) > SCASH_MAX_PAGE )
        scash_fatal("scash_shmem_shared: shared memory size too large"); 

    OMSC_INIT_MAPFILE(file_map_size);

    /* map dsm area */
    scash.dsm_addr = 
	(caddr_t) SCASH_TRUNC((size_t)map_addr, SCASH_PAGE_SIZE);
    mapp = (caddr_t)mmap(scash.dsm_addr, scash.dsm_size, 
			 PROT_READ | PROT_WRITE,
			 MAP_FILE | MAP_SHARED | MAP_FIXED, 
			 scash.map_fd, pagetbl_size);
    
    if(mapp == MAP_FAILED || mapp != scash.dsm_addr) 
	scash_fatal("map dsm failed");
    
    /* map shadow area */
    scash.shadow_addr = (caddr_t)mmap(NULL, scash.dsm_size, 
				      PROT_READ | PROT_WRITE,
				      MAP_FILE | MAP_SHARED, 
				      scash.map_fd, pagetbl_size);

    if(scash.shadow_addr == MAP_FAILED) scash_fatal("map shadow failed");

    scash_memory_lock_init();

    /* page table */
    mapp = (caddr_t)mmap(NULL, pagetbl_size, 
			 PROT_READ | PROT_WRITE,
			 MAP_FILE | MAP_SHARED, 
			 scash.map_fd, 0);

    if(mapp == MAP_FAILED) scash_fatal("map shadow failed");
    scash_page_table = (page_table_t *)mapp;

    /* initalize page table */
    scash_init_pages();
    if(scash_debug) scash_prf("scash_shmem_alloc ... end");

    /* initialize access history */
    scash.prefetch.access_history = NULL;
    scash.prefetch.get_history = 0;
    scash.prefetch.prefetch_on = 0;
}
