/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_lib.c,v 1.6 2006-10-24 13:40:57 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"

void scash_get_access_history();

/*
 *  SCASH Library for C Runtime
 */

int scash_ps() 
{
    return scash.ps;
}

int scash_pe() 
{
    return scash.pe;
}

/* for Omni/lib/libompsm */
int scash_proc()
{
    return 0;
}
int scash_smp()
{
    return 1;
}

/* poll message and call handler */
void
scash_poll()
{
    scash_msg_t*  msg_buff;

    if (scash.ps != 1) {
        while( ENOBUFS != scash_message_recv(&msg_buff) ) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }
}

/* 
 * write address to every node 
 */
static void 
scash_distribute_nbytes(caddr_t gvp,int nbytes) 
{
    int  i;
    volatile char  retval[SCASH_MAX_PE];
    msg_body_t  msg_body;
    caddr_t*  ip;

    if (scash.ps==1)  return;

    ip = (caddr_t*)(msg_body.dist_byte.data);

    for (i = 0 ; i < scash.ps ; i++) {
        retval[i] = SCASH_OFF;
        if (i != scash.pe){
	    msg_body.dist_byte.key = (caddr_t)(retval + i);
	    msg_body.dist_byte.ptr = (caddr_t)gvp;
	    msg_body.dist_byte.length = nbytes;
	    bcopy(gvp,ip,nbytes);
	    scash_message_send(SCASH_MSG_DIST_BYTE, i, 
			       sizeof(scash_msg_dist_byte_t), &msg_body);
	}
    }

    for (i = 0 ; i < scash.ps; i++) {
        while (i != scash.pe && retval[i] == SCASH_OFF) {
	    SCASH_POLL();
        }
    }
}

void 
scash_distribute_caddr(caddr_t*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(caddr_t));
}

void 
scash_distribute_size(size_t*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(size_t));
}

void 
scash_distribute_char(char*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(char));
}
    
void 
scash_distribute_int(int*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(int));
}

void 
scash_distribute_uint(unsigned*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(uint));
}

void 
scash_distribute_double(double*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(double));
}

#define  XX_WORD   (8)
#define  XX_WORD_MASK   (~(XX_WORD - 1))
#define  XX_WORD_ROUNDUP(X)   ((X + XX_WORD -1) & XX_WORD_MASK)

size_t  total_size = 0;

caddr_t
scash_alloc(size_t  s) 
{
    char* old_gp;

    if(scash_debug) scash_prf("scash_alloc s=0x%x",s);
    s = XX_WORD_ROUNDUP(s);

    if (scash.ps==1) {
	return (caddr_t)malloc(s);
    }

    old_gp = scash.dsm_addr + total_size;
    total_size += s;

    scash_distribute_size(&total_size);

    if (total_size >= (size_t)SCASH_MAX_PAGE * (size_t)SCASH_PAGE_SIZE)
	scash_prf("Total size is larger than MAX DSM area");
    return  old_gp;
}

void
scash_page_align() 
{
    total_size = SCASH_PAGE_ROUNDUP(total_size);
    scash_distribute_size(&total_size);
}

caddr_t
scash_alloc_page_align(size_t  s) 
{
    total_size = SCASH_PAGE_ROUNDUP(total_size);
    return scash_alloc(s);
}

/*
 * home management
 */
void
scash_home(int      new_home,
	   caddr_t  pagep, 
	   int      pages)
{
    int   i;
    msg_body_t  msg_body;

    if (scash.ps==1) return;

    scash.home_ok = 1;
    for (i = 0 ; i < scash.ps ; i++) {
      if (i != scash.pe) {
	msg_body.scash_scash_home.home = new_home;
	msg_body.scash_scash_home.pagep = pagep;
	msg_body.scash_scash_home.pagen = pages;
	scash_message_send(SCASH_MSG_SCASH_HOME,
			   i,
			   sizeof(scash_msg_scash_home_t), 
			   &msg_body);
      }
    }

    SCASH_POLL();
    scash_page_set_home(new_home, pagep, pages);

    while(scash.home_ok != scash.ps)
      SCASH_POLL();
}

/* for Omni/lib/libompsm */
void scash_home_fo(int new_home, caddr_t pagep, int pages)
{
    scash_home(new_home, pagep, pages);
}


/* First Touch */
void
scash_home_first(caddr_t pagep, int pages)
{
    int   i;
    msg_body_t  msg_body;
    int start_page;
    
    scash_prf("FirstTouch : %p, %d", pagep, pages);
    if (scash.ps==1)
        return;
    
    start_page = scash_addr2page(pagep);

    for (i = 0; i < pages ; i++) {
	scash_page2home(start_page + i) = FIRST_TOUCH;
    }

    scash.home_ok = 1;
    for (i = 0 ; i < scash.ps ; i++) {
	if (i != scash.pe) {
	    msg_body.scash_scash_home.home = FIRST_TOUCH;
	    msg_body.scash_scash_home.pagep = pagep;
	    msg_body.scash_scash_home.pagen = pages;
	    scash_message_send(SCASH_MSG_SCASH_HOME_FIRST,
				i,
				sizeof(scash_msg_scash_home_t), 
				&msg_body);
	}
    }
    SCASH_POLL();
    scash_prf("FirstTouch end");
}

char *scash_alloc_omni_shmem(size_t  alloc_size)
{
    if (alloc_size > SCASH_OMNI_SHMEM_SIZE) 
	return NULL;
    
    return scash_page_table->omni_shmem;
}

void
scash_senddone_all()
{
    while (!OMSC_IS_SEND_DONE())
	SCASH_POLL();
}

int  scash_exit(int  status) 
{
    OMSC_TERMINATE(status);
    return 0;
}
int  scash_exit_(int status){ return scash_exit(status); }


void scash_profile_print()
{
#ifdef PROFILE
#define CLOCK 1
    scash_prf("barrier : count = %d, time = %lld [us] (1st sync = %lld, flush = %lld, invalidate = %lld)",
	      scash.profile.barrier_cnt, 
	      scash.profile.barrier_time/CLOCK,
	      scash.profile.barrier_1st_sync/CLOCK,
	      scash.profile.barrier_flush/CLOCK,
	      scash.profile.barrier_inval/CLOCK);

    scash_prf("rma : count = %d, time = %lld [us]",
	      scash.profile.rma_cnt, scash.profile.rma_time/CLOCK);
    scash_prf("barrier msg = %d, flush_cnt = %d, inval_cnt = %d",
	      scash.profile.barrier_msg_cnt,
	      scash.profile.barrier_flush_cnt,
	      scash.profile.barrier_inval_cnt);
    /*
    scash_prf("lock_get : count = %d, time = %lld [us]",
	      scash.profile.lock_get_cnt, scash.profile.lock_get_time/CLOCK);
    */
    scash_prf("vm_handler : cnt = %d, time = %lld [us]",
	      scash.profile.handler_cnt, scash.profile.handler_time/CLOCK);
    scash_prf("prefetch_wait = %lld [us]", scash.profile.prefetch_wait/CLOCK);

    scash_prf("total_msg_size = %f MB, barrier_msg_size = %f MB (1st sync : %f MB, flush : %f MB, inval : %f MB",
	      scash.profile.total_msg_size/1048576.0, 
	      scash.profile.barrier_msg_size/1048576.0,
	      scash.profile.barrier_1st_size/1048576.0,
	      scash.profile.barrier_flush_size/1048576.0,
	      scash.profile.barrier_inval_size/1048576.0);
    scash_prf("msg_send : cnt=%d, time=%f, diff_send: cnt=%d, time=%f",
	      scash.profile.msg_cnt,
	      scash.profile.msg_send_time,
	      scash.profile.diff_send_cnt,
	      scash.profile.diff_send_time);
#endif
}
/* for Fortran */
void scash_profile_print_()
{
    scash_profile_print();
}

void scash_profile_reset()
{
#ifdef PROFILE
    scash.profile.barrier_cnt = 0;
    scash.profile.barrier_time = 0;
    scash.profile.barrier_1st_sync = 0;
    scash.profile.barrier_flush = 0;
    scash.profile.barrier_inval = 0;
    scash.profile.inval_msg = 0;
    scash.profile.barrier_msg_cnt = 0;
    scash.profile.barrier_msg_size = 0;
    scash.profile.barrier_1st_size = 0;
    scash.profile.barrier_flush_size = 0;
    scash.profile.barrier_inval_size = 0;
    scash.profile.flush_flag = 0;
    scash.profile.inval_flag = 0;
    scash.profile.first_flag = 0;
    scash.profile.barrier_inval_cnt = 0;
    scash.profile.barrier_flush_cnt = 0;
    
    scash.profile.rma_cnt = 0;
    scash.profile.rma_time = 0;
    scash.profile.barrier_flag = 0;
    scash.profile.lock_get_time = 0;
    scash.profile.lock_get_cnt = 0;
    scash.profile.handler_time = 0;
    scash.profile.rma_wait = 0;
    scash.profile.prefetch_wait = 0;
    scash.profile.total_msg_size = 0;
    scash.profile.get_msg_size = 1;
    scash.profile.handler_time_d = 0.0;
    scash.profile.handler_cnt = 0;

    scash.profile.resend_cnt = 0;
    scash.profile.resend_diff_size = 0.0;

    scash.profile.msg_cnt = 0;
    scash.profile.msg_send_time = 0.0;
    scash.profile.diff_send_cnt = 0;
    scash.profile.diff_send_time = 0.0;
#endif
}
/* for Fortran */
void scash_profile_reset_()
{
    scash_profile_reset();
}

/* EOF */
