/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_lock.c,v 1.1 2004-10-12 05:25:42 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"

static  char  lock_queue[SCASH_MAX_LOCK * SCASH_MAX_PE];

/* initialize scash lock structures */
void
scash_lock_init()
{
    int  i;
    char* qp;

    if(scash_debug) scash_prf("scash_lock_init ...");
    
    qp = lock_queue;

    scash.lock_cnt = 0;
    scash.lock_hint = 0;

    for ( i = 0 ; i < SCASH_MAX_LOCK ; i++) {
        scash.lock[i].id = -1;
	scash.lock[i].head = 0;
	scash.lock[i].q_cnt = 0;
	scash.lock[i].q = (u_char*)qp;
	qp += scash.ps;
    }
}

void
scash_recv_lock(int  id,
	       int  request_node)
{
    int   i, cnt;
    scash_lock_t  *lp;

    for ( i = cnt = 0 ; i < SCASH_MAX_LOCK && cnt < scash.lock_cnt; i++) {
	lp = &scash.lock[ (scash.lock_hint + i) % SCASH_MAX_LOCK];	
	if (lp -> q_cnt)  {
	    cnt++;
	    if (lp -> id == id) {
		lp -> q[(lp->head + lp->q_cnt) % scash.ps ] = request_node;
		lp -> q_cnt++;
		scash.lock_hint = (scash.lock_hint + i) % SCASH_MAX_LOCK;
		return;
	    }
	}
    }

    if (scash.lock_cnt == SCASH_MAX_LOCK)
	scash_fatal("Lock Over Flow");

    for ( lp = &scash.lock[0] ; lp < &scash.lock[SCASH_MAX_LOCK] && lp -> q_cnt ; lp++)
      ;

    if (lp -> q_cnt)
        scash_fatal("scash_recv_lock:Never Reach Here");

    scash.lock_cnt++;
    lp -> id = id;
    lp -> q_cnt = 1;
    lp -> head = 0;
    lp -> q[0] = request_node;

    if (request_node == scash.pe)
        scash.lock_stat = SCASH_LOCK_OK;
    else
        scash_message_send(SCASH_MSG_LOCK_OK, request_node, sizeof(int), &i);
}

void
scash_recv_try_lock(int  id,
		    int  request_node)
{
    int   i, cnt;
    scash_lock_t  *lp;

    for ( i = cnt = 0 ; i < SCASH_MAX_LOCK && cnt < scash.lock_cnt; i++) {
	lp = &scash.lock[ (scash.lock_hint + i) % SCASH_MAX_LOCK];	
	if (lp -> q_cnt)  {
	    cnt++;
	    if (lp -> id == id) {
	      if (request_node == scash.pe)
		  scash.lock_stat = SCASH_LOCK_BUSY;
	      else
		  scash_message_send(SCASH_MSG_LOCK_BUSY, request_node, sizeof(int), &i);
	      return;
	    }
	}
    }

    if (scash.lock_cnt == SCASH_MAX_LOCK)
	scash_fatal("Lock Over Flow");

    for ( lp = &scash.lock[0] ; lp < &scash.lock[SCASH_MAX_LOCK] && lp -> q_cnt ; lp++)
      ;

    if (lp -> q_cnt)
        scash_fatal("scash_recv_try_lock:Never Reach Here");

    scash.lock_cnt++;
    lp -> id = id;
    lp -> q_cnt = 1;
    lp -> head = 0;
    lp -> q[0] = request_node;

    if (request_node == scash.pe)
        scash.lock_stat = SCASH_LOCK_OK;
    else
        scash_message_send(SCASH_MSG_LOCK_OK, request_node, sizeof(int), &i);
}

void
scash_recv_unlock(int  id,
		 int  request_node)
{
    int  i, cnt;
    scash_lock_t  *lp;

    for ( i = cnt = 0 ; i < SCASH_MAX_LOCK && cnt < scash.lock_cnt; i++) {
	lp = &scash.lock[ (scash.lock_hint + i) % SCASH_MAX_LOCK];	
	if (lp->q_cnt) {	    
	    cnt++;
	    if (lp -> id == id) {
	        scash.lock_hint = (scash.lock_hint + i) % SCASH_MAX_LOCK;
		if (lp -> q[ lp -> head] != request_node)
		    scash_fatal("scash_recv_unlock:Unlock Illegal Lock");

		lp -> head = (lp -> head + 1) % scash.ps;
		lp -> q_cnt--;
		if (lp -> q_cnt == 0)
		    scash.lock_cnt--;
		else {
		    if (lp -> q[lp -> head] == scash.pe)
		        scash.lock_stat = SCASH_LOCK_OK;
		    else
		        scash_message_send(SCASH_MSG_LOCK_OK, (int)(lp -> q[lp -> head]), sizeof(int), &i);
		}
		return;
	    }
	}
    }
    scash_fatal("scash_recv_unlock:Never Reach Here");
}

int
scash_try_lock(int  id)
{
    int  dest = (id>>2) % scash.ps;
    msg_body_t   msg_body;

    if (id <= 0) 
	scash_fatal("scash_try_lock:Illegal Lock ID");

    SCASH_POLL();

    msg_body.lock_id = id;
    scash.lock_stat = SCASH_LOCK_OFF;

    if (dest != scash.pe)
	scash_message_send(SCASH_MSG_TRY_LOCK, dest, sizeof(int), &msg_body);
    else
	scash_recv_try_lock(id, dest);

    do {
	SCASH_POLL();
    } while (SCASH_LOCK_OFF == scash.lock_stat);

    return scash.lock_stat;
}

int
scash_lock(int  id)
{
    int  dest = (id>>2) % scash.ps;
    msg_body_t   msg_body;

    if (scash.ps==1)
        return SCASH_SUCCESS;

    if (id <= 0) 
	scash_fatal("scash_lock:Illegal Lock ID");
    SCASH_POLL();

    msg_body.lock_id = id;
    scash.lock_stat = SCASH_LOCK_OFF;
    if (dest != scash.pe)
	scash_message_send(SCASH_MSG_LOCK, dest, sizeof(int), &msg_body);
    else
	scash_recv_lock(id, dest);

    while (SCASH_LOCK_OFF == scash.lock_stat){
	scash_poll();
    }

    return SCASH_SUCCESS;
}

int
scash_unlock(int  id)
{
    int  dest = (id>>2) % scash.ps;
    msg_body_t   msg_body;

    if (scash.ps==1)
        return SCASH_SUCCESS;

    if (id <= 0) {
	scash_fatal("scash_unlock:Illegal Lock ID");
    }

    if (dest == scash.pe) {
	scash_recv_unlock(id, dest);
    } else {
        msg_body.lock_id = id;
	scash_message_send(SCASH_MSG_UNLOCK, dest, sizeof(int), &msg_body);
    }

    SCASH_POLL();

    return SCASH_SUCCESS;
}








