static char rcsid[] = "$Id: omsc_message.c,v 1.8 2006-10-24 13:40:57 ojima Exp $";
/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"
#include  "tlog-omsc.h"

/* 
 * send message 
 */
int
scash_message_send(int    msg_id,
		   int    dest,
		   int    body_size,
		   void*  body)
{
    scash_msg_t*  bufp;
    int  buf_size;
    int  error;

    scash_msg_t*  msg_buff;

    if (scash_debug) {
	scash_prf("scash_send_msg [%s]  to [%d]", 
		  scash_msg_name(msg_id), dest);
    }

    if(scash_tlog_flag) tlog_SEND_IN();

    if (dest==scash.pe) {
	scash_prf("scash_message_send %s, %d, %d", 
		  scash_msg_name(msg_id), body_size, dest);
	exit(1);
    }

    buf_size = sizeof(scash_msg_header_t) + body_size;

    if (buf_size > scash.mtu) {
	scash_prf("scash_send_msg Msg Size Larger than MTU ID=%s, Size=%d,Dest=%d", 
		  scash_msg_name(msg_id), buf_size, dest);
      exit(1);
    }

    while ((error = OMSC_GET_SEND_BUF(dest,(caddr_t*)&bufp, buf_size)) 
	   == ENOBUFS) {
	while(ENOBUFS != scash_message_recv(&msg_buff)) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }

    bufp->header.id = msg_id;
    bufp->header.sender = scash.pe;
    bufp->header.size = buf_size;

    if (body_size > 0) 
	bcopy(body, &bufp->body.body, body_size);

    if ((error = OMSC_SEND()) != 0) {
	scash_fatal("scash_message_send: Send");
    }
    if(scash_tlog_flag) tlog_SEND_OUT();

    return  SCASH_SUCCESS;
}

int  
scash_message_recv(scash_msg_t**  recv_msg)
{
    int   error;
#if defined(__alpha) || defined(__ia64__) || defined(__x86_64__)
    u_long  msg_len;
#else
    u_int msg_len;
#endif

#if  defined(__alpha) || defined(__ia64__) || defined(__x86_64__)
    error = OMSC_RECV((caddr_t*)recv_msg, (u_int*)&msg_len);
#else
    error = OMSC_RECV((caddr_t*)recv_msg, &msg_len);
#endif

    if (error == ENOBUFS) {
	return  error;
    }
    if (error != 0) scash_fatal(" Receive");

    return error;
}

/*
 *  message handler 
 */
void
scash_msg_handler(scash_msg_t*  msg)
{
    int  sender;
    int  fault_page;

    if(scash_debug)
	scash_prf("scash_msg_handler %s ...",scash_msg_name(msg->header.id));

    if(scash_tlog_flag) tlog_MSGH_IN();
    sender = msg->header.sender;

    switch (msg->header.id){
    default: {
	scash_prf("Receive Message [id=%d] from [%d]",
		  msg->header.id, sender);
	scash_fatal("Receive Unknown or Invalid Message");
	break;
    }

    /* 
     * lock messages 
     */
    case SCASH_MSG_LOCK: {
        scash_recv_lock(msg->body.lock_id, sender);
	break;
    }

    case SCASH_MSG_UNLOCK: {
        scash_recv_unlock(msg->body.lock_id, sender);
	break;
    }

    case SCASH_MSG_LOCK_OK: {
	scash.lock_stat = SCASH_LOCK_OK;
	break;
    }

    case SCASH_MSG_TRY_LOCK: {
        scash_recv_try_lock(msg->body.lock_id, sender);
	break;
    }

    case SCASH_MSG_LOCK_BUSY: {
	scash.lock_stat = SCASH_LOCK_BUSY;
	break;
    }

    /*
     * page management 
     */
    /* notify shared */
    case SCASH_MSG_PAGE_SHARED: {
	fault_page = msg->body.fault_page;
	scash_page_cs_set(fault_page, sender,FALSE);
	break;
    }

    case SCASH_MSG_FLUSH_DIFF_SEND:{
	fault_page = msg->body.diff.page;
	scash_page_recv_diff(fault_page,sender, &msg->body.diff.diff);
	break;
    }

    case SCASH_MSG_PAGE_INVALIDATE: {
	scash_page_recv_invalidate(msg->body.invalidate.invalidate.list,
				   msg->body.invalidate.invalidate.count);
	break;
    }

    /* 
     * barrier message
     */
    case SCASH_MSG_BARRIER_OK: {
	unsigned  barrier_cnt;
	int       barrier_id;
        extern int scash_barrier_a2a_recv[4][SCASH_MAX_PE];

	barrier_cnt = msg->body.barrier_a2a.barrier_cnt;
	barrier_id  = msg->body.barrier_a2a.barrier_id;
        scash_barrier_a2a_recv[barrier_cnt % 4][sender] = barrier_id;
	break;
    }

    /* 
     * distribute (remote write) function 
     */
    case SCASH_MSG_DIST_BYTE: {
	msg_body_t  reply_msg_body;
	
	bcopy(msg->body.dist_byte.data, msg->body.dist_byte.ptr, 
	    msg->body.dist_byte.length);

	reply_msg_body.dist_ok.key = msg->body.dist_byte.key;
	scash_message_send(SCASH_MSG_DIST_OK, sender, 
			   sizeof(scash_msg_dist_ok_t), &reply_msg_body);
	break;
    }

    case SCASH_MSG_DIST_OK: {
	*msg->body.dist_ok.key = SCASH_ON;
	break;
    }

    /*
     * home management
     */
    case SCASH_MSG_SCASH_HOME: {
	msg_body_t  reply_msg_body;
	scash_page_set_home(msg->body.scash_scash_home.home,
			    msg->body.scash_scash_home.pagep,
			    msg->body.scash_scash_home.pagen);
	scash_message_send(SCASH_MSG_SCASH_HOME_OK, 
			   sender, 
			   sizeof(int), 
			   &reply_msg_body);
	break;
    }

    case SCASH_MSG_SCASH_HOME_OK: {
	scash.home_ok++;
	break;
    }

    /* first touch */
    case SCASH_MSG_SCASH_HOME_FIRST: {
	scash_page_set_home(FIRST_TOUCH,
		   msg->body.scash_scash_home.pagep,
		   msg->body.scash_scash_home.pagen);
	break;
    }

    case SCASH_MSG_SCASH_HOME_FIRST_OK:{
	scash.home_ok++;
	break;
    }

    case SCASH_MSG_FIRST_REQ:{
	msg_body_t  reply_msg_body;
	if(scash.pe != 0)
	    fprintf(stderr, "<%d> Received SCASH_MSG_FIRST_REQ\n", scash.pe);
	/* first touch */
	if(scash_page2home(msg->body.first.page) == FIRST_TOUCH){
	    scash_page2home(msg->body.first.page) = msg->body.first.home;
	    reply_msg_body.first.home = msg->body.first.home;
	    reply_msg_body.first.page = msg->body.first.page;
	}
	else{
	    reply_msg_body.first.home = scash_page2home(msg->body.first.page);
	    reply_msg_body.first.page = msg->body.first.page;
	}
	
	scash_message_send(SCASH_MSG_FIRST_RET,
			   msg->header.sender,
			   sizeof(int),
			   &reply_msg_body);
	break;
    }

    case SCASH_MSG_FIRST_RET:{
	fprintf(stderr, "<%d> Received SCASH_MSG_FIRST_RET\n", scash.pe);
	exit(1);
	break;
    }
    }

    if(scash_tlog_flag) tlog_MSGH_OUT();
}

/*
 *  for debug 
 */
char *scash_msg_name(int msg_id)
{
    switch(msg_id){
    case  SCASH_MSG_PAGE_SHARED: return "PAGE_SHARED";
    case  SCASH_MSG_PAGE_INVALIDATE: return "PAGE_INVALIDATE";
    case  SCASH_MSG_FLUSH_DIFF_SEND: return "FLUSH_DIFF_SEND";
    case  SCASH_MSG_BARRIER_OK: return "BARRIER_OK";
    case  SCASH_MSG_TRY_LOCK:  return "TRY_LOCK";
    case  SCASH_MSG_LOCK: return "MSG_LOCK";
    case  SCASH_MSG_LOCK_OK: return "LOCK_OK";
    case  SCASH_MSG_LOCK_BUSY: return "LOCK_BUSY";
    case  SCASH_MSG_UNLOCK: return "UNLOCK";
    case  SCASH_MSG_DIST_BYTE: return "DIST_BYTE";
    case  SCASH_MSG_DIST_OK: return "DIST_OK";
    case  SCASH_MSG_SCASH_HOME: return "SCASH_HOME";
    case  SCASH_MSG_SCASH_HOME_OK: return "SCASH_HOME_OK";
    default: return "SCASH_MSG_???";
    }
}
