/* 
 */
#ifndef __omsc_msg_h
#define __omsc_msg_h

enum message_id {
    /* page management */
    SCASH_MSG_PAGE_SHARED,
    SCASH_MSG_PAGE_INVALIDATE,
    SCASH_MSG_FLUSH_DIFF_SEND,

    /* barrier */
    SCASH_MSG_BARRIER_OK,

    /* lock function */
    SCASH_MSG_TRY_LOCK,
    SCASH_MSG_LOCK,
    SCASH_MSG_LOCK_OK,
    SCASH_MSG_LOCK_BUSY,
    SCASH_MSG_UNLOCK,

    SCASH_MSG_SCASH_HOME,
    SCASH_MSG_SCASH_HOME_OK,

    SCASH_MSG_DIST_BYTE,
    SCASH_MSG_DIST_OK,

    /* First touch */
    SCASH_MSG_SCASH_HOME_FIRST,
    SCASH_MSG_SCASH_HOME_FIRST_OK,
    SCASH_MSG_FIRST_REQ,
    SCASH_MSG_FIRST_RET,

    SCASH_MSG_END
};

typedef  struct scash_msg_header_s{
    enum message_id id:8;
    u_int  sender:8;
    u_int    size;
} scash_msg_header_t;

typedef  struct  scash_msg_diff_s {
    u_long   page;
    diff_t   diff;
}scash_msg_diff_t;

typedef  struct  scash_msg_dist_byte_s {
    caddr_t   key;
    caddr_t   ptr;
    int       length;
    char      data[8];
} scash_msg_dist_byte_t ;

typedef  struct  scash_msg_dist_ok_s {
    caddr_t   key;
} scash_msg_dist_ok_t;

typedef  struct  scash_msg_scash_home_s {
   int      home;
   caddr_t  pagep; 
   int      pagen;
} scash_msg_scash_home_t;

typedef  struct  scash_msg_barrier_a2a_s {
   unsigned  barrier_cnt;
   int       barrier_id;
} scash_msg_barrier_a2a_t;

typedef  struct  scash_msg_invalidate_s {
    invalidate_t invalidate;
} scash_msg_invalidate_t;

/* First Touch */
typedef  struct  scash_msg_first {
    int home;
    int page;
} scash_msg_first_t;

typedef union msg_body_u {
    char   body;
    u_int  fault_page;
    int    lock_id;
    scash_msg_diff_t diff;
    scash_msg_dist_byte_t dist_byte;
    scash_msg_dist_ok_t dist_ok;
    scash_msg_scash_home_t scash_scash_home;
    scash_msg_barrier_a2a_t barrier_a2a;
    scash_msg_invalidate_t invalidate;
    scash_msg_first_t first;
} msg_body_t;

typedef  struct scash_msg_s {
    scash_msg_header_t   header;
    msg_body_t   body;
} scash_msg_t;

/* prototype */
extern void scash_msg_handler(scash_msg_t*);
extern int scash_message_send(int, int, int, void*);
extern void scash_message_broadcast(int, int, void*);
extern int scash_message_recv(scash_msg_t**);
extern char *scash_msg_name(int id);

#endif  /*  __omsc_msg_h  */
