/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_page.c,v 1.8 2006-10-24 13:40:57 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>
#include  <pthread.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"
#include  "tlog-omsc.h"

page_t  *scash_pages;
page_table_t *scash_page_table;

u_int *scash_touch_pages;
int scash_touch_count;

u_char *scash_page_flags;

static twin_page_t *scash_twin_free_list = NULL;

static void scash_flush_diff_send(u_int page);
static void scash_diff_index_create(int, int, _s_diff_index_t*);

void scash_init_pages()
{
    scash_pages = scash_page_table->pages;

    /* clean */
    bzero(scash_pages,sizeof(page_t)*scash.dsm_pages);

    scash_touch_count = 0;
    if((scash_touch_pages = (u_int *)malloc(sizeof(u_int)*scash.dsm_pages))
       == NULL)
	scash_fatal("cannot alloc touch_pages");

    if((scash_page_flags = (u_char *)malloc(scash.dsm_pages)) == NULL)
	scash_fatal("cannot alloc page_flags");
    bzero(scash_page_flags,scash.dsm_pages);
     
    /* set default page mapping */
    {
	int home,i;

	for (i = 0 ; i < scash.dsm_pages ; i++) {
	    home = i % scash.ps;
	    scash_page_set_home(home,scash_page2addr(i),1);
	}
    }
}

pthread_mutex_t touch_mutex;
void scash_page_touch(u_int page)
{
    pthread_mutex_lock(&touch_mutex);

    u_char *p = &scash_page_flags[page];
    if((*p&SCASH_PAGE_TOUCHED) == 0){
	*p |= SCASH_PAGE_TOUCHED;
	scash_touch_pages[scash_touch_count++] = page;
    }

    pthread_mutex_unlock(&touch_mutex);
}

void scash_page_touch_clear(u_int page)
{
    u_char *p = &scash_page_flags[page];
    *p &= ~(SCASH_PAGE_TOUCHED|SCASH_PAGE_TWIN);
}

/* set read-only (RO) */
void scash_page_RO(u_int page)
{
    int r;
    u_char *p = &scash_page_flags[page];
    if(((*p)|SCASH_PAGE_RW_MASK) == SCASH_PAGE_RO) return;
    *p = ((*p)&~SCASH_PAGE_RW_MASK)|SCASH_PAGE_RO;
    r = mprotect(scash_page2addr(page), SCASH_PAGE_SIZE, PROT_READ);
    if(r != 0) scash_fatal("scash_page_RO");
}

/* set read-write (RW) */
void scash_page_RW(u_int page)
{
    int r;
    u_char *p = &scash_page_flags[page];

    if(((*p)|SCASH_PAGE_RW_MASK) == SCASH_PAGE_RW) return;
    *p = ((*p)&~SCASH_PAGE_RW_MASK)|SCASH_PAGE_RW;
    r = mprotect(scash_page2addr(page), SCASH_PAGE_SIZE,
		 PROT_READ | PROT_WRITE);
    if(r != 0) scash_fatal("scash_page_RW");
}

void scash_page_INVALID(u_int page)
{
    int r;
    u_char *p = &scash_page_flags[page];

    if(((*p)|SCASH_PAGE_RW_MASK) == SCASH_PAGE_INVALID) return;
    *p = ((*p)&~SCASH_PAGE_RW_MASK);
    r = mprotect(scash_page2addr(page), SCASH_PAGE_SIZE, PROT_NONE);
    if(r != 0) scash_fatal("scash_page_INVALID");
}

/*
 * shared set management
 */
void scash_send_cs_set(u_int page)
{
    scash_message_send(SCASH_MSG_PAGE_SHARED,
		       scash_pages[page].home, sizeof(u_int),
		       &page);
}

void scash_page_cs_set(int  page, int  node, int modified)
{
    page_t *p = &scash_pages[page];
    u_char *cp;

    cp = p->info.cs + (node/8);
    *cp |= (1)<<(node%8);
    p->is_modified |= modified;
}

int scash_page_node_in_cs(int page, int node)
{
    page_t *p = &scash_pages[page];
    u_char *cp;

    cp = p->info.cs + (node/8);
    return (*cp & ((1)<<(node%8)));
}

void scash_page_cs_clear_all(u_int page)
{
    int  i;
    page_t *p = &scash_pages[page];
    for (i = 0 ; i < CS_SIZE ; i++)
        p->info.cs[i] = 0;
    p->is_modified = FALSE;
 }

/* 
 *  prepare copy of page 
 */
void scash_check_page(u_int page)
{
    int home;
#ifdef PROFILE
    u_int64 start, end;
    int diff;
    start = scash_rdtsc();
#endif
    
    if(SCASH_PAGE_IS_LOCAL(page)) return;  /* local page, already have copy */
    if(scash_pages[page].is_valid) return;
    home = scash_pages[page].home;

    /* get page copy */
    if(scash_debug) scash_prf("OMSC_READ to %d, page %d", home, page);
    OMSC_READ(home,
	      scash.shadow_addr_handle[home]+(page*SCASH_PAGE_SIZE),
	      scash.shadow_addr_handle[scash.pe]+(page*SCASH_PAGE_SIZE),
	      SCASH_PAGE_SIZE);

    while(!OMSC_IS_READ_DONE())
	scash_poll();

    /* access history */
    if(scash.prefetch.get_history){
	scash.prefetch.access_history[++scash.prefetch.access_index] = page;
    }
#ifdef PROFILE
    end = scash_rdtsc();
    diff = end-start;
    scash.profile.rma_time += diff;
    scash.profile.rma_cnt++;
#endif
}

/*
 * prepare twin.  
 */
void 
scash_check_twin(u_int  page)
{
    twin_page_t *tp;

    if(scash_pages[page].have_twin) return;	/* already have twin */
    if(SCASH_PAGE_IS_LOCAL(page)) return;	/* local page */

    if((tp = scash_twin_free_list) == NULL){
	if((tp = (twin_page_t *)malloc(sizeof(*tp))) == NULL)
	    scash_fatal("scash_twin_create");
    } else 
	scash_twin_free_list = tp->next;

    //scash_page_RO(page);
    bcopy(scash_page2addr(page), tp->page, SCASH_PAGE_SIZE);
    scash_pages[page].info.twin = tp;
    scash_pages[page].have_twin = TRUE;
    SCASH_PAGE_SET_TWIN(page);
}

/* flush all pages, called from barrier (1st phase) */
/* 
 * flush modified remote page 
 */
void scash_flush_all()
{
    int i;
    u_int page;
    
    /* scan touched pages */
    for(i = 0; i < scash_touch_count; i++){
	page = scash_touch_pages[i];
	if(SCASH_PAGE_IS_REMOTE(page)){
	    /* for remote page */
	    if(scash_pages[page].have_twin){
		twin_page_t *tp;
		scash_flush_diff_send(page);	/* send diff */
		tp = scash_pages[page].info.twin;
		free(tp); /* free twin */
		scash_pages[page].have_twin = FALSE;
	    } else
		;//scash_send_cs_set(page); /* send shared */
	} else {
	    /* for local page */
	    if(SCASH_PAGE_IS_RW(page))
		scash_page_cs_set(page,scash.pe,TRUE);
	    else 
		scash_page_cs_set(page,scash.pe,FALSE);
	}

	/* make it read-only */
	if(SCASH_PAGE_IS_RW(page))
	    scash_page_RO(page);

	scash_poll();
    }
}


/* 
 * user's API: flush(write) the range from "from" to from+len
 */
int
scash_flush(caddr_t from,
	    int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  flush_ptr;
    int      flush_page, flush_size;
    int      page_prot;
    int      offset;

    if (scash.ps==1)
        return SCASH_SUCCESS;

    SCASH_POLL();

    flush_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (flush_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return (-1);

    while(flush_ptr < from + byte_len) {
        page_ru_ptr = 
	  (caddr_t)SCASH_PAGE_TRUNC((off_t)flush_ptr + SCASH_PAGE_SIZE);

	if (page_ru_ptr > from + byte_len)
            flush_size = from + byte_len - flush_ptr;
	else
	    flush_size = page_ru_ptr - flush_ptr;

	flush_page = scash_addr2page(flush_ptr);
	home = scash_page2home(flush_page);

	if (SCASH_PAGE_IS_REMOTE(flush_page)) {
	    SCASH_PAGE_LOCK(flush_page);

	    page_prot = SCASH_PAGE_PROT_ATTR(flush_page);
	    if (page_prot == SCASH_PAGE_RW) {
		OMSC_WRITE(home,
			   scash.shadow_addr_handle[home] + 
			   (flush_ptr - scash.dsm_addr), 
			   scash.shadow_addr_handle[scash.pe] + 
			   (flush_ptr - scash.dsm_addr), 
			   flush_size);
		while(!OMSC_IS_WRITE_DONE()) /* */;

		if(scash_pages[flush_page].have_twin){
		    offset = (flush_ptr - scash.dsm_addr) & (SCASH_PAGE_SIZE-1);
		    bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
			  (char*)scash_pages[flush_page].info.twin + offset,
			  flush_size);
		}
	    } /* if page_prot == RW */ 
	    SCASH_PAGE_UNLOCK(flush_page);
        } /* if (home != scash.pe) */
        flush_ptr += flush_size;
    }
    SCASH_POLL();
    return SCASH_SUCCESS;
}


void
scash_omp_flush(caddr_t  from,  int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  flush_ptr;
    int      flush_page, flush_size;
    int      page_prot;
    int      offset;

    unsigned  *diff_ptr;
    unsigned  *twin_ptr;

    if (scash.ps==1)    return;

    SCASH_POLL();

    flush_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (flush_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return;

    while(flush_ptr < from + byte_len) {
	page_ru_ptr = 
	    (caddr_t)SCASH_PAGE_TRUNC((off_t)flush_ptr + SCASH_PAGE_SIZE);
	
	if (page_ru_ptr > from + byte_len)
	    flush_size = from + byte_len - flush_ptr;
	else
	    flush_size = page_ru_ptr - flush_ptr;

	flush_page = scash_addr2page(flush_ptr);
	home = scash_page2home(flush_page);

	if (SCASH_PAGE_IS_REMOTE(flush_page)) {
	    SCASH_PAGE_LOCK(flush_page);

	    page_prot = SCASH_PAGE_PROT_ATTR(flush_page);
	    if (page_prot == SCASH_PAGE_RO) {
		OMSC_READ(home,
			  scash.shadow_addr_handle[home] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			  scash.shadow_addr_handle[scash.pe] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			  flush_size);
		while(!OMSC_IS_READ_DONE()) /* */;

		if(scash_pages[flush_page].have_twin){
		    offset = (flush_ptr - scash.dsm_addr) & (SCASH_PAGE_SIZE-1);
		    bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
			  (char*)scash_pages[flush_page].info.twin + offset, 
			  flush_size);
		} 
	    }

	    if (page_prot == SCASH_PAGE_RW) {
 		int	diff_len = 0;
 		int	need_flush = 1;

		diff_ptr = (unsigned*)((flush_ptr - scash.dsm_addr) + scash.shadow_addr);

		offset = (flush_ptr - scash.dsm_addr) & (SCASH_PAGE_SIZE-1);
		twin_ptr = (unsigned*)((char*)scash_pages[flush_page].info.twin + offset);

		while( (caddr_t)diff_ptr < ((flush_ptr - scash.dsm_addr) + scash.shadow_addr) + flush_size) {
 		    diff_len = 0;
 		    while( (caddr_t)&diff_ptr[diff_len] < ((flush_ptr - scash.dsm_addr) + scash.shadow_addr) + flush_size) {
 		        if (diff_ptr[diff_len] != twin_ptr[diff_len]) {
 			    diff_len++;
 			} else {
 			    break;
 			}
 		    }

 		    if (diff_len) {
		        OMSC_WRITE(home,
				   scash.shadow_addr_handle[home] + (((caddr_t)diff_ptr) - scash.shadow_addr), 
				   scash.shadow_addr_handle[scash.pe] + (((caddr_t)diff_ptr) - scash.shadow_addr), 
				   diff_len * sizeof(int));
			
 			diff_ptr += diff_len;
 			twin_ptr += diff_len;

 			if (diff_len * sizeof (unsigned) == flush_size) {
 			  need_flush = 0;
 			}
		    } else {
 		        diff_ptr++;
 			twin_ptr++;
  		    }
 		}
 		while(!OMSC_IS_WRITE_DONE()) /* */;

 		if (need_flush) {
 		    OMSC_READ(home,
			      scash.shadow_addr_handle[home] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			      scash.shadow_addr_handle[scash.pe] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			      flush_size);
 		    while(!OMSC_IS_READ_DONE()) /* */;
  		}

		if(scash_pages[flush_page].have_twin){
		    bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
			  (char*)scash_pages[flush_page].info.twin + offset,
			  flush_size);
		}

	    } /* if page_prot == RW */ 
	    SCASH_PAGE_UNLOCK(flush_page);
        } /* if (home != scash.pe) */
        flush_ptr += flush_size;
    }
    SCASH_POLL();
}


/*
 * API: refresh 
 */
int
scash_refresh(caddr_t  from, 
	      int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  refresh_ptr;
    int      refresh_page, refresh_size;
    int      page_prot;
    int      offset;

    //scash_prf("scash_refresh(%p, %d)", from, byte_len);

    SCASH_POLL();

    refresh_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (refresh_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return 1;

    while(refresh_ptr < from + byte_len){
        page_ru_ptr = 
	  (caddr_t)SCASH_PAGE_TRUNC((off_t)refresh_ptr + SCASH_PAGE_SIZE);
	if (page_ru_ptr > from + byte_len)
            refresh_size = from + byte_len - refresh_ptr;
	else
	    refresh_size = page_ru_ptr - refresh_ptr;

	refresh_page = scash_addr2page(refresh_ptr);
	home = scash_page2home(refresh_page);

	if (SCASH_PAGE_IS_REMOTE(refresh_page)) {

	    SCASH_PAGE_LOCK(refresh_page);

	    page_prot = SCASH_PAGE_PROT_ATTR(refresh_page);
	    if (page_prot == SCASH_PAGE_RO || page_prot == SCASH_PAGE_RW) {
		OMSC_READ(home,
			  scash.shadow_addr_handle[home] + 
			  ((caddr_t)refresh_ptr - scash.dsm_addr), 
			  scash.shadow_addr_handle[scash.pe] + 
			  ((caddr_t)refresh_ptr - scash.dsm_addr), 
			  refresh_size);

		while(!OMSC_IS_READ_DONE()) /* */;

		if(scash_pages[refresh_page].have_twin){
		    offset = (refresh_ptr - scash.dsm_addr) & (SCASH_PAGE_SIZE-1);
		    bcopy(scash.shadow_addr + (refresh_ptr - scash.dsm_addr), 
			  (char*)scash_pages[refresh_page].info.twin + offset,
			  refresh_size);
		}
	    } 
	    SCASH_PAGE_UNLOCK(refresh_page);
        } /* if (home != scash.pe) */
        refresh_ptr += refresh_size;
    }
    SCASH_POLL();
    return SCASH_SUCCESS;
}



/* called from barrier, after flush (2nd) */
/* send invalidate messsage */
void
scash_flush_page_invalidate()
{
    int i,page,node,size;
    int count[SCASH_MAX_PE];
    scash_msg_invalidate_t inval[SCASH_MAX_PE];

    for(i = 0; i < scash.ps; i++)
	count[i] = 0;

    /* scan touched pages, check only for home */
    for(i = 0; i < scash_touch_count; i++){
	page = scash_touch_pages[i];
	if(scash_pages[page].home == scash.pe){
	    if(scash_pages[page].is_modified){
		for (node = 0 ; node < scash.ps ; node++){
		    if(node == scash.pe) continue;	/* skip myself */
		    /* for page in shared set, flush it */
		    if(scash_page_node_in_cs(page, node))
			inval[node].invalidate.list[count[node]++] = page;
		    
		    if(count[node] == MAX_INVALIDATE_MSGS){ /* buf is full */
			inval[node].invalidate.count = count[node];
			size = sizeof(int) + sizeof(u_int)*count[node];
			
			scash_message_send(SCASH_MSG_PAGE_INVALIDATE,
					   node, size,
					   &inval[node]);
			count[node] = 0;
		    }
		}
		scash_page_cs_clear_all(page);
	    }
	}

	/* clear touch flag */
	scash_page_touch_clear(page);
	
	scash_poll();
    }
    
    /* send list */
    for(node = 0; node < scash.ps; node++){
	if(count[node] != 0 && node != scash.pe){
	    inval[node].invalidate.count = count[node];
	    size = sizeof(int) + sizeof(u_int)*count[node];
	    
	    scash_message_send(SCASH_MSG_PAGE_INVALIDATE,
			       node, size,
			       &inval[node]);
	}
    }
    /* reset touch_count */
    scash_touch_count = 0;
}


/* called from message handler */
void scash_page_recv_invalidate(u_int *list, int count)
{
    int i, page;
 
    for(i = 0; i < count; i++){
	page = list[i];
	if(SCASH_PAGE_IS_REMOTE(page)){
	    scash_page_INVALID(page);	/* invalidate page */
	    scash_pages[page].is_valid = FALSE;
	    
	    if(scash_pages[page].have_twin)
		free(scash_pages[page].info.twin);
	}
    }
}

/* 
 * home management 
 */
void
scash_page_set_home(int home, caddr_t  start, int pages)
{
    u_int page;
    caddr_t pagep;
    page_t *pp;

    pagep = (caddr_t)(SCASH_TRUNC((off_t)start, SCASH_PAGE_SIZE));

    while(pages-- > 0) {
	page = scash_addr2page(pagep);
	if(scash.pe_node[home] == scash.node){
	    /* page in the same node */
	    /* set RO, and local */
	    scash_page_RO(page);
	    SCASH_PAGE_SET_LOCAL(page);	/* page is local */
	    if(scash.pe == home){
		pp = &scash_pages[page];
		bzero(pp,sizeof(*pp));
		pp->home = home;
		pp->is_valid = TRUE;
	    }
	} else {
	    /* page is in the different node */
	    /* set NONE, and remote */
	    scash_page_INVALID(page);
	    SCASH_PAGE_SET_REMOTE(page);
	    if(scash.proc_id == 0){
		pp = &scash_pages[page];
		bzero(pp,sizeof(*pp));
		pp->home = home;
		pp->is_valid = FALSE;
	    }
	}
	pagep += SCASH_PAGE_SIZE;
    }
}

/*
 * diff
 */
static void
scash_flush_diff_send(u_int page)
{
    int  diff_size;
    _s_diff_index_t  diff_index;
    int  max_diff_size;
    int  i,home;
    int  diff_msg_size = 0;
    u_int   *pagep;
    scash_msg_diff_t  scash_msg_diff;
    diff_block_t   *diff_block;

    pagep = (u_int*)(scash_page2shadowaddr(page));
    home = scash_pages[page].home;
    diff_index.diff_num = 0;

    max_diff_size = scash.mtu - sizeof(scash_msg_header_t) 
	- sizeof(int) - sizeof(int) -sizeof(diff_block_header_t) - 4;

    scash_diff_index_create(page, max_diff_size, &diff_index);

    /*
     *  Send Diff
     */
    scash_msg_diff.page = page;
    scash_msg_diff.diff.length = 0;
    diff_block =  (diff_block_t*)&scash_msg_diff.diff.block[0];
    diff_msg_size = sizeof(scash_msg_header_t) + sizeof(u_long) + sizeof(int);

    if (diff_index.diff_num==0) {
        return;
    }

    i = 0;
    while (i < diff_index.diff_num){
        diff_size = sizeof(diff_block_header_t) + 
	    diff_index.db[i].length * SCASH_WORD_SIZE;
	if (diff_msg_size + diff_size < scash.mtu){
	  diff_msg_size += diff_size;
	  scash_msg_diff.diff.length++;
	  diff_block->header = diff_index.db[i];
	  bcopy((char*)(pagep + diff_index.db[i].pos), 
		diff_block->data, 
		diff_index.db[i].length * SCASH_WORD_SIZE);

	  diff_block = (diff_block_t*)(diff_block->data + 
				       diff_index.db[i].length);

	  i++;
	} else {
	    scash_message_send(SCASH_MSG_FLUSH_DIFF_SEND,
			       home,
			       diff_msg_size - sizeof(scash_msg_header_t),
			       &scash_msg_diff);

	    scash_msg_diff.diff.length = 0;
	    diff_block =  (diff_block_t*)&scash_msg_diff.diff.block[0];
	    diff_msg_size = sizeof(scash_msg_header_t) + 
		sizeof(int) + sizeof(int);
	}
    }

    if (diff_msg_size > sizeof(scash_msg_header_t)+sizeof(int)+sizeof(int)){
        scash_message_send(SCASH_MSG_FLUSH_DIFF_SEND, 
			   home,
			   diff_msg_size - sizeof(scash_msg_header_t),
			   &scash_msg_diff);
    }

    //scash_poll();
    return;
}

static void
scash_diff_index_create(int   page,
			int   max_diff_size,
			_s_diff_index_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_header_t  *blockp = NULL;

    pagep = (u_int*)(scash_page2shadowaddr(page));
    twinp = (u_int*)scash_pages[page].info.twin->page;

    diff->diff_num = 0;

    blockp = &diff->db[0];
#ifdef __ia64__
    blockp->pos = blockp->length = 0;
#else
    blockp->pos = blockp->length = blockp->pad = 0;
#endif /* __ia64__ */

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
      if (*pagep != *twinp ) {
	    if (in_block) {
		blockp->length++;
		if (blockp->length * SCASH_WORD_SIZE >= max_diff_size) {
		    in_block = 0;
		    blockp++;
		}
	    } else {
		diff->diff_num++;
		in_block = 1;
		blockp -> pos = i;
		blockp -> length = 1;
	    }
	} else {
	    if (in_block) {
		in_block = 0;
		blockp++;
	    } 
	}
	pagep++;
	twinp++;
    }
}

void scash_page_recv_diff(u_int page, int pe, diff_t    *diffp)
{
    int       i;
    u_int     *pagep;
    diff_block_t   *blockp;

    if (scash_pages[page].home != scash.pe) {
	scash_prf("SCASH_MSG_FLUSH_DIFF_SEND page[%d] is not mine\n",
		  page);
    }

    pagep = (u_int*)(scash_page2shadowaddr(page));
    blockp = (diff_block_t*)&diffp->block[0];
    for (i = 0 ; i < diffp->length ; i++) {
	bcopy(&blockp->data[0], pagep + blockp->header.pos , 
	      blockp->header.length * sizeof(int));
	blockp = (diff_block_t*)&(blockp->data[blockp->header.length]);
    }
    scash_page_touch(page);
    scash_page_cs_set(page, pe, TRUE);
}

/* EOF */
