/* 
 * $Id: omsc_page.h,v 1.6 2006-10-24 13:40:57 ojima Exp $
 * $PCCC_Release$
 * $PCCC_Copyright$
 */

#include <asm/bitops.h>
#include <linux/kernel.h>
#include "lock_macro.h"

#ifndef _omsc_page_h
#define _omsc_page_h

#define  SCASH_OFF (0)
#define  SCASH_ON  (1)

#define  SCASH_SIG_STACK_SIZE  (SCASH_PAGE_SIZE * 32)
#define  SCASH_WORD_IN_PAGE  (SCASH_PAGE_SIZE/SCASH_WORD_SIZE)

#define  CS_SIZE  (SCASH_MAX_PE/8)

#define  SCASH_ROUNDUP(X,Y)    ((X+Y-1) & (~(Y-1)))
#define  SCASH_PAGE_ROUNDUP(X) ((X+ SCASH_PAGE_SIZE -1) & (~(SCASH_PAGE_SIZE-1)))
#define  SCASH_WORD_ROUNDUP(X) ((X+ SCASH_WORD_SIZE -1) & (~(SCASH_WORD_SIZE-1)))

#define  SCASH_TRUNC(X,Y)     ((X) & (~(Y-1)))
#define  SCASH_PAGE_TRUNC(X)  ((X) & (~(SCASH_PAGE_SIZE-1)))
#define  SCASH_WORD_TRUNC(X)  ((X) & (~(SCASH_WORD_SIZE-1)))

#ifdef __alpha__
#define  SCASH_DEFAULT_ADDR   ((unsigned long)0x80000000) 
#elif __ia64__
#define  SCASH_DEFAULT_ADDR   (0x6000000000000000 + 0x80000000)
#elif __x86_64__
#define  SCASH_DEFAULT_ADDR   ((unsigned long)0x80000000)
#else
#define  SCASH_DEFAULT_ADDR   ((1<<30) + (1<<27))
#endif

#define  SCASH_MAP_FILE   "scash.map" 

#define  SCASH_POLL()     scash_poll()
#define  SCASH_WAITE()    scash_waite()

/*
 * SCASH/OMNI Shared memory
 */
#define   SCASH_OMNI_SHMEM_SIZE  (4 * 4096)

/*
 * Diff 
 */
#ifdef __ia64__
typedef struct  diff_block_header_s{
    u_int   length:16;
    u_int   pos:16;
} diff_block_header_t;
#else
typedef struct  diff_block_header_s{
    u_int   pad:8;
    u_int   length:12;
    u_int   pos:12;
} diff_block_header_t;
#endif

typedef struct diff_block_s {
    diff_block_header_t    header;
#ifdef __ia64__
    u_int     data[4096]; /* might be SCASH_WORD_IN_PAGE : needs check */
#else
    u_int     data[1024];
#endif /* __ia64__ */
} diff_block_t;

typedef struct diff_min_block_s {
    diff_block_header_t    header;
    u_int     data[1];
} diff_min_block_t;

typedef struct diff_s {
    int   length;
    diff_min_block_t  block[1024];
} diff_t;

typedef  struct  _s_diff_index_s {
    int   diff_num;
    diff_block_header_t  db[1024]; 
} _s_diff_index_t;

typedef struct _twin_page_s {
    struct _twin_page_s *next;
    char page[SCASH_PAGE_SIZE];
} twin_page_t;

#define MAX_INVALIDATE_MSGS 1024
typedef struct _invalidate_s {
    int          count;
    unsigned int list[MAX_INVALIDATE_MSGS];
} invalidate_t;

/*
 * DSM Page Table
 */
typedef  struct _page_s {
    volatile u_char lock;
    u_char home:8;		/* page home */
    u_char is_valid:1;	 	/* have valid copy */
    u_char is_modified:1;	/* modified */
    u_char have_twin:1;		/* have twin page */
    u_char is_prefetching:1;    /* prefetching */
    union {
	u_char cs[CS_SIZE];  	/* copy set for home page */
	twin_page_t *twin;      /* twin for remote page */
    } info;
} page_t;

typedef  struct  page_table_s {
    volatile u_char lock;
    char _dummy_pad[16];	/* dummy */
    char omni_shmem[SCASH_OMNI_SHMEM_SIZE];
    page_t pages[1];
} page_table_t;

extern  page_t  *scash_pages;
extern  page_table_t *scash_page_table;
extern  unsigned short  scash_barrier_cnt;

extern  u_char *scash_page_flags;  	/* local */

#define SCASH_PAGE_RD           0x01
#define SCASH_PAGE_WR           0x02
#define SCASH_PAGE_RW_MASK	(SCASH_PAGE_RD|SCASH_PAGE_WR)

#define SCASH_PAGE_INVALID      0x00 
#define SCASH_PAGE_RO           SCASH_PAGE_RD
#define SCASH_PAGE_RW           (SCASH_PAGE_RD|SCASH_PAGE_WR)

#define SCASH_PAGE_PROT_ATTR(page) (scash_page_flags[page]&SCASH_PAGE_RW_MASK)

#define SCASH_PAGE_IS_RO(page) \
 ((scash_page_flags[page]&SCASH_PAGE_RW_MASK) == SCASH_PAGE_RO)
#define SCASH_PAGE_IS_RW(page) \
 ((scash_page_flags[page]&SCASH_PAGE_RW_MASK) == SCASH_PAGE_RW)
   
#define SCASH_PAGE_TOUCHED 	0x80	
#define SCASH_PAGE_TWIN	  	0x40
#define SCASH_PAGE_HAVE_TWIN(page) \
 ((scash_page_flags[page]&SCASH_PAGE_TWIN) != 0)
#define SCASH_PAGE_SET_TWIN(page) \
 (scash_page_flags[page] |= SCASH_PAGE_TWIN)

#define SCASH_PAGE_LOCAL  	0x20	/* have local copy */

#define SCASH_PAGE_IS_LOCAL(page) \
 ((scash_page_flags[page]&SCASH_PAGE_LOCAL) != 0)
#define SCASH_PAGE_IS_REMOTE(page) \
 ((scash_page_flags[page]&SCASH_PAGE_LOCAL) == 0)
#define SCASH_PAGE_SET_LOCAL(page) \
 (scash_page_flags[page] |=SCASH_PAGE_LOCAL)
#define SCASH_PAGE_SET_REMOTE(page) \
 (scash_page_flags[page] &=~SCASH_PAGE_LOCAL)

#define SCASH_PAGE_LOCK(page)
#define SCASH_PAGE_UNLOCK(page)

/* first touch */
#define  FIRST_TOUCH  255

#define  scash_page2home(page) 	 (scash_pages[page].home)
#define  scash_page2addr(P) 	 (scash.dsm_addr + ((P) * SCASH_PAGE_SIZE))
#define  scash_page2shadowaddr(P) (scash.shadow_addr + ((P) * SCASH_PAGE_SIZE))
#define  scash_addr2page(addr)	(((addr) - scash.dsm_addr)/ SCASH_PAGE_SIZE)
#define  scash_pageoffset(addr)      ((off_t)(addr) & (SCASH_PAGE_SIZE - 1))

/* prototype */
void scash_init_pages(void);
void scash_page_touch(u_int page);
void scash_page_touch_clear(u_int page);
void scash_page_RO(u_int page);
void scash_page_RW(u_int page);
void scash_page_INVALID(u_int page);
void scash_send_cs_set(u_int page);
void scash_page_cs_set(int  page, int  node, int modified);
int  scash_page_node_in_cs(int page, int node);
void scash_page_cs_clear_all(u_int page);
void scash_check(u_int page);
void scash_check_twin(u_int  page);
void scash_flush_all(void);
void scash_flush_page_invalidate(void);
void scash_page_send_invalidate(u_int page);
void scash_page_recv_invalidate(u_int *list, int count);
void scash_page_set_home(int home, caddr_t  pagep, int pages);
void scash_page_recv_diff(u_int page, int pe, diff_t    *diffp);

#endif  /* _omsc_page_h */









