#include  <stdio.h>
#include  <stdlib.h>

#include "omsc.h"
#include "omsc_page.h"

void scash_get_access_history();

/* Prefetch */
void scash_prefetch_range(caddr_t pagep, int pages)
{
    u_int st_page, ed_page;
    u_int i, home;

    if(scash.ps == 1) return;
    
    pagep = (caddr_t)(SCASH_TRUNC((unsigned long)pagep, SCASH_PAGE_SIZE));

    st_page = scash_addr2page(pagep); // start page
    ed_page = st_page + pages -1;
    // scash_prf("prefetch : from %d to %d", st_page, ed_page);
    
    // set prefetch flag
    for(i = st_page; i <= ed_page; i++){
	if(SCASH_PAGE_IS_REMOTE(i))
	    scash_pages[i].is_prefetching = 1;
    }
    
    // send RMR request
    for(i = st_page; i <= ed_page; i++){
	// read request
	home = scash_page2home(i);
	if(SCASH_PAGE_IS_REMOTE(i)){
	    omsc_prefetch_page(home,
		      scash.shadow_addr_handle[home]+(i*SCASH_PAGE_SIZE),
		      scash.shadow_addr_handle[scash.pe]+(i*SCASH_PAGE_SIZE),
		      SCASH_PAGE_SIZE);
	}
    }
} 


/* Prefetch using Access History */
void scash_prefetch()
{
    int page;
    int home;

    /* access history is null? */
    if(scash.prefetch.access_history == NULL){
	scash_get_access_history();
	return;
    }
    else{
	scash.prefetch.prefetch_on = 1; /* prefetch flag */

	/* scan access history */
	for(; scash.prefetch.access_history[scash.prefetch.access_index] != BARRIER &&
		scash.prefetch.access_history[scash.prefetch.access_index] != END; 
	    scash.prefetch.access_index++){
	    
	    page = scash.prefetch.access_history[scash.prefetch.access_index];
	    if(page == START) continue;
	    
	    //if(!scash_pages[page].have_twin){ /* have twin ? */
	    if(!SCASH_PAGE_HAVE_TWIN(page)){
		//if(scash_debug); scash_prf("prefetch page : %d", page);

		/* set prefetch flag */
		scash_pages[page].is_prefetching = 1;

		/* send RMR request */
		home = scash_page2home(page);

		if(scash_debug) scash_prf("prefetch to %d, page %d", home, page);
		omsc_prefetch_page(home,
		     scash.shadow_addr_handle[home]+(page*SCASH_PAGE_SIZE),
		     scash.shadow_addr_handle[scash.pe]+(page*SCASH_PAGE_SIZE),
		     SCASH_PAGE_SIZE);
	    }
	}
	if(scash.prefetch.access_history[scash.prefetch.access_index] == END)
	    scash.prefetch.access_index = 0;
	else
	    scash.prefetch.access_index++;
    }
}

void scash_get_access_history()
{
    scash.prefetch.get_history = 1;
    scash.prefetch.prefetch_on = 0;

    if(scash.prefetch.access_history == NULL){
	scash.prefetch.access_history =
	    (int*)malloc(sizeof(int)*scash.dsm_pages);
	scash.prefetch.access_history[0] = START;
	scash.prefetch.access_index = 0;
    }
    else
	return;
}

void scash_prefetch_end()
{
    if(scash.prefetch.get_history){
	scash.prefetch.access_history[++scash.prefetch.access_index] = END;
	scash.prefetch.get_history = 0;
	scash.prefetch.access_index = 0;
    }
    scash.prefetch.prefetch_on = 0;
}

/* for Fortran */
void scash_prefetch_(){ scash_prefetch(); }
void scash_prefetch_end_(){ scash_prefetch_end(); }


void scash_print_access_history()
{
    int i;

    if(scash.prefetch.access_history == NULL)
	return;

    for(i = 0; scash.prefetch.access_history[i] != END; i++){
	if(scash.prefetch.access_history[i] == START)
	    scash_prf("%d : START", i);
	else if(scash.prefetch.access_history[i] == BARRIER)
	    scash_prf("%d : BARRIER", i);
	else
	    scash_prf("%d : page %d", i, scash.prefetch.access_history[i]);
    }
    scash_prf("END");
}
void scash_print_access_history_(){ scash_print_access_history(); }
