/* 
 * $Id: omsc_tlog.c,v 1.5 2006-10-24 13:40:57 ojima Exp $
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
#include  <stdio.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <sys/types.h>

#include  "omsc.h"
#include  "tlog-omsc.h"
#include  "tlog.h"

void scash_tlog_initialize()
{
    double t1,t2,t0,t_diff;
    int i;
    char c;

    if(scash_debug) scash_prf("scash_tlog_initialize ...");

    OMSC_BARRIER();
    if(scash.pe == 0){
	tlog_init("scash.log");
	for(i = 1; i < scash.ps; i++){
	    t1 = tlog_get_time();
	    OMSC_SEND_RAW_DATA(i,&c,1);
	    OMSC_RECV_RAW_DATA(i,(char *)&t0,sizeof(double));
	    t2 = tlog_get_time();
	    t_diff = t1+(t2-t1)/2.0 - t0;
	    OMSC_SEND_RAW_DATA(i,(char *)&t_diff,sizeof(double));
	}
    } else {
	tlog_init0();
	OMSC_RECV_RAW_DATA(0,&c,1);
	t0 = tlog_get_time();
	/* scash_prf("slave t0=%g",t0); */
	OMSC_SEND_RAW_DATA(0,(char *)&t0,sizeof(double));
	OMSC_RECV_RAW_DATA(0,(char *)&t_diff,sizeof(double));
	tlog_adjust_time(t_diff);
    }
    OMSC_BARRIER();
    //tlog_log(scash.pe,TLOG_START);
}

void scash_tlog_finalize()
{
    TLOG_BLOCK *bp;
    char buf[TLOG_BLOCK_SIZE];
    int size,i;
    char c;	/* dummy buffer */

    if(scash_debug) scash_prf("scash_tlog_finalize ...");
    //tlog_log(scash.pe,TLOG_END);
    
    OMSC_BARRIER(); 
    if(scash.pe == 0){	/* master */
	tlog_file_open();
	/* first dump my data */
	/* computation thread */
	for(bp = tlog_handle_table[scash.pe].block_top; 
	    bp != NULL;  bp = bp->next)
	    tlog_dump_data((char *)bp->data);
	/* communication thread */
	for(bp = tlog_handle_table[scash.pe+1].block_top; 
	    bp != NULL;  bp = bp->next)
	    tlog_dump_data((char *)bp->data);

	for(i = 2; i < scash.ps*2; i++){
	    while(1){
		OMSC_SEND_RAW_DATA(i/2,&c,1);
		OMSC_RECV_RAW_DATA(i/2,(char *)&size,sizeof(int));
		if(size == 0) break;
		OMSC_RECV_RAW_DATA(i/2,(char *)buf,TLOG_BLOCK_SIZE);
		tlog_dump_data(buf);
	    }
	}
	tlog_file_close();
    } else {
	/* send log */
	size = TLOG_BLOCK_SIZE;
	/* computation thread */
	for(bp = tlog_handle_table[scash.pe*2].block_top; 
	    bp != NULL;  bp = bp->next){
	    OMSC_RECV_RAW_DATA(0,&c,1);
	    OMSC_SEND_RAW_DATA(0,(char *)&size,sizeof(int));
	    OMSC_SEND_RAW_DATA(0,(char *)bp->data,size);
	}
	size = 0;
	OMSC_RECV_RAW_DATA(0,&c,1);
	OMSC_SEND_RAW_DATA(0,(char *)&size,sizeof(int));

	/* communication thread */
     	size = TLOG_BLOCK_SIZE;
	for(bp = tlog_handle_table[scash.pe*2+1].block_top; 
	    bp != NULL;  bp = bp->next){
	    OMSC_RECV_RAW_DATA(0,&c,1);
	    OMSC_SEND_RAW_DATA(0,(char *)&size,sizeof(int));
	    OMSC_SEND_RAW_DATA(0,(char *)bp->data,size);
	}
	size = 0;
	OMSC_RECV_RAW_DATA(0,&c,1);
	OMSC_SEND_RAW_DATA(0,(char *)&size,sizeof(int));
    }
    OMSC_BARRIER();
}

void tlog_START_EVENT(int id) {  }

void tlog_BARRIER_IN(int id) { tlog_comp_log(scash.pe,TLOG_BARRIER_IN); }
void tlog_BARRIER_OUT(int id) { tlog_comp_log(scash.pe,TLOG_BARRIER_OUT); }

void tlog_MSGH_IN(int id) { tlog_comp_log(scash.pe,TLOG_MSGH_IN); }
void tlog_MSGH_OUT(int id) { tlog_comp_log(scash.pe,TLOG_MSGH_OUT); }

void tlog_SEND_IN(int id) { tlog_comp_log(scash.pe,TLOG_SEND_IN); }
void tlog_SEND_OUT(int id) { tlog_comp_log(scash.pe,TLOG_SEND_OUT); }

void tlog_PAGEF_IN(int id) { tlog_comp_log(scash.pe,TLOG_PAGEF_IN); }
void tlog_PAGEF_OUT(int id) { tlog_comp_log(scash.pe,TLOG_PAGEF_OUT); }

void tlog_BSYNC_IN() { tlog_comp_log(scash.pe, TLOG_BSYNC_IN); }
void tlog_BSYNC_OUT() { tlog_comp_log(scash.pe,TLOG_BSYNC_OUT); }

void tlog_FLUSH_ALL_IN(int id) { tlog_comp_log(scash.pe,TLOG_FLUSH_ALL_IN); }
void tlog_FLUSH_ALL_OUT(int id) { tlog_comp_log(scash.pe,TLOG_FLUSH_ALL_OUT); }

void tlog_PAGE_INV_IN(int id) { tlog_comp_log(scash.pe,TLOG_PAGE_INV_IN); }
void tlog_PAGE_INV_OUT(int id) { tlog_comp_log(scash.pe,TLOG_PAGE_INV_OUT); }

void tlog_PAGEF_EVENT(int id) { tlog_comp_log(scash.pe,TLOG_PAGEF_EVENT); }

void tlog_MPI_SEND_EVENT(int id) { tlog_comm_log(scash.pe,TLOG_MPI_SEND_EVENT); }
void tlog_MPI_RECV_EVENT(int id) { tlog_comm_log(scash.pe,TLOG_MPI_RECV_EVENT); }

void tlog_RMA_IN(int id) { tlog_comm_log(scash.pe,TLOG_RMA_IN); }
void tlog_RMA_OUT(int id) { tlog_comm_log(scash.pe,TLOG_RMA_OUT); }

void scash_tlog_start()
{
    double t0;

    scash_barrier_sync(100);

    t0 = tlog_timestamp();
    tlog_reset_time(t0);

    scash_tlog_flag = TRUE;
    tlog_comp_log(scash.pe,TLOG_START);
    tlog_comm_log(scash.pe,TLOG_START);
    scash_prf("scash_tlog_start end");
}
void scash_tlog_start_(){ scash_tlog_start(); }

void scash_tlog_end()
{
    tlog_comp_log(scash.pe,TLOG_END);
    tlog_comm_log(scash.pe,TLOG_END);
    scash_tlog_flag = FALSE;
}
void scash_tlog_end_(){ scash_tlog_end(); }
