/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: omsc_vm_handler.c,v 1.6 2006-10-24 13:40:57 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#ifdef __x86_64__
#include  <asm/ucontext.h>
#endif
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>
#ifdef  __linux__
#include  <asm/unistd.h>
#else
#include  <unistd.h>
#endif
#include  <pthread.h>

#include  "omsc.h"
#include  "omsc_page.h"
#include  "omsc_message.h"
#include  "tlog-omsc.h"

#ifdef __linux__
void
#ifdef  __alpha__
#define VM_HANDLER_DECL() scash_vm_handler(int  signal, \
					   int  code, \
					   struct sigcontext*  sc) 
#define GET_FAULT_INFO    fault_pc = sc->sc_pc; \
                          fault_addr = (caddr_t)sc->sc_traparg_a0;

#define IS_WRITE_FAULT sc->sc_traparg_a1 == 0x04
#elif defined(__ia64__)
#define VM_HANDLER_DECL() scash_vm_handler(int  signal, \
					   struct siginfo *sinfo,\
					   struct sigcontext *scntxt)
#define GET_FAULT_INFO    fault_pc = scntxt->sc_ip;\
                          fault_addr = (caddr_t)sinfo->si_addr;\
			  sc = scntxt;
#define IS_WRITE_FAULT (sinfo->si_code == SEGV_ACCERR || sinfo->si_code == SEGV_MAPERR)
#elif defined(__x86_64__)
#define VM_HANDLER_DECL() scash_vm_handler(int  signal, \
					   struct siginfo *sinfo,\
					   struct ucontext *uc)
#define GET_FAULT_INFO    fault_pc = uc->uc_mcontext.rip; \
                          fault_addr = sinfo->si_addr;
#define IS_WRITE_FAULT (sinfo->si_code == SEGV_ACCERR || sinfo->si_code == SEGV_MAPERR)
#else
#define VM_HANDLER_DECL() scash_vm_handler(int  signal,int  code)
#define GET_FAULT_INFO    sc = (struct sigcontext*)&code; \
			  fault_pc = sc->eip; \
                          fault_addr = (caddr_t)sc -> cr2;
#define IS_WRITE_FAULT sc->trapno == 0x0e && sc->err & 0x2
#endif
#else  /* not __linux__ */
#define VM_HANDLER_DECL() scash_vm_handler(int  signal,\
					   int  code,\
					   struct sigcontext*  sc,\
					   caddr_t fault_addr)
#define IS_WRITE_FAULT (sc->sc_trapno == T_PAGEFLT && sc->sc_err & 0x2)
#endif

VM_HANDLER_DECL()
{
    unsigned long  fault_page;
    unsigned long  fault_pc;
    caddr_t  fault_addr;
    u_char prot_attr;
    //struct sigcontext*  sc;
    msg_body_t msg_body;
    scash_msg_t* msg_buff;
    u_int msg_length;
    int error;
    int home_node;

#ifdef PROFILE
    u_int64 start, end;
    int diff;
    u_int64 prefetch_start, prefetch_end;
    double d_start, d_end;

    start = scash_rdtsc();
    d_start = scash_dtime();
#endif

    if(scash_debug) scash_prf("scash_vm_handler ...");
    if(scash_tlog_flag){
	tlog_PAGEF_EVENT();
	tlog_PAGEF_IN();
    }

    scash_poll();
    
    GET_FAULT_INFO;

    if (fault_addr < scash.dsm_addr || 
	fault_addr >= scash.dsm_addr + scash.dsm_size) {
#ifdef __ia64__
	scash_prf("VM Fault out of DSM area PC=[%08lx]:ADDR=[%p]\n", fault_pc, fault_addr);
#elif defined(__x86_64__)
	scash_prf("VM Fault out of DSM area PC=[%p]:ADDR=[%p]\n", fault_pc, fault_addr); 
#else
	scash_prf("VM Fault out of DSM area PC=[%08x]:ADDR=[%08x]",
		  (unsigned)fault_pc, (unsigned)fault_addr);
#endif
	scash_exit(1);
    }

    fault_addr = 
	(caddr_t)SCASH_TRUNC((unsigned long)fault_addr, SCASH_PAGE_SIZE);
    fault_page = 
	((unsigned long)(fault_addr - scash.dsm_addr))/ SCASH_PAGE_SIZE;

    if(scash_debug) scash_prf("fault_page=0x%x ...",fault_page);

    /* now prefetching */
    if(scash_pages[fault_page].is_prefetching){
#ifdef PROFILE
	prefetch_start = scash_rdtsc();
#endif
	while(scash_pages[fault_page].is_prefetching){
	    /* scash_prf("waiting prefetch ..."); */
	    sched_yield();
	}
	scash_page_RO(fault_page);
	scash_pages[fault_page].is_prefetching = FALSE;

#ifdef PROFILE
	prefetch_end = scash_rdtsc();
	diff = prefetch_end - prefetch_start;
	scash.profile.prefetch_wait += diff;
#endif
    }

    /* Lock Fault Page */
    SCASH_PAGE_LOCK(fault_page);
    prot_attr = SCASH_PAGE_PROT_ATTR(fault_page);

    /* First Touch */
    home_node = scash_page2home(fault_page);
    if(home_node == FIRST_TOUCH) {
	if(scash_debug)scash_prf("first_touch to page %d", fault_page);
	if(scash.pe == 0){
	    home_node = 0;
	    scash_page2home(fault_page) = 0;
	    scash_home(0, fault_addr, 1);
	}
	/* mutual exclusion */
	else{
	    /* send request to node 0 */
	    msg_body.first.page = fault_page;
	    msg_body.first.home = scash.pe;
	    scash_message_send(SCASH_MSG_FIRST_REQ, 0, sizeof(scash_msg_first_t), &msg_body);
	    
	    /* receive answer from node 0 */
	    for(;;){
		while((error = scash_message_recv(&msg_buff)) == ENOBUFS)
		    /* wait */;
		
		if(msg_buff->header.id == SCASH_MSG_FIRST_RET){
		    OMSC_RELEASE_BUF();
		    break;
		}
		else{
		    scash_msg_handler(msg_buff);
		    OMSC_RELEASE_BUF();
		}
	    }  
	    
	    /* first touch */
	    home_node = msg_buff->body.first.home;
	    scash_page2home(fault_page) = home_node;
	    if(scash_page2home(fault_page) == scash.pe){
		scash_home(scash.pe, fault_addr, 1);
	    }
	    else{
		printf("home node of page %4d is pe %2d\n", fault_page, scash.pe);
		fflush(stdout);
	    }
	}
    }

#if 1
    if(prot_attr == SCASH_PAGE_RO){ // Write Access Fault
	scash_page_RW(fault_page);
	scash_check_twin(fault_page);
	scash_page_touch(fault_page);
    }
    else if(prot_attr == SCASH_PAGE_INVALID){ // Read Access Fault
	scash_page_RO(fault_page);
	scash_check_page(fault_page);
	scash_page_touch(fault_page);
    }
    else{
	scash_fatal("unknown page fault ???");
    }
#endif

#if 0
    if (IS_WRITE_FAULT) {
	/* Write Access Fault */
	if(scash_debug) scash_prf("W Falut: page_attr %d ...",prot_attr);

	switch(prot_attr){
	case SCASH_PAGE_RW:
	default:
	    scash_fatal("unknown W Falut ???");
	    break;

	case SCASH_PAGE_INVALID:
	    scash_check_page(fault_page);
	    /* FALL THROUGH */
	case SCASH_PAGE_RO:
	    if(scash_debug) scash_prf("W Fault: page_attr RO/INV ...");
	    scash_check_twin(fault_page);
	    scash_page_RW(fault_page);
	    scash_page_touch(fault_page);
	}
    } else {
	/* Read ACCESS Fault */
	if(scash_debug) scash_prf("R Falut: page_attr %d ...",prot_attr);

	switch(prot_attr){
	case SCASH_PAGE_RW:
	case SCASH_PAGE_RO:
	default:
	    scash_fatal("unknown R Falut");
	    break;

	case SCASH_PAGE_INVALID:
	    scash_check_page(fault_page);
	    scash_page_RO(fault_page);
	    scash_page_touch(fault_page);
	    break;
	}
    }
#endif

    /* Unlock Fault Page */ 
    SCASH_PAGE_UNLOCK(fault_page);

    scash_poll();

    if(scash_tlog_flag) tlog_PAGEF_OUT();
    if(scash_debug) scash_prf("scash_vm_handler ... end");

#ifdef PROFILE
    end = scash_rdtsc();
    d_end = scash_dtime();
    diff = end - start;
    scash.profile.handler_time += diff;
    scash.profile.handler_time_d += d_end - d_start;
    scash.profile.handler_cnt++;
#endif
}

void  
scash_sig_handler_set()
{

#ifdef __NetBSD__
    scash.sig_stack.ss_flags = 0;
    scash.sig_stack.ss_size = SCASH_SIG_STACK_SIZE;
    scash.sig_stack.ss_sp = (caddr_t)malloc(SCASH_SIG_STACK_SIZE);

#ifdef MMAP_STACK
    scash.sig_stack.ss_sp = mmap(0, SCASH_SIG_STACK_SIZE, 
				 PROT_READ|PROT_WRITE,
				 MAP_ANON|MAP_PRIVATE, 
				 scash.mmap, 0);
#endif   /* MMAP_STACK */
    if (sigaltstack( &scash.sig_stack, NULL) < 0)
	scash_exit (1);

#endif  /* __NetBSD__ */
#ifdef  __linux__
    struct _sigaltstack {
        void *ss_sp;
        int ss_flags;
        size_t ss_size;
    };

    struct _sigaltstack  sig_stack;

    sig_stack.ss_flags = 0;
    sig_stack.ss_size = SCASH_SIG_STACK_SIZE;
    sig_stack.ss_sp = (char *)valloc(SCASH_SIG_STACK_SIZE);
    if (sigaltstack((struct sigaltstack *)&sig_stack, NULL) < 0) {
      perror("sigaltstack");
      exit(1);
    }
#endif

    sigemptyset(&scash.sig_mask);
    sigaddset(&scash.sig_mask, SIGBUS);
    sigaddset(&scash.sig_mask, SIGSEGV);
    sigaddset(&scash.sig_mask, SIGIO);
    sigaddset(&scash.sig_mask, SIGQUIT);
    scash.sig_act.sa_mask = scash.sig_mask;

#ifdef __linux__
#ifdef __ia64__
    scash.sig_act.sa_flags = 0;
#else
#ifdef __x86_64__
    scash.sig_act.sa_flags = SA_SIGINFO | SA_ONSTACK;
#else
    scash.sig_act.sa_flags = SA_STACK;
#endif
#endif
#ifdef __i386__
    scash.sig_act.sa_restorer = 0;
#endif
#else
    scash.sig_act.sa_flags = SA_ONSTACK;
#endif
    if(scash_debug) scash_prf("scash_sig_handler_set ...\n");

    scash.sig_act.sa_handler = (void (*)(int))scash_vm_handler;
    sigaction(SIGSEGV, &scash.sig_act, NULL);
    if (sigaction(SIGSEGV, &scash.sig_act, NULL) < 0)
        perror("sigaction");
}

/* EOF */
