/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash.c,v 1.6 2004-08-24 18:58:50 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>

#ifdef USE_PM
#include  <score_resource.h>
#endif

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

scash_t   scash;	 /* global SCASH structure */

scash_stats_t   scash_stats[SCASH_MAX_STATS];
unsigned char   *scash_upmap;

unsigned  scash_option;

static  char  lock_queue[SCASH_MAX_LOCK * SCASH_MAX_PE];

/*
 * initialize scash system
 */
void 
scash_initialize(int argc, char **argv, int shmem_size) 
{
    extern void scash_tlog_initialize();

    if(scash_debug) scash_prf("scash_initialize ...");
    OMSC_INITIALIZE(argc, argv);

    scash_init(shmem_size);

    if(scash_debug) scash_prf("scash_initialize ... end");

#ifdef USE_MPI_TCP
    omsc_start_mpi_handler();
    /* From here, MPI cannot be called directly. */
#endif

#ifdef USE_MPI_TCP
    if(scash_tlog_flag) scash_tlog_initialize();
#endif
}

/* initialize scash lock structures */
void
scash_lock_init()
{
    int  i;
    char* qp;

    if(scash_debug) scash_prf("scash_lock_init ...");
    
    qp = lock_queue;

    scash.lock_cnt = 0;
    scash.lock_hint = 0;

    for ( i = 0 ; i < SCASH_MAX_LOCK ; i++) {
        scash.lock[i].id = -1;
	scash.lock[i].head = 0;
	scash.lock[i].q_cnt = 0;
	scash.lock[i].q = (u_char*)qp;
	qp += scash.ps;
    }
}

/* initialize proc private data */
void
scash_pe_init()
{
    int   i;
    extern  void scash_page_lock_init();

    if(scash_debug) scash_prf("scash_pe_init ...");

    scash.pid = getpid();
    scash.seq = 0;
    scash.logf = stderr;
    scash.map_file[0] = 0x0;
    scash.page_file[0] = 0x0;

#ifdef SCASH_FAULT_TRACE
    scash.tracef = fopen(SCASH_TRACE_FILE, "w");
    scash.fault_trace = 0; 
#endif
    
    scash_page_lock_init();
    scash_option = SCASH_OPT_DEFAULT;

#ifdef USE_PM
    scash_score_init();
#endif
    scash_pa_init();

    dirty_page = (dirty_page_t*)malloc(sizeof(dirty_page_t));
    dirty_page->lock_bit = 0;
    scash_dp_init(); /* called twice? */

#ifdef SCASH_INVALIDATE_POOL
    scash_invalidate_pool_init();
#endif

    scash_lock_init();
    scash_sig_handler_set();
    scash.update_cnt = 0;

    scash.debug_level = -1;
    scash.barrier.master = 0;
    scash.barrier.id = -1;
    scash.barrier.cnt = 0;
    scash.barrier.f = 0;
    scash_barrier_tree_node(scash.ps, scash.pe, 
			    &scash.barrier.gap, &scash.barrier.upper);

    scash_init_stats_all();
    scash.stat_cnt = 0;
    scash.stats = scash_stats;
    for (i=0 ; i < SCASH_MAX_PE ; i++)
	scash.last_seq[i] = 0;

    for (i=0 ; i < SCASH_MAX_PE ; i++)
	scash.shadow_addr_handle[i] = 0;

    for (i=0 ; i < SCASH_MAX_PAGE/32 ; i++)
	scash.flushed_page[i] = 0;

    sigemptyset(&scash.sig_blockio);
    sigaddset(&scash.sig_blockio, SIGIO);

    if(scash_debug) scash_prf("scash_pe_init ... end");
}

/* poll message and call handler */
void
scash_poll()
{
    scash_msg_t*  msg_buff;
#ifdef SCASH_STATS
    u_int64  ccnt;
    ccnt = scash_rdtsc();
    scash.stats->poll_cnt++;
#endif

    if (scash.ps != 1) {
        while( ENOBUFS != scash_message_recv(&msg_buff) ) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }
#ifdef SCASH_STATS
    scash.stats->poll_clock += (scash_rdtsc() - ccnt)/CPU_CLOCK;
#endif
}

void
scash_wait()
{
    scash_msg_t*  msg_buff;

#ifdef  SCASH_STATS
    u_int64  ccnt;
    ccnt = scash_rdtsc();
    scash.stats->wait_cnt++;
#endif

    while( ENOBUFS == scash_message_recv(&msg_buff) )
	SCASH_IDLE();

    SCASH_BUSY();
    scash_msg_handler(msg_buff);
    OMSC_RELEASE_BUF();

    while( ENOBUFS != scash_message_recv(&msg_buff) ) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
#ifdef  SCASH_STATS
    scash.stats->waite_clock += (scash_rdtsc() - ccnt)/CPU_CLOCK;
#endif
}

/*
 *  SCASH Library for C Runtime
 */
static void
scash_shmem_init(caddr_t  addr, int  size)
{
    int  retval;

    if(scash_debug) scash_prf("scash_shmem_init addr=0x%x size=0x%x...",addr,size);

    retval = scash_shmem_alloc(&addr, size);
    if (retval != SCASH_SUCCESS) {
        scash_prf("could not allocate dsm");
        scash_exit(1);
    }
}

/* 
 * internal main initialize routine 
 */
void
scash_init(int  shmem_size)
{
    void  scash_barrier_init(int , int, int);

    if(scash_debug) scash_prf("scash_init shmem_size=0x%x ...",shmem_size);
    
    if (OMSC_NUM_PROC() == 1) { /* single processor case */
        scash.pe = 0;
        scash.ps = 1;
        scash.smp = 1;
        scash.proc = 0;
        scash.box = 0;
	return;
    }

    OMSC_INIT();

#ifdef USE_PM
    if(scash_tlog_flag) scash_tlog_initialize();
#endif

    scash_barrier_init(0, OMSC_NUM_PROC()-1, 0);
    OMSC_BARRIER();

    /* initialize processor private data */
    scash_pe_init();
    OMSC_BARRIER();

    /* initialize memory */
    scash_shmem_init((caddr_t)SCASH_DEFAULT_ADDR, shmem_size);
    OMSC_BARRIER();

    /*
     * We should flush kernel buffer caches if we pindown
     * large physical memory.
     */
    if(scash_debug) scash_prf("scash_init ... sync");
    sync();
    /* OMSC_BARRIER();*/ /* scash_barrier_a2a(1); */
    /* scash_barrier_a2a(1);*/
    OMSC_BARRIER();
    
    if(scash_debug) scash_prf("scash_init ... end");
}

/* write address to every node */
static void 
scash_distribute_nbytes(caddr_t gvp,int nbytes) 
{
    int  i;
    volatile char  retval[SCASH_MAX_PE];
    msg_body_t  msg_body;
    caddr_t*  ip;

    if (scash.ps==1)  return;

    ip = (caddr_t*)(msg_body.dist_byte.data);

    for (i = 0 ; i < scash.ps ; i++) {
        retval[i] = SCASH_OFF;
        if (i != scash.pe){
	    msg_body.dist_byte.key = (caddr_t)(retval + i);
	    msg_body.dist_byte.ptr = (caddr_t)gvp;
	    msg_body.dist_byte.length = nbytes;
	    bcopy(gvp,ip,nbytes);
	    scash_message_send(SCASH_MSG_DIST_BYTE, i, 
			       sizeof(scash_msg_dist_byte_t), &msg_body);
	}
    }

    for (i = 0 ; i < scash.ps; i++) {
        while (i != scash.pe && retval[i] == SCASH_OFF) {
	    SCASH_POLL();
        }
    }
}

void 
scash_distribute_caddr(caddr_t*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(caddr_t));
}

void 
scash_distribute_size(size_t*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(size_t));
}

void 
scash_distribute_char(char*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(char));
}
    
void 
scash_distribute_int(int*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(int));
}

void 
scash_distribute_uint(unsigned*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(uint));
}

void 
scash_distribute_double(double*  gvp) 
{
    scash_distribute_nbytes((caddr_t)gvp,sizeof(double));
}


#define  XX_WORD   (8)
#define  XX_WORD_MASK   (~(XX_WORD - 1))
#define  XX_WORD_ROUNDUP(X)   ((X + XX_WORD -1) & XX_WORD_MASK)

size_t  total_size = 0;

caddr_t
scash_alloc(size_t  s) 
{
    char *old_gp;

    if(scash_debug) scash_prf("scash_alloc s=0x%x",s);
    s = XX_WORD_ROUNDUP(s);

    if (scash.ps==1) {
	return (caddr_t)malloc(s);
    }

    old_gp = scash.dsm_addr + total_size;
    total_size += s;

    scash_distribute_size(&total_size);

    if (total_size >= (size_t)SCASH_MAX_PAGE * (size_t)SCASH_PAGE_SIZE)
	scash_prf("Total size is larger than MAX DSM area");
    return  old_gp;
}

void
scash_page_align() 
{
    total_size = SCASH_PAGE_ROUNDUP(total_size);
    scash_distribute_size(&total_size);
}

caddr_t
scash_alloc_page_align(size_t  s) 
{
    total_size = SCASH_PAGE_ROUNDUP(total_size);
    return scash_alloc(s);
}

 void
_scash_home_fo(int      new_home,
	       caddr_t  pagep, 
	       int      pages,
	       unsigned  char*  cur_home,
	       int      barrier) 
{
    u_int  page;
    u_int  current_home;

    pagep = (caddr_t)(SCASH_TRUNC((unsigned long)pagep, SCASH_PAGE_SIZE));

    while(pages--) {
	page = scash_addr2page(pagep);
	current_home = *cur_home++;

	if (new_home != current_home) {
	    if (scash.proc==0) {
	        scash_page2home(page) = new_home;
	    }

	    if (new_home/scash.smp != current_home/scash.smp) {
	        if (new_home/scash.smp == scash.pe/scash.smp) {
		    if (scash.proc==0) {
		        page_dir[page].attr = SCASH_PAGE_RO;
			bzero((void*)page_dir[page].cs, 
			      CS_LEN * sizeof(u_int));
		    }
		    scash_mprot_ro(scash_page2addr(page));	
		} 
		if (current_home/scash.smp == scash.pe/scash.smp) {
		    if (scash.proc==0) {
		        page_dir[page].attr = SCASH_PAGE_UNMAPPED;
		    }
		    scash_mprot_none(scash_page2addr(page));
		}
	    }
	}
	pagep += SCASH_PAGE_SIZE;
    }

    if (barrier) {
        scash_barrier_a2a(1);
    }
}

void
scash_home_scash(int      new_home,
		 caddr_t  pagep, 
		 int      pages)
{
    u_int page;
    int   current_home;

    pagep = (caddr_t)(SCASH_TRUNC((unsigned long)pagep, SCASH_PAGE_SIZE));

    while(pages--) {
	page = scash_addr2page(pagep);
	current_home = scash_page2home(page);
	if (scash.proc==0) {
	    scash_page2home(page) = new_home;
	}
	if (new_home/scash.smp == scash.pe/scash.smp) {
	    if (scash.proc==0) {
	        page_dir[page].attr = SCASH_PAGE_RO;
		bzero((void*)page_dir[page].cs, CS_LEN * sizeof(u_int));
	    }
	    scash_mprot_ro(scash_page2addr(page));	
	} else {
	    if (scash.proc==0) {
	        page_dir[page].attr = SCASH_PAGE_UNMAPPED;
	    }
	    scash_mprot_none(scash_page2addr(page));
	}
	pagep += SCASH_PAGE_SIZE;
    }
}

 void
_scash_home(int      new_home,
	    caddr_t  pagep, 
	    int      pages,
	    int      barrier) 
{
    u_int page;
    int   current_home;

    pagep = (caddr_t)(SCASH_TRUNC((unsigned long)pagep, SCASH_PAGE_SIZE));

    while(pages--) {
	page = scash_addr2page(pagep);
	current_home = scash_page2home(page);
	if (scash.proc==0) {
	    scash_page2home(page) = new_home;
	}
	if (new_home/scash.smp == scash.pe/scash.smp) {
	    if (scash.proc==0) {
	        page_dir[page].attr = SCASH_PAGE_RO;
		bzero((void*)page_dir[page].cs, CS_LEN * sizeof(u_int));
	    }
	    scash_mprot_ro(scash_page2addr(page));	
	} else {
	    if (scash.proc==0) {
	        page_dir[page].attr = SCASH_PAGE_UNMAPPED;
	    }
	    scash_mprot_none(scash_page2addr(page));
	}
	pagep += SCASH_PAGE_SIZE;
    }

    if (barrier) {
        scash_barrier_a2a(1);
    }
}

void
scash_fhome(caddr_t  fp)
{
    int    hf, i;
    int    page, page_cnt, read_page = 1024;
    char   base_buf[1024];

    hf = open(fp, O_RDONLY,  0644);
    read(hf, (void*)&page_cnt, sizeof(int));

    for (page = 0 ; page < page_cnt ; page+= 1024) {
        if (page_cnt - page < read_page)
	    read_page = page_cnt - page;
	read(hf, (void*)base_buf, read_page);

	for (i=0 ; i < read_page ; i++) {
	    scash_home(base_buf[i], 
		       scash.dsm_addr + (page + i) * SCASH_PAGE_SIZE, 1);
	}
    }
    close(hf);
}

void
scash_home(int      new_home,
	   caddr_t  pagep, 
	   int      pages)
{
    int   i;
    msg_body_t  msg_body;

    if (scash.ps==1) return;

    scash.home_ok = 1;
    for (i = 0 ; i < scash.ps ; i++) {
      if (i != scash.pe) {
	msg_body.scash_scash_home.home = new_home;
	msg_body.scash_scash_home.pagep = pagep;
	msg_body.scash_scash_home.pagen = pages;
	scash_message_send2(SCASH_MSG_SCASH_HOME,
			    i,
			    sizeof(scash_msg_scash_home_t), 
			    &msg_body);
      }
    }

    SCASH_POLL();
    scash_home_scash(new_home, pagep, pages);

    while(scash.home_ok != scash.ps)
      SCASH_POLL();

    SCASH_POLL();
}

/* assign home for Omni */
void
scash_home_fo(int      new_home,
	      caddr_t  pagep, 
	      int      pages)
{
    int   i;
    msg_body_t  msg_body;
    unsigned  char  cur_home[SCASH_MSG_HOME_LEN];
    int  ch_pages;
    int  page;

    pagep = (caddr_t)(SCASH_TRUNC((unsigned long)pagep, SCASH_PAGE_SIZE));

    while (pages){
        ch_pages = pages;
        if ( ch_pages > SCASH_MSG_HOME_LEN)
	    ch_pages = SCASH_MSG_HOME_LEN;

	page = scash_addr2page(pagep);
	for (i = 0 ; i < ch_pages ; i++) {
	    cur_home[i] = scash_page2home(page + i);
	}

        for (i = 0 ; i < scash.ps ; i++) {
	    if (i != scash.pe) {
	        msg_body.scash_omni_home.home = new_home;
		msg_body.scash_omni_home.pagep = pagep;
		msg_body.scash_omni_home.pagen = ch_pages;
		bcopy(cur_home, msg_body.scash_omni_home.cur_home, 
		      SCASH_MSG_HOME_LEN);
		scash_message_send(SCASH_MSG_OMNI_HOME,
				   i,
				   sizeof(scash_msg_omni_home_t), 
				   &msg_body);
	    }
	}
	_scash_home_fo(new_home, pagep, ch_pages, cur_home, 0);
	pages -= ch_pages;
	pagep += ch_pages * SCASH_PAGE_SIZE;
    }
}

void  
scash_barrier_init(int left, int right, int me)
{
    int  i, j;
    extern  int  scash_barrier_exec_recv[4][SCASH_MAX_PE];

    for (i = 0 ; i < 4 ; i++) {
	for (j = 0 ; j < SCASH_MAX_PE ; j++) {
	    scash_barrier_exec_recv[i][j] = -1;
	}
    }
}

char *scash_alloc_omni_shmem(size_t  alloc_size)
{
    if (alloc_size > SCASH_OMNI_SHMEM_SIZE) 
	return NULL;
    
    return  scash.update_map_addr + SCASH_UPMAP_SIZE;
}

void
scash_senddone_all()
{
    while (!OMSC_IS_SEND_DONE())
	SCASH_POLL();
}

int  scash_exit(int  status) 
{  
  extern void scash_tlog_finialize();
  if(scash_tlog_flag) scash_tlog_finialize();
  return OMSC_TERMINATE(status);
}

/* 
 * send & recv raw data
 */
void scash_send_raw_data(int dst,int nbytes,void *datap)
{
    scash_message_send(SCASH_MSG_RAW_DATA,dst,nbytes,datap);
}

void scash_recv_raw_data(int nbytes,void *datap)
{
    int ret;
    char *msg;
    scash_msg_header_t *header;

    while(1){
	ret = scash_message_recv((scash_msg_t**)&msg);
	if(ret == ENOBUFS) continue;
	header = (scash_msg_header_t *)msg;
	if(header->id != SCASH_MSG_RAW_DATA){
	  scash_prf("scash_recv_raw_data: bad msg");
	  exit(1);
	}
	msg += sizeof(scash_msg_header_t);
	if(nbytes != 0) bcopy(msg,datap,nbytes);
	OMSC_RELEASE_BUF();
	return;
    }
}

/* EOF */


