/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
#ifndef __scash_h
#define __scash_h

#include <sys/types.h>
#include <signal.h>

#define TRUE 1
#define FALSE 0

#if defined(__alpha__) || defined(__ia64__)
typedef  unsigned long u_int64;
typedef  long int64;
#else
typedef  unsigned long long u_int64;
typedef  long long int64;
#endif

#include "scash_omsc.h"

#include "scash_queue.h"
#include "scash_stats.h"

/*
 *  Page Size depends the CPU's MMU. 
 */
#ifdef i386
#define  SCASH_PAGE_SIZE  (4096)
#endif
#ifdef __alpha__
#define  SCASH_PAGE_SIZE  (8192)
#endif
#ifdef __ia64__
#define  SCASH_PAGE_SIZE  (16384)
#endif
#ifdef __x86_64__
#define  SCASH_PAGE_SIZE  (4096)
#endif

#define  SCASH_WORD_SIZE  (4)
#define  SCASH_SUCCESS  0
#define  SCASH_MAX_PE   (64)
#ifdef __ia64__
#define  SCASH_MAX_PAGE (160 * 1024)
#else
#define  SCASH_MAX_PAGE (128 * 1024)
#endif /* __ia64__ */
#define  SCASH_MAX_LOCK (256)

#define  SCASH_LOCK_BUSY (0x00)
#define  SCASH_LOCK_OK   (0x01)
#define  SCASH_LOCK_OFF  (0xFFFF)

/* barrier structure */
typedef  struct  scash_barrier_s {
    int    master;
    int    id;
    int    cnt;
    int    upper;
    int    gap;
    int    f;
} scash_barrier_t;

/* lock structure */
typedef  struct  scash_lock_s {
    int    id;
    u_int  head;
    u_int  q_cnt;
    u_char *q;
} scash_lock_t;

typedef struct scash_s {
    int    pid;   /* Process ID */
    int    ps;    /* Partition Size */
    int    pe;    /* PE (Node) Number */
    int    smp;   /* Number of Processors in a SMP box */
    int    proc;  /* Proc id in SMP box */
    int    box;	  /* id for SMP box */
    int    mtu;   /* MTU of SCASH Network */
    u_int  time_stamp; 

    int     dsm_size;
    int     dsm_pages;
    caddr_t dsm_addr;
    caddr_t twin_addr;
    caddr_t update_map_addr;
    caddr_t shadow_addr;

    addrHandle shadow_addr_handle[SCASH_MAX_PE];
    addrHandle page_map_addr_handle[SCASH_MAX_PE];

    caddr_t diff_rma_addr;
    int     diff_rma_size;
    u_int   diff_receive_addr[SCASH_MAX_PE];
    u_int   diff_dest_addr[SCASH_MAX_PE];

    volatile int     home_ok;

    u_int   seq;
    FILE*   logf;
    FILE*   tracef;
    int     map_fd;

    int             stat_cnt;
    scash_stats_t*  stats;

    scash_barrier_t  barrier;
    volatile int     lock_stat;
    int              lock_cnt;
    int              lock_hint;
    scash_lock_t     lock[SCASH_MAX_LOCK];
    int  debug_level;
    u_int    msg_seq;
    volatile int    update_cnt;

    struct   sigaltstack sig_stack;
    struct   sigaction   sig_act;
    sigset_t sig_mask;
    sigset_t sig_blockio;

    u_int    last_seq[SCASH_MAX_PE];
    int      fault_trace;
    char     map_file[512];
    char     page_file[512];
    volatile u_int flushed_page[SCASH_MAX_PAGE/32];
} scash_t;

typedef  struct  scash_omni_update_map_s {
  int      node;
  caddr_t  update_addr;
  int      update_size;
} scash_omni_update_map;

struct scash_msg_refresh_s;

extern int scash_debug; /* debug flag */
extern int scash_tlog_flag; /* for loging */

/*
 * Public Interfaces
 */

#ifdef  __cplusplus
extern "C" {
#endif
    void  scash_initialize(int, char**, int);
    void scash_init(int);   /* For mpc++ runtime compatibility */
    int  scash_init_old(int, caddr_t*);

    caddr_t scash_alloc(size_t);
    caddr_t scash_alloc_page_align(size_t);
    caddr_t scash_alloc_omni_shmem(size_t);
    void  scash_page_align();

    void  scash_barrier(int);
    void  scash_barrier_invalidate(int);
    void  scash_barrier_update(int);
    void  scash_barrier_dynamic_home(int);

    int   scash_lock();
    int   scash_try_lock();
    int   scash_unlock();

    void  scb_init(int, int, int);
    int  scb_exec(int);

    void   scash_update(caddr_t, int);
    void   scash_sync_update(caddr_t, int);
    int    scash_omni_update(char*, int);
    int    scash_omni_update2(scash_omni_update_map*, int);

    void   scash_update_word(caddr_t);
    void   scash_sync_update_word(caddr_t);

    void  scash_fatal(char*, int);
    void scash_prf(char *msg,...);

    void  scash_home(int, caddr_t, int);
    void  scash_home_fo(int, caddr_t, int);  /* scash_home special for omni runtime */

    void  scash_stats_init();
    void  scash_stats_print();

    int  scash_flush(caddr_t, int);
    int  scash_flush_word(void*);
    int  scash_flush_acq(void*);
    void scash_invalidate_all();
    void scash_omp_flush(char*, int);

    int  scash_refresh(caddr_t, int);
    int  scash_refresh_word(void*);

    void  scash_poll();
    void  scash_wait();

    void  scash_profile_on();
    void  scash_profile_off();
    void  scash_profile_close();

    void  scash_fhome(caddr_t);
    void  scash_fprot(caddr_t);

    double  scash_dtime();
    u_int64  scash_rdtsc();
#ifdef  __cplusplus
}
#endif

#ifdef __cplusplus
extern void scash_distribute(caddr_t*);
extern void scash_distribute(int*);
extern void scash_distribute(unsigned*);
extern void scash_distribute(float*);
extern void scash_distribute(double*);
#endif

extern void scash_distribute_caddr(caddr_t*);
extern void scash_distribute_char(char*);
extern void scash_distribute_int(int*);
extern void scash_distribute_uint(unsigned*);

extern void scash_distribute_var(char*, int) ;
extern void scash_distribute_diff_area();
extern void scash_page_mprot_correct(int);
extern void mpc_stats_print();

extern void scash_barrier_page_update_new(int);

extern  scash_t  scash;
extern  scash_stats_t  scash_stats[];
extern  unsigned  scash_option;

static inline int  scash_ps() {  return scash.ps; }
static inline int  scash_pe() {  return scash.pe; }
static inline int  scash_smp() {  return scash.smp; }
static inline int  scash_proc() {  return scash.proc; }
static inline int  scash_box() {  return scash.box; }

int scash_exit(int status);

#endif  /* __scash_h*/



