/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_alloc.c,v 1.5 2004-08-24 17:25:43 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

#ifdef USE_PM
extern int score_ckpt_force_checkpoint(int flag, int fd);
#endif

/***
   +----------+   0x00000000
   |          |
   | Text     |
   |          |
   +----------+
   |          |
   | Data     |
   |          |
   +----------+
   |          |
   |          |
   +----------+
   |          |
   | Shared   |
   | Lib      |
   |          |
   +----------+
   |          |
   |          |
   +----------+  SCASH_DEFAULT_ADDR (1<<30 + 1<<27)
   |          |
   |  Shadow  |
   |          |
   +----------+
   |          |  Protect NONE(1 Page)
   +----------+
   |          |
   |  DSM     |
   |          |
   +----------+
   |          |  Protect NONE(1 Page)
   +----------+
   |          |
   |  Twin    |
   |          |
   +----------+
   | UdateMAP | MAX_PE * MAX_PAGE
   +----------+
   | OmniShmem| 
   +----------+
   |          |  Protect NONE(1 Page)
   +----------+
   | DIFF RMA |
   | Buffer   |  (WE NEVER ALLOCATE THIS BUFFER NOW!!)
   +----------+
   |          |
   |  Stack   |
   |          |
   +----------+
***/

static caddr_t
scash_mmap(caddr_t  map_addr,
	   size_t map_size,
	   off_t map_offset)
{
    caddr_t  mapp;

#ifdef  __linux__
    mapp = (caddr_t)mmap(map_addr, map_size, 
			 PROT_READ | PROT_WRITE, 
			 MAP_FILE | MAP_SHARED | MAP_FIXED, 
			 scash.map_fd, map_offset);

#else
    mapp = mmap(map_addr, map_size, 
		PROT_READ|PROT_WRITE, 
		MAP_FILE|MAP_SHARED, 
		scash.map_fd, map_offset);
#endif
    printf("scash_mmap size=%p, map_addr=%p, mapp=%p\n",
	   map_size, map_addr, mapp);

    return mapp;
}

int
scash_shmem_alloc(caddr_t* map_addr, u_int  map_size)
{
    int   i;
    caddr_t  buff;
    unsigned long  file_map_size;
    caddr_t  mapp;

#ifdef USE_PM
    extern  int sc_create_temporary_file2(char *, int *, char*);
    extern  int sc_create_temporary_file(char *, int *);
#endif
    
    if(scash_debug) scash_prf("scash_shmem_alloc ...");

    if (*map_addr == NULL)
	*map_addr = (caddr_t)SCASH_DEFAULT_ADDR;

    scash.dsm_size = SCASH_ROUNDUP(map_size, SCASH_PAGE_SIZE);
    scash.dsm_pages = scash.dsm_size/(SCASH_PAGE_SIZE);

    /* decide address */
    scash.shadow_addr = 
	(caddr_t) SCASH_TRUNC((unsigned long)(*map_addr), SCASH_PAGE_SIZE);
    scash.update_map_addr = scash.shadow_addr + scash.dsm_size;
    scash.dsm_addr = scash.shadow_addr + scash.dsm_size + SCASH_PAGE_SIZE 
                   + SCASH_UPMAP_SIZE + SCASH_OMNI_SHMEM_SIZE;
    scash.twin_addr = scash.dsm_addr + scash.dsm_size + SCASH_PAGE_SIZE;

    file_map_size = scash.dsm_size + 
                    SCASH_PAGE_ROUNDUP(SCASH_UPMAP_SIZE) + 
                    scash.dsm_size + 
                    SCASH_PAGE_ROUNDUP((sizeof(page_map_t) * SCASH_MAX_PAGE)) +
                    SCASH_OMNI_SHMEM_SIZE;

    if ( (scash.dsm_size/SCASH_PAGE_SIZE) > SCASH_MAX_PAGE )
        scash_fatal("scash_shmem_shared:shared memory size too large", 1); 

    
    if (scash.proc==0) {
	/* if proc 0 in SMP box, allocate map file */
        if ( OMSC_TMP_FILE( SCASH_MAP_FILE, &scash.map_fd, scash.map_file) 
	     != 0 )
	    scash_fatal("scash_shmem_alloc:can not create map file", 1); 

	/* initialize file */
        buff = (caddr_t)malloc(SCASH_PAGE_SIZE);
        if (buff == NULL)
            scash_fatal("scash_shmem_alloc:can not allocate map file buffer", 
			1); 
        bzero(buff, SCASH_PAGE_SIZE);
        for (i = 0 ; i <= file_map_size ; i += SCASH_PAGE_SIZE) {
            if (write(scash.map_fd, buff, SCASH_PAGE_SIZE) < SCASH_PAGE_SIZE) {
	        scash_fatal("scash_shmem_alloc:can not write map file", 1); 
	    }
        }
        free(buff);

	for (i = 1 ; i < scash.smp ; i++) {
	  scash_message_send(SCASH_MSG_MAP_FILE, scash.pe + i,
			     512, scash.map_file);
	}
    } else {
        while (strlen(scash.map_file)==0)
	    SCASH_POLL();

        scash.map_fd = open(scash.map_file, O_RDWR);
	if (scash.map_fd < 0)
	    scash_fatal("scash_shmem_shared:can not open map file", 1);
    }

#ifdef ENABLE_CHECKPOINT
    score_ckpt_force_checkpoint(1, scash.map_fd);
#endif

    /* allocate shadow area */
    mapp = scash_mmap(scash.shadow_addr, scash.dsm_size, 0);
    if (mapp != scash.shadow_addr) {
	scash_prf("SCASH shadow addr incorrect %p:%p\n", 
		  scash.shadow_addr,mapp);
    }

    scash_memory_lock_init();

    /* allocate UPMAP OMNI_SHMEM */
    mapp = scash_mmap(scash.update_map_addr,
		      SCASH_UPMAP_SIZE + SCASH_OMNI_SHMEM_SIZE, 
		      scash.dsm_size);

    if (mapp != scash.update_map_addr) {
        scash_prf("SCASH Update Map addr incorrect %p:%p\n",
		  scash.update_map_addr, mapp);
    }

    if (scash.proc==0)
        bzero(scash.update_map_addr, SCASH_UPMAP_SIZE);

    /* allocate DSM area */
    *map_addr = scash_mmap(scash.dsm_addr, scash.dsm_size, 0);

    if (*map_addr != scash.dsm_addr) {
      scash_prf("SCASH dsm map addr incorrect %p:%p\n",
		scash.dsm_addr, *map_addr);
    }

    if (scash.proc==0)
        bzero(*map_addr, scash.dsm_size);

    /* allocate TWIN area */
    mapp = scash_mmap(scash.twin_addr, scash.dsm_size, 
		      scash.dsm_size + SCASH_UPMAP_SIZE + 
		      SCASH_OMNI_SHMEM_SIZE + SCASH_PAGE_SIZE);

    if (mapp != scash.twin_addr) {
      scash_prf("SCASH twin map addr incorrect %p:%p\n", 
		scash.twin_addr, mapp);
    }

    if (scash.proc==0)
        bzero(scash.twin_addr, scash.dsm_size);

#ifdef ENABLE_CHECKPOINT
    score_ckpt_force_checkpoint(0, scash.map_fd);
#endif

    if (scash.proc==0)
        bzero(scash.shadow_addr, scash.dsm_size);

    for (i = 0 ; i < scash.dsm_pages ; i++) {
        scash_page2base(i) = scash_page2home(i) = i % scash.ps;
        scash_page2ocount(i) = 0;
	if (scash_page2home(i) / scash.smp == scash.box) {
	    if (scash.proc==0) {
	        page_dir[i].attr = SCASH_PAGE_RO; 
	    }
	    scash_mprot_ro(scash.dsm_addr + (SCASH_PAGE_SIZE * i));
	} else {
	    scash_mprot_none(scash.dsm_addr + (SCASH_PAGE_SIZE * i));
	}
    }

    if(scash_debug) scash_prf("scash_shmem_alloc: check barrier ...");

    /* Hack Hack Hack and Check !! */
    OMSC_BARRIER(); /*  scash_barrier_a2a(1); */
    for (i = 0 ; i < scash.dsm_pages ; i++) {
      if (scash_page2home(i) != i % scash.ps)
	  scash_prf(" home[%d] = %d(%d)\n",
		    i, scash_page2home(i), scash.ps);
      if (scash_page2base(i) != i % scash.ps)
	  scash_prf(" base[%d] = %d(%d)\n", 
		    i, scash_page2base(i), scash.ps);
    }

    if(scash_debug) scash_prf("scash_shmem_alloc ... end");

    return SCASH_SUCCESS;
}

int
scash_dealloc()
{
    scash_fatal("scash_dealloc", 1);
    return EPERM;
}
