/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_barrier.c,v 1.5 2004-08-24 18:58:50 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

unsigned  scash_barrier_exec_cnt = 0;
int       scash_barrier_exec_recv[4][SCASH_MAX_PE];

unsigned  short  barrier_cnt;

#define  SCASH_BARRIER_A2A

void
scash_barrier_tree_node(int  ps,
			int  pe,
			int  *gap,
			int  *upper_node)
{
    int   node;

    *upper_node = 0;
    node = *gap = ps/2;
    
    while (*gap) {
	*gap = *gap/2;
	if (node == pe)
	    return;

	*upper_node = node;
	if (node > pe)
	    node -= *gap;
	else
	    node += *gap;
    }
}

static int  
barrier_check_cnt(int barrier_id)
{
    int  i, check_cnt = 0;

    for (i=0; i < scash.ps ; i++) {
	if (scash_barrier_exec_recv[scash_barrier_exec_cnt % 4][i] 
	    == barrier_id) {
	    check_cnt++;
	} else 
	    if (scash_barrier_exec_recv[scash_barrier_exec_cnt % 4][i] != -1) {
		scash_prf("scash invalid barrier id %d(%d)!!", 
			  scash_barrier_exec_recv[scash_barrier_exec_cnt % 4][i], barrier_id); 
	}
    }
    return check_cnt;
}

/*
 *  All to All barrier 
 */
void
scash_barrier_a2a(int  barrier_id)
{
    scash_msg_t*  msg_buff;
    int    i, error;
    extern void scash_senddone_all();
    scash_msg_barrier_a2a_t  barrier_msg;

    if(scash_debug) scash_prf("scash_barrier_a2a id=%d ...",barrier_id);

    while (!OMSC_IS_SEND_DONE()){
        while (ENOBUFS != scash_message_recv(&msg_buff)) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }
    if(scash_debug) scash_prf("scash_barrier_a2a send ...");

    scash_barrier_exec_cnt++;

    scash_barrier_exec_recv[scash_barrier_exec_cnt % 4][scash.pe] = barrier_id;

    barrier_msg.barrier_cnt = scash_barrier_exec_cnt;
    barrier_msg.barrier_id  = barrier_id;

    scash_message_broadcast(SCASH_MSG_BARRIER_OK, sizeof(barrier_msg), &barrier_msg);

    while (barrier_check_cnt(barrier_id) != scash.ps) {
	while ((error = scash_message_recv(&msg_buff)) == ENOBUFS)
	    ;
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
	/* if(scash_debug) scash_prf("scash_barrier_a2a count=%d ...",
				  barrier_check_cnt(barrier_id)); */
    }

    for (i=0; i < scash.ps ; i++) {
        scash_barrier_exec_recv[scash_barrier_exec_cnt % 4][i] = -1;
    }

    while (ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }

    if(scash_debug) scash_prf("scash_barrier_a2a ... end");
}

/*
 * 1(Master) to all Others(Client) barrier sync
 */
void
scash_barrier_master(int  barrier_id)
{
    scash_msg_t*  msg_buff;
    u_int  msg_length;
    int  error;

    scash.barrier.id = barrier_id;
    scash.barrier.cnt++;

    for (;;) {
	if (scash.barrier.cnt == scash.ps) {
	    scash.barrier.cnt = 0;
	    scash_message_broadcast(SCASH_MSG_BARRIER_OK, 
				    sizeof(int), &barrier_id);
	    break;
	} 
	while ((error = scash_message_recv(&msg_buff)) == ENOBUFS)
	    ;

	if (msg_buff->header.id == SCASH_MSG_BARRIER_REQ) {
	    OMSC_RELEASE_BUF();
	    scash.barrier.cnt++;
	} else if (msg_buff->header.id == SCASH_MSG_BARRIER_OK) {
	    exit(1);
	} else { 
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }

    while (ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
}

void
scash_barrier_client(int  barrier_id)
{
    scash_msg_t*  msg_buff;
    u_int  msg_length;
    int  error;

    if (scash.barrier.id != -1) {
	scash_prf("scash_barrier: Another waiting barrier %d",
		  scash.barrier.id);
	scash_fatal("scash_barrier:Another waiting barrier", 1);
    }

    scash.barrier.id = barrier_id;
    scash_message_send(SCASH_MSG_BARRIER_REQ, scash.barrier.master, sizeof(int), &barrier_id);

    for (;;) {
	while ( (error = scash_message_recv(&msg_buff)) == ENOBUFS)
	    ;

	if (msg_buff->header.id == SCASH_MSG_BARRIER_OK) {
	    scash.barrier.cnt = 0;
	    scash.barrier.id = -1;
	    OMSC_RELEASE_BUF();
	    break;
	} else {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }

    while (ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
}

void
scash_barrier_tree(int  barrier_id)
{
    scash_msg_t*  msg_buff;
    u_int  msg_length;
    int  error;
    int  send_f = 1; 

    if (barrier_id < 0) {
	scash_fatal("scash_barrier:illegal barrier_id", 1);
    }
    if (scash.ps == 1)
	return;

    while(!OMSC_IS_SEND_DONE()) /* */;

    if (scash.pe == 0) {
	scash.barrier.id = barrier_id;
	scash_message_send(SCASH_MSG_BARRIER_REQ, scash.ps/2, sizeof(int), &barrier_id);
    }

    for (;;) {
	if (scash.pe == 0) {
	    if (scash.barrier.cnt==1) {
		scash.barrier.cnt = 0;
		goto endloop;
	    }
	} else {
	    if (scash.barrier.cnt==1) {
		if (scash.pe % 2) {
		    scash_message_send(SCASH_MSG_BARRIER_REQ, scash.barrier.upper, sizeof(int), &barrier_id);
		    scash.barrier.cnt = 0;
		    goto endloop;
		} else {
		    if (send_f) {
			scash_message_send(SCASH_MSG_BARRIER_REQ, 
					   scash.pe - scash.barrier.gap, sizeof(int), &barrier_id);
			scash_message_send(SCASH_MSG_BARRIER_REQ, 
					   scash.pe + scash.barrier.gap, sizeof(int), &barrier_id);
			send_f = 0;
		    }
		}
	    } else if (scash.barrier.cnt==3){
		scash_message_send(SCASH_MSG_BARRIER_REQ, 
				   scash.barrier.upper, sizeof(int), &barrier_id);
		scash.barrier.cnt = 0 ;
		goto endloop;
	    }
	}
	while ( (error = scash_message_recv(&msg_buff)) == ENOBUFS)
	    ;
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }

endloop:
    while (ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
}

void
scash_barrier_sync(int  barrier_id, int num)
{
    if (barrier_id < 0) {
	scash_fatal("scash_barrier:illegal barrier_id", 1);
    }

    if (scash.ps == 1)
	return;

    if(scash_tlog_flag) tlog_BSYNC_IN(num);

#ifdef SCASH_BARRIER_A2A
    scash_barrier_a2a(barrier_id);
#else
#ifdef SCASH_BARRIER_1TOALL
    while(!OMSC_IS_SEND_DONE()) /* */;

    if (scash.barrier.master == scash.pe)
	scash_barrier_master(barrier_id);
    else
	scash_barrier_client(barrier_id);
#else
#ifdef SCASH_BARRIER_TREE
    while(!OMSC_IS_SEND_DONE()) /* */;
    scash_barrier_tree(barrier_id);
#endif  /* SCASH_BARRIER_TREE */
#endif  /* SCASH_BARRIER_1TOALL */
#endif  /* SCASH_BARRIER_A2A */
    if(scash_tlog_flag) tlog_BSYNC_OUT(num);
}

static void 
scash_invalidate_home()
{
    int   page;
    int   new_home;
    scash_msg_new_home_base_t  home_base;

    while (update_page_head) {
	page = update_page_head - page_dir;
	new_home = scash_page2home(page);

	if ((page_dir[page].attr & SCASH_PAGE_HOME_WRITE) == 0) {
	    if (page_dir[page].attr & SCASH_PAGE_BASE_WRITE) {
	        new_home = scash_page2base(page);
	    } else if (page_dir[page].diff[1] > 0) {
	        new_home = page_dir[page].diff[0];
	    }
	}

	if (scash.pe != new_home) {
	    OMSC_WRITE(new_home, 
		       scash.shadow_addr_handle[new_home] + 
		       page * SCASH_PAGE_SIZE, 
		       scash.shadow_addr_handle[scash.pe] + 
		       page * SCASH_PAGE_SIZE, 
		       SCASH_PAGE_SIZE);

	    scash_message_send(SCASH_MSG_HOME, new_home, sizeof(int), &page);

	    if (new_home != scash_page2base(page) && 
		scash.pe != scash_page2base(page)) {
	        home_base.page = page;
	        home_base.home = new_home;

		scash_message_send(SCASH_MSG_NEW_HOME_BASE, 
				   scash_page2base(page), 
				   sizeof(scash_msg_new_home_base_t), 
				   (void*)&home_base);
	    }
	    scash_page2home(page) = new_home;
	    scash_page2ocount(page) = barrier_cnt + 1;
	    SCASH_PAGE_SET_UNMAPPED(page_dir[page].attr);
	    scash_mprot_none(scash_page2addr(page));
	    while(!OMSC_IS_SEND_DONE()) /* */;
	}
	scash_page_invalidate(page, new_home);
	page_dir[page].diff[0] = page_dir[page].diff[1] = 0;
	SCASH_PAGE_CLEAR_BASE_WRITE(page_dir[page].attr);
	SCASH_PAGE_CLEAR_HOME_WRITE(page_dir[page].attr);
	SCASH_PAGE_CLEAR_UPDATE(page_dir[page].attr);
	update_page_head = update_page_head -> update_next;
	page_dir[page].update_next = NULL;
	SCASH_POLL();
    }
    update_page_cnt = 0;
}

void 
scash_barrier_invalidate_home_prot()
{
    int  barrier_id = 1;

    scash_flush_all_home();
    scash_barrier_sync(barrier_id,1);
    scash_invalidate_home();
    scash_barrier_sync(barrier_id,1);
}

/*
 * flush page on invalidate protocol
 */
void
scash_barrier_invalidate(int  barrier_id)
{
#ifdef  SCASH_STATS
    u_int64  clock_cnt[5];
#endif
    extern void  scash_flush_page_invalidate();

    SCASH_RDTSC(clock_cnt[0]);

#ifdef not
    do {
        SCASH_POLL();
    } while(scash.update_cnt);
#endif

    SCASH_RDTSC(clock_cnt[1]);
    scash_barrier_sync(barrier_id,1);

    scash.stats->barrier_cnt++;

    SCASH_RDTSC(clock_cnt[2]);
    scash_flush_all();

    scash_barrier_sync(barrier_id,2);
    scash_flush_page_invalidate();

    scash_barrier_sync(barrier_id,3);

    SCASH_RDTSC(clock_cnt[3]);
    SCASH_POLL();

    dirty_page->inbox_lock[scash.proc] = 0;

#ifdef SCASH_STATS
    clock_cnt[4] = scash_rdtsc();
    scash.stats->barrier_clock[0] += (u_int64)(clock_cnt[4] - clock_cnt[0])/(unsigned long long)(CPU_CLOCK);
#endif  /* SCASH_STATS */

    update_page_cnt = 0;
    barrier_cnt++;
}

void
scash_barrier(int  barrier_id)
{
    if (scash.ps==1) return;	/* single processor case */

    if(scash_tlog_flag) tlog_BARRIER_IN();

    switch (scash_opt_barrier()) {
    case SCASH_OPT_BARRIER_INVALIDATE:
        scash_barrier_invalidate(barrier_id);
	break;

#ifdef not_used  /* we are currently support only invalidate protocol */
    case SCASH_OPT_BARRIER_UPDATE:
        scash_barrier_update(barrier_id);
	break;
    case SCASH_OPT_BARRIER_DYNAMIC_HOME:
        scash_barrier_dynamic_home(barrier_id);
	break;
    case SCASH_OPT_BARRIER_HYBRID:
        scash_barrier_hybrid(barrier_id);
	break;
#else
    default:
	scash_fatal("scash_barrier: only invalidate mode is supported", 1);
#endif	
    }

#ifdef ENABLE_CHECKPOINT
    if (score_ckpt_leave_uncheckpointable() != 0) {
      /*
       * scash_prf("score_ckpt_leave_uncheckpointable() error.");
      */
    }
#endif

    scash_barrier_sync(barrier_id,4);

#ifdef ENABLE_CHECKPOINT
    if (score_ckpt_enter_uncheckpointable() != 0) {
	scash_prf("score_ckpt_enter_uncheckpointable() error.");
    }
#endif

    if(scash_tlog_flag) tlog_BARRIER_OUT();
}




