/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_diff.c,v 1.2 2004-04-05 08:29:22 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

int
scash_diff_create(int   page,
		  diff_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_t  *blockp;
    int    diff_byte_size = sizeof(int);

#ifdef  SCASH_STATS
    int   diff_cnt = 0;
    scash.stats->diff_cnt++;
#endif

    pagep = (u_int*)(scash_page2shadowaddr(page));
    twinp = (u_int*)(scash_page2twinaddr(page));

    diff->length = 0;

    blockp = (diff_block_t*)&diff->block[0];
    blockp->header.pos = blockp->header.length = 0;

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
	if (*pagep != *twinp ) {
#ifdef  SCASH_STATS
	    diff_cnt++;
#endif
	    if (in_block) {
		blockp->data[ blockp->header.length++ ] = *pagep;
		diff_byte_size += sizeof(u_int);
	    } else {
		diff->length++;
		in_block = 1;
		blockp -> header.pos = i;
		blockp -> data[0] = *pagep;
		blockp -> header.length = 1;
		diff_byte_size += sizeof(diff_min_block_t);
	    }
	} else {
	    if (in_block) {
		u_int*   pp;
		in_block = 0;
		pp = &(blockp->data[ blockp->header.length ]);
		blockp = (diff_block_t*)pp;
	    } 
	}
	pagep++;
	twinp++;
    }
#ifdef SCASH_STATS 
    if (diff_cnt==0) {
        scash.stats->diff_size[0]++;
    } else if (diff_cnt <=4){
        scash.stats->diff_size[1]++;
    } else if (diff_cnt <=16){
        scash.stats->diff_size[2]++;
    } else if (diff_cnt <=64){
        scash.stats->diff_size[3]++;
    } else if (diff_cnt <=256){
        scash.stats->diff_size[4]++;
    } else if (diff_cnt <=512){
        scash.stats->diff_size[5]++;
    } else if (diff_cnt <=768){
        scash.stats->diff_size[6]++;
    } else {
        scash.stats->diff_size[7]++;
    }
#endif
    return  diff_byte_size;
}

int
scash_diff_copy_create(int   page,
		       diff_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_t  *blockp;
    int    diff_byte_size = sizeof(int);

    pagep = (u_int*)(scash_page2shadowaddr(page));
    twinp = (u_int*)(scash_page2twinaddr(page));

    diff->length = 0;

    blockp = (diff_block_t*)&diff->block[0];
    blockp->header.pos = blockp->header.length = 0;

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
	if (*pagep != *twinp ) {
	    if (in_block) {
		blockp->data[ blockp->header.length++ ] = *pagep;
		diff_byte_size += sizeof(u_int);
	    } else {
		diff->length++;
		in_block = 1;
		blockp -> header.pos = i;
		blockp -> data[0] = *pagep;
		blockp -> header.length = 1;
		diff_byte_size += sizeof(diff_min_block_t);
	    }
	    *twinp = *pagep;
	} else {
	    if (in_block) {
		u_int*   pp;
		in_block = 0;
		pp = &(blockp->data[ blockp->header.length ]);
		blockp = (diff_block_t*)pp;
	    } 
	}
	pagep++;
	twinp++;
    }

    return  diff_byte_size;
}

int
scash_diff_rmw(int  page)
{
    int     i;
    u_int   *shadowp;
    u_int   *twinp;
    caddr_t  diffp = NULL;
    int      in_diff = 0;
    u_int    diff_offset;
    int      node = scash_page2home(page);
    int      diff_size = 0;

    shadowp = (u_int*)(scash_page2shadowaddr(page));

    twinp = (u_int*)(scash_page2twinaddr(page));

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
	if (*shadowp != *twinp) {
	    diff_size++;
	    if (in_diff == 0) {
		in_diff = 1;
		diffp = (caddr_t)shadowp;
	    }
	} else {
	    if (in_diff) {
		diff_offset = diffp - scash.shadow_addr;
		OMSC_WRITE( node,
			    scash.shadow_addr_handle[node] + diff_offset, 
			    scash.shadow_addr_handle[scash.pe] + diff_offset, 
			    (caddr_t)shadowp - diffp);
		in_diff = 0;
	    } 
	}
	shadowp++;
	twinp++;
    }

    if (in_diff) {
	diff_offset = diffp - scash.shadow_addr;
	OMSC_WRITE(node, 
		   scash.shadow_addr_handle[node] + diff_offset, 
		   scash.shadow_addr_handle[scash.pe] + diff_offset,
		   (caddr_t)shadowp - diffp);
    }
    while (!OMSC_IS_WRITE_DONE()) /* */;
    return diff_size;
}

int
scash_diff_create_range(int   page,
			int   start_offset,
			int   word_len,
			diff_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_t  *blockp;
    u_int    diff_byte_size = sizeof(int);
    int      diff_cnt;

    pagep = (u_int*)(scash_page2shadowaddr(page)) + start_offset;
    twinp = (u_int*)(scash_page2twinaddr(page)) + start_offset;

    diff->length = 0;

    blockp = (diff_block_t*)&diff->block[0];
    blockp->header.pos = blockp->header.length = 0;

    for (i = 0 ; i < word_len ; i++) {
	if (*pagep != *twinp ) {
	    diff_cnt++;

	    if (in_block) {
		blockp->data[ blockp->header.length++ ] = *pagep;
		diff_byte_size += sizeof(u_int);
	    } else {
		diff->length++;
		in_block = 1;
		blockp -> header.pos = i + start_offset;
		blockp -> header.length = 1;
		blockp -> data[0] = *pagep;
		diff_byte_size += sizeof(diff_min_block_t);
	    }
	} else {
	    if (in_block) {
		u_int*   pp;
		in_block = 0;
		pp = &(blockp->data[ blockp->header.length ]);
		blockp = (diff_block_t*)pp;
	    } 
	}
	pagep++;
	twinp++;
    }

#ifdef SCASH_STATS 
    if (diff_cnt==0) {
        scash.stats->diff_size[0]++;
    } else if (diff_cnt <=4){
        scash.stats->diff_size[1]++;
    } else if (diff_cnt <=16){
        scash.stats->diff_size[2]++;
    } else if (diff_cnt <=64){
        scash.stats->diff_size[3]++;
    } else if (diff_cnt <=256){
        scash.stats->diff_size[4]++;
    } else if (diff_cnt <=512){
        scash.stats->diff_size[5]++;
    } else if (diff_cnt <=768){
        scash.stats->diff_size[6]++;
    } else {
        scash.stats->diff_size[7]++;
    }
#endif
    return  diff_byte_size;
}

int
scash_diff_copy_create_range(int   page,
			     int   start_offset,
			     int   word_len,
			     diff_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_t  *blockp;
    u_int    diff_byte_size = sizeof(int);
    int      diff_cnt;

    pagep = (u_int*)(scash_page2shadowaddr(page)) + start_offset;
    twinp = (u_int*)(scash_page2twinaddr(page)) + start_offset;

    diff->length = 0;

    blockp = (diff_block_t*)&diff->block[0];
    blockp->header.pos = blockp->header.length = 0;

    for (i = 0 ; i < word_len ; i++) {
	if (*pagep != *twinp ) {
	    diff_cnt++;

	    if (in_block) {
		blockp->data[ blockp->header.length++ ] = *pagep;
		diff_byte_size += sizeof(u_int);
	    } else {
		diff->length++;
		in_block = 1;
		blockp -> header.pos = i + start_offset;
		blockp -> header.length = 1;
		blockp -> data[0] = *pagep;
		diff_byte_size += sizeof(diff_min_block_t);
	    }
	    *twinp = *pagep;
	} else {
	    if (in_block) {
		u_int*   pp;
		in_block = 0;
		pp = &(blockp->data[ blockp->header.length ]);
		blockp = (diff_block_t*)pp;
	    } 
	}
	pagep++;
	twinp++;
    }

    return  diff_byte_size;
}

void
scash_diff_index_create(int   page,
			int   max_diff_size,
			_s_diff_index_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_header_t  *blockp = NULL;

    pagep = (u_int*)(scash_page2shadowaddr(page));
    twinp = (u_int*)(scash_page2twinaddr(page));

    diff->diff_num = 0;

    blockp = &diff->db[0];
#ifdef __ia64__
    blockp->pos = blockp->length = 0;
#else
    blockp->pos = blockp->length = blockp->pad = 0;
#endif /* __ia64__ */

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
      if (*pagep != *twinp ) {
	    if (in_block) {
		blockp->length++;
		if (blockp->length * SCASH_WORD_SIZE >= max_diff_size) {
		    in_block = 0;
		    blockp++;
		}
	    } else {
		diff->diff_num++;
		in_block = 1;
		blockp -> pos = i;
		blockp -> length = 1;
	    }
	} else {
	    if (in_block) {
		in_block = 0;
		blockp++;
	    } 
	}
	pagep++;
	twinp++;
    }
}


void
scash_diff_index_create_copy(int   page,
			     int   max_diff_size,
			     _s_diff_index_t*  diff)
{
    int   i;
    u_int   *pagep;
    u_int   *twinp;
    int      in_block = 0;
    diff_block_header_t  *blockp = NULL;

    pagep = (u_int*)(scash_page2shadowaddr(page));
    twinp = (u_int*)(scash_page2twinaddr(page));

    diff->diff_num = 0;

    blockp = &diff->db[0];
#ifdef __ia64__
    blockp->pos = blockp->length = 0;
#else
    blockp->pos = blockp->length = blockp->pad = 0;
#endif /* */

    for (i = 0 ; i < SCASH_WORD_IN_PAGE ; i++) {
      if (*pagep != *twinp ) {
	    if (in_block) {
		blockp->length++;
		if (blockp->length * SCASH_WORD_SIZE >= max_diff_size) {
		    in_block = 0;
		    blockp++;
		}
	    } else {
		diff->diff_num++;
		in_block = 1;
		blockp -> pos = i;
		blockp -> length = 1;
	    }
	} else {
	    if (in_block) {
		in_block = 0;
		blockp++;
	    } 
	}
        *pagep = *twinp;
	pagep++;
	twinp++;
    }
}

