/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */

static char rcsid[] = "$Id: scash_flush.c,v 1.5 2004-08-24 18:58:50 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

int 
scash_page_current_home(int  page)
{
    int   home = scash_page2home(page);
    int   base = scash_page2base(page);

    /* if (home == scash.pe || base == scash.pe || scash_page2ht(page) != 0) */
    if (home == scash.pe || base == scash.pe)
      return home;

    OMSC_READ(base,
	      scash.page_map_addr_handle[base] + sizeof(page_map_t) * page,
	      scash.page_map_addr_handle[scash.pe] + 
	      sizeof(page_map_t) * (SCASH_MAX_PAGE -1),
	      sizeof(page_map_t));
      
    while(!OMSC_IS_READ_DONE()) /* */;

    page_map[page].home = page_map[SCASH_MAX_PAGE - 1].home;
    page_map[page].ocount = barrier_cnt;
    return page_map[SCASH_MAX_PAGE - 1].home;
}

static int
scash_flush_diff_send(int  home, 
		      int  dp,
		      int  msg_id,
		      int  copy_flag) 
{
    int  diff_size;
    scash_msg_t*  msg_buff;
    u_int  msg_length;

    _s_diff_index_t  diff_index;
    int  max_diff_size;
    int  i;
    int  diff_msg_size = 0;
    u_int   *pagep;
    scash_msg_diff_t  scash_msg_diff;
    diff_block_t   *diff_block;
    extern  void scash_diff_index_create(int, int, _s_diff_index_t*);
    extern  void scash_diff_index_create_copy(int, int, _s_diff_index_t*);

    pagep = (u_int*)(scash_page2shadowaddr(dp));
    diff_index.diff_num = 0;

    /* Message Header = 12  Page    length       diff_block_header   */
    max_diff_size = scash.mtu - sizeof(scash_msg_header_t) 
	- sizeof(int) - sizeof(int) -sizeof(diff_block_header_t) - 4;

    if (copy_flag)
        scash_diff_index_create_copy(dp, max_diff_size, &diff_index);
    else
        scash_diff_index_create(dp, max_diff_size, &diff_index);

    /*
     *  Send Diff
     */
    scash_msg_diff.page = dp;
    scash_msg_diff.diff.length = 0;
    diff_block =  (diff_block_t*)&scash_msg_diff.diff.block[0];
    diff_msg_size = sizeof(scash_msg_header_t) + sizeof(u_long) + sizeof(int);

    if (diff_index.diff_num==0) {
	scash_message_send2(SCASH_MSG_FLUSH_DIFF_SEND, 
			    home,
			    diff_msg_size - sizeof(scash_msg_header_t),
			    &scash_msg_diff);
        return SCASH_SUCCESS;
    }

    i = 0;
    while (i < diff_index.diff_num){
        diff_size = sizeof(diff_block_header_t) + 
	    diff_index.db[i].length * SCASH_WORD_SIZE;
	if (diff_msg_size + diff_size < scash.mtu){
	  diff_msg_size += diff_size;
	  scash_msg_diff.diff.length++;
	  diff_block->header = diff_index.db[i];
	  bcopy((char*)(pagep + diff_index.db[i].pos), 
		diff_block->data, 
		diff_index.db[i].length * SCASH_WORD_SIZE);

	  diff_block = (diff_block_t*)(diff_block->data + 
				       diff_index.db[i].length);

	  i++;
	} else {
	    scash_message_send2(SCASH_MSG_FLUSH_DIFF_SEND,
				home,
				diff_msg_size - sizeof(scash_msg_header_t),
				&scash_msg_diff);

	    scash_msg_diff.diff.length = 0;
	    diff_block =  (diff_block_t*)&scash_msg_diff.diff.block[0];
	    diff_msg_size = sizeof(scash_msg_header_t) + 
		sizeof(int) + sizeof(int);
	}
    }

    if (diff_msg_size > sizeof(scash_msg_header_t)+sizeof(int)+sizeof(int)){
        scash_message_send2(SCASH_MSG_FLUSH_DIFF_SEND, 
			    home,
			    diff_msg_size - sizeof(scash_msg_header_t),
			    &scash_msg_diff);
    }

    while(ENOBUFS != scash_message_recv(&msg_buff)){
        scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
    return  SCASH_SUCCESS;
}

int
scash_flush_diff_size(int  home, 
		      int  dp)
{
    scash_msg_diff_size_t   diff_size;

    diff_size.page = dp;
    diff_size.size = scash_diff_rmw(dp);
    return scash_message_send(SCASH_MSG_DIFF_SIZE, home, 
			      sizeof(scash_msg_diff_size_t), 
			      (void*)&diff_size);
}


/* flush all pages, called from barrier */
void
scash_flush_all()
{
    int   dp;
    int   home; 
    int   dirty_page_idx;
    int   dirty_page_cnt = dirty_page->total;
    scash_msg_t*  msg_buff;
    u_int  msg_length;
    int   proc;
    
    if(scash_tlog_flag) tlog_FLUSH_ALL_IN();
    
    for (dirty_page_idx = 0; dirty_page_idx < dirty_page_cnt;
	 dirty_page_idx++){

	dp = dirty_page->dp[dirty_page_idx];
	home = scash_page2home(dp);

	if (home/scash.smp == scash.box) {
	    scash_page2ocount(dp) = barrier_cnt+1;
	    if (scash.pe == home) {
	        scash_flushed_page_set(dp);
	        SCASH_PAGE_SET_RO(page_dir[dp].attr);
		SCASH_PAGE_CLEAR_TWIN(page_dir[dp].attr);
		SCASH_PAGE_CLEAR_BASE_WRITE(page_dir[dp].attr);
		SCASH_PAGE_CLEAR_HOME_WRITE(page_dir[dp].attr);
	    }
	    scash_mprot_ro(scash_page2addr(dp));
	} else {
	    if (dirty_page_idx % scash.smp == scash.proc) {
	        scash_flush_diff_send(home, dp, SCASH_MSG_FLUSH_DIFF_SEND, 0);
	    }
	}
	while(ENOBUFS != scash_message_recv(&msg_buff)) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }

    while(ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }

    if (scash.smp > 1) {
        dirty_page->inbox_lock[scash.proc] = 1;
	do {
	    dp = 0;
	    for (proc = 0 ; proc < scash.smp ; proc++) {
	        if (dirty_page->inbox_lock[proc]) {
		    dp++;
		}
	    }
	    while(ENOBUFS != scash_message_recv(&msg_buff)) {
	        scash_msg_handler(msg_buff);
		OMSC_RELEASE_BUF();
	    }
	} while (dp != scash.smp);
    }

    if (scash.proc==0) {
        scash_dp_init();
    }

    if(scash_tlog_flag) tlog_FLUSH_ALL_OUT();
}

void
scash_flush_all_home()
{
    int   dp, attr;
    int   base, home; 
    int   dirty_page_idx;
    scash_msg_t*  msg_buff;
    u_int  msg_length;

    for (dirty_page_idx = 0 ; dirty_page_idx < dirty_page->total ; dirty_page_idx++) {
	SCASH_RDTSC(ccnt[0]);
	dp = dirty_page->dp[dirty_page_idx];
	home = scash_page2home(dp);
	base = scash_page2base(dp);

	if (home == scash.pe) {
	    SCASH_PAGE_SET_RO(page_dir[dp].attr);
	    scash_mprot_ro( scash_page2addr(dp) );
	    SCASH_PAGE_CLEAR_TWIN(page_dir[dp].attr);
	    SCASH_PAGE_SET_HOME_WRITE(page_dir[dp].attr);
	    scash_page_invalidate(dp, home);

	} else {
	    attr = scash_mprot(dp);
	    scash_flush_diff_size(home, dp);
	    if (attr != SCASH_PAGE_UNMAPPED) {
		SCASH_PAGE_SET_RO(page_dir[dp].attr);
		scash_mprot_ro(scash_page2addr(dp));
		SCASH_PAGE_CLEAR_TWIN(page_dir[dp].attr);
	    }
	}
	while(ENOBUFS != scash_message_recv(&msg_buff)) {
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }
    while(!OMSC_IS_SEND_DONE()) /* */;
    scash_dp_init();
}

void
scash_send_refresh(int  dest,
		  msg_body_t*  msg_body)
{
    scash_msg_t*  recv_msg;
    int  recv_length;
    int  msg_id = 0;

    scash_message_send(SCASH_MSG_REFRESH, dest, 
		       sizeof(scash_msg_refresh_t), msg_body);

    while (msg_id != SCASH_MSG_FRESH) {
	while( ENOBUFS == scash_message_recv(&recv_msg))
	    ;
	msg_id = recv_msg->header.id;
	scash_msg_handler(recv_msg);
	OMSC_RELEASE_BUF();
    }
}

int
scash_refresh_word(void*  rp)
{
    caddr_t   pp = (caddr_t)rp;
    u_int refresh_page, page_prot;
    u_int     home;

    if (scash.ps==1)
        return SCASH_SUCCESS;

    SCASH_POLL();

    if (pp < scash.dsm_addr || pp > scash.dsm_addr + scash.dsm_size) {
      scash_prf("scash_refresh:Invalid Address[%p]\n", pp);
	scash_fatal("scash_refresh:Invalid Address", 1);
	return 1;
    }

    refresh_page = scash_addr2page(pp);
    home = scash_page2home(refresh_page);
    if (home/scash.smp == scash.box)
        return SCASH_SUCCESS;

    scash_page_lock(refresh_page, 1);

    page_prot = scash_mprot(refresh_page);
    if ( (page_prot == SCASH_PAGE_RO || page_prot == SCASH_PAGE_RW)) {
	unsigned* wp;
	int  gap;

	OMSC_READ(home,
		  scash.shadow_addr_handle[home] + (pp - scash.dsm_addr),
		  scash.shadow_addr_handle[scash.pe] + (pp - scash.dsm_addr),
		  4);

	while(!OMSC_IS_READ_DONE()) /* */;

	wp = (unsigned *)pp;
	gap = (u_int*)scash.twin_addr - (u_int*)scash.dsm_addr;
	*(wp + gap) = *wp;
    }
    scash_page_unlock(refresh_page);
    SCASH_POLL();
    return SCASH_SUCCESS;
}

/*
 * API: refresh 
 */
int
scash_refresh(caddr_t  from, 
	      int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  refresh_ptr;
    int      refresh_page, refresh_size;
    int      page_prot;

    SCASH_POLL();

    refresh_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (refresh_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return 1;

    while(refresh_ptr < from + byte_len){
        page_ru_ptr = 
	  (caddr_t)SCASH_PAGE_TRUNC((off_t)refresh_ptr + SCASH_PAGE_SIZE);
	if (page_ru_ptr > from + byte_len)
            refresh_size = from + byte_len - refresh_ptr;
	else
	    refresh_size = page_ru_ptr - refresh_ptr;

	refresh_page = scash_addr2page(refresh_ptr);
	home = scash_page2home(refresh_page);

	if (home/scash.smp != scash.box) {

	    scash_page_lock(refresh_page, 1);

	    page_prot = scash_mprot(refresh_page);
	    if (page_prot == SCASH_PAGE_RO || page_prot == SCASH_PAGE_RW) {
		OMSC_READ(home,
			  scash.shadow_addr_handle[home] + 
			  ((caddr_t)refresh_ptr - scash.dsm_addr), 
			  scash.shadow_addr_handle[scash.pe] + 
			  ((caddr_t)refresh_ptr - scash.dsm_addr), 
			  refresh_size);

		while(!OMSC_IS_READ_DONE()) /* */;

		bcopy(scash.shadow_addr + (refresh_ptr - scash.dsm_addr), 
		      scash.twin_addr + (refresh_ptr - scash.dsm_addr), 
		      refresh_size);
	    } 
	    scash_page_unlock(refresh_page);
        } /* if (home != scash.pe) */
        refresh_ptr += refresh_size;
    }
    SCASH_POLL();
    return SCASH_SUCCESS;
}

void
scash_recv_refresh(int  sender, 
		  scash_msg_refresh_t  *refresh)
{
    int  i, gap;
    msg_body_t msg_body;
    scash_msg_fresh_t  *fresh = &msg_body.fresh;

    fresh->length = refresh->length;
    gap = (u_int*)scash.shadow_addr - (u_int*)scash.dsm_addr;

    for (i = 0 ; i < refresh->length ; i++) {
	fresh->addr[i] = refresh->addr[i];
	fresh->data[i] = *(refresh->addr[i] + gap);
    }
    scash_message_send(SCASH_MSG_FRESH, sender, 
		       sizeof(scash_msg_fresh_t), &msg_body);
}

void
scash_send_flush(int  dest, 
		msg_body_t  *msg_body)
{
    scash_message_send(SCASH_MSG_FLUSH, dest, 
		       sizeof(scash_msg_flush_t), msg_body);
}

/* 
 * user's API: flush(write) the range from "from" to from+len
 */
int
scash_flush(caddr_t from,
	    int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  flush_ptr;
    int      flush_page, flush_size;
    int      page_prot;

    if (scash.ps==1)
        return SCASH_SUCCESS;

    SCASH_POLL();

    flush_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (flush_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return (-1);

    while(flush_ptr < from + byte_len) {
        page_ru_ptr = 
	  (caddr_t)SCASH_PAGE_TRUNC((off_t)flush_ptr + SCASH_PAGE_SIZE);

	if (page_ru_ptr > from + byte_len)
            flush_size = from + byte_len - flush_ptr;
	else
	    flush_size = page_ru_ptr - flush_ptr;

	flush_page = scash_addr2page(flush_ptr);
	home = scash_page2home(flush_page);

	if (home/scash.smp != scash.box) {

	    scash_page_lock(flush_page, 1);

	    page_prot = scash_mprot(flush_page);
	    if (page_prot == SCASH_PAGE_RW) {
		OMSC_WRITE(home,
			   scash.shadow_addr_handle[home] + 
			   (flush_ptr - scash.dsm_addr), 
			   scash.shadow_addr_handle[scash.pe] + 
			   (flush_ptr - scash.dsm_addr), 
			   flush_size);
		while(!OMSC_IS_WRITE_DONE()) /* */;

		bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
		      scash.twin_addr + (flush_ptr - scash.dsm_addr), 
		      flush_size);
	    } /* if page_prot == RW */ 
	    scash_page_unlock(flush_page);
        } /* if (home != scash.pe) */
        flush_ptr += flush_size;
    }
    SCASH_POLL();
    return SCASH_SUCCESS;
}


/* reset all pages to UNMAPPED */
void
scash_invalidate_all()
{
    int  page;

    SCASH_POLL();

    for (page = 0 ; page < scash.dsm_pages  ; page++ ) {
        if (scash_page2home(page) == scash.pe)
            continue;

        SCASH_PAGE_SET_UNMAPPED(page_dir[page].attr);
        scash_mprot_none(scash_page2addr(page));
    }
    SCASH_POLL();
}

void
scash_omp_flush(caddr_t  from,  int  byte_len)
{
    int      home;
    caddr_t  page_ru_ptr;
    caddr_t  flush_ptr;
    int      flush_page, flush_size;
    int      page_prot;
  
    unsigned  *diff_ptr;
    unsigned  *twin_ptr;

    if (scash.ps==1)    return;

    SCASH_POLL();

    flush_ptr = (caddr_t)(SCASH_WORD_TRUNC((off_t)from));

    if (flush_ptr < scash.dsm_addr || 
	from + byte_len > scash.dsm_addr + scash.dsm_size)
        return;

    while(flush_ptr < from + byte_len) {
      page_ru_ptr = 
	(caddr_t)SCASH_PAGE_TRUNC((off_t)flush_ptr + SCASH_PAGE_SIZE);

      if (page_ru_ptr > from + byte_len)
	flush_size = from + byte_len - flush_ptr;
      else
	flush_size = page_ru_ptr - flush_ptr;

	flush_page = scash_addr2page(flush_ptr);
	home = scash_page2home(flush_page);

	if (home/scash.smp != scash.box) {

	    scash_page_lock(flush_page, 1);

	    page_prot = scash_mprot(flush_page);
	    //scash_prf("scash_omp_home page_prot[%d]\n", page_prot);
	    if (page_prot == SCASH_PAGE_RO) {
		OMSC_READ(home,
			  scash.shadow_addr_handle[home] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			  scash.shadow_addr_handle[scash.pe] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			  flush_size);
	      while(!OMSC_IS_READ_DONE()) /* */;

	      bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
		    scash.twin_addr + (flush_ptr - scash.dsm_addr), 
		    flush_size);
	    } 

	    if (page_prot == SCASH_PAGE_RW) {
 		int	diff_len = 0;
 		int	need_flush = 1;

		diff_ptr = (unsigned*)((flush_ptr - scash.dsm_addr) + scash.shadow_addr);
		twin_ptr = (unsigned*)((flush_ptr - scash.dsm_addr) + scash.twin_addr);

		while( (caddr_t)diff_ptr < ((flush_ptr - scash.dsm_addr) + scash.shadow_addr) + flush_size) {
 		    diff_len = 0;
 		    while( (caddr_t)&diff_ptr[diff_len] < ((flush_ptr - scash.dsm_addr) + scash.shadow_addr) + flush_size) {
 		        if (diff_ptr[diff_len] != twin_ptr[diff_len]) {
 			    diff_len++;
 			} else {
 			    break;
 			}
 		    }
 		    if (diff_len) {
		        OMSC_WRITE(home,
				   scash.shadow_addr_handle[home] + (((caddr_t)diff_ptr) - scash.shadow_addr), 
				   scash.shadow_addr_handle[scash.pe] + (((caddr_t)diff_ptr) - scash.shadow_addr), 
				   diff_len * sizeof(int));

 			diff_ptr += diff_len;
 			twin_ptr += diff_len;
 			if (diff_len * sizeof (unsigned) == flush_size) {
 			  need_flush = 0;
 			}
		    } else {
 		        diff_ptr++;
 			twin_ptr++;
  		    }
 		}
 		while(!OMSC_IS_WRITE_DONE()) /* */;

 		if (need_flush) {
 		    OMSC_READ(home,
			      scash.shadow_addr_handle[home] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			      scash.shadow_addr_handle[scash.pe] + ((caddr_t)flush_ptr - scash.dsm_addr), 
			      flush_size);
 		    while(!OMSC_IS_READ_DONE()) /* */;
  		}
  		bcopy(scash.shadow_addr + (flush_ptr - scash.dsm_addr), 
  		      scash.twin_addr + (flush_ptr - scash.dsm_addr), 
		      flush_size);
	    } /* if page_prot == RW */ 
	    scash_page_unlock(flush_page);
        } /* if (home != scash.pe) */
        flush_ptr += flush_size;
    }
    SCASH_POLL();
}

/* called from barrier, after flush */
void
scash_flush_page_invalidate()
{
    int   home; 
    int   flush_page;
    scash_msg_t*  msg_buff;
    u_int  msg_length;

    if(scash_tlog_flag) tlog_PAGE_INV_IN();
    
    for (flush_page = 0 ; flush_page < scash.dsm_pages ; flush_page++) {
        home = scash_page2home(flush_page);
	if (scash.pe == home) {
	    if (scash_flushed_page(flush_page)) {
	        scash_flushed_page_clear(flush_page); 
		scash_page_invalidate(flush_page, home);
	    }
	} else {
	    if (scash_flushed_page(flush_page)) {
	        scash_prf("scash_flush_page_invalidate:: Page=%d, Home=%d\n", 
			  flush_page, home);
		for(;;)
		    SCASH_POLL();
	    }
	}
    }

    while(ENOBUFS != scash_message_recv(&msg_buff)) {
        scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
    }
    if(scash_tlog_flag) tlog_PAGE_INV_OUT();
}

/* EOF */

