/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_handler.c,v 1.4 2004-07-23 07:57:08 ojima Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#ifdef __x86_64__
#include  <asm/ucontext.h>
#endif
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>
#ifdef  __linux__
#include  <asm/unistd.h>
#else
#include  <unistd.h>
#endif

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

#ifdef __linux__
void
#ifdef  __alpha__
scash_vm_handler(int  signal,
		 int  code,
		 struct sigcontext*  sc)
#elif defined(__ia64__)
scash_vm_handler(int  signal,
		 struct siginfo *sinfo,
		 struct sigcontext *scntxt)
#elif defined(__x86_64__)
scash_vm_handler(int  signal,
		 struct siginfo *sinfo,
		 struct ucontext *uc)
#else
scash_vm_handler(int  signal,
		 int  code)
#endif
#else
scash_vm_handler(int  signal,
		 int  code,
		 struct sigcontext*  sc,
		 caddr_t fault_addr)
#endif
{
    unsigned long  fault_page;
    unsigned long  fault_pc;
    caddr_t  fault_addr;
    page_t *page_p;
    u_int  page_attr;
    int  home_node;
    struct sigcontext*  sc;

    if(scash_debug) scash_prf("scash_vm_handler ...");
    if(scash_tlog_flag){
	tlog_PAGEF_EVENT();
	tlog_PAGEF_IN();
    }
    SCASH_POLL();
    
#if __linux__
#ifdef  __alpha__
    fault_pc = sc->sc_pc;
    fault_addr = (caddr_t)sc->sc_traparg_a0;
#elif defined(__ia64__)
    fault_pc = scntxt->sc_ip;
    fault_addr = (caddr_t)sinfo->si_addr;
    sc = scntxt;
#elif defined(__x86_64__)
    fault_pc = uc->uc_mcontext.rip;
    fault_addr = sinfo->si_addr;
#else
    sc = (struct sigcontext*)&code;
    fault_pc = sc->eip;
    fault_addr = (caddr_t)sc -> cr2;
#endif
#endif
    if (fault_addr < scash.dsm_addr || 
	fault_addr >= scash.dsm_addr + scash.dsm_size) {
#ifdef __ia64__
	scash_prf("VM Fault out of DSM area PC=[%08lx]:ADDR=[%p]\n", fault_pc, fault_addr);
#elif defined(__x86_64__)
	scash_prf("VM Fault out of DSM area PC=[%p]:ADDR=[%p]\n", fault_pc, fault_addr); 
#else
	scash_prf("VM Fault out of DSM area PC=[%08x]:ADDR=[%08x]",
		  (unsigned)fault_pc, (unsigned)fault_addr);
#endif
	scash_exit(1);
    }

    fault_addr = (caddr_t)SCASH_TRUNC((unsigned long)fault_addr, SCASH_PAGE_SIZE);
    fault_page = ((unsigned long)(fault_addr - scash.dsm_addr))/ SCASH_PAGE_SIZE;
    home_node = scash_page2home(fault_page);

    if(scash_debug) scash_prf("fault_page=0x%x ...",fault_page);

    /* Lock Fault Page */
    scash_page_lock(fault_page, 1);

    page_p = page_dir + fault_page;
    page_attr = SCASH_PAGE_ATTR(page_p->attr);
#ifdef  __linux__
#ifdef __alpha__
    if (sc->sc_traparg_a1 == 0x04) {
#elif defined(__ia64__) || defined(__x86_64__)
    if (sinfo->si_code == SEGV_ACCERR || sinfo->si_code == SEGV_MAPERR) {
#else
    if (sc->trapno == 0x0e && sc->err & 0x2) {
#endif
#else
    if (sc->sc_trapno == T_PAGEFLT && sc->sc_err & 0x2) {
#endif
      /* Write Access Fault */
      if (page_attr == SCASH_PAGE_RW) {
	  if(scash_debug) scash_prf("W Falut: page_attr RW ...");
	  scash_mprot_rw(fault_addr);
	  if (home_node==scash.pe) {
	      scash_page_cs_set(fault_page, scash.pe);
	  } else {
	      scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, 
				  home_node, sizeof(fault_page), &fault_page);
	  }
	  goto unlock_and_return;
      } else if (page_attr == SCASH_PAGE_RO) {
	  if(scash_debug) scash_prf("W Fault: page_attr RO ...");
	  scash_dp_insert(fault_page);
	  scash_mprot_rw(fault_addr);
	  SCASH_PAGE_SET_RW(page_p->attr);
	  if (home_node==scash.pe) {
	      scash_page_cs_set(fault_page, scash.pe);
	  } else {
	      scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, 
				  home_node, sizeof(fault_page), &fault_page);
	      scash_twin_create(fault_page);
	      SCASH_PAGE_SET_TWIN(page_p->attr);
	  }
      } else { /* SCASH_PAGE_UNMAPPED */
	  if(scash_debug) scash_prf("W Fault: page_attr UNMAPPED ...");
	  scash_dp_insert(fault_page);
	  scash_mprot_rw(fault_addr);
	  SCASH_PAGE_SET_RW(page_p->attr);
	  
	  if (home_node/scash.smp == scash.box) {
	      if (home_node==scash.pe ) {
		  scash_page_cs_set(fault_page, scash.pe);
	      } else {
		  scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, 
				      home_node, sizeof(fault_page), &fault_page);
	      }
	  } else {
	      if (scash_opt_page_copy() == SCASH_OPT_RMA_PAGE_COPY) {
		  scash_page_rma_copy(fault_page, home_node);
	      } else {
		  scash_page_copy(fault_page, home_node);
	      }
	      scash_twin_create(fault_page);
	      SCASH_PAGE_SET_TWIN(page_p->attr);
	  }
      }
    } else {  	/* Read ACCESS Fault */
	if (page_attr == SCASH_PAGE_RW) {
	    if(scash_debug) scash_prf("R Falut: page_attr RW ...");
	    scash_mprot_rw(fault_addr);
	    if (home_node == scash.pe) {
		scash_page_cs_set(fault_page, scash.pe);
	    } else {
	        scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ,home_node,
				    sizeof(fault_page), &fault_page);
	    }
	} else if (page_attr == SCASH_PAGE_RO) {
	    if(scash_debug) scash_prf("R Fault: page_attr RO ...");
	    scash_mprot_ro(fault_addr);
	    if (home_node == scash.pe) {
		scash_page_cs_set(fault_page, scash.pe);
	    } else {
	        scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, home_node, 
				    sizeof(fault_page), &fault_page);
	    }
	} else if (page_attr == SCASH_PAGE_UNMAPPED) {
	    if(scash_debug) scash_prf("R Fault: page_attr UNMAPPED ...");
	    scash_mprot_ro(fault_addr);
	    SCASH_PAGE_SET_RO(page_p->attr);
	    if (home_node/scash.smp == scash.box) {
	        if (home_node==scash.pe) {
		    scash_page_cs_set(fault_page, scash.pe);
		} else {
		    scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, home_node, 
					sizeof(fault_page), &fault_page);
		}
	    } else {  
	        if (scash_opt_page_copy() == SCASH_OPT_RMA_PAGE_COPY) {
		    scash_page_rma_copy(fault_page, home_node);
		} else {
		    scash_page_copy(fault_page, home_node);
		}
	    }
	}
    }

unlock_and_return:
    /* Unlock Fault Page */ 
    scash_page_unlock(fault_page);
    SCASH_POLL();

    if(scash_tlog_flag) tlog_PAGEF_OUT();
    if(scash_debug) scash_prf("scash_vm_handler ... end");
}

void  
scash_sig_handler_set()
{

#ifdef __NetBSD__
    scash.sig_stack.ss_flags = 0;
    scash.sig_stack.ss_size = SCASH_SIG_STACK_SIZE;
    scash.sig_stack.ss_sp = (caddr_t)malloc(SCASH_SIG_STACK_SIZE);

#ifdef MMAP_STACK
    scash.sig_stack.ss_sp = mmap(0, SCASH_SIG_STACK_SIZE, 
				 PROT_READ|PROT_WRITE,
				 MAP_ANON|MAP_PRIVATE, 
				 scash.mmap, 0);
#endif   /* MMAP_STACK */
    if (sigaltstack( &scash.sig_stack, NULL) < 0)
	scash_exit (1);

#endif  /* __NetBSD__ */
#ifdef  __linux__
    struct _sigaltstack {
        void *ss_sp;
        int ss_flags;
        size_t ss_size;
    };

    struct _sigaltstack  sig_stack;

    sig_stack.ss_flags = 0;
    sig_stack.ss_size = SCASH_SIG_STACK_SIZE;
    sig_stack.ss_sp = (char *)valloc(SCASH_SIG_STACK_SIZE);
    if (sigaltstack((struct sigaltstack *)&sig_stack, NULL) < 0) {
      perror("sigaltstack");
      exit(1);
    }
#endif

    sigemptyset(&scash.sig_mask);
    sigaddset(&scash.sig_mask, SIGBUS);
    sigaddset(&scash.sig_mask, SIGSEGV);
    sigaddset(&scash.sig_mask, SIGIO);
    sigaddset(&scash.sig_mask, SIGQUIT);
    scash.sig_act.sa_mask = scash.sig_mask;

#ifdef __linux__
#ifdef __ia64__
    scash.sig_act.sa_flags = 0;
#else
#ifdef __x86_64__
    scash.sig_act.sa_flags = SA_SIGINFO | SA_STACK;
#else
    scash.sig_act.sa_flags = SA_STACK;
#endif
#endif
#ifdef __i386__
    scash.sig_act.sa_restorer = 0;
#endif
#else
    scash.sig_act.sa_flags = SA_ONSTACK;
#endif
    if(scash_debug) scash_prf("scash_sig_handler_set ...\n");

    scash.sig_act.sa_handler = (void (*)(int))scash_vm_handler;
    sigaction(SIGSEGV, &scash.sig_act, NULL);
    if (sigaction(SIGSEGV, &scash.sig_act, NULL) < 0)
        perror("sigaction");
}
/* EOF */
