static char rcsid[] = "$Id: scash_message.c,v 1.4 2004-08-24 17:25:43 msato Exp $";
/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

/* 
 * send message 
 */
int
scash_message_send(int    msg_id,
		   int    dest,
		   int    body_size,
		   void*  body)
{
    scash_msg_t*  bufp;
    int  buf_size;
    int  error;

    if (scash_debug) {
	scash_prf("scash_send_msg [%s:%x]  to [%d]", 
		  scash_msg_name(msg_id), scash.seq, dest);
    }
    if(scash_tlog_flag) tlog_SEND_IN();

    if (dest==scash.pe) {
	scash_prf("scash_message_send %s, %d, %d",
		  scash_msg_name(msg_id), body_size, dest);
	exit(1);
    }

    buf_size = sizeof(scash_msg_header_t) + body_size;

    if (buf_size > scash.mtu){
	scash_prf("scash_send_msg Msg Size Larger than MTU ID=%s, Size=%d,Dest=%d", 
		  scash_msg_name(msg_id), buf_size, dest);
      exit(1);
    }

    while ((error = OMSC_GET_SEND_BUF(dest, (caddr_t*)&bufp, buf_size)) 
	   == ENOBUFS)
	;
    if (error)
	scash_prf("scash_send_msg [%s:%x]  to [%d:%d] %d", 
		  scash_msg_name(msg_id), scash.seq, dest, body_size, error);

    bufp->header.id = msg_id;
    bufp->header.sender = scash.pe;
    bufp->header.size = buf_size;

    if (body_size > 0)
	bcopy(body, &bufp->body.body, body_size);

    if ((error = OMSC_SEND()) != 0) {
/*
	if(error == EINVAL)
	    scash_prf("error EINVAL");
	else if(error == EPERM)
	    scash_prf("error EPERM");
*/
	scash_fatal("scash_message_send: Send", error);
    }
    if(scash_tlog_flag) tlog_SEND_OUT();

    return  SCASH_SUCCESS;
}

int
scash_message_send2(int    msg_id,
		    int    dest,
		    int    body_size,
		    void*  body)
{
    scash_msg_t*  bufp;
    int  buf_size;
    int  error;

    scash_msg_t*  msg_buff;
    u_int  msg_length;

    if (scash_debug) {
	scash_prf("scash_send_msg2 [%s:%x]  to [%d]\n", 
		  scash_msg_name(msg_id), scash.seq, dest);
    }
    if(scash_tlog_flag) tlog_SEND_IN();

    if (dest==scash.pe) {
	scash_prf("scash_message_send %s, %d, %d\n", 
		  scash_msg_name(msg_id), body_size, dest);
      exit(1);
    }

    buf_size = sizeof(scash_msg_header_t) + body_size;

    if (buf_size > scash.mtu) {
	scash_prf("scash_send_msg Msg Size Larger than MTU ID=%s, Size=%d,Dest=%d", 
		  scash_msg_name(msg_id), buf_size, dest);
      exit(1);
    }

    while ((error = OMSC_GET_SEND_BUF(dest,(caddr_t*)&bufp, buf_size)) 
	   == ENOBUFS) {
	while(ENOBUFS != scash_message_recv(&msg_buff)) {
	scash_msg_handler(msg_buff);
	OMSC_RELEASE_BUF();
      }
    }

    bufp->header.id = msg_id;
    bufp->header.sender = scash.pe;
    bufp->header.size = buf_size;

    if (body_size > 0) 
	bcopy(body, &bufp->body.body, body_size);

    if ((error = OMSC_SEND()) != 0) {
	scash_fatal("scash_message_send: Send", error);
    }
    if(scash_tlog_flag) tlog_SEND_OUT();

    return  SCASH_SUCCESS;
}

int
scash_message_multicast(int    msg_id,
			int*   dest,
			int    dest_cnt,
			int    body_size,
			void*  body)
{
    int   i;

    for (i = 0 ; i < dest_cnt ; i++) {
	scash_message_send(msg_id, dest[i], body_size, body);
    }
    return  SCASH_SUCCESS;
}

int
scash_message_multicast2(int    msg_id,
			 int*   dest,
			 int    dest_cnt,
			 int    body_size,
			 void*  body)
{
    int   i;

    for (i = 0 ; i < dest_cnt ; i++) {
	scash_message_send2(msg_id, dest[i], body_size, body);
    }
    return  SCASH_SUCCESS;
}

int
scash_message_broadcast(int    msg_id,
			int    body_size,
			void*  body)
{
    int   box, proc, dest;

    if(scash_debug) 
	scash_prf("scash_message_broadcast id=%s, size=%d ...",
		  scash_msg_name(msg_id),body_size);

    if (scash.pe % 2) {
      for (proc = 0 ; proc < scash.smp ; proc++) {
	for (box = 0 ; box < scash.ps/scash.smp ; box++) {
	  dest = box * scash.smp + proc;
	  if (dest!=scash.pe)
	    scash_message_send2(msg_id, dest, body_size, body);
	}
      }
    } else {
      for (proc = 0 ; proc < scash.smp ; proc++) {
	for (box = (scash.ps/scash.smp)-1; box >= 0 ; box--) {
	  dest = box * scash.smp + proc;
	  if (dest!=scash.pe)
	    scash_message_send2(msg_id, dest, body_size, body);
	}
      }
    }
    return SCASH_SUCCESS; 
}

int  
scash_message_recv(scash_msg_t**  recv_msg)
{
    int   error;
#if defined(__alpha) || defined(__ia64__) || defined(__x86_64__)
    u_long  msg_len;
#else
    u_int msg_len;
#endif

#if  defined(__alpha) || defined(__ia64__) || defined(__x86_64__)
    error = OMSC_RECV((caddr_t*)recv_msg, &msg_len);
    /* *msg_length = msg_len; */
#else
    error = OMSC_RECV((caddr_t*)recv_msg, &msg_len);
#endif

    if (error == ENOBUFS) {
	return  error;
    }
    if (error != 0) scash_fatal(" Receive", error);
    /* if (scash_debug) scash_prf("recv ..."); */
    return error;
}
