/* 
 */
#ifndef __scash_msg_h
#define __scash_msg_h

typedef  struct scash_msg_header_s{
    u_short  id;
    u_short  sender;
    u_int    size;
} scash_msg_header_t;

typedef  struct  scash_msg_copy_page_s {
    char  page[SCASH_PAGE_SIZE];
    u_int copy_page;
}scash_msg_copy_page_t;

typedef  struct  scash_msg_diff_s {
    u_long   page;
    diff_t   diff;
}scash_msg_diff_t;

typedef  struct  scash_msg_refresh_s {
    int      length;
    u_int*   addr[SCASH_MAX_REFRESH];
}scash_msg_refresh_t;

typedef  struct  scash_msg_fresh_s {
    int      length;
    u_int*   addr[SCASH_MAX_REFRESH];
    u_int    data[SCASH_MAX_REFRESH];
}scash_msg_fresh_t;

typedef  struct  scash_msg_flush_s {
    int      length;
    u_int*   addr[SCASH_MAX_REFRESH];
    u_int    data[SCASH_MAX_REFRESH];
} scash_msg_flush_t;

typedef  struct  scash_msg_addr_handle_s {
    addrHandle   shadow;
    addrHandle   page_map;
} scash_msg_addr_handle_t;

typedef  struct  scash_msg_diff_update_s {
    u_int    page;
    u_int    offset;
    u_int    len;
    diff_t   diff;
} scash_msg_diff_update_t;

typedef  struct  scash_msg_invalidate_s {
    int     page;
    int     new_home; 
} scash_msg_invalidate_t;

typedef  struct  scash_msg_new_home_s {
    int   page_num;
    char  page_data[SCASH_PAGE_SIZE];
} scash_msg_new_home_t;

typedef  struct  scash_msg_new_home_base_s {
    int   page;
    int   home;
} scash_msg_new_home_base_t;

typedef  struct  scash_msg_diff_size_s {
    int   page;
    int   size;
} scash_msg_diff_size_t;

typedef  struct  scash_msg_dist_byte_s {
    caddr_t   key;
    caddr_t   ptr;
    int       length;
    char      data[8];
} scash_msg_dist_byte_t ;

typedef  struct  scash_msg_dist_ok_s {
    caddr_t   key;
} scash_msg_dist_ok_t;

#define  SCASH_MSG_HOME_LEN    256

typedef  struct  scash_msg_omni_home_s {
   int      home;
   caddr_t  pagep; 
   int      pagen;
   unsigned char   cur_home[SCASH_MSG_HOME_LEN];
} scash_msg_omni_home_t;

typedef  struct  scash_msg_scash_home_s {
   int      home;
   caddr_t  pagep; 
   int      pagen;
} scash_msg_scash_home_t;

typedef  struct  scash_msg_barrier_a2a_s {
   unsigned  barrier_cnt;
   int       barrier_id;
} scash_msg_barrier_a2a_t;

typedef union msg_body_u {
    char   body;
    u_int  fault_page;
    int    lock_id;
    scash_msg_copy_page_t copy_page;
    scash_msg_diff_t diff;
    scash_msg_invalidate_t   invalidate;
    scash_msg_refresh_t    refresh;
    scash_msg_fresh_t      fresh;
    scash_msg_flush_t      flush;
    scash_msg_addr_handle_t addr_handle;
    scash_msg_new_home_t  new_home;
    scash_msg_new_home_base_t  new_home_base;
    scash_msg_diff_size_t  diff_size;
    scash_msg_dist_byte_t dist_byte;
    scash_msg_dist_ok_t dist_ok;
    scash_msg_omni_home_t scash_omni_home;
    scash_msg_scash_home_t scash_scash_home;
    scash_msg_barrier_a2a_t barrier_a2a;
} msg_body_t;

typedef  struct scash_msg_s {
    scash_msg_header_t   header;
    msg_body_t   body;
}scash_msg_t;


#define  SCASH_MSG_PAGE_REQ    0x0001
#define  SCASH_MSG_PAGE_SEND   0x0002
/* #define  SCASH_MSG_PAGE_RMA_REQ    0x0003 */
/* #define  SCASH_MSG_PAGE_RMA_SEND   0x0004 */
#define  SCASH_MSG_PAGE_SHARE_REQ   0x0005
#define  SCASH_MSG_PAGE_INVALIDATE   0x0006
/* #define  SCASH_MSG_INVALIDATE_POOL   0x0007 */
#define  SCASH_MSG_FLUSH_DIFF_SEND   0x0008
/* #define  SCASH_MSG_FLUSH_UPDATE_DIFF_SEND   0x0009 */
/* #define  SCASH_MSG_DIFF_UPDATE 0x000a */
/* #define  SCASH_MSG_DIFF_MULTI   0x000b */
/* #define  SCASH_MSG_FLUSH_UPDATE_DIFF_SEND_NEW   0x000c */
/* #define  SCASH_MSG_FLUSH_DIFF_SEND_NEW   0x000d */
/* #define  SCASH_MSG_FLUSH_DIFF_RMA   0x000e */
#define  SCASH_MSG_DIFF_SIZE     0x000f
#define  SCASH_MSG_BARRIER_REQ   0x0010
#define  SCASH_MSG_BARRIER_OK    0x0011
#define  SCASH_MSG_TRY_LOCK      0x0012
#define  SCASH_MSG_LOCK      0x0013
#define  SCASH_MSG_LOCK_OK   0x0014
#define  SCASH_MSG_LOCK_BUSY   0x0015
#define  SCASH_MSG_UNLOCK    0x0016
#define  SCASH_MSG_REFRESH   0x0017
#define  SCASH_MSG_FRESH     0x0018
#define  SCASH_MSG_FLUSH     0x0019
/* #define  SCASH_MSG_FLUSH2    0x001a */
/* #define  SCASH_MSG_FLUSH_OK  0x001b */
#define  SCASH_MSG_MLOCK_ADDR 0x001c
/* #define  SCASH_MSG_UPDATE_W   0x001d */
/* #define  SCASH_MSG_UPDATE     0x001e */
/* #define  SCASH_MSG_UPDATE_OK  0x001f */
/* #define  SCASH_MSG_PAGE_UPDATE   0x0020 */
/* #define  SCASH_MSG_BASE_WRITE   0x0021 */
#define  SCASH_MSG_NEW_HOME   0x0022
#define  SCASH_MSG_HOME   0x0023
#define  SCASH_MSG_NEW_HOME_BASE   0x0024
/* #define  SCASH_MSG_FLUSH_PAGE_RMA   0x0025 */
/* #define  SCASH_MSG_FLUSH_PAGE_SEND   0x0026 */
#define  SCASH_MSG_DIST_BYTE   0x0027
#define  SCASH_MSG_DIST_OK   0x0029
/* #define  SCASH_MSG_DIST_LONG   0x0027 */
/* #define  SCASH_MSG_DIST_SHORT   0x0028 */
/* #define  SCASH_MSG_DIST_RET   0x0029 */
#define  SCASH_MSG_OMNI_HOME   0x002a
#define  SCASH_MSG_SCASH_HOME   0x002b
#define  SCASH_MSG_SCASH_HOME_OK  0x002c
#define  SCASH_MSG_MAP_FILE   0x002d
#define  SCASH_MSG_PAGE_FILE   0x002e
#define  SCASH_MSG_DIRTY_PAGE  0x002f

#define  SCASH_MSG_RAW_DATA   0x00FF

/* #define  SCASH_MSG_RMA_PINGPONG  0xffff */

extern void scash_msg_handler(scash_msg_t*);
extern int scash_message_send(int, int, int, void*);
extern int scash_message_broadcast(int, int, void*);
extern int scash_message_recv(scash_msg_t**);
extern char *scash_msg_name(int id);

void scash_send_raw_data(int dst,int nbytes,void *datap);
void scash_recv_raw_data(int nbytes,void *datap);

#endif  /*  __scash_msg_h  */



