/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */

static char rcsid[] = "$Id: scash_msg_handler.c,v 1.5 2004-08-24 18:58:50 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

#define     SCASH_DIFF_APPLY   pagep = (u_int*)(scash_page2shadowaddr(fault_page));\
	    blockp = (diff_block_t*)&diffp->block[0];\
	    for (i = 0 ; i < diffp->length ; i++) {\
		bcopy(&blockp->data[0], pagep + blockp->header.pos , blockp->header.length * sizeof(int));\
		blockp = (diff_block_t*)&(blockp->data[blockp->header.length]);\
	    }

void
scash_msg_handler(scash_msg_t*  msg)
{
    int  sender;
    int  fault_page;
    msg_body_t  reply_msg_body;

    if(scash_debug)
	scash_prf("scash_handler %s ...",scash_msg_name(msg->header.id));

    if(scash_tlog_flag) tlog_MSGH_IN();

    switch (msg->header.id) {

    case SCASH_MSG_LOCK: {
        scash_recv_lock(msg->body.lock_id, msg->header.sender);
	break;
    }

    case SCASH_MSG_UNLOCK: {
        scash_recv_unlock(msg->body.lock_id, msg->header.sender);
	break;
    }

    case SCASH_MSG_LOCK_OK: {
	scash.lock_stat = SCASH_LOCK_OK;
	break;
    }

    case SCASH_MSG_TRY_LOCK: {
        scash_recv_try_lock(msg->body.lock_id, msg->header.sender);
	break;
    }

    case SCASH_MSG_LOCK_BUSY: {
	scash.lock_stat = SCASH_LOCK_BUSY;
	break;
    }

    case SCASH_MSG_PAGE_REQ: {
	sender = msg->header.sender;
	fault_page = msg->body.fault_page;
	scash_page_cs_set(fault_page, sender);
	reply_msg_body.copy_page.copy_page = fault_page;
	bcopy(scash_page2shadowaddr(fault_page), 
	      reply_msg_body.copy_page.page, SCASH_PAGE_SIZE);
	scash_message_send(SCASH_MSG_PAGE_SEND, sender, 
			   sizeof(scash_msg_copy_page_t), &reply_msg_body);
	break;
    }

    case SCASH_MSG_PAGE_SHARE_REQ: {
	sender = msg->header.sender;
	fault_page = msg->body.fault_page;
	scash_page_cs_set(fault_page, sender);
	break;
    }

    case SCASH_MSG_PAGE_INVALIDATE: {
	fault_page = msg->body.invalidate.page;
	scash_mprot_none(scash_page2addr(fault_page));

	/* SCASH_SMP */
	while( scash_inbox_try_lock(0,  (void*)(&dirty_page->attr_lock))) 
	    ;

	SCASH_PAGE_SET_UNMAPPED(page_dir[fault_page].attr);
	SCASH_PAGE_CLEAR_TWIN(page_dir[fault_page].attr);

	/* SCASH_SMP */
	if (msg->body.invalidate.new_home != scash_page2home(fault_page)) {
	    scash_prf("SCASH_MSG_PAGE_INVALIDATE new_home of page[%d] is %d\n", 
		      fault_page, scash_page2home(fault_page));
	}

	scash_page2ocount(fault_page) = barrier_cnt + 1;
	if (page_dir[fault_page].attr & SCASH_PAGE_RO) {
	    scash_prf("Error on Page %d AAA\n", fault_page);
	}
	/* SCASH_SMP */
	scash_inbox_unlock(0, (void*)(&dirty_page->attr_lock));

	if (page_dir[fault_page].attr & SCASH_PAGE_RO) {
	    scash_prf("Error on Page %d BBB", fault_page);
	}

	break;
    }

    case SCASH_MSG_FLUSH_DIFF_SEND:{
	int       i;
	diff_t    *diffp;
	u_int     *pagep;
	diff_block_t   *blockp; 

	fault_page = msg->body.diff.page;
	diffp = &msg->body.diff.diff;

	if (scash_page2home(fault_page) != scash.pe) {
	    scash_prf("SCASH_MSG_FLUSH_DIFF_SEND page[%d] is not mine[%d:%d]\n",
		  fault_page,
		  scash_page2home(fault_page), msg->header.sender);
	}

	scash_flushed_page_set(fault_page);

#ifdef __SCASH_SINGLE_PE__
	scash_page_invalidate(fault_page, new_home);
#endif
	SCASH_DIFF_APPLY
	break;
    }

    case SCASH_MSG_BARRIER_REQ: {
	scash.barrier.id = msg->body.fault_page;
	scash.barrier.cnt++;
	break;
    }

    case SCASH_MSG_BARRIER_OK: {
	unsigned  barrier_cnt;
	int       barrier_id;
        extern unsigned scash_barrier_exec_recv[4][SCASH_MAX_PE];

	barrier_cnt = msg->body.barrier_a2a.barrier_cnt;
	barrier_id  = msg->body.barrier_a2a.barrier_id;
        scash_barrier_exec_recv[barrier_cnt % 4][msg->header.sender] = barrier_id;
	break;
    }

    case SCASH_MSG_REFRESH: {
        scash_recv_refresh(msg->header.sender, &msg->body.refresh);
        break;
    }

    case SCASH_MSG_FRESH: {
        scash_msg_fresh_t   *fresh = &msg->body.fresh;
        int   i, t_gap, s_gap ;
        s_gap = (u_int*)scash.shadow_addr - (u_int*)scash.dsm_addr;
        t_gap = (u_int*)scash.twin_addr - (u_int*)scash.dsm_addr;

	for (i = 0 ; i < fresh->length ; i++) {
	    *(fresh->addr[i] + s_gap) = fresh->data[i];
	    *(fresh->addr[i] + t_gap) = fresh->data[i];
	}
        break;
    }

    case SCASH_MSG_FLUSH: {
        scash_msg_flush_t   *flush = &msg->body.flush;
	int   i, gap = (u_int*)scash.shadow_addr - (u_int*)scash.dsm_addr;

	for (i = 0 ; i < flush->length ; i++)
	    *(flush->addr[i] + gap) = flush->data[i];
        break;
    }

    case SCASH_MSG_MLOCK_ADDR: {
	sender = msg->header.sender;
	scash.shadow_addr_handle[sender] = msg->body.addr_handle.shadow;
	scash.page_map_addr_handle[sender] = msg->body.addr_handle.page_map;
        break;
    }

    case SCASH_MSG_DIFF_SIZE: {
	int       page, size;

	sender = msg->header.sender;
	page = msg->body.diff_size.page;
	size = msg->body.diff_size.size;
	
	if (sender == scash_page2base(page))
	    SCASH_PAGE_SET_BASE_WRITE(page_dir[page].attr);

	if ( size > page_dir[page].diff[1] ) {
	    page_dir[page].diff[0] = sender;
	    page_dir[page].diff[1] = size;
	}

	if (SCASH_PAGE_GET_UPDATE(page_dir[page].attr) == 0) {
	    SCASH_PAGE_SET_UPDATE(page_dir[page].attr);
	    page_dir[page].update_next = update_page_head;
	    update_page_head = &page_dir[page];
	    update_page_cnt++;
	}
        break;
    }

    case SCASH_MSG_DIST_BYTE: {
	bcopy(msg->body.dist_byte.data, msg->body.dist_byte.ptr, 
	    msg->body.dist_byte.length);

	reply_msg_body.dist_ok.key = msg->body.dist_byte.key;
	scash_message_send(SCASH_MSG_DIST_OK, msg->header.sender, 
			   sizeof(scash_msg_dist_ok_t), &reply_msg_body);
	break;
    }

    case SCASH_MSG_DIST_OK: {
	*msg->body.dist_ok.key = SCASH_ON;
	break;
    }

    case SCASH_MSG_OMNI_HOME: {
	extern void _scash_home_fo(int, caddr_t, int, unsigned char*, int);
	unsigned  char  cur_home[SCASH_MSG_HOME_LEN];

      bcopy( msg->body.scash_omni_home.cur_home, cur_home, SCASH_MSG_HOME_LEN);
      _scash_home_fo(msg->body.scash_omni_home.home,
		     msg->body.scash_omni_home.pagep,
		     msg->body.scash_omni_home.pagen,
		     cur_home,
		     0);
      break;
    }

    case SCASH_MSG_SCASH_HOME: {
      extern void scash_home_scash(int, caddr_t, int);

      scash_home_scash(msg->body.scash_scash_home.home,
		       msg->body.scash_scash_home.pagep,
		       msg->body.scash_scash_home.pagen);

      scash_message_send(SCASH_MSG_SCASH_HOME_OK, 
			 msg->header.sender, 
			 sizeof(int), 
			 &reply_msg_body);
      break;
    }

    case SCASH_MSG_SCASH_HOME_OK: {
      scash.home_ok++;
      break;
    }

    case SCASH_MSG_MAP_FILE:{
        bcopy(&msg->body.body, scash.map_file, 512);
	break;
    }

    case SCASH_MSG_PAGE_FILE:{
        bcopy(&msg->body.body, scash.page_file, 512);
	break;
    }

    case SCASH_MSG_DIRTY_PAGE:{
        scash_dp_insert(msg->body.fault_page);
	break;
    }

    case SCASH_MSG_PAGE_SEND:
    default:
	scash_prf("Receive Message [id=%d] from [%d]",
		  msg->header.id, msg->header.sender);
	scash_fatal("Receive Unknown or Invalid Message", 1);
	break;
    }

    if(scash_tlog_flag) tlog_MSGH_OUT();
}

char *scash_msg_name(int msg_id)
{
    switch(msg_id){
    case  SCASH_MSG_PAGE_REQ: return "PAGE_REQ";
    case  SCASH_MSG_PAGE_SEND: return "PAGE_SEND";
    case  SCASH_MSG_PAGE_SHARE_REQ: return "PAGE_SHARE_REQ";
    case  SCASH_MSG_PAGE_INVALIDATE: return "PAGE_INVALIDATE";
    case  SCASH_MSG_FLUSH_DIFF_SEND: return "FLUSH_DIFF_SEND";
    case  SCASH_MSG_DIFF_SIZE: return "DIFF_SIZE";
    case  SCASH_MSG_BARRIER_REQ: return "BARRIER_REQ";
    case  SCASH_MSG_BARRIER_OK: return "BARRIER_OK";
    case  SCASH_MSG_TRY_LOCK:  return "TRY_LOCK";
    case  SCASH_MSG_LOCK: return "MSG_LOCK";
    case  SCASH_MSG_LOCK_OK: return "LOCK_OK";
    case  SCASH_MSG_LOCK_BUSY: return "LOCK_BUSY";
    case  SCASH_MSG_UNLOCK: return "UNLOCK";
    case  SCASH_MSG_REFRESH: return "REFRESH";
    case  SCASH_MSG_FRESH: return "FRESH";
    case  SCASH_MSG_FLUSH: return "FLUSH";
    case  SCASH_MSG_MLOCK_ADDR: return "MLOCK_ADDR";
    case  SCASH_MSG_NEW_HOME: return "NEW_HOME";
    case  SCASH_MSG_HOME: return "HOME";
    case  SCASH_MSG_NEW_HOME_BASE:   return "NEW_HOME_BASE";
    case  SCASH_MSG_DIST_BYTE: return "DIST_BYTE";
    case  SCASH_MSG_DIST_OK: return "DIST_OK";
    case  SCASH_MSG_OMNI_HOME: return "OMNI_HOME";
    case  SCASH_MSG_SCASH_HOME: return "SCASH_HOME";
    case  SCASH_MSG_SCASH_HOME_OK: return "SCASH_HOME_OK";
    case  SCASH_MSG_MAP_FILE: return "MAP_FILE";
    case  SCASH_MSG_PAGE_FILE: return "PAGE_FILE";
    case  SCASH_MSG_DIRTY_PAGE: return "DIRTY_PAGE";
    case  SCASH_MSG_RAW_DATA: return "RAW_DATA";
    default: return "SCASH_MSG_???";
    }
}
