/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_page.c,v 1.4 2004-08-24 18:58:50 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <sys/mman.h>
#include  <errno.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"

page_t   *page_dir;    /* Shared  */
page_map_t *page_map;  /* Shared  */

u_int    update_page_cnt;
page_t*  update_page_head;
page_protocol_t  page_protocol;

#ifdef SCASH_INVALIDATE_POOL
scash_invalidate_pool_t  scash_invalidate_pool[SCASH_MAX_PE];
#endif

inline void
scash_page_cs_clear_all(int  page)
{
    int  i;

    for (i = 0 ; i < CS_LEN ; i++)
        page_dir[page].cs[i] = 0;
}

int
scash_page_new_home(int  page, int update_node)
{
    scash_page2ocount(page) = barrier_cnt+1;

    if (SCASH_PAGE_HOME_WRITE & page_dir[page].attr) {
	return scash_page2home(page);
    } 

    if (SCASH_PAGE_BASE_WRITE & page_dir[page].attr) {
	return scash_page2home(page) =  scash_page2base(page);
    } 
    return scash_page2home(page) =  update_node;
}

void 
scash_page_copy(int  fault_page,
		int  home)
{
    int  ret;
    scash_msg_t*  reply_msg;
    u_int        msg_length;

    if(scash_debug) 
	scash_prf("scash_page_copy fault=%d, home=%d ...",fault_page,home);

#ifdef SCASH_CHECK
    if (home == scash.pe)
        scash_fatal("scash_page_copy", 2);
#endif

    ret = scash_message_send2(SCASH_MSG_PAGE_REQ, home, 
			      sizeof(fault_page), &fault_page);
    for(;;) {
	ret = scash_message_recv(&reply_msg);
	if (ret == ENOBUFS)
	    continue;

	if (reply_msg->header.id == SCASH_MSG_PAGE_SEND) {
	    if (fault_page != reply_msg->body.copy_page.copy_page)  {
	        scash_prf("scash_page_copy:receive Invalid Page !! from [%d]",
			  home);
	    }
	    bcopy(reply_msg->body.copy_page.page, 
		  scash_page2shadowaddr(fault_page), SCASH_PAGE_SIZE);
	    OMSC_RELEASE_BUF();
	    break;
	} else {
	    scash_msg_handler(reply_msg);
	    OMSC_RELEASE_BUF();
	}
    }
}

void 
scash_page_rma_copy(int  fault_page, 
		    int  home)
{
    scash_msg_t*  msg_buff;
    u_int  length;
    int  ret;
    int  error;

    if(scash_debug) 
	scash_prf("scash_page_rma_copy fault=%d, home=%d ...",fault_page,home);

#ifdef SCASH_HOME_REPLACE
    scash_page2ht(fault_page) += 1;
#endif

    ret = scash_message_send2(SCASH_MSG_PAGE_SHARE_REQ, home, 
			      sizeof(fault_page), &fault_page);

    OMSC_READ(home,
	      scash.shadow_addr_handle[home]+(fault_page*SCASH_PAGE_SIZE),
	      scash.shadow_addr_handle[scash.pe]+(fault_page*SCASH_PAGE_SIZE),
	      SCASH_PAGE_SIZE);
    
    while(!OMSC_IS_READ_DONE()) {
	while ( ENOBUFS != scash_message_recv(&msg_buff)){
	    scash_msg_handler(msg_buff);
	    OMSC_RELEASE_BUF();
	}
    }
}

void 
scash_twin_create(int  fault_page)
{
    bcopy(scash_page2addr(fault_page), 
	  scash_page2twinaddr(fault_page), SCASH_PAGE_SIZE);
}

void
scash_page_init(int page_idx) 
{
    page_t*  page = page_dir + page_idx;
    
    scash_page2base(page_idx) = 0;
    scash_page2home(page_idx) = 0;
    scash_page2ocount(page_idx) = 0;

    scash_page_cs_clear_all(page_idx);
    page->attr = SCASH_PAGE_UNMAPPED;
    // page->fm = 0;
    page->update_next = NULL;
}

void
scash_page_cs_set(int  page, 
		  int  node)
{
    void*  copy_set;

    copy_set = (void*)(page_dir[page].cs + (node / 32));
    scash_inbox_try_lock(node % 32, copy_set);
    scash_inbox_unlock(0, (void*)(&dirty_page->cs_lock));
}


void
scash_page_invalidate(int page, int new_home)
{
    scash_msg_t*  send_buff;
    scash_msg_t*  recv_buff;
    u_int  recv_length;
    int   error;
    int   i;

    for (i = 0 ; i < scash.ps ; i++) {
        if (scash_page_cs(page, i)) {
	    if (i/scash.smp != scash.box && 
		i/scash.smp != new_home/scash.smp) {

	        while(OMSC_GET_SEND_BUF(i, (caddr_t*)&send_buff, 
					sizeof(scash_msg_header_t) + 
					sizeof(scash_msg_invalidate_t))
		      == ENOBUFS) {
		    if (ENOBUFS != scash_message_recv(&recv_buff)) {
		        scash_msg_handler(recv_buff);
			OMSC_RELEASE_BUF();
		    }
		}

		send_buff->header.id = SCASH_MSG_PAGE_INVALIDATE;
		send_buff->header.sender = scash.pe;
		send_buff->header.size = sizeof(scash_msg_header_t) 
		    + sizeof(scash_msg_invalidate_t);
		send_buff->body.invalidate.page = page;
		send_buff->body.invalidate.new_home = new_home;

		if ((error = OMSC_SEND()) != 0) {
		    scash_fatal("scash_page_invalidate: Send", error);
		}
	    }
	}
    }
    scash_page_cs_clear_all(page);
}

void 
scash_pa_init()
{
    int  i;
    int  page_f;
    char  dummy[1024];
    caddr_t   map_addr=0;
    unsigned  map_size;

    
    map_size = (SCASH_MAX_PAGE * sizeof(page_t))
	+ (SCASH_MAX_PAGE * sizeof(page_map_t))
	+ (SCASH_PAGE_SIZE * 8)
	+ sizeof(dirty_page_t);

    map_size = SCASH_PAGE_ROUNDUP(map_size);

    if(scash_debug) 
	scash_prf("scash_pa_init(pe=%d,proc=%d,smp=%d) map_size=0x%x...",
		  scash.pe,scash.proc,scash.smp,map_size);

    if (scash.proc == 0) {
	if ( OMSC_TMP_FILE( SCASH_PAGE_FILE, &page_f, scash.page_file) != 0 )
	    scash_fatal("scash_pa_init:can not create page table file", 1); 

	bzero(dummy, 1024);
	for (i = 0 ; i < map_size ; i+= 1024) {
	    write(page_f, dummy, 1024);
	}
	for (i = 1 ; i < scash.smp ; i++) {
	    scash_message_send2(SCASH_MSG_PAGE_FILE, scash.pe + i, 
				512, scash.page_file);
	}
    } else { /* scash.proc != 0 */
	while(strlen(scash.page_file)==0)
	    SCASH_POLL();
	page_f = open(scash.page_file, O_RDWR);
    }

    map_addr += SCASH_DEFAULT_ADDR - map_size;

    dirty_page = (dirty_page_t*)mmap(map_addr, 
				     map_size, 
				     PROT_READ|PROT_WRITE, 
				     MAP_FILE | MAP_SHARED | MAP_FIXED,
				     page_f, 
				     0);

    if(scash_debug) scash_prf("scash_pa_init: dirty_page=0x%x...",dirty_page);

#ifdef ENABLE_CHECKPOINT
    score_ckpt_checkpoint_file(scash.page_file);
#endif

    if (scash.proc == 0) scash_dp_init();

    page_dir = (page_t*)(&dirty_page[1]); 

    page_map = (page_map_t*)(page_dir + SCASH_MAX_PAGE + 128);
  
    if (scash.proc==0) {
	for (i=0 ; i < SCASH_MAX_PAGE ; i++)
	    scash_page_init(i);
    }

    if(scash_debug) scash_prf("scash_pa_init ... end");
}

void 
scash_page_mprot_correct(int  page)
{
    switch (page_dir[page].attr & SCASH_PAGE_ATTR_MASK) {
    case SCASH_PAGE_RW:{
	scash_mprot_rw(scash_page2addr(page));
	break;
    }
    case SCASH_PAGE_RO:{
	scash_mprot_ro(scash_page2addr(page));
	break;
    }
    case SCASH_PAGE_INVALID:{
	scash_mprot_none(scash_page2addr(page));
	break;
    }
    }
}

void
scash_prot_init()
{
    int   map_size;
    int   i;

    map_size = SCASH_MAX_PAGE / 8;

    page_protocol.default_map = (u_char*)malloc(map_size);

    for ( i = 0 ; i < map_size ; i++)
	page_protocol.default_map[i] = 255;

    page_protocol.cur_map = page_protocol.default_map;
    for ( i = 0 ; i < 1024 ; i++)
	page_protocol.map[ i ] = page_protocol.default_map;
}

void
scash_prot_read(caddr_t   pmp)
{
    int  pmf;
    int  map_size, i;
    int  barrier_cnt, barrier;
    int  buf_len;

    map_size = scash.dsm_pages / 8;
    if (scash.dsm_pages % 8)
	map_size++;

    pmf = open(pmp, O_RDONLY, 0644);
    read(pmf, (void *)&barrier_cnt, 4);

    if (barrier_cnt > 1024)
	scash_prf("too much protocol map");

    for (i=0 ; i < barrier_cnt ; i++) {
	read(pmf, (void *)&barrier, 4);
	read(pmf, (void *)&buf_len, 4);
	if (buf_len > map_size)
	    scash_prf("protocol map too large");
	page_protocol.map[barrier] = (u_char*)malloc( map_size );
	read(pmf, (void *)page_protocol.map[barrier], buf_len);
    }
}

volatile char page_lock_ref[SCASH_MAX_PAGE];

void
scash_page_lock_init()
{
    bzero((char*)page_lock_ref, SCASH_MAX_PAGE);
}

void 
scash_page_lock(int page,
		int flag)
{
    if (page_lock_ref[page]) {
	page_lock_ref[page]++; 
	return;
    }

    while( scash_inbox_try_lock(0, (void*)(page_map + page))) {
	if (flag) {
	    SCASH_POLL();
	}
    }
    page_lock_ref[page]++; 
}

void 
scash_page_unlock(int page)
{
    if (--page_lock_ref[page] == 0) {
	scash_inbox_unlock(0, (void*)(page_map + page));
    } 
    if (page_lock_ref[page] < 0) {
	scash_prf("Error scash_page_unlock %d",  page_lock_ref[page]);
    }
}

void
check_page_prot(int writer,
		int reader,
		int  page,
		int  f)
{
    if (page_dir[page].attr & SCASH_PAGE_RO || f) {
	int *lip;
	lip = (int*)scash_alloc_omni_shmem(1024);
  	scash_prf("<%d> [%d:%d]Error on Page Prot %d [%d:%d]CCC", 
		  scash.stats->barrier_cnt, writer, reader, page,
		  lip[512],lip[513]);
    }
}









