/* 
 * $Id: scash_private.h,v 1.7 2005-04-07 04:49:01 ojima Exp $
 */

#include <asm/bitops.h>
#include <linux/kernel.h>

#ifndef __scash_private_h
#define __scash_private_h

#define  SCASH_OFF (0)
#define  SCASH_ON  (1)

#define  SCASH_SIG_STACK_SIZE  (SCASH_PAGE_SIZE * 32)
#define  SCASH_WORD_IN_PAGE  (SCASH_PAGE_SIZE/SCASH_WORD_SIZE)

#define  SCASH_MAX_MSG  (SCASH_MAX_PE+8)
#define  PAGE_DIR_LEN  (SCASH_MAX_PE/8)
#define  CS_LEN  (SCASH_MAX_PE/32)
#define  SCASH_MAX_INVALIDATE_MSG    (128)
#define  SCASH_MAX_STATS    (4)
#define  SCASH_UPMAP_SIZE    (SCASH_MAX_PE * SCASH_MAX_PAGE)

#define  SCASH_VMREAD_MARK_OFFSET1  (scash.dsm_size + SCASH_UPMAP_SIZE - 4)
#define  SCASH_VMREAD_MARK_OFFSET2  (scash.dsm_size + SCASH_UPMAP_SIZE - 8)

#ifdef  SCASH_INVALIDATE_POOL
#define  SCASH_INVALIDATE_THRESHOLD    (scash.ps / 4)
#else
#define  SCASH_INVALIDATE_THRESHOLD    (0)
#endif
#define  SCASH_MAX_REFRESH    (32)

#define  SCASH_ROUNDUP(X,Y)    ((X+Y-1) & (~(Y-1)))
#define  SCASH_PAGE_ROUNDUP(X) ((X+ SCASH_PAGE_SIZE -1) & (~(SCASH_PAGE_SIZE-1)))
#define  SCASH_WORD_ROUNDUP(X) ((X+ SCASH_WORD_SIZE -1) & (~(SCASH_WORD_SIZE-1)))

#define  SCASH_TRUNC(X,Y)     ((X) & (~(Y-1)))
#define  SCASH_PAGE_TRUNC(X)  ((X) & (~(SCASH_PAGE_SIZE-1)))
#define  SCASH_WORD_TRUNC(X)  ((X) & (~(SCASH_WORD_SIZE-1)))

#ifdef __alpha__
#define  SCASH_DEFAULT_ADDR   ((unsigned long)0x80000000) 
#elif __ia64__
#define  SCASH_DEFAULT_ADDR   (0x6000000000000000 + 0x80000000)
#elif __x86_64__
#define  SCASH_DEFAULT_ADDR   (0x6000000000000000 + 0x80000000)
#else
#define  SCASH_DEFAULT_ADDR   ((1<<30) + (1<<27))
#endif

#define  SCASH_MAP_FILE   "scash.map" 
#define  SCASH_PAGE_FILE  "scash.pagetbl"
#define  SCASH_TRACE_FILE "/var/scash/trace/trace.out" 
#define  CPU_CLOCK   (1800) 

#define  SCASH_POLL()     scash_poll()
#define  SCASH_WAITE()    scash_waite()

#ifdef  SCASH_STATS
#define  SCASH_RDTSC(X)   X = scash_rdtsc()
#else
#define  SCASH_RDTSC(X)          {}
#endif

#ifdef SCASH_IDLE_DETECT
#define  SCASH_IDLE()   scash_idle()
#define  SCASH_BUSY()   scash_busy()
#else
#define  SCASH_IDLE()   {}
#define  SCASH_BUSY()   {}
#endif

/*
 * SCASH/OMNI macros
 */
#define   SCASH_OMNI_SHMEM_SIZE  (4 * 4096)

/*
 * SCASH OPTION
 */
#define  SCASH_OPT_BARRIER_MASK        0x07
#define  SCASH_OPT_BARRIER_INVALIDATE  0x01
#define  SCASH_OPT_BARRIER_UPDATE      0x02
#define  SCASH_OPT_BARRIER_DYNAMIC_HOME  0x03
#define  SCASH_OPT_BARRIER_HYBRID      0x04
#define  SCASH_OPT_RMA_PAGE_COPY     0x10

#ifdef SCASH_HOME_REPLACE
#define  SCASH_OPT_DEFAULT      (SCASH_OPT_BARRIER_DYNAMIC_HOME | SCASH_OPT_RMA_PAGE_COPY)
#else
#define  SCASH_OPT_DEFAULT      (SCASH_OPT_BARRIER_INVALIDATE | SCASH_OPT_RMA_PAGE_COPY) 
#endif

#define  scash_opt_barrier()   (scash_option & 0x07)
#define  scash_opt_page_copy()   (scash_option & 0x10)

/*
 * Dirty Page 
 */
typedef  struct  dirty_page_s{
    int    total;
    volatile int    lock_bit;
    volatile char   inbox_lock[16];
    volatile int    cs_lock;
    volatile int    attr_lock;
    unsigned int    dp[SCASH_MAX_PAGE];
}dirty_page_t;

/*
 * Diff 
 */
#define  SCASH_RMA_DIFF_PAGES    (8)
#define  SCASH_RMA_DIFF_SIZE     (SCASH_RMA_DIFF_PAGES * SCASH_PAGE_SIZE)

typedef struct {
    u_int   pe:10;
    u_int   id:22;
}diff_id_t;

#ifdef __ia64__
typedef struct  diff_block_header_s{
    u_int   length:16;
    u_int   pos:16;
} diff_block_header_t;
#else
typedef struct  diff_block_header_s{
    u_int   pad:8;
    u_int   length:12;
    u_int   pos:12;
} diff_block_header_t;
#endif

typedef struct diff_block_s {
    diff_block_header_t    header;
#ifdef __ia64__
    u_int     data[4096]; /* might be SCASH_WORD_IN_PAGE : needs check */
#else
    u_int     data[2048];
#endif /* __ia64__ */
} diff_block_t;

typedef struct diff_min_block_s {
    diff_block_header_t    header;
    u_int     data[1];
} diff_min_block_t;

typedef struct diff_s {
    int   length;
    diff_min_block_t  block[1024];
} diff_t;

typedef  struct  _s_diff_index_s {
    int   diff_num;
    diff_block_header_t  db[1024]; 
} _s_diff_index_t;

/*
 * DSM Page Array
 */
typedef  struct _page_s {
    volatile u_int  cs[CS_LEN];
    struct _page_s  *update_next; 
    volatile u_char  attr;
    short    diff[2];
} page_t;

typedef struct _page_map_s {
#if (defined(__i386__) | defined(__alpha__) | defined(__ia64__) | defined(__x86_64__))
    unsigned lock:1;     /* LSB is used for Lock bit */
    unsigned ht:8;       /* Home Track */
#endif
    unsigned base:8;
    unsigned home:8;
    unsigned ocount:7;
#if (defined(__sparc__))
    unsigned ht:8;     /* Home Track */
    unsigned lock:1;     /* LSB is useed for Lock bit */
#endif
} page_map_t;

typedef  struct _page_protocol_s {
    u_char*  default_map;
    u_char*  map[1024];
    u_char*  cur_map;
    int      flag;
} page_protocol_t;

#define  SCASH_PAGE_UNMAPPED      0x00
#define  SCASH_PAGE_INVALID       0x01
#define  SCASH_PAGE_RO            0x02
#define  SCASH_PAGE_RW            0x03

#define  SCASH_PAGE_ATTR_MASK     0x03
#define  SCASH_PAGE_ATTR(X)       ((X) & SCASH_PAGE_ATTR_MASK)
#define  SCASH_PAGE_SET_RW(X)         ((X) |= SCASH_PAGE_RW)
#define  SCASH_PAGE_SET_RO(X)         ((X) = ((X) & (~SCASH_PAGE_ATTR_MASK)) | SCASH_PAGE_RO)
#define  SCASH_PAGE_SET_INVALID(X)    ((X) = ((X) & (~SCASH_PAGE_ATTR_MASK)) | SCASH_PAGE_INVALID)
#define  SCASH_PAGE_SET_UNMAPPED(X)   ((X) = ((X) & (~SCASH_PAGE_ATTR_MASK)) | SCASH_PAGE_UNMAPPED)

#define  SCASH_PAGE_TWIN          0x04
#define  SCASH_PAGE_SET_TWIN(X)   (X |= SCASH_PAGE_TWIN)
#define  SCASH_PAGE_CLEAR_TWIN(X)   (X &= ~SCASH_PAGE_TWIN)

#define  SCASH_PAGE_UPDATE          0x08
#define  SCASH_PAGE_SET_UPDATE(X)   (X |= SCASH_PAGE_UPDATE)
#define  SCASH_PAGE_CLEAR_UPDATE(X) (X &= ~SCASH_PAGE_UPDATE)
#define  SCASH_PAGE_GET_UPDATE(X)   (X & SCASH_PAGE_UPDATE)

#define  SCASH_PAGE_BASE_WRITE          0x10
#define  SCASH_PAGE_SET_BASE_WRITE(X)   (X |= SCASH_PAGE_BASE_WRITE)
#define  SCASH_PAGE_CLEAR_BASE_WRITE(X)   (X &= ~SCASH_PAGE_BASE_WRITE)

#define  SCASH_PAGE_HOME_WRITE          0x20
#define  SCASH_PAGE_SET_HOME_WRITE(X)   (X |= SCASH_PAGE_HOME_WRITE)
#define  SCASH_PAGE_CLEAR_HOME_WRITE(X) (X &= ~SCASH_PAGE_HOME_WRITE)

#define  SCASH_PAGE_HOME_BASE          0x40
#define  SCASH_PAGE_SET_HOME_BASE(X)   (X |= SCASH_PAGE_HOME_BASE)
#define  SCASH_PAGE_CLEAR_HOME_BASE(X) (X &= ~SCASH_PAGE_HOME_BASE)

#define  SCASH_PAGE_FM_MASK          (0xff)
#define  SCASH_PAGE_FM(X,Y)          ((X) & (1<<Y))
#define  SCASH_PAGE_FM_SET(X,Y)      ((X) |= (1 << Y))
#define  SCASH_PAGE_FM_CLEAR(X, Y)   ((X) &= (~(1 << Y)))
#define  SCASH_PAGE_FM_CLEAR_ALL(X)  ((X) &= 0x00)

typedef  struct  scash_invalidate_pool_s {
    u_int  cnt;
    u_int  page[SCASH_MAX_INVALIDATE_MSG];
    int  new_home[SCASH_MAX_INVALIDATE_MSG];
} scash_invalidate_pool_t;

#ifdef  __cplusplus
extern "C" {
#endif
    void  scash_prot_read(caddr_t);
    void  mpc_init_stats();
    int  scash_shmem_alloc(caddr_t*, u_int);
    void scash_flush_all();
    void scash_flush_all_home();
    void scash_flush_all_hybrid();
    void scash_flush_update_all();
    void scash_flush_update_all_new();
    void scash_flush_invalidate_pool();
    void scash_update_barrier(int);
    int  scash_barrier_f();
    void scash_barrier_f_set(int);
    void scash_barrier_sync(int, int);
    void scash_barrier_tree_node(int, int, int*, int*);
    void scash_page_invalidate(int, int);
    void scash_page_invalidate_all();
    void scash_page_cs_set(int, int);
    void scash_page_cs_clear(int, int);
    int  scash_page_new_home(int, int);

    void scash_pe_init();
    void scash_mpc_init();
    void scash_score_init();
    void scash_page_copy(int, int);
    void scash_page_rma_copy(int, int);
    void scash_page_init(int);
    void scash_twin_create(int);
    void scash_pa_init();
    void scash_prot_init();
    void scash_dp_init();
    void scash_dp_insert(u_int);
    void scash_wn_init();
    void scash_interval_init();
    void scash_invalidate_pool_init();
    void scash_sig_handler_set();
    int  scash_diff_create(int, diff_t*);
    int  scash_diff_copy_create(int, diff_t*);
    int  scash_diff_create_range(int, int, int, diff_t*);
    int  scash_diff_copy_create_range(int, int, int, diff_t*);
    int  scash_diff_rmw(int);
    int  scash_diff_size(diff_t*);
    int  scash_barrier_cnt();
    void scash_barrier_a2a(int);

    int  scash_message_send(int, int, int, void*);
    int  scash_message_send2(int, int, int, void*);
    int  scash_message_multicast(int, int*, int, int,void*);
    int  scash_message_multicast2(int, int*, int, int,void*);

    void   scash_recv_lock();
    void   scash_recv_try_lock();
    void   scash_recv_unlock();

    int    scash_vmread_done();
    
    void   scash_recv_refresh(int, struct scash_msg_refresh_s*);
    void   scash_memory_lock_init();

    void   scash_receive_update_word(caddr_t, int);
    void   scash_diff_update(caddr_t, int);

    void   scash_receive_update(int, caddr_t, int);
    void   scash_receive_diff_update(int, caddr_t, u_int);

    void   scash_page_lock(int, int);
    void   scash_page_unlock(int);

    int    sc_barrier(void);
#ifdef  __cplusplus
}
#endif

extern  page_t  *page_dir;
extern  page_map_t  *page_map;
extern  dirty_page_t  *dirty_page;

extern  unsigned short  barrier_cnt;

extern  u_int    update_page_cnt;
extern  page_t*  update_page_head;
extern  page_protocol_t  page_protocol;

extern  dirty_page_t  *dirty_page;
extern  scash_invalidate_pool_t  scash_invalidate_pool[];

#define  scash_mprot_rw(addr) 	mprotect(addr, SCASH_PAGE_SIZE, PROT_READ | PROT_WRITE)
#define  scash_mprot_ro(addr) 	mprotect(addr, SCASH_PAGE_SIZE, PROT_READ)
#define  scash_mprot_none(addr) 	mprotect(addr, SCASH_PAGE_SIZE, PROT_NONE)
#define  scash_mprot(page) 	(page_dir[page].attr & SCASH_PAGE_ATTR_MASK)

#define  scash_page2base(P) 	 (page_map[P].base)
#define  scash_page2home(P) 	 (page_map[P].home)
#define  scash_page2ocount(P) 	 (page_map[P].ocount)
#define  scash_page2ht(P) 	 (page_map[P].ht)
#define  scash_page2addr(P) 	 (scash.dsm_addr + (P * SCASH_PAGE_SIZE))
#define  scash_page2twinaddr(P) 	 (scash.twin_addr + (P * SCASH_PAGE_SIZE))
#define  scash_page2shadowaddr(P) (scash.shadow_addr + (P * SCASH_PAGE_SIZE))
#define  scash_addr2page(P)       ((((caddr_t)P) - scash.dsm_addr)/ SCASH_PAGE_SIZE)
#define  scash_pageoffset(P)      ((u_int)(P) & (SCASH_PAGE_SIZE - 1))

#define  scash_page_protocol(P)       (page_protocol.cur_map[P/8] & (1 << (P % 8)))
#define  scash_prot_is_update(P)      (page_protocol.cur_map[P/8] & (1 << (P % 8)))
#define  scash_page_protocol_set(P)   (page_protocol.cur_map[P/8] |= (1 << (P % 8)))
#define  scash_page_protocol_clear(P) (page_protocol.cur_map[P/8] &= ~(1 << (P % 8)))

#define  scash_update_map(P, N)      scash.update_map_addr[N * scash.dsm_pages + P]
#define  scash_update_map_offset(P, N)      ((scash.dsm_size + N * scash.dsm_pages + P) & ~(0x03))
#define  scash_update_map_clear(P, N)      scash.update_map_addr[N * scash.dsm_pages + P] = 0;
#define  scash_update_map_on(P, N)      scash.update_map_addr[N * scash.dsm_pages + P] = 1;

static inline int
scash_page_cs(int  page, 
	      int  node) 
{
    return (page_dir[page].cs[node/32]) & (1 << (node % 32));
}

#include "lock_macro.h"

/* #include <pm_lock.h> */
#ifndef smp_mb__before_clear_bit
#define smp_mb__before_clear_bit()
#endif

#ifndef smp_mb__after_clear_bit
#define smp_mb__after_clear_bit()
#endif

static inline void
scash_inbox_init_lock(int  bp, void  *lockp)
{
  clear_bit(bp, lockp);
}

static  inline  void
scash_inbox_unlock(int  bp, void  *lockp)
{
  clear_bit(bp, lockp);
}

static  inline  void
scash_inbox_spin_lock(int  bp, void  *lockp)
{
  while ( test_and_set_bit(bp, lockp) )
    ;
}

static  inline  int
scash_inbox_try_lock(int  bp, void  *lockp)
{
  return test_and_set_bit(bp, lockp);
}

static inline int
scash_flushed_page(int  p)
{
  return( scash.flushed_page[p/32] & (1 << (p%32)) );
}

static inline void
scash_flushed_page_set(int  p)
{
  scash.flushed_page[p/32] |= (1 << (p%32));
}

static inline void
scash_flushed_page_clear(int  p)
{
  scash.flushed_page[p/32] &= ~(1 << (p%32));
}

#endif  /* __scash_private_h */









