/* 
 * $Id: scash_queue.h,v 1.1 2004-02-27 18:19:11 msato Exp $
 */

#ifndef __scash_queue_h
#define __scash_queue_h

#define QUEUE_HEAD_S(name, type)				\
struct name {							\
	struct type *qh_first;	/* first element */		\
	struct type **qh_last;	/* addr of last next element */	\
}

#define QUEUE_HEAD_T(name, type)			\
struct name {						\
	type *qh_first;	/* first element */		\
	type **qh_last;	/* addr of last next element */	\
}

#define QUEUE_ENTRY(type)					\
struct {							\
	struct type *qe_next;	/* next element */		\
}

/*
 * Simple queue functions.
 */
#define	QUEUE_INIT(head) {					\
	(head)->qh_first = NULL;				\
	(head)->qh_last = &(head)->qh_first;			\
}

#define QUEUE_INSERT_HEAD(head, elm, field) {			\
	if (((elm)->field.qe_next = (head)->qh_first) == NULL)	\
		(head)->qh_last = &(elm)->field.qe_next;	\
	(head)->qh_first = (elm);				\
}

#define QUEUE_INSERT_TAIL(head, elm, field) {			\
	(elm)->field.qe_next = NULL;				\
	*(head)->qh_last = (elm);				\
	(head)->qh_last = &(elm)->field.qe_next;		\
}

#define QUEUE_REMOVE_HEAD(head, elm, field) {			\
	if (((head)->qh_first = (elm)->field.qe_next) == NULL)	\
		(head)->qh_last = &(head)->qh_first;		\
}

#define scash_first_elm(head)	(head.qh_first)
#define	scash_queue_init(head)   QUEUE_INIT(head)
#define scash_enqueue(head, elm, field)    QUEUE_INSERT_TAIL(head, elm, field)
#define scash_dequeue(head, elm, field)    QUEUE_REMOVE_HEAD(head, elm, field)

/*
 * Circular queue definitions.
 */
#define CIRCLEQ_HEAD_S(name, type)					\
struct name {								\
	struct type *cqh_first;		/* first element */		\
	struct type *cqh_last;		/* last element */		\
}

#define CIRCLEQ_HEAD_T(name, type)					\
struct name {								\
	type *cqh_first;		/* first element */		\
	type *cqh_last;		        /* last element */		\
}

#define CIRCLEQ_ENTRY(type)						\
struct {								\
	struct type *cqe_next;		/* next element */		\
	struct type *cqe_prev;		/* previous element */		\
}

/*
 * Circular queue functions.
 */
#define	CIRCLEQ_INIT(head) {						\
	(head)->cqh_first = (void *)(head);				\
	(head)->cqh_last = (void *)(head);				\
}

#define	CIRCLEQ_INIT_V(head) {						\
	head.cqh_first = (void *)(head);				\
	head.cqh_last = (void *)(head);				\
}

#define CIRCLEQ_INSERT_AFTER(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm)->field.cqe_next;		\
	(elm)->field.cqe_prev = (listelm);				\
	if ((listelm)->field.cqe_next == (void *)(head))		\
		(head)->cqh_last = (elm);				\
	else								\
		(listelm)->field.cqe_next->field.cqe_prev = (elm);	\
	(listelm)->field.cqe_next = (elm);				\
}

#define CIRCLEQ_INSERT_BEFORE(head, listelm, elm, field) {		\
	(elm)->field.cqe_next = (listelm);				\
	(elm)->field.cqe_prev = (listelm)->field.cqe_prev;		\
	if ((listelm)->field.cqe_prev == (void *)(head))		\
		(head)->cqh_first = (elm);				\
	else								\
		(listelm)->field.cqe_prev->field.cqe_next = (elm);	\
	(listelm)->field.cqe_prev = (elm);				\
}

#define CIRCLEQ_INSERT_HEAD(head, elm, field) {				\
	(elm)->field.cqe_next = (head)->cqh_first;			\
	(elm)->field.cqe_prev = (void *)(head);				\
	if ((head)->cqh_last == (void *)(head))				\
		(head)->cqh_last = (elm);				\
	else								\
		(head)->cqh_first->field.cqe_prev = (elm);		\
	(head)->cqh_first = (elm);					\
}

#define CIRCLEQ_INSERT_TAIL(head, elm, field) {				\
	(elm)->field.cqe_next = (void *)(head);				\
	(elm)->field.cqe_prev = (head)->cqh_last;			\
	if ((head)->cqh_first == (void *)(head))			\
		(head)->cqh_first = (elm);				\
	else								\
		(head)->cqh_last->field.cqe_next = (elm);		\
	(head)->cqh_last = (elm);					\
}

#define	CIRCLEQ_REMOVE(head, elm, field) {				\
	if ((elm)->field.cqe_next == (void *)(head))			\
		(head)->cqh_last = (elm)->field.cqe_prev;		\
	else								\
		(elm)->field.cqe_next->field.cqe_prev =			\
		    (elm)->field.cqe_prev;				\
	if ((elm)->field.cqe_prev == (void *)(head))			\
		(head)->cqh_first = (elm)->field.cqe_next;		\
	else								\
		(elm)->field.cqe_prev->field.cqe_next =			\
		    (elm)->field.cqe_next;				\
}

#define	scash_cq_init(head)                             CIRCLEQ_INIT(head)
#define	scash_cq_rm(head, elm, field)                   CIRCLEQ_REMOVE(head, elm, field)
#define scash_cq_ins_after(head, listelm, elm, field)   CIRCLEQ_INSERT_AFTER(head, listelm, elm, field)
#define scash_cq_ins_before(head, listelm, elm, field)  CIRCLEQ_INSERT_BEFORE(head, listelm, elm, field)

#endif /* #ifndef __scash_queue_h */
