/* 
 * $PCCC_Release$
 * $PCCC_Copyright$
 */
static char rcsid[] = "$Id: scash_score.c,v 1.2 2004-10-12 05:25:42 msato Exp $";

#include  <stdio.h>
#include  <fcntl.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <errno.h>
#include  <signal.h>
#include  <sys/types.h>
#include  <errno.h>

#include <score.h>
#include <score_nodeset.h>
#include <score_options.h>
#include <score_trace.h>

#include  <pm.h>

#include  <score_resource.h>

SCORE_RSRC_ZEROCOPY;
SCORE_RSRC_NUM_NETS(1);

extern   pmContext*  score_pmnet[];
extern   int  score_num_pmnet;
extern   int  score_self_node;
extern   int  score_self_proc;
extern   int  score_self_host;
extern   int  score_num_proc;
extern   int  score_num_node;
extern   int  score_num_host;

pmContext*   scash_pmnet;

#ifdef __SCORE_2__
extern "C" {
    char* _score_get_opt(char*);
};
#endif

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"


#include <score.h>
#include <score_options.h>
#include <score_signal.h>
#include <score_dumpregs.h>

void score_attach_debugger( char*, int );
void score_set_signal_handler( int, void(*)() );
void score_setup_sysycall( void );

extern int  sc_barrier(void);

/* scash exception handler */
static void scash_exception( int sig, int code, 
			     struct sigcontext *sc, char* addr)
{
  if(score_get_opt("regdump") != NULL) {
#ifdef i386
    struct sigcontext *sbx = (struct sigcontext *)&code;
    score_dump_registers(sig, code, sbx);
#endif /* i386  */
#ifdef __alpha__
    score_dump_registers(sig, code, sc);
#endif
  }
  score_attach_debugger( "Exception signal", sig );
  fclose( stdout );
  fclose( stderr );
  _exit( sig );			/* do not call exit() */
  /* do not reach here */
}

static void set_exception_handlers( void ) {
  score_set_signal_handler(SIGHUP , (void(*)(void)) scash_exception);
  score_set_signal_handler(SIGILL , (void(*)(void)) scash_exception);
  score_set_signal_handler(SIGFPE , (void(*)(void)) scash_exception);
  score_set_signal_handler(SIGBUS , (void(*)(void)) scash_exception);
  score_set_signal_handler(SIGSEGV , (void(*)(void)) scash_exception);
#ifndef linux
  score_set_signal_handler(SIGSYS , (void(*)(void)) scash_exception);
#endif /* linux */
}

void 
scash_score_initialize(int argc, 
		       char **argv)
{
  munlockall();
  score_initialize(argc, argv);		/* initialize score */
  sc_barrier();
}

int  score_runtime_resource() { return 0; }

void
scash_score_init()
{
    int   ret;
    pmContextConfig  pm_config;
    int  i, mtu;

    scash.pe = score_self_node;
    scash.ps = score_num_node;
    scash.smp = score_num_proc;   /* num of proc in box */
    scash.proc = score_self_proc; /* proc num in box */
    scash.box = scash.pe/scash.smp;

    if (scash.ps > SCASH_MAX_PE) {
        scash_fatal("scash_score_init::# of PE is too large");
    }

    scash_pmnet = score_pmnet[0];
    scash.mtu = 1 << 30;

    for (i = 0 ; i < scash.ps ; i++) {
        if (i == scash.pe)
	    continue;
	pmGetMtu(scash_pmnet, i, &mtu);

	if (mtu < scash.mtu)
	    scash.mtu = mtu;
    }
    ret = pmCacheSize((2 * SCASH_MAX_PAGE * SCASH_PAGE_SIZE) + 
		      SCASH_UPMAP_SIZE +
		      SCASH_PAGE_ROUNDUP((sizeof(page_map_t) * SCASH_MAX_PAGE)));

    ret = pmGetContextConfig(scash_pmnet, &pm_config);
    if (!(pm_config.option & (PM_OPT_REMOTE_READ | PM_OPT_REMOTE_WRITE))) {
	scash_fatal("scash_score_init::No remote memory access facility");
    }
}

/* called before pe_init */
int 
scash_score_num_proc()
{
    return score_num_node;
}

void 
scash_pm_release_buffer()
{
    (void) pmReleaseReceiveBuffer(scash_pmnet);
}

int 
scash_pm_is_send_done()
{
    return pmIsSendDone(scash_pmnet) != EBUSY;
}

void 
scash_pm_write(int dst_node, addrHandle dst_addr,addrHandle src_addr,int size)
{
    while( pmWrite(scash_pmnet, dst_node, dst_addr, src_addr,size))/* */;
}

int 
scash_pm_is_write_done()
{
    return pmIsWriteDone(scash_pmnet) != EBUSY;
}

int 
scash_pm_read(int src_node, addrHandle src_addr, addrHandle dst_addr, int size)
{
    while( pmRead(scash_pmnet, src_node, src_addr, dst_addr,size))/* */;
}

int 
scash_pm_is_read_done()
{
    return pmIsReadDone(scash_pmnet) != EBUSY;
}

int 
scash_pm_get_send_buf(int dst, caddr_t *bufpp, int size)
{
    return pmGetSendBuffer(scash_pmnet,dst,bufpp,size);
}

int 
scash_pm_trunc_buf(int size)
{
    return pmTruncateBuffer(scash_pmnet,size);
}

int 
scash_pm_send()
{
    return pmSend(scash_pmnet);
}

int 
scash_pm_recv(caddr_t *msg, u_int *size)
{
    return pmReceive(scash_pmnet,msg,size);
}

/* notify idle & busy */
void 
scash_idle()
{
    score_become_idle();
}

scash_busy()
{
    score_become_busy();
}

/* memory lock function */
int
scash_mlock_cnt()
{
    int   i;
    int   cnt = 0;

    for (i = 0 ; i < scash.ps ; i++)
	if (scash.shadow_addr_handle[i])
	    cnt++;

    return cnt;
}

void
scash_memory_lock_init()
{
    int  ret;
    u_int  lock_size;
    msg_body_t msg_body;

    lock_size = SCASH_PAGE_ROUNDUP(scash.dsm_size);

    ret = pmMLock( scash_pmnet,
		   PM_NODE_ANY,
		   scash.shadow_addr, 
		   lock_size, 
		   &scash.shadow_addr_handle[scash.pe]);
    if (ret) {
        scash_fatal("scash_mlock_init:can not lock shadow area");
    }

    ret = pmMLock( scash_pmnet, 
		   PM_NODE_ANY, 
		   (caddr_t)page_map , 
		   sizeof(page_map_t) * SCASH_MAX_PAGE, 
		   &scash.page_map_addr_handle[scash.pe]);

    if (ret) 
        scash_fatal("scash_mlock_init:can not lock page map area");

    msg_body.addr_handle.shadow = scash.shadow_addr_handle[scash.pe];
    msg_body.addr_handle.page_map = scash.page_map_addr_handle[scash.pe];

    ret = scash_message_broadcast(SCASH_MSG_MLOCK_ADDR, 
				  sizeof(scash_msg_addr_handle_t), &msg_body);

    while (scash_mlock_cnt() != scash.ps) {
	scash_wait();
    }
}

void scash_init_stats_all() { }
u_int64 scash_rdtsc() { }

/* EOF */




