#include  <stdio.h>
#include  <unistd.h>
#include  <stdlib.h>
#include  <string.h>
#include  <errno.h>
#include  <sys/types.h>

#include  "scash.h"
#include  "scash_private.h"
#include  "scash_message.h"
#include  "tlog-omsc.h"
#include  "tlog.h"

void scash_tlog_initialize()
{
    double t1,t2,t0,t_diff;
    int i;

    if(scash_debug) scash_prf("scash_tlog_initialize ...");

    /* OMSC_BARRIER(); */
    /* scash_barrier_sync(1,1); */
    if(scash.pe == 0){
	tlog_init("scash.log");
	for(i = 1; i < scash.ps; i++){
	    t1 = tlog_get_time();
	    scash_send_raw_data(i,0,NULL);
	    scash_recv_raw_data(sizeof(double),&t0);
	    t2 = tlog_get_time();
	    t_diff = t1+(t2-t1)/2.0 - t0;
	    scash_send_raw_data(i,sizeof(double),&t_diff);
	    /* scash_prf("i=%d,t1=%g,t2=%g,t0=%g,diff=%g",
	       i,t1,t2,t0,t_diff); */
	}
    } else {
	tlog_init0();
	scash_recv_raw_data(0,NULL);
	t0 = tlog_get_time();
	/* scash_prf("slave t0=%g",t0); */
	scash_send_raw_data(0,sizeof(double),&t0);
	scash_recv_raw_data(sizeof(double),&t_diff);
	/* scash_prf("t_diff=%g",t_diff); */
	tlog_adjust_time(t_diff);
    }
    /* OMSC_BARRIER(); */
    scash_barrier_sync(2,1);
    tlog_log(scash.pe,TLOG_START);
    /* scash_prf("tlog_init_end t=%g",tlog_get_time()); */
}

void scash_tlog_finialize()
{
    TLOG_BLOCK *bp;
    char buf[TLOG_BLOCK_SIZE];
    int size,i;

    if(scash_debug) scash_prf("scash_tlog_finalize ...");
    tlog_log(scash.pe,TLOG_END);
    
    /* OMSC_BARRIER(); */
    scash_barrier_sync(1,1);
    if(scash.pe == 0){	/* master */
	tlog_file_open();
	/* first dump my data */
	for(bp = tlog_handle_table[scash.pe].block_top; 
	    bp != NULL;  bp = bp->next)
	    tlog_dump_data((char *)bp->data);
	for(i = 1; i < scash.ps; i++){
	    while(1){
		scash_send_raw_data(i,0,NULL);
		scash_recv_raw_data(sizeof(int),&size);
		/* scash_prf("dump from pe=%d, size=%d",i,size); */
		if(size == 0) break;
		scash_recv_raw_data(TLOG_BLOCK_SIZE,buf);
		tlog_dump_data(buf);
	    }
	}
	tlog_file_close();
    } else {
	/* send log */
	size = TLOG_BLOCK_SIZE;
	for(bp = tlog_handle_table[scash.pe].block_top; 
	    bp != NULL;  bp = bp->next){
	    scash_recv_raw_data(0,NULL);
	    scash_send_raw_data(0,sizeof(int),&size);
	    scash_send_raw_data(0,size,(char *)bp->data);
	}
	size = 0;
	scash_recv_raw_data(0,NULL);
	scash_send_raw_data(0,sizeof(int),&size);
    }
    /* OMSC_BARRIER(); */
    scash_barrier_sync(2,1);
}

void tlog_START_EVENT(int id) {  }

void tlog_BARRIER_IN(int id) { tlog_log(scash.pe,TLOG_BARRIER_IN); }
void tlog_BARRIER_OUT(int id) { tlog_log(scash.pe,TLOG_BARRIER_OUT); }

void tlog_MSGH_IN(int id) { tlog_log(scash.pe,TLOG_MSGH_IN); }
void tlog_MSGH_OUT(int id) { tlog_log(scash.pe,TLOG_MSGH_OUT); }

void tlog_SEND_IN(int id) { tlog_log(scash.pe,TLOG_SEND_IN); }
void tlog_SEND_OUT(int id) { tlog_log(scash.pe,TLOG_SEND_OUT); }

void tlog_PAGEF_IN(int id) { tlog_log(scash.pe,TLOG_PAGEF_IN); }
void tlog_PAGEF_OUT(int id) { tlog_log(scash.pe,TLOG_PAGEF_OUT); }

void tlog_BSYNC_IN(int id) {
    switch(id){
    case 1:
	tlog_log(scash.pe,TLOG_BSYNC_IN);
	break;
    case 2:
	tlog_log(scash.pe,TLOG_BSYNC2_IN);
	break;
    case 3:
	tlog_log(scash.pe,TLOG_BSYNC3_IN);
	break;
    case 4:
	tlog_log(scash.pe,TLOG_BSYNC4_IN);
	break;
    }
}

void tlog_BSYNC_OUT(int id) {
    switch(id){
    case 1:
	tlog_log(scash.pe,TLOG_BSYNC_OUT);
	break;
    case 2:
	tlog_log(scash.pe,TLOG_BSYNC2_OUT);
	break;
    case 3:
	tlog_log(scash.pe,TLOG_BSYNC3_OUT);
	break;
    case 4:
	tlog_log(scash.pe,TLOG_BSYNC4_OUT);
	break;
    }
}

void tlog_FLUSH_ALL_IN(int id) { tlog_log(scash.pe,TLOG_FLUSH_ALL_IN); }
void tlog_FLUSH_ALL_OUT(int id) { tlog_log(scash.pe,TLOG_FLUSH_ALL_OUT); }

void tlog_PAGE_INV_IN(int id) { tlog_log(scash.pe,TLOG_PAGE_INV_IN); }
void tlog_PAGE_INV_OUT(int id) { tlog_log(scash.pe,TLOG_PAGE_INV_OUT); }

void tlog_PAGEF_EVENT(int id) { tlog_log(scash.pe,TLOG_PAGEF_EVENT); }
void tlog_RMA_EVENT(int id) { tlog_log(scash.pe,TLOG_RMA_EVENT); }

void tlog_MPI_SEND_EVENT(int id) { tlog_log(scash.pe,TLOG_MPI_SEND_EVENT); }
void tlog_MPI_RECV_EVENT(int id) { tlog_log(scash.pe,TLOG_MPI_RECV_EVENT); }
void tlog_MPI_RMA_EVENT(int id) { tlog_log(scash.pe,TLOG_MPI_RMA_EVENT); }
