#!/usr/bin/perl
#
# npb-score.pl (Nas Parallel Benchmark(SCore + SCASH) Test Tool)
#
# $Id: npb-score.pl,v 1.1.1.1 2004-02-06 18:16:53 msato Exp $
# $RWC_Release$
# $RWC_Copyright$
#

#
# Constant Definition
#

# Benchmark suits name
$benchmark_name = "NPB2.3-omp";

# All Benchmark name list
$all_bench = "bt,cg,ep,ft,is,lu,mg,sp";

# All Class name list
$all_class = "S,W,A,B,C";

# Procs, Threads number default value decision
if (defined $ENV{OMPC_NUM_PROCS}) {
    $default_procs = $ENV{OMPC_NUM_PROCS};
} else {
    $default_procs = 1;
};

if (defined $ENV{OMP_NUM_THREADS}) {
    $default_thread = $ENV{OMP_NUM_THREADS};
} else {
    $default_thread = 1;
};

# Byte order swap flag decision
if (defined $ENV{OMPF77_BYTE_SWAP_IO}) {
    $default_byte_swap = $ENV{OMPF77_BYTE_SWAP_IO};
} else {
    open PIPE, "uname -a|";
    while (<PIPE>) {
        if (($_ =~ /i86pc/) || ($_ =~ /alpha/)) {
            $default_byte_swap = "yes";
        }
        else {
	    $default_byte_swap = undef;
        }
    }
};

# Table for Default values
%default_table = (
		  BENCHDIR	=>	"./",      # Benchmark Direcotry
		  BENCHMARKS	=>	$all_bench,# Benchmark Name
		  CLASSES	=>	"S",       # Class name
		  PROCS	=>	$default_procs,    # Processor number
		  THREADS =>	$default_thread,   # Thread number
		  MAKE =>	"/usr/bin/make",   # command make
		  ACTION =>	"verify",          # Benchtool master action
		  VERBOSE =>	0,                 # Verbose message output flag
		  PERFVIEW =>   0,                 # Perfview format output flag
		  PERFVIEW_FILE => "./npb",        # Perfview format output filename prefix
		  BYTE_SWAP     => $default_byte_swap, # Byte order swap flag
		  );

#
# Subroutines
#

#
# compile action
#
sub do_compile {
    my ($benchmark, $class, %table) = @_;

    my $exe_file = "$table{BENCHDIR}/bin/$benchmark\.$class";
    if (-e "$exe_file") {
	warn "Executable binary is already exists\n";
	return $false;
    }

    my $compile_command = "$table{MAKE} -omniconfig=scash_smp $benchmark CLASS=$class";
    chdir $table{BENCHDIR};
    open LOG, "$compile_command 2>&1 |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}
    }
    close LOG;

    unless (-e "$exe_file") {
	return $false;
    }
    return $true;
}

#
# re-compile action
#
sub do_recompile {
    my ($benchmark, $class, %table) = @_;

    my $exe_file = "$table{BENCHDIR}/bin/$benchmark\.$class";
    if (-e "$exe_file") {
	unlink("$exe_file");
    }	

    chdir $table{BENCHDIR};
    my $clean_command = "$table{MAKE} clean";
    open LOG, "$clean_command 2>&1 |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}
    }
    close LOG;

    my $compile_command = "$table{MAKE} -omniconfig=scash_smp $benchmark CLASS=$class";
    chdir $table{BENCHDIR};
    open LOG, "$compile_command 2>&1 |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}
    }
    close LOG;

    unless (-e "$exe_file") {
	return $false;
    }
    return $true;
}

#
# run action
#
sub do_run {
    my ($benchmark, $class, %table) = @_;

    my $exe_file = "$table{BENCHDIR}/bin/$benchmark\.$class";
    unless (-x "$exe_file") {
	warn "Exec file is not exist.\n";
	return $false;
    }

    my $valid = $false;
    my $cluster = $table{WITH_SCORE};
    my $ncluster;
    if (($cluster eq "0") || ($cluster eq "1")) {
	$ncluster = "4x";
    }
    elsif ($cluster eq "") {
	$ncluster = "8x";
    }
    else {
	warn "value of --with-score must 0 or 1 or null";
	exit $error_exit;
    }

    my $code = $false;
    open LOG, "unset OMPC_NUM_PROCS; unset OMP_NUM_THREADS; scrun -nodes=$ncluster$proc_num $exe_file 2>&1 |";
    while (<LOG>) {
	if ($table{VERBOSE}){
	    print $_;
	}

	if ($_ =~ /Benchmark Completed/) {
	    $code = $true;
	    next;
	}
    }

    return $code;
}

#
# verify action
#
sub do_verify {
    my ($benchmark, $class, $time, $proc_num, %table) = @_;

    my $exe_file = "$table{BENCHDIR}/bin/$benchmark\.$class";

    unless (-x "$exe_file") {
	warn "Exec file is not exist.\n";
	return $false;
    }

    my $valid = $false;
    my $cluster = $table{WITH_SCORE};
    my $ncluster;
    if (($cluster eq "0") || ($cluster eq "1")) {
	$ncluster = "4x";
    }
    elsif ($cluster eq "") {
	$ncluster = "8x";
    }
    else {
	warn "value of --with-score must 0 or 1 or null";
	exit $error_exit;
    }

    open LOG, "unset OMPC_NUM_PROCS; unset OMP_NUM_THREADS; scrun -nodes=$ncluster$proc_num $exe_file 2>&1 |";
    while (<LOG>) {
	if ($table{VERBOSE}){
	    print $_;
	}

	if ($_ =~ /Time in seconds\s*=\s*([0-9.]+)/) {
	    $$time = $1;
	    next;
	}
	elsif ($_ =~ /Verification\s+=\s+SUCCESSFUL/) {
	    $valid = $true;
	    next;
	}
	elsif ($_ =~ /^.*NaN.*$/) {
	    $valid = $false;
	    last;
	}
    }

    return $valid;
}

# EOF
