#
# rwcp.pl (Omni F-Test Test Tool)
#
# $Id: rwcp.pl,v 1.1.1.1 2004-02-06 18:16:53 msato Exp $
# $RWC_Release$
# $RWC_Copyright$
#

#
# Constant Definition
#

# Benchmark suits name
$benchmark_name = "F-test";

# All Benchmark name list
$all_bench = "none";
$allbench = "BT,CG,EP,FT,LU,MG,SP";

# All Class name list
$all_class = "none";

# Procs, Threads number default value decision
if (defined $ENV{OMPC_NUM_PROCS}) {
    $default_procs = $ENV{OMPC_NUM_PROCS};
} else {
    $default_procs = 1;
};

if (defined $ENV{OMP_NUM_THREADS}) {
    $default_thread = $ENV{OMP_NUM_THREADS};
} else {
    $default_thread = 1;
};

# Byte order swap flag decision
if (defined $ENV{OMPF77_BYTE_SWAP_IO}) {
    $default_byte_swap = $ENV{OMPF77_BYTE_SWAP_IO};
} else {
    $default_byte_swap = "no";
};

# Table for Default values
%default_table = (
		  BENCHDIR	=>	"./",      # Benchmark Direcotry
		  BENCHMARKS	=>	"none",    # Benchmark Name
		  CLASSES       =>      "none",       # Class name
		  PROCS	  =>	$default_procs,    # Processor number 
		  THREADS =>	$default_thread,   # Thread number
		  MAKE    =>	"make",            # command make
		  ACTION  =>	"verify",          # Benchtool master action
		  VERBOSE =>	0,                 # Verbose message output flag
		  BYTE_SWAP   => $default_byte_swap, # Byte order swap flag
		  );
#
# Subroutines
#

sub make_command {
    my (%table) = @_;

    my $verbose = $table{VERBOSE};
    my $make_command = $table{MAKE};
    my $benchdir = $table{BENCHDIR};
    chdir $benchdir;

    @benchlist = split /,/, $allbench;

    open LOG, "$make_command  2>&1 |";
    while ($bench = shift @benchlist) {
	while (<LOG>) {
	    if ($verbose) {
		print $_;
	    }
	    if ($_ =~ /^Making $bench\.\.$/) {
		last;
	    }
	}

	while (<LOG>) {
	    if ($verbose) {
		print $_;
	    }
	    if ($_ =~ /^Making $bench done\.$/) {
		last;
	    }
	}
    }

    if (defined shift @benchlist) {
	close LOG;
	return $false;
    }

    @benchlist = split /,/, $allbench;
    while ($bench = shift @benchlist) {
	while (<LOG>) {
	    if ($verbose) {
		print $_;
	    }
	    if ($_ =~ /^Run $bench\.\.$/) {
		last;
	    }
	}

	while (<LOG>) {
	    if ($verbose) {
		print $_;
	    }
	    if ($_ =~ /^Testing $bench OK\.$/) {
		last;
	    }
	}
    }
    close LOG;

    if (defined shift @benchlist) {
	return $false;
    }

    return $true;
}

#
# compile action
#
sub do_compile {
    my ($benchmark, $class, %table) = @_;

    return make_command(%table);
}

#
# re-compile action
#
sub do_recompile {
    my ($benchmark, $class, %table) = @_;

    my $make = $table{MAKE};
    my $verbose = $table{VERBOSE};
    unless ($verbose) {
	$output = ">& /dev/null";
    }
    chdir $table{BENCHDIR};
    system("$make clean $output");

    return make_command(%table);
}

#
# run action
#
sub do_run {
    my ($benchmark, $class, %table) = @_;

    return make_command(%table);
}

#
# verify action
#
sub do_verify {
    my ($benchmark, $class, $times, %table) = @_;

    return make_command(%table);
}

# EOF
