#
# specomp.pl (SPEComp-omp2001 Benchmark Test Tool)
#
# $Id: specomp.pl,v 1.1.1.1 2004-02-06 18:16:54 msato Exp $
# $RWC_Release$
# $RWC_Copyright$
#

#
# Constant Definition
#

# Benchmark suits name
$benchmark_name = "SPEComp-omp2001";

# All Benchmark name list
$all_bench = "wupwise,swim,mgrid,applu";

# All Class name list
$all_class = "l,m,s";

# Procs, Threads number default value decision
if (defined $ENV{OMPC_NUM_PROCS}) {
    $default_procs = $ENV{OMPC_NUM_PROCS};
} else {
    $default_procs = 1;
};

if (defined $ENV{OMP_NUM_THREADS}) {
    $default_thread = $ENV{OMP_NUM_THREADS};
} else {
    $default_thread = 1;
};

# Byte order swap flag decision
if (defined $ENV{OMPF77_BYTE_SWAP_IO}) {
    $default_byte_swap = $ENV{OMPF77_BYTE_SWAP_IO};
} else {
    open PIPE, "uname -a|";
    while (<PIPE>) {
        if (($_ =~ /i86pc/) || ($_ =~ /alpha/)) {
            $default_byte_swap = "yes";
        }
        else {
	    $default_byte_swap = undef;
        }
    }
};

# Environment variable SPEC decision
if (defined $ENV{SPEC}) {
    $default_spec = $ENV{SPEC};
} else {
    $default_spec = "./";
}

# Table for Default values
%default_table = (
		  BENCHDIR	=>	"./",      # Benchmark Direcotry
		  BENCHMARKS	=>	$all_bench,# Benchmark Name
		  CLASSES	=>	"s",       # Class name
		  PROCS	   =>	$default_procs,    # Processor number
		  THREADS  =>	$default_thread,   # Thread number
		  MAKE     =>	"make",            # command make
		  ACTION   =>	"verify",          # Benchtool master action
		  VERBOSE  =>	0,                 # Verbose message output flag
		  PERFVIEW =>   0,                 # Perfview format output flag
		  PERFVIEW_FILE => "./specomp-pv", # Perfview format output filename prefix
		  BYTE_SWAP => $default_byte_swap, # Byte order swap flag

# SPEComp specific FACTOR is below. These values is passed runspec as arguments.
# See also SPEComp document.

		  SPEC => $default_spec,
		  SPEC_CONFIG     => "",
		  SPEC_VERBOSE    => 0,

		  IGNORE_ERRORS   => "",
		  ITERATIONS      => 0,
		  OUTPUT_FORMAT   => "",
		  RATE            => 0,
		  RAWFORMAT       => "",
		  REPORTABLE      => 0,
		  TUNE            => "",
		  USERS           => 0,

		  DELETEWORK      => 0,

		  EXTENSION       => "",
		  MACHINE         => "",
		  MAKE_NO_CLOBBER => 0,
		  MAXCOMPARES     => "",
		  SETPROCGROUP    => 0,
		  SIZE            => "",
		  TABLE           => 0,
		  UNBUFFER        => 0,
		  USERNAME        => "",
		  VERSION         => ,
		  );

#
# Subroutines
#

sub make_command {
    my ($action, $benchmark, $class, %table) = @_;

    if ($table{SPEC_CONFIG} ne "") {
	$spec_config = "--config=$table{SPEC_CONFIG}";
    }
    if ($table{IGNORE_ERRORS} ne "") {
	$ignore_errors = "--ignore_erros";
    }
    if ($table{ITERATIONS}) {
	$iterations = "--iterations=$table{ITERATIONS}";
    }
    if ($table{OUTPUT_FORMAT} ne "") {
	$output_form = "--output_format=$table{OUTPUT_FORMAT}";
    }
    if ($table{RATE}) {
	$rate = "--rate"
    }
    if ($table{RAWFORMAT} ne "") {
	$rawformat = "--rawformat=$table{RAWFORMAT}";
    }
    if ($table{REPORTABLE}) {
	$reportable = "--reportable";
    }
    if ($table{TUNE}) {
	$tune = "--tune=$table{TUNE}";
    }
    if ($table{USERS}) {
	$users = "--users=$table{USERS}";
    }
    if ($table{DELETEWORKS}) {
	$deleteworks = "--deleteworks";
    }
    if ($table{EXTENSION} ne "") {
	$extension = "--extension=$table{EXTENSION}";
    }
    if ($table{MACHINE} ne "") {
	$machine = "--machine=$table{MACHINE}";
    }
    if ($table{MAKE_NO_CLOBBER}) {
	$make_no_clobber = "--make_no_clobber";
    }
    if ($table{MAXCOMPARES} ne "") {
	$maxcompares = "--maxcompares=$table{MAXCOMPARES}";
    }
    if ($table{SETPROCGROUP}) {
	$setprocgroup = "--setprocgroup";
    }
    if ($table{SIZE} ne "") {
	$size = "--size=$table{SIZE}";
    }
    if ($table{TABLE}) {
	$spectable = "--table";
    }
    if ($table{UNBUFFER}) {
	$unbuffer = "--unbuffer";
    }
    if ($table{USERNAME} ne "") {
	$usrename = "--username=$table{USERNAME}";
    }
    if ($table{SPEC_VERBOSE}) {
	$spec_verbose = "--verbose=$table{SPEC_VERBOSE}";
    }
    if ($table{VERSION}) {
	$version = "--version";
    }

    my $bench = "$benchmark\_$class";
    my $command = "./runspec $spec_config --action=$action $output_form $size $iterations $rate $ignore_erros $deleteworks $make_no_clobber $setprocgroup $spectable $tune $users $username $unbuffer $extension $machine $rawformat $maxcompares $spec_verbose $version $bench $output";

    return $command;
}

#
# compile action
#
sub do_compile {
    my ($benchmark, $class, %table) = @_;
    my $code = $false;

    my $spec_bin_dir = "$table{BENCHDIR}/bin";
    $ENV{PATH} = "$spec_bin_dir:$ENV{PATH}";
    $ENV{SPEC} = $table{SPEC};
    chdir $spec_bin_dir;

    my $compile_command = make_command("build", $benchmark, $class, %table);
    open LOG, "$compile_command |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}

	if ($_ =~ /^Build Complete/) {
	    $code = $true;
	}

	if ($_ =~ /^\*\*\* Error building/) {
	    $code = $false;
	}
    }
    close LOG;

    return $code;
}

#
# re-compile action
#
sub do_recompile {
    my ($benchmark, $class, %table) = @_;

    my $spec_bin_dir = "$table{BENCHDIR}/bin";
    $ENV{PATH} = "$spec_bin_dir:$ENV{PATH}";
    $ENV{SPEC} = $table{SPEC};
    chdir $spec_bin_dir;

    my $compile_command = make_command("nuke", $benchmark, $class, %table);
    open LOG, "$compile_command |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}
    }
    close LOG;

    return do_compile($benchmark, $class, %table);
}

#
# run action
#
sub do_run {
    my ($benchmark, $class, %table) = @_;
    my $dummy;

    return do_verify($benchmark, $class, \$dummy, %table);
}

#
# verify action
#
sub do_verify {
    my ($benchmark, $class, $times, %table) = @_;

    my $spec_bin_dir = "$table{BENCHDIR}/bin";
    $ENV{PATH} = "$spec_bin_dir:$ENV{PATH}";
    $ENV{SPEC} = $table{SPEC};
    chdir $spec_bin_dir;

    my $valid = $false;
    my $run_command = make_command("validate", $benchmark, $class, %table);
    open LOG, "$run_command |";
    while (<LOG>) {
	if ($table{VERBOSE}) {
	    print $_;
	}

	if ($_ =~ /^\s*format: raw -> (.*)$/) {
	    $logfile = $1;
	}
    }
    close LOG;

    open BENCHLOG, $logfile || die "Can't open file";
    my $prefix = "spec\\.omp2001\\.results\\.\\d+\\_$benchmark\_$class\\.base\\.\\d+";
    while (<BENCHLOG>) {
	if ($_ =~ /^$prefix\.reported\_time:\s+([\d\.]+)/) {
	    $$times = $1;
	}
	elsif ($_ =~ /^$prefix\.valid:\s+(\d)/) {
	    $valid = $1;
	}
    }
    close BENCHLOG;

    return $valid;
}

# EOF
